/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef BURST_H
#define BURST_H

//#include "ns3/drop-tail-queue.h"
#include "ns3/packet.h"
#include "ns3/head_queue.h"
#include <sys/queue.h>
#include <sys/types.h>
#include <inttypes.h>
	
namespace ns3 {

class PhyPdu;



class Burst;
LIST_HEAD (burst_list, Burst);

struct burst_link {
	Burst *le_next;
	Burst **le_prev;
};

/**
 * This class describes a burst
 */
class Burst
{	
  //friend class WimaxBurstTimer;

 public:
  /**
   * Creates a burst
   * @param phypdu The PhyPdu where it is located
   */
  Burst (PhyPdu *phypdu);

  /**
   * Delete the object
   */
  ~Burst ();

  /**
   * Return the CID for this burst
   * @return the CID for this burst
   */
  int getCid( );


  /**
   * Get and set cdma_code and opportunity from/to burst
   * @return the CID for this burst
   */
  inline uint8_t getB_CDMA_CODE( ) { return b_cdma_code_; }
  inline uint8_t getB_CDMA_TOP( ) { return b_cdma_top_; }
  inline void setB_CDMA_CODE( uint8_t cdma_code ) { b_cdma_code_ = cdma_code; }
  inline void setB_CDMA_TOP( uint8_t cdma_top ) { b_cdma_top_ = cdma_top; }

  /**
   * Return the burst duration in units of OFDM symbols (1 symbol x 1 subchannel)
   * @return the burst duration 
   */
  int getDuration( );

  /**
   * Set the duration of the burst in units of OFDM symbols
   * @param duration The burst duration
   */
  void setDuration (int duration);
  
  /**
   * Return the burst start time in units of symbol duration
   * @return the burst start time
   */
  int getStarttime( );

  /**
   * Return the Interval Usage Code
   * @return the burst start time
   */
  int getIUC( );

//rpi add for OFDMA 

  /** 
   * return the subchannel offset
   */
  int getSubchannelOffset (); 
  
  /** 
   * return the number of subchannels
   */
  int getnumSubchannels ();

  /** 
   * set the number of subchannels
   */
   void setnumSubchannels (int numsubchannels);

  /** 
   * set the subchannel offset- @param subchannel offset
   */
   void setSubchannelOffset (int subchanneloffset);

// rpi end add for OFDMA 

  /**
   * Set burst CID
   * @param cid The burst CID
   */
  void setCid( int cid );

  /**
   * Set burst start time in units of symbol duration
   * @param starttime the burst start time
   */
  void setStarttime( int starttime );

  /**
   * Set burst IUC
   * @param iuc The burst IUC
   */
  void setIUC( int iuc );



  /**
   * Enqueue a packet to be schedule in this burst
   * @param p The packet to queue
   */
  void enqueue (Ptr<Packet> p);

  /**
   * Dequeue a packet from the queue
   * @param p The packet to enqueue
   */
  Ptr<Packet> dequeue ();

  /**
   * Return the queue size in bytes
   */
  int getQueueLength();// { return queue_->byteLength();}



  /**
   * Remove the given packet from the queue
   * @param p The packet to remove
   */
  void remove(Ptr<Packet> p);

  //rpi added for dlsubframe timer 
 
  Ptr<Packet> lookup (int n);
 
  int getQueueLength_packets();  

  // rpi 

 // Chain element to the list
  inline void insert_entry_head(struct burst_list *head) {
    LIST_INSERT_HEAD(head, this, link);
  }
  
  // Chain element to the list
  inline void insert_entry(Burst *elem) {
    LIST_INSERT_AFTER(elem, this, link);
  }

  // Return next element in the chained list
  Burst* next_entry(void) const { return link.le_next; }

  // Remove the entry from the list
  inline void remove_entry() { 
    LIST_REMOVE(this, link); 
  }
  
  /*
   * Pointer to next in the list
   * would be protected but pybindgen complains
   */
  burst_link link;
  //LIST_ENTRY(Burst); //for magic draw
  
  /**
   * Packets to be sent during this burst
   * This would be protected, but the python bindings generator can't handle
   * protected variables
   */
  HeadQueue* queue_;
  
  
protected:
  /**
   * A timer use to transmit the burst
   */
  //WimaxBurstTimer timer_;



  /** 
   * Return the PhyPdu 
   * @return the PhyPdu 
   */
  inline PhyPdu *getPhyPdu () { return phypdu_; }



 private:
  /**
   * The CID for the burst. If a broadcast or multicast is used, then Mac SDUs for different SSs can be included in the burst.
   */
  int cid_;
  
  /**
   * The burst duration in units of OFDM Symbols
   */
  int duration_;
  
  /**
   * The start time of the burst in units of Symbol Duration
   */
  int starttime_;
  
  /**
   * The profile ID
   */
  int iuc_;

  /**
   * The PhyPdu where it is located
   */
  PhyPdu *phypdu_;

// rpi addition for ofdma

  /** 
   *  subchannel offset-
   */
   int subchannelOffset_; 

  /** 
   * number of subchannels
   */
   int numsubchannels_; 

  /** 
   * cdma code and top
   */
   uint8_t b_cdma_code_;
   uint8_t b_cdma_top_;
   uint32_t unique_id_;
};

}

#endif
