/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2007 University of Washington
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 //based heavily on DropTailQueue.cc

#include "ns3/log.h"
#include "ns3/enum.h"
#include "ns3/uinteger.h"
#include "ns3/wimax2-common.h"
#include "head_queue.h"

NS_LOG_COMPONENT_DEFINE ("HeadQueue");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (HeadQueue);

TypeId HeadQueue::GetTypeId (void) 
{
  static TypeId tid = TypeId ("ns3::HeadQueue")
    .SetParent<Queue> ()
    .AddConstructor<HeadQueue> ()
    .AddAttribute ("Mode", 
                   "Whether to use Bytes (see MaxBytes) or Packets (see MaxPackets) as the maximum queue size metric.",
                   EnumValue (PACKETS),
                   MakeEnumAccessor (&HeadQueue::SetMode),
                   MakeEnumChecker (BYTES, "Bytes",
                                    PACKETS, "Packets"))
    .AddAttribute ("MaxPackets", 
                   "The maximum number of packets accepted by this HeadQueue.",
                   UintegerValue (100),
                   MakeUintegerAccessor (&HeadQueue::m_maxPackets),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("MaxBytes", 
                   "The maximum number of bytes accepted by this HeadQueue.",
                   UintegerValue (100 * 65535),
                   MakeUintegerAccessor (&HeadQueue::m_maxBytes),
                   MakeUintegerChecker<uint32_t> ())
    ;
  return tid;
}

HeadQueue::HeadQueue () :
  Queue (),
  m_packets (),
  m_bytesInQueue (0),
  packetsInQueue (0)
{
  resetIterator();
  NS_LOG_FUNCTION_NOARGS ();
  m_mode = PACKETS;
  m_maxPackets = 100;		//taken from typeid declaration, which was taken from ns3::Queue class
  m_maxBytes = 100*65535; 	//taken from typeid declaration, which was taken from ns3::Queue class
}

HeadQueue::~HeadQueue ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

  void 
HeadQueue::SetMode (enum Mode mode)
{
  NS_LOG_FUNCTION (mode);
  m_mode = mode;
}

HeadQueue::Mode
HeadQueue::GetMode (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_mode;
}

bool
HeadQueue::EnqueueHead(Ptr<Packet> p) {	

	if (m_mode == PACKETS && (m_packets.size () >= m_maxPackets))
    {
      NS_LOG_LOGIC ("Queue full (at max packets) -- droppping pkt");
      Drop (p);
      return false;
    }

  if (m_mode == BYTES && (m_bytesInQueue + p->GetSize () > m_maxBytes))
    {
      NS_LOG_LOGIC ("Queue full (packet would exceed max bytes) -- droppping pkt");
      Drop (p);
      return false;
    }
    
    m_bytesInQueue += p->GetSize ();
    m_packets.push_front(p);

	NS_LOG_LOGIC ("Number packets " << m_packets.size () );
    NS_LOG_LOGIC ("Number bytes " << m_bytesInQueue);
	packetsInQueue++;
	
	//this is to make this function play nice with DoEnqueue, since we don't have the ns3
	//defined function to take care of some housekeeping that it expects
	Enqueue(p);
	m_packets.pop_back();
	
  return true;

}

Ptr<Packet> HeadQueue::head() {
	if (GetNPackets() > 0)	return m_packets.front();
	return Ptr<Packet>(NULL);
}

  
bool 
HeadQueue::DoEnqueue (Ptr<Packet> p)
{
  NS_LOG_FUNCTION (this << p);

  if (m_mode == PACKETS && (m_packets.size () >= m_maxPackets))
    {
      NS_LOG_LOGIC ("Queue full (at max packets) -- droppping pkt");
      Drop (p);
      return false;
    }

  if (m_mode == BYTES && (m_bytesInQueue + p->GetSize () > m_maxBytes))
    {
      NS_LOG_LOGIC ("Queue full (packet would exceed max bytes) -- droppping pkt");
      Drop (p);
      return false;
    }

  m_bytesInQueue += p->GetSize ();
  m_packets.push_back(p);

  NS_LOG_LOGIC ("Number packets " << m_packets.size ());
  NS_LOG_LOGIC ("Number bytes " << m_bytesInQueue);
  packetsInQueue++;
  return true;
}

uint32_t
HeadQueue::GetQueueLength() {
	return m_packets.size();
}

uint32_t
HeadQueue::GetNPackets() {
	return m_packets.size ();
//	return packetsInQueue;
}

void HeadQueue::Remove(Ptr<Packet> p) {
	std::deque<Ptr<Packet> >::iterator itter;
	for (itter = m_packets.begin(); itter != m_packets.end(); ++itter) {
		if (p = *itter) {
			packetsInQueue--;
			return;
		}
	}
	
}


Ptr<Packet>
HeadQueue::DoDequeue (void)
{
  NS_LOG_FUNCTION (this);
  
   
  if (m_packets.empty()) 
    {
      NS_LOG_LOGIC ("Queue empty");
      return 0;
    }

  Ptr<Packet> p = m_packets.front ();
  m_packets.pop_front ();
  m_bytesInQueue -= p->GetSize ();

  NS_LOG_LOGIC ("Popped " << p);

  NS_LOG_LOGIC ("Number packets " << m_packets.size ());
  NS_LOG_LOGIC ("Number bytes " << m_bytesInQueue);
  packetsInQueue--;
  return p;
}


Ptr<const Packet>
HeadQueue::DoPeek (void) const
{
  NS_LOG_FUNCTION (this);
  
  
  if (m_packets.empty()) 
    {
      NS_LOG_LOGIC ("Queue empty");
      return 0;
    }

  Ptr<Packet> p = m_packets.front ();

  NS_LOG_LOGIC ("Number packets " << m_packets.size ());
  NS_LOG_LOGIC ("Number bytes " << m_bytesInQueue);

  return p;
}

Ptr<Packet> HeadQueue::lookup(int i) {
	std::deque<Ptr<Packet> >::iterator itter;
	for (itter = m_packets.begin(); itter != m_packets.end(); ++itter) {
		if (i == 0) return *itter;
		i--;
	}
	Ptr<Packet> p = Ptr<Packet>(NULL);
	return p;
}

Ptr<Packet> HeadQueue::getNext() {
	pos++;
	return *pos;
}

void HeadQueue::resetIterator() {
	pos = m_packets.begin();
}

void HeadQueue::DoDispose() {
	Queue::DoDispose();
}

// Ptr<Packet> HeadQueue::createEmptyPacket() {
// 	Ptr<Packet> p(new Packet);
// 	All_Headers info;
//  	Generic_MAC_Header *GMH = &info.GMH;
//  	GMH->setHeaderType(0);		
//  	GMH->setEncryptionControl(0); 
//  	GMH->setARQFB(0);	
//  	GMH->setExt(0);		
//  	GMH->setFrag(0);		
//  	GMH->setPck(0);		
//  	GMH->setFFBGM(0);	
//  	GMH->setCI(0); 		
//  	GMH->setEks(0);
//  	GMH->setPduLen(6); 
//  	GMH->setCID(BROADCAST_CID);	  
//  	GMH->setHCS(0);	 
//  	GMH->GMH.len = 0;
//  	p->AddHeader(info);
// 	return p;
// }

} // namespace ns3

//#include "ns3/test.h"
//
//namespace ns3 {
//
//class HeadQueueTestCase : public TestCase
//{
//public:
//  HeadQueueTestCase ();
//  virtual void DoRun (void);
//};
//
//HeadQueueTestCase::HeadQueueTestCase ()
//  : TestCase ("Sanity check on the drop tail queue implementation")
//{}
//void  
//HeadQueueTestCase::DoRun (void)
//{
//  Ptr<HeadQueue> queue = CreateObject<HeadQueue> ();
//  NS_TEST_EXPECT_MSG_EQ (queue->SetAttributeFailSafe ("MaxPackets", UintegerValue (3)), true, 
//                         "Verify that we can actually set the attribute");
//  
//  Ptr<Packet> p1, p2, p3, p4;
//  p1 = Create<Packet> ();
//  p2 = Create<Packet> ();
//  p3 = Create<Packet> ();
//  p4 = Create<Packet> ();
//
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 0, "There should be no packets in there");
//  queue->Enqueue (p1);
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 1, "There should be one packet in there");
//  queue->Enqueue (p2);
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 2, "There should be two packets in there");
//  queue->Enqueue (p3);
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 3, "There should be three packets in there");
//  queue->Enqueue (p4); // will be dropped
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 3, "There should be still three packets in there");
//
//  Ptr<Packet> p;
//
//  p = queue->Dequeue ();
//  NS_TEST_EXPECT_MSG_EQ ((p != 0), true, "I want to remove the first packet");
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 2, "There should be two packets in there");
//  NS_TEST_EXPECT_MSG_EQ (p->GetUid (), p1->GetUid (), "was this the first packet ?");
//
//  p = queue->Dequeue ();
//  NS_TEST_EXPECT_MSG_EQ ((p != 0), true, "I want to remove the second packet");
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 1, "There should be one packet in there");
//  NS_TEST_EXPECT_MSG_EQ (p->GetUid (), p2->GetUid (), "Was this the second packet ?");
//
//  p = queue->Dequeue ();
//  NS_TEST_EXPECT_MSG_EQ ((p != 0), true, "I want to remove the third packet");
//  NS_TEST_EXPECT_MSG_EQ (queue->GetNPackets (), 0, "There should be no packets in there");
//  NS_TEST_EXPECT_MSG_EQ (p->GetUid (), p3->GetUid (), "Was this the third packet ?");
//
//  p = queue->Dequeue ();
//  NS_TEST_EXPECT_MSG_EQ ((p == 0), true, "There are really no packets in there");
//
//  return ;
//}

//static class HeadQueueTestSuite : public TestSuite
//{
//public:
//  HeadQueueTestSuite ()
//    : TestSuite ("head-queue", UNIT)
//  {
//    AddTestCase (new HeadQueueTestCase ());
//  }
//} g_HeadQueueTestSuite;

//} // namespace ns3
