/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "dlsubframetimer.h"
#include "ns3/framemap.h"
#include "ns3/subframe.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/contentionslot.h"
#include "ns3/wimax2-ss-net-device.h"



namespace ns3 {

/**
 * Creates a timer to handle the subframe transmission
 * @param subframe The DlSubframe
 */
DlSubFrameTimer::DlSubFrameTimer (DlSubFrame *subframe): Wimax2Timer(0), OFDMSymbol_(DL_PREAMBLE), newsubframe_(true)
{
  assert (subframe);
  subframe_ = subframe;
  busy_ = 0;
}

/**
 * Reset the timer
 */
void DlSubFrameTimer::reset ()
{
  //  burstIndex_ = 0;
  //  newburst_ = true;
  OFDMSymbol_ = DL_PREAMBLE;   
  newsubframe_ = true; 
  if (busy_ == 1) stop();
}

void DlSubFrameTimer::start(double time) {
  debug10 ("STARTING DL SUB FRAME TIMER with time %f\n", time);
  assert(busy_ == 0);
  reset();
  busy_ = 1;
  paused_ = 0;
  stime = NOW;
  rtime = time;
  assert(rtime >= 0.0);
  //intr = Simulator::ScheduleWithContext(mac->GetNode()->GetId(), Seconds(rtime), &Wimax2Timer::handle, this, &intr/*function,args*/);
  if (intr.IsRunning()) {
  	Simulator::Cancel(intr);
  	}
  if (time == 0)
    	//intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(0), &Wimax2Timer::handle, this/*function,args*/);
    	Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(0), &Wimax2Timer::handle, this/*function,args*/);
  else
	//intr = Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
	Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(rtime), &Wimax2Timer::handle, this/*function,args*/);
}

/**
 * we handle an event the same way we handle expiration
 */
void DlSubFrameTimer::handle() {
	//debug10("DL SUBFRAME TIMER EXPIRED\n");
	busy_ = 0;
	expire ();
}

/**
 * When it expires, the timer will handle the next packet to send
 */
void DlSubFrameTimer::expire()
{
  if (!mac) {
    mac= subframe_->map_->getMac();
  }

  FrameMap *map = mac->getMap();
  if(mac->getNodeType() == STA_MN ) {
    mac->GetObject<Wimax2SSNetDevice>()->StartSubframe(true); //notify SSNetDevice we're in DL Subframe
    return;
  }

  //if we get past the above condition, we're in a BS
  mac->getPhy()->setMode(OFDM_SEND);

  for (int index = 0 ; index < map->getDlSubframe()->getPdu ()->getNbBurst(); index++) 
    {
    
      Burst *b = subframe_->getPdu()->getBurst(index);
      if (b->getIUC()==DIUC_END_OF_MAP) break;
		
      //change modulation
      int iuc = b->getIUC();
      Ofdm_mod_rate rate = subframe_->getProfile (iuc)->getEncoding();
      mac->getPhy()->setModulation (rate);

      int len = b->getQueueLength_packets();
      debug10("Entering dlsubframe timer loop with len = %d\n",len);
	  while(len > 0) {
	  	 len--;
		  Ptr<Packet> p = b->dequeue();


		 PHY_INFO_Header pih;
		 pih.num_subchannels = b->getnumSubchannels();
		 pih.subchannel_offset = b->getSubchannelOffset();
		 pih.num_OFDMSymbol = b->getDuration();
		 pih.OFDMSymbol_offset = b->getStarttime();
		 pih.direction = DL_; //since we're in the DL subframetimer, outgoing packets will be in this direction
		 pih.modulation_ = rate;

		 p->AddHeader(pih);

		  
		  debug2(" loop dlsubframe timer -Symbol_off[%d]\t symbol #[%d]\t subch_off[%d]\t subch_#[%d], burst_index :%d\n",
			 pih.OFDMSymbol_offset,pih.num_OFDMSymbol, pih.subchannel_offset, pih.num_subchannels, index);
			 
			 double transmission_time = pih.OFDMSymbol_offset*mac->getPhy()->getSymbolTime();
			 debug10("transmitting the packet p dequeued from b starting at %f seconds from now\n", transmission_time);
			 
			 Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(transmission_time), &Wimax2NetDevice::transmit, mac, p);
			 
			 //this packet is sent out at time (start of subframe) + symbol length * symbol offset
			 
	}
    }
    
	reset();
}

} //namespace ns3
