/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef SUBFRAME_H
#define SUBFRAME_H

#include "profile.h"
//#include "contentionslot.h"
#include "phypdu.h"
#include <list>
//#include "ulsubframetimer.h"
//#include "dlsubframetimer.h"

namespace ns3 {

class FrameMap;
class DlSubFrameTimer;
class UlSubFrameTimer;

/**
 * Abstract class for subframe
 */
class SubFrame 
{
 public:
  /**
   * Constructor
   * @param map The frame 
   */
    // -#- @map(transfer_ownership=false) -#-
  SubFrame (FrameMap *map);

  /**
   * Destructor
   */
  ~SubFrame ();

  /**
   * Add a downlink profile
   * @param f The frequency of the profile
   * @param enc The encoding used in the profile
   * @return a new profile with the given caracteristics
   */
  Profile * addProfile (int f, Ofdm_mod_rate enc);
  
  /**
   * Return the DL profile with the given UID
   * @return the DL profile with the given UID
   */
  Profile * getProfile (int iuc); 

  /**
   * Remove a downlink profile
   * @param p The profile to remove
   */
    // -#- @p(transfer_ownership=true) -#-
  void removeProfile (Profile *p);

  /**
   * Remove all downlink profiles
   */
  void removeProfiles ();

  /**
   * Return the number of profiles for this subframe
   */
  int getNbProfile ();

  /**
   * Return the head of the profile list
   */
  Profile * getFirstProfile ();

  /**
   * Return the Configuration Change count
   */
  int getCCC ();

  /**
   * Increment the configuration change count
   */
  void incrCCC ();

 //protected: uncomment if python bindings genrator can handle protected variables in the future
  /**
   * The frame where it is located
   */
  FrameMap *map_;

  /**
   * List of burst profiles for downlink
   */
  struct profile_list profile_list_;
  
  /**
   * Number of element in the PhyPDU list
   */
  int nbProfile_;

  /**
   * Configuration change count
   */
  int ccc_;
};



/**
 * This class defines a downlink subframe
 */
class DlSubFrame : public SubFrame
{
  friend class DlSubFrameTimer;
 public:
  /**
   * Constructor
   * @param map The frame 
   */
    // -#- @map(transfer_ownership=false) -#-
  DlSubFrame (FrameMap *map);

  /**
   * Destructor
   */
  ~DlSubFrame ();

  /**
   * Return the DL phypdu
   * @return the phypdu
   */
  inline PhyPdu * getPdu () { return phypdu_; }

  /**
   * Set the channel ID
   * @param id The channel id
   */
  inline void setChannelID (int id) { channel_ = id; }

  /**
   * Return the channel ID
   * @return The channel ID
   */
  inline int getChannelID () { return channel_; }
  
  /**
   * Set the start time in units of PS
   * @param time The subframe start time
   */
  inline void setAbsStarttime (double time) { abs_starttime_ = time;}

  /**
   * Get the start time in units of PS
   * @return The subframe start time
   */
  inline double getAbsStarttime () { return abs_starttime_;}

  /**
   * Return the timer handling the subframe
   */
  inline DlSubFrameTimer *getTimer () { return timer_; }
  
  void replaceTimer();
  DlSubFrameTimer* replaceTimerWithoutDelete();
    // -#- @in(transfer_ownership=true) -#-
  void deleteSavedTimer(DlSubFrameTimer *in);

  
  void replaceTimer(double time);


 private:

  /**
   * List of uplink PhyPDU composing the uplink subframe
   */
  PhyPdu * phypdu_;

  /**
   * The downlink channel id
   */
  int channel_;
  
  /**
   * Start time of the subframe in unit of PS
   */
  double abs_starttime_;

  /**
   * The timer for handling bursts
   */
  DlSubFrameTimer *timer_;
};

class BwContentionSlot;
class RngContentionSlot;
/**
 * This class defines a uplink subframe
 */
class UlSubFrame : public SubFrame
{
  friend class UlSubFrameTimer;
 public:
  /**
   * Constructor
   * @param map The frame 
   */
    // -#- @map(transfer_ownership=false) -#-
  UlSubFrame (FrameMap *map);

  /**
   * Destructor
   */
  ~UlSubFrame ();

  /**
   * Return the DL phypdu
   * @return the phypdu
   */
  PhyPdu * getFirstPdu (); 

  std::list<PhyPdu*> * getPduList(); 

  
  /**
   * Add an uplink phy pdu
   * @param pos The position of the PhyPdu in the subframe
   * @param preamble The size of the preamble in units of XX
   * @return newly created PhyPdu
   */
  PhyPdu * addPhyPdu (uint32_t pos, int preamble);

  /**
   * Remove a Phy PDU
   * @param pdu The Phy PDU to remove
   */
  void removePhyPdu (PhyPdu *p);

  /**
   * Return the number of PhyPdu
   */
  uint32_t getNbPdu ();  

  /**
   * Return the burst located at the given index
   * @param pos The position of the burst
   */
  PhyPdu* getPhyPdu(uint32_t pos);

  /*
   * Return the contention slot for BW requests
   * @return the pointer to the contention slot for BW requests
   */
   BwContentionSlot* getBw_req( ); 

  /*
   * Return the contention slot for Ranging requests
   * @return the pointer to the contention slot for Ranging requests
   */
  inline RngContentionSlot* getRanging( ) { return ranging_; }

  /**
   * Set the start time in units of PS
   * @param time The subframe start time
   */
  inline void setStarttime (int time) { starttime_ = time;}

  /**
   * Get the start time in units of PS
   * @return The subframe start time
   */
  inline int getStarttime () { return starttime_;}

  /**
   * Set the channel ID
   * @param id The channel id
   */
  inline void setChannelID (int id) { channel_ = id; }

  /**
   * Return the channel ID
   * @return The channel ID
   */
  inline int getChannelID () { return channel_; }

  /**
   * Return the timer handling the subframe
   */
  inline UlSubFrameTimer *getTimer () { return timer_; }
  
  void replaceTimer();
  
  void replaceTimer(double time);

// protected: - uncomment if and when pybindgen gains support for protected member variables

  
  /**
   * Information about the ranging contention
   */
  RngContentionSlot *ranging_;

 private:
  /**
   * Information about the bandwidth request contention
   */
  BwContentionSlot *bw_req_;

  /**
   * List of uplink PhyPDU composing the uplink subframe
   */
   std::list<PhyPdu*> phypdu_list_;
//   PhyPdu *phypdu_list_;

  /**
   * Number of phypdu
   */
  int nbPhyPdu_;
  
  /**
   * Start time of the subframe in unit of PS
   */
  int starttime_;

  /**
   * The downlink channel id
   */
  int channel_;

  /**
   * The timer for handling bursts
   */
  UlSubFrameTimer *timer_;

};

}
#endif
