/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */
//#include <cstdlib> 
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/log.h"
//#include "ns3/wimax2-scheduler.h"
#include "ns3/wimax2-bs-scheduler.h"
#include "ns3/ulsubframetimer.h"
#include "ns3/dlsubframetimer.h"
#include "ns3/destclassifier.h"
#include "ns3/wimaxneighborentry.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/pointer.h"
#include "ns3/wimax2-channel.h"
//#include "cmu-trace.h"
//#include "globalparams_wimax.h"
//#include <rtp.h>

NS_LOG_COMPONENT_DEFINE ("Wimax2BSNetDevice");


namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (Wimax2BSNetDevice);


TypeId Wimax2BSNetDevice::GetTypeId (void) {
    static TypeId tid = TypeId ("ns3::Wimax2BSNetDevice")
    .SetParent<Wimax2NetDevice> ()
    .AddConstructor<Wimax2BSNetDevice> ()
    .AddAttribute  ("TxQueue", "A queue used to hit the hooks necessary for PyViz.", PointerValue(), MakePointerAccessor (&Wimax2BSNetDevice::m_queue), MakePointerChecker<Queue>())
    .AddTraceSource("Rx", "Receive trace", MakeTraceSourceAccessor (&Wimax2BSNetDevice::m_traceRx))
    .AddTraceSource("Tx", "Transmit trace", MakeTraceSourceAccessor (&Wimax2BSNetDevice::m_traceTx))
    ;
    return tid;
}

/**
 * Creates a Wimax2NetDevice
 */
Wimax2BSNetDevice::Wimax2BSNetDevice() : Wimax2NetDevice (), cl_head_(0), cl_tail_(0), ctrlagent_(0)
{
  NS_LOG_FUNCTION_NOARGS ();
  type_ = STA_BS; //type of MAC. In this case it is for SS
  
  //Default configuration
  addClassifier (new DestClassifier ());
  scheduler_ = new Wimax2BSScheduler();
  scheduler_->setMac (this); //register the mac
  init_default_connections ();
  LIST_INIT (&t17_head_);
  LIST_INIT (&scan_stations_);
  LIST_INIT (&fast_ranging_head_);
  bw_peer_ = NULL;
  bw_node_index_ = 0;
  sendDCD = false;
  dlccc_ = 0;
  sendUCD = false;
  ulccc_ = 0;
  reg_SS_number = 0;
  map_ = new FrameMap (this);
  
  //create timers for DL/UL boundaries
  dl_timer_ = new DlTimer (this);
  ul_timer_ = new UlTimer (this);

  //initialize queue used to hit PyViz hooks
  m_queue = CreateObject<DropTailQueue> ();
}

/**
 * Initialize default connections
 */
void Wimax2BSNetDevice::init_default_connections ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Connection * con;

  //create initial ranging and padding connection
  con = new Connection (CONN_INIT_RANGING);
  con->enable_fragmentation(false);
  con->enable_packing(false);
  connectionManager_->add_connection (con, IN_CONNECTION); //uplink
  con = new Connection (CONN_INIT_RANGING);
  con->enable_fragmentation(false);
  con->enable_packing(false);
  connectionManager_->add_connection (con, OUT_CONNECTION); //downlink
  con = new Connection (CONN_PADDING);
  connectionManager_->add_connection (con, IN_CONNECTION);
  con = new Connection (CONN_PADDING);
  connectionManager_->add_connection (con, OUT_CONNECTION);

  //we need to create a Broadcast connection and AAS init ranging CIDs
  con = new Connection (CONN_BROADCAST);
  con->enable_fragmentation(false);
  con->enable_packing(false);
  connectionManager_->add_connection (con, OUT_CONNECTION);
  con = new Connection (CONN_AAS_INIT_RANGING);
  con->enable_fragmentation(false);
  con->enable_packing(false);
  connectionManager_->add_connection (con, IN_CONNECTION);
}

/**
 * Initialize the MAC
 */
void Wimax2BSNetDevice::init ()
{
  NS_LOG_FUNCTION_NOARGS ();

  //schedule the first frame by using a random backoff to avoid
  //synchronization between BSs.
  
  UniformVariable unfvar(0,getFrameDuration());
  
  double stime = getFrameDuration () + unfvar.GetValue(0, getFrameDuration ());
  dl_timer_->Schedule(Seconds (stime));

  //also start the DCD and UCD timer
  dcdtimer_ = new WimaxDCDTimer (this);
  ucdtimer_ = new WimaxUCDTimer (this);
  nbradvtimer_ = new WimaxMobNbrAdvTimer (this);
  dcdtimer_->start (macmib_->dcd_interval);
  ucdtimer_->start (macmib_->ucd_interval);
  nbradvtimer_->start (macmib_->nbr_adv_interval+stime);
  int nbPS = (int) floor((getFrameDuration()/getPhy()->getPS()));
  int nbPS_left = nbPS - phymib_.rtg - phymib_.ttg;
  int nbSymbols = (int) floor((getPhy()->getPS()*nbPS_left)/getPhy()->getSymbolTime());  // max num of OFDM symbols available per frame. 
  int nbSubcarrier = getPhy()->getFFT();
  int nbSubchannel = getPhy()->getNumsubchannels (UL_);


  intpower_ = (double**) malloc (nbSymbols*sizeof (double*));
  for (int i = 0 ; i<nbSymbols ; i++) {
    intpower_[i] = (double*) malloc (nbSubcarrier*sizeof (double));
    for (int j = 0 ; j< nbSubcarrier ; j++)
      intpower_[i][j] = 0.0 ; 
  }
  basepower_ = (double**) malloc (nbSymbols*sizeof (double*));
  for (int i = 0 ; i<nbSymbols; i++) {
    basepower_[i] = (double*) malloc (nbSubchannel*sizeof (double));
    for (int j = 0 ; j<nbSubchannel ; j++)
      basepower_[i][j] = 0.0 ; 
  }
   

  //init the scheduler
  scheduler_->init();
  //init the channel for collision detection sync (will only run once if multiple BS's try to initialize it)
  getPhy()->getChannel()->initializeCollisionDetection();

  //setup the data structure that stores incoming packet fragments
  //note that this must be after the scheduler is initialized so everything needed is set (max DL and UL durations)
  initialize_grid();
}

/**
 * Set the control agent
 */
void Wimax2BSNetDevice::setCtrlAgent (Wimax2ControlAgent *agent)
{
  NS_LOG_FUNCTION_NOARGS ();
  assert (agent);
  ctrlagent_ = agent;
}

/**** Packet processing methods ****/

/*
 * Process packets going out
 * @param p The packet to send out
 */
void Wimax2BSNetDevice::sendDown(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  debug10("Entering (BS) %s sendDown\n", addr_s());
	accepting_ = false;
  //We first send it through the CS
  int cid = -1;
  All_Headers wimaxHdr;//, trash;

//   if (!notify_upper_) {
//     assert (!PeekPointer(pktBuf_));
//     pktBuf_ = p;
//     return;
//   } 

  cid = classify (p);
  p->RemoveHeader(wimaxHdr);






  if (cid == -1) {
    debug2 ("At %f in Mac %d drop packet because no classification were found\n", NOW, id_index);
    drop(p, "CID");
    p = 0;
    return;
    //Packet::free (p);
  } else {
    //Enqueue the packet 
    Connection *connection = connectionManager_->get_connection (cid, OUT_CONNECTION);
    if (connection == NULL) {
      debug2 ("Warning: At %f in Mac %d connection with cid = %d does not exist. Please check classifiers\n", NOW, id_index, cid);
      //update_watch (&loss_watch_, 1);
      drop(p, "CID");
      return;
    }
    else 
    {
			bool need_drop = false;
 			if(!connection->getArqStatus())
			{
				if(connection->queueLength() >= macmib_->queue_length)
				{
					debug10("dropping due to connection->queueLength (%d) > macmib_->queue_length (%d)\n",connection->queueLength(),macmib_->queue_length);
				//	update_watch (&loss_watch_, 1);
					need_drop = true;
					drop (p, "QWI");
					return;
				}
			}
			else if(connection->getArqStatus() && (connection->getArqStatus()->isArqEnabled() == 1))
			{
				int data_packet_size = wimaxHdr.GMH.GMH.len;
				int num_arq_blocks = (int) ceil ((double)data_packet_size/arq_block_size_);

				debug2("data_packet_size=%d\t num_arq_blocks=%d\t qlength=%d max_qlength=%d\n",
							data_packet_size,num_arq_blocks,connection->queueLength(),macmib_->queue_length);

				if((connection->queueLength() + num_arq_blocks) >= macmib_->queue_length)
				{
					//update_watch (&loss_watch_, 1);
					need_drop = true;
					drop (p, "QWI");
					return;
				}
			}
			
			if( !need_drop)
			{
				if (connection->getArqStatus () == NULL) 
				{
					//End RPI
					//update mac header information
					//set header information
					//hdr_mac802_16 *wimaxHdr = HDR_MAC802_16(p);
					//we have to stick it into this queue then remove it to hit the hooks PyViz wants due to the need to register it as a "Point to Point like device" (so that we can easily work with just a PyViz plugin rather than modifying src/tools/visualizer/model/pyviz.cc)
					m_queue->Enqueue(p);
					m_queue->Dequeue(); //this actually triggers the hook
					m_traceTx(p, declassify(cid));//, wimaxHdr.dest);
					wimaxHdr.GMH.GMH.ht = 0;
					wimaxHdr.GMH.GMH.ec = 1;
					wimaxHdr.GMH.GMH.type_mesh = 0;
					wimaxHdr.GMH.GMH.type_arqfb = 0;
					wimaxHdr.GMH.GMH.type_ext = 0;
					wimaxHdr.GMH.GMH.type_frag = 0;
					wimaxHdr.GMH.GMH.type_pck = 0;
					wimaxHdr.GMH.GMH.type_fbgm = 0;
					wimaxHdr.GMH.GMH.ci = 0;
					wimaxHdr.GMH.GMH.eks = 0;
					wimaxHdr.GMH.GMH.cid = cid; //default
					wimaxHdr.GMH.GMH.hcs = 0;
					wimaxHdr.GMH.GMH.len += wimaxHdr.GMH.GetSerializedSize();
					p->AddHeader(wimaxHdr);
					connection->enqueue (p);
					debug2 ("At %f (BS) in Mac %d, SENDDOWN, Enqueue packet to cid :%d queue size :%d (max :%d)\n", NOW, id_index, cid,connection->queueLength (), macmib_->queue_length);
					//Begin RPI
				}
				else if((connection->getArqStatus () != NULL) && (connection->getArqStatus ()->isArqEnabled () == 1)) 
				{
					//We will have to divide the packet to ARQ Blocks
					//In NS2 it will be implemented by duplicating the packet but changing the size of the packet

					int packet_size = wimaxHdr.GMH.GMH.len;
					int number_of_packs= (packet_size/arq_block_size_); 
					//hdr_mac802_16 *wimaxHdr = HDR_MAC802_16(p);
					m_queue->Enqueue(p);
					m_queue->Dequeue(); //this actually triggers the hook
					m_traceTx(p, declassify(cid));//, wimaxHdr.dest);
					wimaxHdr.GMH.GMH.ht = 0;
					wimaxHdr.GMH.GMH.ec = 1;
					wimaxHdr.GMH.GMH.type_mesh = 0;
					wimaxHdr.GMH.GMH.type_arqfb = 0;
					wimaxHdr.GMH.GMH.type_ext = 0;
					wimaxHdr.GMH.GMH.type_frag = 0;
					wimaxHdr.GMH.GMH.type_pck = 0;
					wimaxHdr.GMH.GMH.type_fbgm = 0;
					wimaxHdr.GMH.GMH.ci = 0;
					wimaxHdr.GMH.GMH.eks = 0;
					wimaxHdr.GMH.GMH.cid = cid; //default
					wimaxHdr.GMH.GMH.hcs = 0;
					wimaxHdr.GMH.GMH.len = arq_block_size_;
					p->AddHeader(wimaxHdr);
					
					for(int i=1; i<= number_of_packs; i++)
					{
						if(i == 1)
							connection->getArqStatus ()->arqSend(p, connection, FRAG_FIRST);
						else
							connection->getArqStatus ()->arqSend(p, connection, FRAG_CONT);					
						connection ->enqueue (p->Copy());
					}
					
					wimaxHdr.GMH.GMH.len = (packet_size - (number_of_packs * arq_block_size_));
					p->AddHeader(wimaxHdr);
					if(number_of_packs)
					{
						connection->getArqStatus ()->arqSend(p, connection, FRAG_LAST);
					}
					else
					{
						connection->getArqStatus ()->arqSend(p, connection, FRAG_NOFRAG);
					}
					connection ->enqueue (p);
					debug2 ("ARQ BS Dividing Packet into Blocks Packet Size=%d, Number of packets = %d, ARQ_BLOCK_SIZE = %d,Last Packet Size= %d\n",packet_size,number_of_packs+1,arq_block_size_, wimaxHdr.GMH.GMH.len) ;
				}
			}
		}
	}

	//inform upper layer that it can send another packet
	accepting_ = true;
	//resume (NULL);  
}

/*
 * Transmit a packet to the physical layer
 * @param p The packet to send out
 */
void Wimax2BSNetDevice::transmit(Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  debug10("Entering (BS) %s transmit\n", addr_s());

  debug2 ("At %f (BS) in Mac %d, sending packet, TRANSMIT, (size :%d)\n", NOW, id_index, p->GetSize());


  debug10("About to start sending a packet of size %d to the PHY layer\n", p->GetSize());

  //pass it down
  phy_->Send(p);
  //Simulator::Schedule(Seconds(ah.txtime), &Wimax2Phy::Send, phy_, p, 0, Mac48Address::GetBroadcast(), m_address, Ptr<Wimax2NetDevice>(this));
  
}

/*
 * Process incoming packets (from other hosts)
 * @param p The incoming packet
 * This is the old sendUp function that asumed entire packets were recieved at once
 * It is left in for reference with other older code
 */
//void Wimax2BSNetDevice::sendUp (Ptr<Packet> p, uint8_t subchannel, Time received)
//{
//  NS_LOG_FUNCTION_NOARGS ();
//  debug10("Entering (BS) %s sendUp\n",addr_s());
//  All_Headers wimaxHdr;
//  p->RemoveHeader(wimaxHdr);
//  p->PeekHeader(wimaxHdr.GMH);
//  p->PeekHeader(wimaxHdr.SMH);
//  p->RemoveHeader(wimaxHdr.CRH);
//
//  debug10("wimaxHdr.GMH.GMH.ht = %d\n",(int)wimaxHdr.GMH.GMH.ht);
//  debug10("wimaxHdr.SMH.BRH.ht = %d\n",(int)wimaxHdr.SMH.BRH.ht);
//  if (wimaxHdr.GMH.GMH.ht == 0)
//	  debug10 ("At %f (BS) in Mac %s, SENDUP, receive first bit..over at :%f(txtime :%f) (size :%d), cid :%d\n", NOW, HelperFunctions::getMacAddrString(m_address).c_str(), NOW+wimaxHdr.txtime,wimaxHdr.txtime, wimaxHdr.GMH.GMH.len, wimaxHdr.GMH.GMH.cid);
//  else if (wimaxHdr.SMH.BRH.type == 0)
//	  debug10 ("At %f (BS) in Mac %s, SENDUP, receive first bit..over at :%f(txtime :%f) (size :%d), cid :%d\n", NOW, HelperFunctions::getMacAddrString(m_address).c_str(), NOW+wimaxHdr.txtime,wimaxHdr.txtime, 6, wimaxHdr.SMH.BRH.cid);
//  else 
//	  debug10 ("At %f (BS) in Mac %s, SENDUP, receive first bit..over at :%f(txtime :%f) (size :%d), cid :%d\n", NOW, HelperFunctions::getMacAddrString(m_address).c_str(), NOW+wimaxHdr.txtime,wimaxHdr.txtime, 6, wimaxHdr.CRH.data.cid);
//  if (wimaxHdr.GMH.GMH.ht == 0) {
//  	 if (wimaxHdr.GMH.GMH.rsv1 == 1) wimaxHdr.mac_packet = 1;
//	 if (p->GetSize() > 6) {
//		Type_Check tc;
//		p->PeekHeader(tc);
//		debug2 (" generic mngt: %d\n", tc.type);
//		//cout << "Received GMH:\n";
//		//wimaxHdr.GMH.Print(cout);
//	  }
//    } else {
//      //hdr_mac802_16 *wimaxHdr_tmp = HDR_MAC802_16(p);
//      //gen_mac_header_t header_tmp = wimaxHdr_tmp->header;
//
//      cdma_req_header_t *req_tmp = &wimaxHdr.CRH.data;
//      if (req_tmp->type == 0x3) { //cmda-init
//        debug2 (" cdma_bw_req code :%d, top :%d, ssid :%d\n", req_tmp->code, req_tmp->top, req_tmp->br);
//      } else if (req_tmp->type == 0x2) { //cmda-bw-req
//        debug2 (" cdma_init_req code :%d, top :%d, ssid :%d\n", req_tmp->code, req_tmp->top, req_tmp->br);
//      } else { //bw-req
//        debug2 (" bwreq\n");
//      }
//    }
//
//
//  //create a timer to wait for the end of reception
//  //since the packets are received by burst, the beginning of the new packet 
//  //is the same time as the end of this packet..we process this packet 1ns 
//  //earlier to make room for the new packet.
//  int cid;
//  if (wimaxHdr.GMH.GMH.ht == 0) {
//	  gen_mac_header_t *header = &wimaxHdr.GMH.GMH;
//	  cid = header->cid;
//  }
//  else if (wimaxHdr.CRH.data.type == 0) {
//  	 cid = wimaxHdr.SMH.BRH.cid;
//  }
//  else cid = wimaxHdr.CRH.data.cid;
//  Connection *con = connectionManager_->get_connection (cid, IN_CONNECTION);
//
//  cdma_req_header_t *req_tmp = &wimaxHdr.CRH.data;
//
//  if (req_tmp->type == 0x3) {
//      debug10 (" BS, SENDUP, received by pass cdma_bw_req code :%d, top :%d, ssid :%d\n", req_tmp->code, req_tmp->top, req_tmp->br);
//      p->AddHeader(wimaxHdr);
//      addPacket(p);
//      return;
//  } else if (req_tmp->type == 0x2) {
//      debug10 (" BS, SENDUP, received by pass cdma_init_req code :%d, top :%d, ssid :%d\n", req_tmp->code, req_tmp->top, req_tmp->br);
//      p->AddHeader(wimaxHdr);
//      addPacket(p);
//      return;
//  }
//
//  if( (wimaxHdr.PIH.OFDMSymbol_offset == 0 && wimaxHdr.PIH.num_OFDMSymbol == 0) || wimaxHdr.GMH.GMH.cid == BROADCAST_CID)   // this kind of packets are treated diff(OFDM types) basically it a bw req packet.
//    {
//      if (con == NULL) {
//	//This packet is not for us
//	//debug2 ("At %f in Mac %d Connection null\n", NOW, id_index);
//	//update_watch (&loss_watch_, 1);
//
//		p = 0; 
//		debug2(" packet getting dropped header.ht :%d", wimaxHdr.GMH.GMH.ht);
//		//pktRx_=NULL;
//		return;
//      } 
//      
//      if(contPktRxing_)
//	 {
//// 	  // if there is a collision and the power of the currently recvd packet is less by factor 10 thn discard it. 
////            
// 	  if (head_pkt_ !=NULL) {
////             
//// 	    if(head_pkt_->p->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
//// 	      //Chakchai  => disable collission
//// 	      //	      Packet::free(p);
//// 	    }
////         else if(head_pkt_->p->txinfo_.RxPr / p->txinfo_.RxPr <= p->txinfo_.CPThresh)
//// 	    {
//// 			head_pkt_->timer.stop();
//// 			removePacket(head_pkt_);
//// 			p->AddHeader(wimaxHdr);
//// 			addPacket(p);
//// 			p->RemoveHeader(wimaxHdr);
//// 	    } else {
//// 		if(txtime(p) > head_pkt_->timer.expire())
//// 		  {
//// 		    head_pkt_->timer.stop();
//// 		    removePacket(head_pkt_);
//// 		    ch->error() = 1;
//// 		    //update_watch (&loss_watch_, 1);
//// 		    collision_ = true;
//// 		    p->AddHeader(wimaxHdr);
//// 		    addPacket(p);	
//// 		    p->RemoveHeader(wimaxHdr);
//// 
//// 		  }
//// 
//// 	      }
////             
////        
//// 	  }
//  
//          drop (p, "COL");
//          debug2(" packet dropped due to collision"); 
//          //update_watch (&loss_watch_, 1);
//          return;
//	}
//}
//      else 
//
//        {
//          //contPktRxing_ = true;
//          p->AddHeader(wimaxHdr);
//          addPacket(p);
//		}  
//    }
//  
//  else {
//
//    if (con == NULL) {
//      //This packet is not for us
//      debug2 ("At %f in Mac %d Connection null -- packet getting dropped at SS not for us \n", NOW, id_index);
//      //wimaxHdr.phy_info.num_subchannels;
//      //wimaxHdr.phy_info.subchannel_offset;
//      //wimaxHdr.phy_info.num_OFDMSymbol;
//      //wimaxHdr.phy_info.OFDMSymbol_offset; 
//      
//      		//update_watch (&loss_watch_, 1);   //-not commented in ns2
//      		
//      //intpower_ +=  p->txinfo_.RxPr;
//      //for (int i= 0; i++; i<1024)
//      //intpower_[i] += p->txinfo_.RxPr_OFDMA[i]; 
//    
//      // remeber to take care of collision, tht is if a packet destined to this is already being recvd and another packets destined to this comes thn thr will be collision    
//
//      p = 0;  //removed jut to chk
//      //pktRx_=NULL;
//      return;
//    }  // free and end if con == NULL 
//    
//    // collision check .  rpi
//
//    bool collision = false; 
//
//    //collision = IsCollision ( wimaxHdr, 0.0);  removed replace it later
//
//    if (collision == true) {
//      collision = false; 
//      //p->RemoveHeader(wimaxHdr);
//      //hdr_cmn *hdr = HDR_CMN(p);
//      wimaxHdr.error = 1;   // we do not discard this packet but just mark it and r=transmit coz, this packet and the other collided packet might not have symbols and carrriers exactly same, but they can collide, we shld also keep track of this packets transmission, because other packets overlapping with this can collide. 
//      
//      // set to the symbol and corresponding subcarriers over which this packet was transmitted to -1, so tht for future packets we can check for collision and also after the other collided packet is received we can chk for collision.    
//      //collision_ = true; 
//     
//     //addPowerinfo(wimaxHdr, -1.0,true);  // enable , diabled just to compile // will be in the base power 
//      
//      // set end
//      
//    } // else {
////       addPowerinfo(wimaxHdr, BASE_POWER,true ); // enable , disabled just to compile 
////     }
//  
////    addPowerinfo(wimaxHdr, 0.0,false);   // enable, disabled just to compile  // separated the variabled for interference and collision detection. 
//
//    // collision check rpi 
//	p->AddHeader(wimaxHdr);
//    addPacket(p);    // adding the collided packet also , which will be dropped in the receive function, this will help for collision detection of packets which might over lap with this.
//
//  } // ends else for bwreq  replace this } later when testing over 
//
//}

/**
 * Process the fully received packet
 * This will be called after Wimax2NetDevice::process_incoming_packet finishes processing fragmentation and/or packing
 * which in turn is called after assemble_incoming_packet is run by the scheduler, based on when the BS-Scheduler assumes packets will be fully recieved 
 * The old version of this function is included in a comment below for reference.  It handled assembly/fragmentation, power calculations, etc.
 */

void Wimax2BSNetDevice::receive (Ptr<Packet> p) {

	NS_LOG_FUNCTION_NOARGS();
	debug10("entering (BS) %s receive\n", addr_s());

	assert(p);

	All_Headers ah;
	p->RemoveHeader(ah);

	if (ah.GMH.GMH.ht == 0) { //GMH-headed packet, so not pure BW-REQ or contention
		Connection *in_c = getCManager()->get_connection (ah.GMH.GMH.cid, IN_CONNECTION);
		if (isMACPacket(p)) {
			debug2("BS %s received a GMH-Headed packet bound for the MAC layer of size %d at time %f.\n", addr_s(), p->GetSize(), NOW);
			p->AddHeader(ah);
			process_mac_packet(in_c, p);
			return;
		} else { //send it up to the higher layers
			//we need to get rid of any excess packing zeros added to fill transmission allocations
			//as the higher layers won't always ignore them correctly
			ah.GMH.GMH.len -= ah.GMH.GetSerializedSize(); //since the GMH was already removed
			if ( p->GetSize() > ah.GMH.GMH.len) p->RemoveAtEnd(p->GetSize() - ah.GMH.GMH.len);
			m_traceRx(p, declassify(ah.GMH.GMH.cid));
			UL_Protocol_Header uph;
			p->RemoveHeader(uph);

			debug2("BS %s recieved a packet bound for the higher layers of size %d at time %f.\n", addr_s(), p->GetSize(), NOW);
			Simulator::ScheduleWithContext(this->GetNode()->GetId(), Seconds(0), &Wimax2NetDevice::Receive, this->GetObject<Wimax2NetDevice>(), p, uph.protocol, addr(), declassify(ah.GMH.GMH.cid));
//			Receive(p, uph.protocol, addr(), declassify(ah.GMH.GMH.cid));
			return;
		}
	} else { //this is a Signaling MAC Header of some sort
		switch(ah.SMH.BRH.type) {
		case 0: //BW-REQ Header
			debug2("BS %s received a BW-REQUEST on CID %d for %d bytes at time %f\n", addr_s(), ah.SMH.BRH.cid, ah.SMH.BRH.br, NOW);
			getCManager()->get_connection(ah.SMH.BRH.cid, IN_CONNECTION)->setBw(ah.SMH.BRH.br);
			break;
			return;
		case 1: { //CDMA-INIT HEADER - note braces are to allow for variable declarations without scope issues
			Connection *in_c = getCManager()->get_connection (ah.CRH.data.cid, IN_CONNECTION);
			debug2("BS %s received a CDMA-INIT-REQUEST with top %d and code %d at time %f\n", addr_s(), ah.CRH.data.top, ah.CRH.data.code, NOW);
			//this is sent as an anonymous initial raning request, so we generate a responce with parameters to send a real one
			Ptr<Packet> irrsp = getPacket(); //"initial ranging response packet"

			All_Headers ah_ir;
			irrsp->RemoveHeader(ah_ir);

			RANGING_RESPONSE_FRAME rrf;
                        ah_ir.GMH.GMH.len += rrf.GetSerializedSize();
                        mac802_16_rng_rsp_frame *rrfd = &rrf.data;
                        
                        rrfd->type = MAC_RNG_RSP;
                        rrfd->uc_id = 0;
                        rrfd->rng_status = 3; //success - send rng_req
                        rrfd->basic_cid = INITIAL_RANGING_CID;
                        rrfd->primary_cid = INITIAL_RANGING_CID;
                        rrfd->top = ah.CRH.data.top;
                        rrfd->code = ah.CRH.data.code;
                        rrfd->ss_mac_address = Mac48Address::GetBroadcast(); //since we don't yet know this, but we will after we get the RNG_REQ
                        ah_ir.dest = rrfd->ss_mac_address;
                        irrsp->AddHeader(rrf);
                        irrsp->AddHeader(ah_ir); 
		
			//enqueue the response on the outgoing connection
			getCManager()->get_connection(BROADCAST_CID, OUT_CONNECTION)->enqueue(irrsp);
      			//in_c->enqueue (irrsp);

//			in_c = getCManager()->get_connection(CDMA_ALLOCATION_IE, IN_CONNECTION);
			
			in_c->setCDMA(2);
			in_c->setCDMA_SSID_FLAG(ah.CRH.data.br, 2);
			in_c->setCDMA_SSID_SSID(ah.CRH.data.br, ah.CRH.data.br);
			in_c->setCDMA_SSID_CODE(ah.CRH.data.br, ah.CRH.data.code);
			in_c->setCDMA_SSID_TOP(ah.CRH.data.br, ah.CRH.data.top);
			in_c->setCDMA_SSID_CID(ah.CRH.data.br, in_c->get_cid());
			in_c->setCDMA_code(ah.CRH.data.code);
			in_c->setCDMA_top(ah.CRH.data.top);
			return;
			}	
		case 3: //CDMA-BW-REQ HEADER
			Connection *in_c = getCManager()->get_connection (ah.CRH.data.cid, IN_CONNECTION);
			debug2("BS %s received a CDMA-BW-REQUEST (polling request) with top %d and code %d at time %f\n", addr_s(), ah.CRH.data.top, ah.CRH.data.code, NOW);
			//what this means is that we got a contention-based request to allocate space for an actual bw-req 
			in_c->setCDMA(1);
			in_c->setCDMA_code(ah.CRH.data.code);
			in_c->setCDMA_top(ah.CRH.data.top);
			return;
		}
	}

}



//old version
//void Wimax2BSNetDevice::receive (Ptr<Packet> pktRx_)
//{
//  NS_LOG_FUNCTION_NOARGS ();
//	debug10("Entering (BS) %s receive\n", addr_s());
//	assert (PeekPointer(pktRx_));
//	
//	All_Headers ah;
//	pktRx_->RemoveHeader(ah);
//
//	//cdma_flag => 1 (bw-req), 2 (initial-ranging-req)
//	int cdma_flag = 0;
//
//
//  debug10 ("At %f (BS) in Mac %s, RECEIVE, packet received \n", NOW, addr_s());
//  
//  //debug10 (" phyinfo header - symbol offset :%d, numsymbol :%d\n", wimaxHdr.PIH.OFDMSymbol_offset,wimaxHdr.PIH.num_OFDMSymbol);
//
//    if (ah.GMH.GMH.ht == 0) { //if we have a normal frame with a GMH on it...
//    		ah.GMH = ah.GMH;
//    		if (pktRx_->GetSize() != 0) { //if this packet still contains data...
//    			Type_Check tc;
//    			pktRx_->PeekHeader(tc);
//    			debug10 ( "generic mngt :%d\n", tc.type);
//    		} else {
//    			debug10("bad packet received, dropping\n");
//    			drop(pktRx_, "BAD");
//    		}
//    		 		
//    		//setup arq_ie list in ah
//    		if (ah.GMH.GMH.type_arqfb) {
//    			ARQ_FB_IE_HEADER afih;
//    			while (true) {
//    				pktRx_->RemoveHeader(afih);
//    				ah.arq_ie[ah.num_of_acks++] = afih.arqfbie;
//    				if (afih.last_) break;
//    			}
//    		
//    		}
//    		
//    } else { //we do not have a generic mac header => singaling header (bandwdith request or ranging)
//      pktRx_->AddHeader(ah.GMH);
//      CDMA_REQUEST_HEADER req_tmp;
//      pktRx_->RemoveHeader(req_tmp);
//      ah.CRH = req_tmp;
//
//      //int cid = header_tmp.cid;
//      int cid = req_tmp.data.cid;
////      debug2 ("CON1 :%d\n", cid);
//	  Connection *con_tmp;
//	  if (req_tmp.data.type == 0x2 || req_tmp.data.type == 0x3) {
//	  	//done in here so we don't get an error if it's not a cdma req 
//	  	//(and the cid is therefore different)
//      	con_tmp = connectionManager_->get_connection (cid, IN_CONNECTION);
//      }
//
//      if (req_tmp.data.type == 0x2 && con_tmp!=NULL) { //if this is a CDMA_BW_REQ header
//        debug2 (" cdma_bw_req code :%d, top :%d, ssid :%d\n", req_tmp.data.code, req_tmp.data.top, req_tmp.data.br);
//        cdma_flag = 1;
//        debug10 ("At %f Connection %d, old :%d, set to 1\n", NOW, con_tmp->get_cid(), con_tmp->getCDMA());
//
//        con_tmp->setCDMA(1);
//        con_tmp->setCDMA_code(req_tmp.data.code);
//        con_tmp->setCDMA_top(req_tmp.data.top);
//        pktRx_ = 0;
//       // Packet::free(pktRx_);
//        return;
//      } else if (req_tmp.data.type == 0x3 && con_tmp!=NULL) { //if this is a CDMA-INIT header
//			debug2 (" cdma_init_req code :%d, top :%d, ssid :%d\n", req_tmp.data.code, req_tmp.data.top, req_tmp.data.br);
//			cdma_flag = 2;
//			//recieved CDMA-INIT header, respond with anonymous RNG_RSP with status set to success so we get a RNG_REQ
//			Ptr<Packet> new_p = getPacket();
//			All_Headers new_ah;
//			new_p->RemoveHeader(new_ah);
//			RANGING_RESPONSE_FRAME rrf;
//			
//			new_ah.GMH.GMH.len += rrf.GetSerializedSize();
//			mac802_16_rng_rsp_frame *rrfd = &rrf.data;
//			
//			rrfd->type = MAC_RNG_RSP;
//			rrfd->uc_id = 0;
//			rrfd->rng_status = 3; //success - send rng_req
//			rrfd->basic_cid = INITIAL_RANGING_CID;
//			rrfd->primary_cid = INITIAL_RANGING_CID;
//			rrfd->top = req_tmp.data.top;
//			rrfd->code = req_tmp.data.code;
//			rrfd->ss_mac_address = Mac48Address::GetBroadcast(); //since we don't yet know this, but we will after we get the RNG_REQ
//			new_ah.dest = rrfd->ss_mac_address;
//			new_p->AddHeader(rrf);
//			new_p->AddHeader(new_ah);
//			sendDown(new_p);
//	
//			//we send back this RNG_RSP to tell the SS that it will be able to send in the frame after 
//			//this RNG_RSP is recieved by the SS, so we need to allocate a BW oppurtunity for that frame
//	
//			//debug10 ("At %f Connection %d, old :%d, set to 2\n", NOW, con_tmp->get_cid(), con_tmp->getCDMA());
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA, con_tmp, 2);
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_SSID_FLAG, con_tmp, req_tmp.data.br, 2);
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_SSID_SSID, con_tmp, req_tmp.data.br, req_tmp.data.br);
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_SSID_CODE, con_tmp, req_tmp.data.br, req_tmp.data.code);		
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_SSID_TOP, con_tmp, req_tmp.data.br, req_tmp.data.top);
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_SSID_CID, con_tmp, req_tmp.data.br, con_tmp->get_cid());
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_code, con_tmp, 0);
//			Simulator::Schedule(Seconds(0.005), &Connection::setCDMA_top, con_tmp, 0);	
//			
//	//         con_tmp->setCDMA(2);
//	//         con_tmp->setCDMA_SSID_FLAG(req_tmp.data.br, 2);
//	//         con_tmp->setCDMA_SSID_SSID(req_tmp.data.br, req_tmp.data.br);
//	//         con_tmp->setCDMA_SSID_CODE(req_tmp.data.br, req_tmp.data.code);
//	//         con_tmp->setCDMA_SSID_TOP(req_tmp.data.br, req_tmp.data.top);
//	//         con_tmp->setCDMA_SSID_CID(req_tmp.data.br, con_tmp->get_cid());
//	//         con_tmp->setCDMA_code(0);
//	//         con_tmp->setCDMA_top(0);
//	//         reg_SS_number++;
//			pktRx_ = 0;
//	//        Packet::free(pktRx_);
//			return;
//      } else { //if this is a non-CDMA BW Request header...
//        debug2 (" bwreq with CID %d\n", ah.SMH.BRH.cid);
//      }
//    }
//
//  // commenting here to chk without collision start 
//
//  //if( (ah.PIH.OFDMSymbol_offset == 0 && ah.PIH.num_OFDMSymbol == 0) || ah.GMH.GMH.cid == BROADCAST_CID)   // this kind of packets are treated diff(OFDM types) basically it a bw req packet.
//  if (ah.GMH.GMH.ht != 0 || ah.GMH.GMH.len == 6 || ah.GMH.GMH.cid == BROADCAST_CID)
//    {
//      if (ah.error)
//		{
//			  if (collision_) {
//			debug2 ("\t drop new pktRx..collision\n");
//			drop (pktRx_, "COL");
//			collision_ = false;
//		  } 
//			  else {
//				debug2("\t error in the packet, the Mac does not process\n");
//				pktRx_ = 0;
//		  }
//	
//		  //update drop stat
//		  //update_watch (&loss_watch_, 1);
//		  contPktRxing_ = false;
//		  pktRx_ = NULL;
//			  return;
//		}          
//      //else
//      contPktRxing_ = false; 
//    }
//  else {   //bwreq packet treatment else 
//  
//    //collision_ = false; // bwreq packet recvd , so no packet being transmitted. 
//    //drop the packet if corrupted
//    if (ah.error) {
//      //  if (collision_) {
//      //   debug2 ("\t drop new pktRx..collision\n");
//     // addPowerinfo(ah.GMH, 0.0,true);   
//      //   drop (pktRx_, "COL");
//      //   collision_ = false;
//      // } else {
//      debug2("error in the packet, the Mac does not process");
//      //  addPowerinfo(ah.GMH, 0.0,false);  // when we drop or discard a packet we have to reinitialise the intpower array to 0.0 , which shows tht there is no packet being recvd over the symbols and subcarriers.   // interference initialisation not reqd cause it is initialised whenever a packet is recvd
//      pktRx_ = 0;
//      //update drop stat
//      
//      
//      //update_watch (&loss_watch_, 1);
//      
//      pktRx_ = NULL;
//      return;
//    }
//
//    // chk here for collision of the packet which was received before the packet that caused actual collision, so we have to chk whether collision had occur on this packet or not. , and the collsion was caused  
//
//    bool collision = false; 
//    if (collision); //to stop compiler from complaining about unused variable
//
//    //collision = IsCollision(ah.GMH, BASE_POWER);  // removed -replace it later
//// 
////     if (collision == true)
////       {
//// 	collision = false; 
//// 	addPowerinfo(ah.GMH, 0.0,true);  
//// 	drop (pktRx_, "COL"); 
//// 	//update drop stat
//// 	update_watch (&loss_watch_, 1);
//// 	pktRx_ = NULL;
//// 	debug2("Drop this packet because of collision.\n");
//// 	return;
////       }
//
//    //addPowerinfo(ah.GMH, 0.0,true);  // ENABLE LATER
//
//    //commenting here to chk without collision  ends
//
//    // chk collision ends
//
//    //   SINR calcualtions
//
//    //removed for testing without Rxinpwr in packet header
//
//    uint32_t total_subcarriers=0; 
//    int num_data_subcarrier = getPhy()->getNumDataSubcarrier (UL_);         
//    uint32_t num_data_subcarrier = getPhy()->getNumSubcarrier (UL_);         
//    uint32_t num_subchannel      = getPhy()->getNumsubchannels (UL_); 
//    uint32_t num_symbol_per_slot = getPhy()->getSlotLength (UL_); 
//
//    // calculate the total subcarriers needed
//    if(ah.PIH.num_OFDMSymbol != 0)
//      { 
//             
//	//total_subcarriers = ah.GMH.phy_info.num_subchannels * ah.GMH.phy_info.num_OFDMSymbol * num_data_subcarrier ;
//     
//	if(ah.PIH.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition , chk whether broacastcid reqd ir not. 
//	  {
//	    if(ah.PIH.num_OFDMSymbol > num_symbol_per_slot) 
//	      {
//		// for the first 3 symbols 
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
//		  //   for(uint32_t j = (ah.GMH.phy_info.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers += (num_subchannel*num_data_subcarrier) - ((ah.PIH.subchannel_offset)*num_data_subcarrier ) ;
//		  }
//		// except the last 3 and first 3 whatever is thr
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset+num_symbol_per_slot ; i< (ah.PIH.OFDMSymbol_offset) + (ah.PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
//		  //for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers +=  num_subchannel*num_data_subcarrier;
//		  } 
//
//		// last 3 
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol-num_symbol_per_slot ; i<  ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol ; i++)
//		  //for(uint32_t j = 0 ; j<(ah.GMH.phy_info.num_subchannels%num_subchannel)*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers += ((ah.PIH.num_subchannels - (num_subchannel - (ah.PIH.subchannel_offset)) )%num_subchannel)*num_data_subcarrier;	
//		  }
//	      }
//	    else 
//	      {
//
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
//		  //for(uint32_t j = (ah.GMH.phy_info.subchannel_offset)*num_data_subcarrier ; j<ah.GMH.phy_info.num_subchannels*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers += ah.PIH.num_subchannels*num_data_subcarrier ;//-  (ah.GMH.phy_info.subchannel_offset)*num_data_subcarrier ;	
//		  }	   
//	      }
//	  }
//	else
//	  {
//	    // for the first num_symbol_per_slot symbols 
//	    if(ah.PIH.num_OFDMSymbol > 1) 
//	      {
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + 1) ; i++)
//		  //   for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers += num_subchannel*num_data_subcarrier - (ah.PIH.subchannel_offset)*num_data_subcarrier ;
//		  }
//		// except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//          
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset+1 ; i< (ah.PIH.OFDMSymbol_offset) + (ah.PIH.num_OFDMSymbol-1) ; i++)
//		  //for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers +=  num_subchannel*num_data_subcarrier ;
//		  }  
//
//		// last num_symbol_per_slot 
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol-1 ; i< ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol ; i++)
//		  // for(uint32_t j = 0 ; j<(ah.PIH.num_subchannels%num_subchannel)*num_data_subcarrier ; j++ )
//		  {
//                    total_subcarriers += ((ah.PIH.num_subchannels - (num_subchannel - (ah.PIH.subchannel_offset)) )%num_subchannel)*num_data_subcarrier ; 
//		  }
//
//	      }
//	    else
//	      {
//          
//		for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + 1) ; i++)
//		  // for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; j<ah.PIH.num_subchannels*num_data_subcarrier ; j++ )
//		  {
//		    total_subcarriers +=    ah.PIH.num_subchannels*num_data_subcarrier;// - (ah.PIH.subchannel_offset)*num_data_subcarrier ; 
//		  }	
//
//	      }
//             
//	  }
//      } 
//    else
//      total_subcarriers = ah.PIH.num_subchannels * num_data_subcarrier ;  
//
//    // total subcarrier calculation ends 
//
//    debug2(" total_subcarriers = %d \n", total_subcarriers); 
//
//    double *signalpower = (double *) new double [total_subcarriers] ;  
//    double *interferencepower = (double *) new double [total_subcarriers]; 
//    double *SINR = (double *) new double [total_subcarriers]; 
//    for(uint32_t i = 0; i< total_subcarriers ; i++)
//      {
//	SINR[i] = 0.0;
//	signalpower[i] = 0.0;
//	interferencepower[i] = 0.0;
//      }
//
//    int n=0,m=0; 
//
//    if(ah.PIH.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition 
//      {
//	if(ah.PIH.num_OFDMSymbol > num_symbol_per_slot) 
//	  {
//	    // for the first num_symbol_per_slot symbols 
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
//	      for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}
//	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset+num_symbol_per_slot ; i< (ah.PIH.OFDMSymbol_offset) + (ah.PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
//	      for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		} 
//
//	    // last num_symbol_per_slot 
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol-num_symbol_per_slot ; i<  ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol ; i++)
//	      for(uint32_t j = 0 ; j<((ah.PIH.num_subchannels - (num_subchannel - (ah.PIH.subchannel_offset)) )%num_subchannel)*num_data_subcarrier; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}
//	  }
//	else 
//	  {
//
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset ; 
//		i<  (ah.PIH.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
//
//	      for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; 
//		  j<(((ah.PIH.subchannel_offset)*num_data_subcarrier) + ah.PIH.num_subchannels*num_data_subcarrier); 
//		  j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}	    
//	  }
//      }
//
//    else
//
//      {
//        // for the first num_symbol_per_slot symbols 
//	if(ah.PIH.num_OFDMSymbol > 1) 
//	  {
//            for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + 1) ; i++)
//	      for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}
//	    
//            // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//            for(uint32_t i = (ah.PIH.OFDMSymbol_offset+1) ; i<  (ah.PIH.OFDMSymbol_offset) + (ah.PIH.num_OFDMSymbol-1) ; i++)
//	      for(uint32_t j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}
//
//	    // last num_symbol_per_slot 
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol-1 ; i< ah.PIH.OFDMSymbol_offset +ah.PIH.num_OFDMSymbol ; i++)
//	      for(uint32_t j = 0 ; j<((ah.PIH.num_subchannels - (num_subchannel - (ah.PIH.subchannel_offset)) )%num_subchannel)*num_data_subcarrier; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}
//
//          }
//	else
//	  {
//	    for(uint32_t i = ah.PIH.OFDMSymbol_offset ; i<  (ah.PIH.OFDMSymbol_offset + 1) ; i++)
//	      for(uint32_t j = (ah.PIH.subchannel_offset)*num_data_subcarrier ; j<(((ah.PIH.subchannel_offset)*num_data_subcarrier) + ah.PIH.num_subchannels*num_data_subcarrier) ; j++ )
//		{
//		  interferencepower[n++]= intpower_[i][j] + BASE_POWER /*+ pktRx_->txinfo_.RxIntPr[i][j]*/; 
//		  //signalpower[m++]= pktRx_->txinfo_.RxPr_OFDMA[j];
//		}	
//	  }
//      }
//
//    int counter = 0;
//    for(uint32_t k = 0; k < total_subcarriers; k++)
//      {
//	if( signalpower[k] > DOUBLE_INT_ZERO) // not equals to zero
//	  {
//	    SINR[counter] =  signalpower[k]/interferencepower[counter];
//	    counter++;
//	  }
//      }
//    debug2("in MAC BS, [%d] subcarriers signal power are calculated.\n ", counter);
//   
//
//    double SIR = 0.0;
//    //debug2("In MAC BS: the total number of subcarrier/3 is [%d]\n", total_subcarriers/3);
// 
//
//    //process packet
//    gen_mac_header_t *header = &(ah.GMH.GMH);
//    if (header); //to stop compiler from complaining about unused variable
//    Ptr<Wimax2Phy> phy = getPhy();
//
//    //debug2(" checking for intereference power \n" );
// 
//    // BLER calculation
//
//    GlobalParams_Wimax* global ; 
//
//    global =  GlobalParams_Wimax::Instance();
//
//    uint32_t num_of_slots,num_of_complete_block,max_block_size,last_block_size , index, num_subcarrier_block ;
//
//    double BLER= 0.0, beta=0.0, eesm_sum = 0.0, rand1 =0.0; 
//
//    num_of_slots = (int) ceil((double)ah.GMH.GMH.len / (double)phy->getSlotCapacity(ah.PIH.modulation_, UL_));
//
//    max_block_size=phy->getMaxBlockSize(ah.PIH.modulation_);   // get max  block size in number of slots
//
//    last_block_size = num_of_slots % max_block_size; 
//
//    num_of_complete_block = (int) floor(num_of_slots / max_block_size);
//
//    int ITU_PDP = getITU_PDP (); 
//
//    bool pkt_error = false; 
//
//    debug2(" num of complete blocks = %d , ITU_PDP = %d, num of slots = %d, max block size = %d, last block size = %d \n" , num_of_complete_block, ITU_PDP, num_of_slots, max_block_size, last_block_size ); 
//
//
//    if (num_of_complete_block == 0) { 
//      //num_of_complete_block =1;
//      index =  phy->getMCSIndex( ah.PIH.modulation_ ,  last_block_size);
//      
//      beta = global->Beta[ITU_PDP][index]; 
//      debug2(" beta = %.2f = \n index = %d\n" , beta,index ); 
//
//      num_subcarrier_block = num_symbol_per_slot * getPhy()->getNumSubcarrier (UL_) * last_block_size; 
//
//      /* if(num_subcarrier_block > total_subcarriers) num_subcarrier_block = total_subcarriers; 
//	 for( int i = 0; i< num_subcarrier_block ; i++) 
//	 {    
//	 debug2( " SINR (%d) = %g , eesm_sum = %g %g\n", i, SINR[i], eesm_sum, exp(-(SINR[i]/beta)));    
//	 eesm_sum = eesm_sum + exp( -(SINR[i]/beta));
//	 }*/
//
//      //printf("\n n [%d] --- counter [%d]\n",n, counter);
//      for(int i=0; i<counter; i++)
//	{
//	  eesm_sum = eesm_sum + exp( -(SINR[i]/beta));		
//	}
//
//	
//      if (num_subcarrier_block == 0) {
//	fprintf(stderr, "ERROR: num_subcarrier_block = 0\n");
//	exit (1);
//      }
//      if (eesm_sum >= BASE_POWER) {
//	SIR =  (-beta) * log(eesm_sum/counter);
//	SIR=10*log10(SIR);
//	BLER = global->TableLookup(index, SIR);
//      } else {
//		//eesm = 0 when SINR is large (MS close to BS) and exp returns 0
//		BLER = 0;
//      }
//      debug2(" BLER-BS one block = %.6f = \n" , BLER ); 
//
//      int rand_num = ((rand() % 100) +1 ) ; 
//      // debug2(" random num = %d = \n" , rand_num ); 
//      rand1 = rand_num/100.00; 
//
//     //if (!phymib_.disableInterference && BLER > rand1) 
//      if (!phymib_.disableInterference && BLER > 0.96) 
//	pkt_error = true;
//      
//      // getBLER(ah.PIH.modulation_, SINR, last_block_size);
//    }
//    else {
//
//      // First the complete blocks are checked fro error 
//
//      index =  phy->getMCSIndex( ah.PIH.modulation_ ,  max_block_size);
//      
//      beta = global->Beta[ITU_PDP][index];  
//
//      debug2(" beta = %.2f index = %d \n" , beta,index ); 
// 
//      num_subcarrier_block = num_symbol_per_slot * getPhy()->getNumSubcarrier (UL_) * max_block_size; 
//	
//
//      if(num_subcarrier_block > total_subcarriers) num_subcarrier_block = total_subcarriers; 
//
//      /* for( int i = 0; i< num_subcarrier_block ; i++)     
//	 eesm_sum = eesm_sum + exp( -(SINR[i]/beta));*/
//
//      for(int i=0; i<counter; i++)
//	{
//	  eesm_sum = eesm_sum + exp( -(SINR[i]/beta));		
//	}
//
//
//      if (num_subcarrier_block == 0) {
//	fprintf(stderr, "ERROR: num_subcarrier_block = 0\n");
//	exit (1);
//      }
//      if (eesm_sum >=BASE_POWER) {
//	//SIR =  (-beta) * log( eesm_sum/(num_subcarrier_block) );
//	SIR =  (-beta) * log(eesm_sum/counter);
//	SIR=10*log10(SIR);
//	debug2(" SIR-BS = %.6f = \n" , SIR );
//	BLER = global->TableLookup(index, SIR);
//      } else {
//	//eesm = 0 when SINR is large (MS close to BS) and exp returns 0
//	BLER = 0;
//      } 
//      debug2(" BLER-BS complete blocks = %.6f = \n" , BLER ); 
// 
//      int rand_num = ((rand() % 100) +1 ) ; 
//
//      //debug2(" random num = %d = \n" , rand_num ); 
// 
//      rand1 = rand_num/100.00; 
//
//       //if (!phymib_.disableInterference && BLER > rand1) 
//      if (!phymib_.disableInterference && BLER > 0.96) 
//	pkt_error = true; 
//
//      // Chk if thr is any last block, compute whether it is in error or not. 
//      if(last_block_size > 0)
//	{
//	  eesm_sum=0;
//	  index =  phy->getMCSIndex( ah.PIH.modulation_ ,  last_block_size);
//      
//	  beta = global->Beta[ITU_PDP][index];  
// 
//	  num_subcarrier_block = num_symbol_per_slot * getPhy()->getNumSubcarrier (UL_) * last_block_size; 
//
//	  if(num_subcarrier_block > total_subcarriers) num_subcarrier_block = total_subcarriers; 
//
//
//	  /*	  for( int i = 0; i< num_subcarrier_block ; i++)     
//	    eesm_sum = eesm_sum + exp( -(SINR[i]/beta));*/
//
//	  for(int i=0; i<counter; i++)
//	    {
//	      eesm_sum = eesm_sum + exp( -(SINR[i]/beta));		
//	    }
//
//
//	  if (num_subcarrier_block == 0) {
//	    fprintf(stderr, "ERROR: num_subcarrier_block = 0\n");
//	    exit (1);
//	  }
//	  if (eesm_sum >=BASE_POWER) {
//	    //SIR =  (-beta) * log( eesm_sum/(num_subcarrier_block) );
//	    SIR =  (-beta) * log(eesm_sum/counter);
//	    SIR=10*log10(SIR);
//	    debug2(" SIR-BS = %.6f = \n" , SIR );
//	    BLER = global->TableLookup(index, SIR);
//	  } else {
//	    //eesm = 0 when SINR is large (MS close to BS) and exp returns 0
//	    BLER = 0;
//	  } 
//
//#if 0
//	  int rand_num = ((rand() % 100) +1 ) ; 
//
//	  //debug2(" random num = %d = \n" , rand_num ); 
// 
//	  rand1 = rand_num/100.00; 
//
//	  if (!phymib_.disableInterference && BLER > rand1) 
//	    pkt_error = true; 
//#endif
//	 if (!phymib_.disableInterference && BLER > 0.96) 
//	    pkt_error = true; 
//	}
//
// 
//    }
//
//    //debug2( " BLER = %.25f" , BLER); 
//
//    // deleting dynamic allocations  //removed for testing without Rxinpwr in packet header  
//
//    delete [] SINR;
//    delete [] signalpower;
//    delete [] interferencepower; 
//
//
//    if( pkt_error == true)
//      {
//		//addPowerinfo(ah.GMH, 0.0,true);   	   
//		debug2("error in the packet, drop this packet. the Mac does not process\n");
//		pktRx_ = 0;
//	//	Packet::free(pktRx_);
//		pkt_error = false;  
//		//update drop stat
//		//update_watch (&loss_watch_, 1);
//		pktRx_ = NULL;
//		return;    
//	  }
//
//
// 
//    //removed for testing without Rxinpwr in packet header  
//
//    //BLER calcualtion ends
//
//
//
//  } //bwreq packet treatment else ends here.  replace { when testing ends 
//
//
//  //addPowerinfo(ah.GMH, 0.0,true); 
//  // normal packet processing starts. 
//  Connection *con;
//  if (ah.GMH.GMH.ht == 1) {
//  	if (ah.SMH.BRH.type ==0) con = connectionManager_->get_connection (ah.SMH.BRH.cid, IN_CONNECTION);
//  	else con = connectionManager_->get_connection (ah.CRH.data.cid, IN_CONNECTION);
//  }
//  
//  if (ah.GMH.GMH.ht == 0) {
//
//  debug2 ("normal packet processing\n");
//
//  gen_mac_header_t *header = &ah.GMH.GMH;
//  int cid = header->cid;
//  con = connectionManager_->get_connection (cid, IN_CONNECTION);
//  if(con==NULL) 
//  	return;
//
//  //update rx time of last packet received
//  PeerNode *peer;
//  if (type_ == STA_MN)
//    peer = getPeerNode_head(); //MN only has one peer
//  else
//    peer = con->getPeerNode(); //BS can have multiple peers
//
//  if (peer) {
//    peer->setRxTime (NOW);
//    
//    //collect receive signal strength stats
////     if (pktRx_->txinfo_.RxPr*1e3 == 0) {
////       fprintf(stderr, "ERROR: pktRx_->txinfo_.RxPr*1e3 = 0\n");
////     }
//    //peer->getStatWatch()->update(10*log10(pktRx_->txinfo_.RxPr*1e3));
//    //debug2 ("At %f in Mac %d weighted RXThresh: %e rxp average %e\n", NOW, id_index, macmib_->lgd_factor_*macmib_->RXThreshold_, pow(10,peer->getStatWatch()->average()/10)/1e3);
//    //double avg_w = pow(10,(peer->getStatWatch()->average()/10))/1e3;
//
////handover stuff below - add back in later
//    
////     if ( avg_w < (macmib_->lgd_factor_*macmib_->RXThreshold_)) {
//// #ifdef USE_802_21
////       double probability = ((macmib_->lgd_factor_*macmib_->RXThreshold_)-avg_w)/((macmib_->lgd_factor_*macmib_->RXThreshold_)-macmib_->RXThreshold_);
////       Mac::send_link_going_down (peer->getAddr(), addr(), -1, (int)(100*probability), LGD_RC_LINK_PARAM_DEGRADING, eventId_++);
//// #endif
////       if (peer->getPrimary(IN_CONNECTION)!=NULL) { //check if node registered
//// 	peer->setGoingDown (true);      
////       }
////     }
////     else {
////       if (peer->isGoingDown()) {
//// #ifdef USE_802_21
//// 	Mac::send_link_rollback (addr(), getPeerNode_head()->getAddr(), eventId_-1);
//// #endif
//// 	peer->setGoingDown (false);
////       }
////     }
//  }
// 
//  //this flag is used to update fragmentaion for piggybacking scenario
//  int flag_pig = 0;
//
//  //Process GM subheader (piggybacking)
//  int bw_update_dueto_gm = 0;
//  int more_bw = 0;
//  int old_bw = 0;
//  if (header->type_fbgm) {
//  	pktRx_->RemoveHeader(ah.GMSH);
//    grant_subheader_s *tmp_gss = &ah.GMSH.data;
//    //grant_map_ugs_s *tmp_gmus = reinterpret_cast<grant_map_ugs_s*>(tmp_gss);
//    //more_bw = tmp_gss->piggyback_req;
//    int temp_val = tmp_gss->piggyback_req;
//    more_bw = temp_val;
//    //more_bw = (reinterpret_cast<grant_subheader_s*>(tmp_gmus))->piggyback_req;
//    old_bw = con->getBw();
//    flag_pig = 2;
//    debug10 ("BS: Piggybacking, At %f Connection %d, flag :%d, old_bw :%d, add_bw :%d\n", NOW, con->get_cid(), header->type_fbgm, con->getBw(), more_bw);
//    if (more_bw > old_bw) {
//      bw_update_dueto_gm = more_bw;
//      con->setBw(bw_update_dueto_gm);
//    }
//
//    debug10 ("    Final bw_update con->getBw :%d\n",con->getBw());
//
//  } else {
//    bw_update_dueto_gm = 0;
//  }
//  //End piggybacking process
//
//  //Begin RPI 
//  // New function for reassembly will be implemented for ARQ, Fragmentation and Packing
//  if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1 
//      && con->isFragEnable () == true && con->isPackingEnable () == true 
//      /*&& HDR_CMN(pktRx_)->ptype()!=PT_MAC*/){
//      pktRx_->AddHeader(ah);
//    process_mac_pdu_witharqfragpack(con, pktRx_);
//    //update_throughput (&rx_data_watch_, 8*ch->GMH.GMH.len); 
//    //update_throughput (&rx_traffic_watch_, 8*ch->GMH.GMH.len);
//    return;		
//  }	  
//  //End RPI 
//  
//  //process reassembly
//  debug2 ("Reading fragmentation flags: %d (%d) for CID %d\n", header->type_frag, header->type_frag & 0x1, cid); 
//
//  if (header->type_frag) {
//  	pktRx_->RemoveHeader(ah.FSH);
//    bool drop_pkt = true;
//    bool frag_error = false;
//    debug2 ("\tfrag type = %d (type noflag :%d, first :%d, cont :%d, last :%d)\n", ah.FSH.data.fc & 0x3, FRAG_NOFRAG, FRAG_FIRST,FRAG_CONT, FRAG_LAST);
//    switch (ah.FSH.data.fc & 0x3) {
//    case FRAG_NOFRAG: 
//      if (con->getFragmentationStatus()!=FRAG_NOFRAG)
//	con->updateFragmentation (FRAG_NOFRAG, 0, 0); //reset
//      drop_pkt = false;
//      break; 
//    case FRAG_FIRST: 
//      //when it is the first fragment, it does not matter if we previously
//      //received other fragments, since we reset the information
//      assert (ah.FSH.data.fsn == 0);
//      debug2 ("\tReceived first fragment fsn :%d cid :%d", ah.FSH.data.fsn&0x7, cid);
//      if (con->getFragmentationStatus()!=FRAG_NOFRAG) {
//	debug2 ("...Error: expecting CONT or LAST\n");
//      } else {
//	debug2 ("...Ok\n");
//      }
//
//      if (header->type_fbgm) {
//        con->updateFragmentation (FRAG_FIRST, 0, ah.GMH.GMH.len-(HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE+flag_pig));
//      } else {
//        con->updateFragmentation (FRAG_FIRST, 0, ah.GMH.GMH.len-(HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE));
//      }
//
//      break; 
//    case FRAG_CONT: 
//      debug2 ("\tReceived cont fragment fsn :%d cid :%d", ah.FSH.data.fsn&0x7, cid);
//      if ( (con->getFragmentationStatus()!=FRAG_FIRST
//	    && con->getFragmentationStatus()!=FRAG_CONT)
//	   || ((ah.FSH.data.fsn&0x7) != (con->getFragmentNumber ()+1)%8) ) {
//	frag_error = true;
//	debug2 ("...Error: status :%d, fsn :%d, expected fsn :%d\n", 
//		con->getFragmentationStatus(), ah.FSH.data.fsn&0x7, (con->getFragmentNumber ()+1)%8);
//	con->updateFragmentation (FRAG_NOFRAG, 0, 0); //reset
//      } else {
//	debug2 ("...Ok\n");
//        if (header->type_fbgm) {
//	  con->updateFragmentation (FRAG_CONT, ah.FSH.data.fsn&0x7, con->getFragmentBytes()+ah.GMH.GMH.len-(HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE+flag_pig));	
//	} else {
//	  con->updateFragmentation (FRAG_CONT, ah.FSH.data.fsn&0x7, con->getFragmentBytes()+ah.GMH.GMH.len-(HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE));	
//	}
//      }
//      break; 
//    case FRAG_LAST: 
//      debug2 ("\tReceived last fragment fsn :%d cid :%d", ah.FSH.data.fsn&0x7, cid);
//      if ( (con->getFragmentationStatus()==FRAG_FIRST || con->getFragmentationStatus()==FRAG_CONT)
//	   && ((ah.FSH.data.fsn&0x7) == (con->getFragmentNumber ()+1)%8) ) {
//	ah.GMH.GMH.len += con->getFragmentBytes()-HDR_MAC802_16_FRAGSUB_SIZE;
//	drop_pkt = false;
//	debug2 ("...Ok\n");
//      } else {
//	debug2 ("...Error: status :%d (nofrag :%d, first :%d, cont :%d, last :%d), fsn :%d, expected fsn :%d\n", 
//		con->getFragmentationStatus(), FRAG_NOFRAG, FRAG_FIRST, FRAG_CONT, FRAG_LAST ,ah.FSH.data.fsn&0x7, (con->getFragmentNumber ()+1)%8);
//	frag_error = true;
//      }     
//      con->updateFragmentation (FRAG_NOFRAG, 0, 0); //reset
//      break; 
//    default:
//      fprintf (stderr,"Error, unknown fragmentation type\n");
//      exit (-1);
//    }
//    //if we got an error, or it is a fragment that is not the last, free the packet
//    if (drop_pkt) {
//      if (frag_error) {
//	//update drop stat
//	//update_watch (&loss_watch_, 1);
//	debug10 ("Drop packets FRG error\n");
//	drop (pktRx_, "FRG"); //fragmentation error
//      } else {
//	//silently discard this fragment.
//	//Packet::free(pktRx_);
//      }
//      pktRx_=NULL;
//      return;
//    } 
//  }
//  
//  }
//
//  //Check if this is a bandwidth request
//  //hdr_mac802_16 *ah.GMH_tmp = HDR_MAC802_16(pktRx_);
//  //gen_mac_header_t header_tmp = wimaxHdr_tmp->header;
//
//  bw_req_header_t *req_tmp = &ah.SMH.BRH;
//  //req_tmp = (bw_req_header_t *)&header_tmp;
//
//  if (ah.GMH.GMH.ht == 0) {
//    if ( (req_tmp->type == 0x3) || (req_tmp->type == 0x2) ) {
//      debug10 ("Impossible; PANIC\n");
//
//#ifndef BWREQ_PATCH
//    } else {
//      if (con->getBw() > 0) {
//	debug10 ("setBW left-over At %f Connection %d, old=%d, rcv=%d, left=%d\n", NOW, con->get_cid(), con->getBw(), ch->GMH.GMH.len, con->getBw()-ch->GMH.GMH.len);
//	if (con->getBw()-ch->GMH.GMH.len >0) {
//	  con->setBw(con->getBw()-ch->GMH.GMH.len);
//	} else {
//	  con->setBw(0);
//	}
//      }
//#endif
//    }
//  }
//  //end checking bw-req packet
//
//  /* => Old
//  //check if this is a bandwidth request
//  if (header.ht == 0 && con->getBw() >0) {
//  debug2 ("At %f Connection %d, old=%d, rcv=%d, left=%d\n", NOW, con->get_cid(), con->getBw(), ch->GMH.GMH.len, con->getBw()-ch->GMH.GMH.len);
//  con->setBw(con->getBw()-ch->GMH.GMH.len);
//  }
//  */
//	pktRx_->AddHeader(ah);
//  //We check if it is a MAC packet or not
//  if (isMACPacket(pktRx_)) {
//    process_mac_packet (con, pktRx_);
//    //update_throughput (&rx_traffic_watch_, 8*ch->GMH.GMH.len);
//    //mac_log(pktRx_);
//    pktRx_=0;
//  }
//  else {
//  	pktRx_->RemoveHeader(ah);
//    //update_throughput (&rx_data_watch_, 8*ch->GMH.GMH.len);    
//    //update_throughput (&rx_traffic_watch_, 8*ch->GMH.GMH.len);
//    debug2( "Sending Recieved packet up to IP Layer in BS\n");
//    Mac48Address origin = declassify(ah.GMH.GMH.cid);
//    m_traceRx(pktRx_);//, origin);
//    UL_Protocol_Header uph;
//    pktRx_->RemoveHeader(uph);
//    uint16_t proto_num = uph.protocol;
//    Receive(pktRx_, proto_num, addr(), origin);
////    m_rxCallback (this, pktRx_, proto_num, origin);
//  }
//}

// bool Wimax2BSNetDevice::isMACPacket(Ptr<Packet> p) {
// 	All_Headers ah;
// 	p->PeekHeader(ah);
// 	if (ah.gm_sm != 1) return true;
// 	if (ah.type_arqfb) return true;
// 	if (!(*ah.type_arqfb) && !(*ah.type_ext) && !(*ah.type_frag) && !(*ah.type_pck) && !(*ah.type_fbgm)) return true;
// 	return false;
// }

//Begin RPI
/**
 * The function is used to process the MAC PDU when ARQ,Fragmentation and Packing are enabled
 * @param con The connection by which it arrived
 * @param p The packet to process
 */
void Wimax2BSNetDevice::process_mac_pdu_witharqfragpack (Connection *con, Ptr<Packet> pkt)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ptr<Packet> mac_pdu, p_current, p, mac_sdu, p_previous, p_temp;
  //hdr_cmn *ch_pdu, *ch_current, *ch, *ch_sdu;
  All_Headers wimaxHdr_pdu, wimaxHdr_current, wimaxHdr, wimaxHdr_sdu;
  double tmp;
  uint8_t isInOrder;
  uint32_t seqno;
  bool mac_sdu_gen = false;
  int pdu_size = 0;	
  
  debug2("ARQ BS: Entering the process_mac_pdu_witharqfragpack function \n");	 
  //The packet received is always MAC PDU
  mac_pdu = pkt->Copy();
  mac_pdu->RemoveHeader(wimaxHdr_pdu);
  pdu_size = wimaxHdr_pdu.GMH.GMH.len; 	

	// We shall first handle the acknowledgements received
	if(wimaxHdr_pdu.GMH.GMH.type_arqfb == 1)
	{
		for(uint16_t i=0; i < wimaxHdr_pdu.num_of_acks; i++)
		{
			Connection *connection = this->getCManager ()->get_connection(wimaxHdr_pdu.arq_ie[i].cid, OUT_CONNECTION);
			if(connection)
			{ 
				if(connection->getArqStatus () && connection->getArqStatus ()->isArqEnabled () == 1)
				{
					mac_pdu->AddHeader(wimaxHdr_pdu);
					debug2("ARQ : BS Feedback in Data Payload Received: Has a feedback: Value of i:%d , Value of number of acks:%d \n" , i, wimaxHdr_pdu.num_of_acks);
					connection->getArqStatus ()->arqRecvFeedback(mac_pdu, i, connection);
					mac_pdu->RemoveHeader(wimaxHdr_pdu);
				}
			}  
		}
		wimaxHdr_pdu.GMH.GMH.len -= (wimaxHdr_pdu.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE);
		mac_pdu->RemoveAtStart(wimaxHdr_pdu.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE);
	}
  		
	//Add loss on data connection
	UniformVariable unfvr(0,1);
	tmp = unfvr.GetValue(0, 1);
	if (tmp < data_loss_) 
	{
		debug2 ("ARQ: BS process_mac_pdu_witharqfragpack %f Drop data loss-- packet size [%d] %f %f\n", NOW,wimaxHdr_pdu.GMH.GMH.len, tmp, data_loss_);
		//update_watch(&loss_watch_,1); 
		drop (pkt, "DROP_MAC_PACKET_ERROR");
		pkt = NULL;    
		return;
	} 
	else 
	{
		debug2 ("ARQ: BS process_mac_pdu_witharqfragpack %f No drop data loss %f %f\n", NOW, tmp, data_loss_);
	}
  
	// We need to change the size of the first packet to its original size as in transmission logic 
	wimaxHdr_pdu.GMH.GMH.len -= HDR_MAC802_16_SIZE+HDR_MAC802_16_FRAGSUB_SIZE; /*Size of generic Mac Header was set (Note one packet header and first ARQ Block is present)*/

	//remove CRC
	if(wimaxHdr_pdu.GMH.GMH.ci) {
		wimaxHdr_pdu.GMH.GMH.len -= HDR_MAC802_16_CRC_SIZE;	
		mac_pdu->RemoveAtEnd(HDR_MAC802_16_CRC_SIZE);
	}

	// Initial Condition
	p_previous = mac_pdu;	
	PACKING_SUBHEADER psh_temp, psh_prev;
	p_temp = mac_pdu->Copy();
	p_temp->RemoveHeader(psh_temp);

 	/*Maybe there are a lot of ARQ blocks in this incoming packet, but they belong to different MAC SDU.
 	   We first update the size of first ARQ block. and use a loop to remove all the overheads.*/
 	debug2("ARQ BS: before update, the ch_pdu size: %d \n",wimaxHdr_pdu.GMH.GMH.len); 
	while(psh_temp.data.fc > 1)
	{
		// Update the Mac PDU Size
		debug2("ARQ:BS p-previous :%p p-temp : %p \n", PeekPointer(p_previous), PeekPointer(p_temp));	
		debug2 ("ARQ : BS type of ARQ to update size:  %d, %d\n ",psh_temp.data.fc, (((psh_temp.data.sn) - (psh_prev.data.sn)) &0x7FF));	

		//Initial conditions
		psh_prev = psh_temp;
		p_temp->RemoveAtStart(psh_temp.data.length);
		p_temp->RemoveHeader(psh_temp);
	}	
	debug2("ARQ BS: Finally the pdu size: %d \n",p_temp->GetSize()); 

	/*Distribute all the ARQ blocks in this MAC SDU to transmission queue and re-transmission queue.
	   according to if they are in-order or not.*/
	   mac_pdu->AddHeader(wimaxHdr_pdu);
	if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) 
	{
		isInOrder = 1;
		con->getArqStatus ()->arqReceive(mac_pdu,con,&isInOrder);
		debug2("Receive: ARQ BS :ORDER is:%d\n",isInOrder);
		if (isInOrder == 0) /*In middle part*/
		{
			// packet is out of order, so queue it in retransmission queue at receiver end since it is not being used
			debug2("Receive: ARQ BS : Out of order packet, buffering in retransmission queue\n");
			con->getArqStatus ()->arq_retrans_queue_->Enqueue(mac_pdu->Copy());
		}
		else if (isInOrder == 1)  /*in order*/
		{
			// packet is in order, so queue it in transmission queue at receiver end since it is not being used
			debug2("Receive: ARQ BS : In order packet, buffering in arq transmission queue\n");
			con->getArqStatus ()->arq_trans_queue_->Enqueue(mac_pdu->Copy());
		}
		else /*outside of ARQ window.*/
		{
			debug2("Receive: ARQ BS : Not buffering in arq trans/retrans queue\n");
		}	 
	}
	
   /*for the second and following ARQ blocks.  Distribution go ahead.*/
	p_current = p_temp;
  
	while(p_current->GetSize() > 0)
	{
		p_current->PeekHeader(wimaxHdr_current);
//		wimaxHdr_current= HDR_MAC802_16(p_current);
//		ch_current = HDR_CMN(p_current);

		//Time to perform ARQ for the new packet
		if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) 
		{
			isInOrder = 1;
			con->getArqStatus ()->arqReceive(p_current,con,&isInOrder);
	               debug2("Receive: ARQ BS cid=%d:ORDER is:%d\n",con->get_cid(),isInOrder);
			if (isInOrder == 0) 
			{
				// packet is out of order, so queue it in retransmission queue at receiver end since it is not being used
                                debug2("Receive: ARQ BS cid=%d: Out of order packet, buffering in retransmission queue\n",con->get_cid());
				con->getArqStatus ()->arq_retrans_queue_->Enqueue(p_current->Copy());
			}
			else if(isInOrder == 1) 
			{
				// packet is in order, so queue it in transmission queue at receiver end since it is not being used
	                       debug2("Receive: ARQ BS cid=%d: In order packet, buffering in arq transmission queue\n",con->get_cid());
				con->getArqStatus ()->arq_trans_queue_->Enqueue(p_current->Copy());
			}
			else 
			{
				debug2("Receive: ARQ BS :  Not buffering in arq trans/retrans queue\n"); 	
			} 
		}
		p_current->RemoveAtStart(wimaxHdr_current.GMH.GMH.len);
	}
 
	// Here we will check if packets in retransmission queue can be fit into transmission queue now that a lost packet may have been retransmitted
	primary:
	con->getArqStatus ()->arq_retrans_queue_->resetIterator();
   
	while(con->getArqStatus ()->arq_retrans_queue_)
	{
		secondary:
		p = con->getArqStatus ()->arq_retrans_queue_->getNext ();

		/*There are no packets in ARQ re-transmission queue.*/
		if(!p)
		{
	               debug2("ARQ BS cid=%d: Retransmission queue is empty- so nothing to fit or packets are there but of no use\n",con->get_cid());
			break;
		}  		 

		//Time to perform ARQ for the new packet
		if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) 
		{
			isInOrder = 1;
			con->getArqStatus ()->arqReceiveBufferTransfer(p,con,&isInOrder);
	               debug2("ARQ BS cid=%d: isInOrder=%d\n", con->get_cid(), isInOrder);
			if(isInOrder == 2) /*Just outside of ARQ window. We simply remove it since it is of no use.*/
			{
				con->getArqStatus ()->arq_retrans_queue_->Remove(p);
				debug2("ARQ BS remove out-order ARQ blocks.\n");
				p = 0;
				goto secondary;		
			}
			else if (isInOrder == 0) 
			{
				/*we need to find the expected block. So the other will stay in re-transmission queue. Wait for the next search.*/
				goto secondary;
			}
			else  /*In order.*/
			{
				con->getArqStatus ()->arq_trans_queue_->Enqueue(p->Copy());
				con->getArqStatus ()->arq_retrans_queue_->Remove(p);
				p = 0;
				
				while(con->getArqStatus ()->arq_retrans_queue_)
				{
					p = con->getArqStatus ()->arq_retrans_queue_->getNext ();

					if(!p)
					{
		                                debug2("ARQ BS cid=%d: Retransmission2 queue is empty- so nothing to fit or packets are there but of no use\n",con->get_cid());
						break;
					}   		
					
					//Time to perform ARQ for the new packet
					if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1) 
					{
						isInOrder = 1;
						con->getArqStatus ()->arqReceiveBufferTransfer(p,con,&isInOrder);
		                                debug2("ARQ BS cid=%d: isInOrder2=%d\n", con->get_cid(), isInOrder);
						if(isInOrder == 2)
						{
							con->getArqStatus ()->arq_retrans_queue_->Remove(p);
							p = 0;
							goto primary; 		
						}	
						
						if (isInOrder == 0)
						{
							goto primary;
						}
						else /*in order.*/
						{
							con->getArqStatus ()->arq_trans_queue_->Enqueue(p->Copy());
							con->getArqStatus ()->arq_retrans_queue_->Remove(p);
							p = 0;
						}
					}
				}
				goto primary;
				//break;	
			}
		}		    	
	}

   
	//The advantage of seperating the packets in transmission and retransmission queue is the all the packets in transmission queue will be in order.
	while(1) 
	{
		con->getArqStatus ()->arq_trans_queue_->resetIterator();
		p = con->getArqStatus ()->arq_trans_queue_->getNext ();

		if(!p) 
		{
			debug2("ARQ BS: Transmission Queue is Empty \n ");	
			break;
		}	
		p->PeekHeader(wimaxHdr);
//		wimaxHdr= HDR_MAC802_16(p);
//		ch = HDR_CMN(p);

		// This will be first part of the MAC SDU
		if(wimaxHdr.PSH.data.fc == FRAG_FIRST) 
		{
			mac_sdu = p->Copy();
			mac_sdu->PeekHeader(wimaxHdr_sdu);
//			wimaxHdr_sdu= HDR_MAC802_16(mac_sdu);
//			ch_sdu = HDR_CMN(mac_sdu);
			seqno = wimaxHdr_sdu.PSH.data.sn; 
			debug2("ARQ BS: This is the first packet in transmission queue CID:[%d] and FRAG_FIRST Fsn: %d Size :%d\n", wimaxHdr.GMH.GMH.cid, wimaxHdr_sdu.PSH.data.sn, wimaxHdr_sdu.GMH.GMH.len);

			while(1) 
			{
				p =  con->getArqStatus ()->arq_trans_queue_->getNext ();

				if(!p)
				{
					debug2("ARQ BS: Except for the first Block, all other ARQ Blocks for this MAC SDU are not in ARQ transmission queue. \n");
					break;
				}
				p->PeekHeader(wimaxHdr);
//				wimaxHdr= HDR_MAC802_16(p);
//				ch = HDR_CMN(p);

				
				if((wimaxHdr.PSH.data.fc == FRAG_CONT) && (((wimaxHdr.PSH.data.sn - seqno) &0x7FF)==1)) 
				{
					debug2("ARQ:BS Generating the MAC SDU CID %d Current SN: %d Previous SN: %d Frag Type: %d Size: %d, %d \t Trans-Q length is [%d] Re-trans-Q length is %d \n ", con->get_cid(), wimaxHdr.PSH.data.sn, seqno, wimaxHdr.PSH.data.fc, wimaxHdr.GMH.GMH.len, wimaxHdr_sdu.GMH.GMH.len, con->getArqStatus ()->arq_trans_queue_->GetNPackets(), con->getArqStatus ()->arq_retrans_queue_->GetNPackets()); 	
					
					wimaxHdr_sdu.GMH.GMH.len += wimaxHdr.GMH.GMH.len;
					seqno++;	
				}
				else if(wimaxHdr.PSH.data.fc == FRAG_LAST && (((wimaxHdr.PSH.data.sn - seqno) &0x7FF)==1)) 
				{
					debug2("ARQ: BS Generating the MAC SDU CID %d Current SN: %d Previous SN: %d Frag Type: %d Size: %d, %d \t Trans-Q length is [%d] Re-trans-Q length is %d \n ", con->get_cid(), wimaxHdr.PSH.data.sn, seqno, wimaxHdr.PSH.data.fc, wimaxHdr.GMH.GMH.len, wimaxHdr_sdu.GMH.GMH.len, con->getArqStatus ()->arq_trans_queue_->GetNPackets(), con->getArqStatus ()->arq_retrans_queue_->GetNPackets());
					wimaxHdr_sdu.GMH.GMH.len += wimaxHdr.GMH.GMH.len;
					
					mac_sdu_gen = true; 
					break;		
				}
				else 
				{
					debug2("ARQ BS:This is a unsual case FSN: %d  and Frag_Type: %d \n",wimaxHdr.PSH.data.sn, wimaxHdr.PSH.data.fc);		 
					break;
				}		
			}
			
			if(mac_sdu_gen == true)
			{
				
				debug2("ARQ BS: Created the Mac SDU SIZE: %d\n ", wimaxHdr_sdu.GMH.GMH.len);
				//update_throughput (&rx_data_watch_, 8*pdu_size);    
				//update_throughput (&rx_traffic_watch_, 8*pdu_size);
				debug2("send to upper layer, Mac SDU size [%d]  time[%f] for MAC [%s] Last block FSN is [%d]\n",wimaxHdr_sdu.GMH.GMH.len, NOW, HelperFunctions::getMacAddrString(addr()).c_str(),seqno);
				wimaxHdr_sdu.direction = UL; 
				
			    	UL_Protocol_Header uph;
			    	mac_sdu->RemoveHeader(uph);
			    	uint16_t proto_num = uph.protocol;
			
//				m_rxCallback (this, mac_sdu, proto_num, classifier_list_.lh_first->declassify(wimaxHdr_sdu.GMH.GMH.cid));
				//note there might be a bug here if a more robust classification system is used that actually uses more than one entry in the classifier list
				Mac48Address origin = declassify(wimaxHdr_sdu.GMH.GMH.cid);
				m_traceRx(mac_sdu, origin);//, origin);
				Receive(mac_sdu, proto_num, addr(), origin);
				mac_sdu_gen = false;
				// Go ahead and delete all the packets pertaining to this MAC SDU from the ARQ transmission queue
				do 
				{
					p_temp =con->getArqStatus ()->arq_trans_queue_->Dequeue();
					p_temp = 0;
	                debug2("ARQ BS cid=%d: To keep a count on how many times Dequeue function is called \n",con->get_cid());
					p = con->getArqStatus ()->arq_trans_queue_->head(); /*Enqueue new block at the tail. Need to move the first ARQ block to continue.*/
					if(!p)
					{
						goto end;
					}
					p->PeekHeader(wimaxHdr);
//					wimaxHdr= HDR_MAC802_16(p);
//					ch = HDR_CMN(p);
					debug2("ARQ BS: Next Packet to Dequeue FSN: %d FRAG_STATUS: %d \n",wimaxHdr.PSH.data.sn, wimaxHdr.PSH.data.fc);
				} while(wimaxHdr.PSH.data.fc != FRAG_FIRST);
			debug2("ARQ:BS Generating the MAC SDU CID %d Trans-Q length is [%d] Re-trans-Q length is [%d] \n ",
							 con->get_cid() ,con->getArqStatus ()->arq_trans_queue_->GetNPackets(), con->getArqStatus ()->arq_retrans_queue_->GetNPackets()); 	

			}
			else  // Either no packets in trans queue or all the mac sdu that could be generated are sent above
			{
				debug2("ARQ BS: Either no packets in trans queue or all the mac sdu that could be generated are sent above \n");
				goto end; 
			}
		}
		else if(wimaxHdr.PSH.data.fc == FRAG_NOFRAG) // This will be the only part of the MAC SDU
		{
			mac_sdu = p->Copy();
			mac_sdu->RemoveHeader(wimaxHdr_sdu);
			//ch_sdu = HDR_CMN(mac_sdu);
			seqno = wimaxHdr_sdu.PSH.data.sn; 
                       debug2("ARQ BS cid=%d: This is the only fragment  of MAC SDU FRAG_NOFRAG Fsn: %d Size :%d\n",con->get_cid(),wimaxHdr_sdu.PSH.data.sn, wimaxHdr_sdu.GMH.GMH.len);
                         debug2("ARQ BS cid=%d: Created the Mac SDU SIZE: %d\n ",con->get_cid(), wimaxHdr_sdu.GMH.GMH.len);
			
			//update_throughput (&rx_data_watch_, 8*pdu_size);    
			//update_throughput (&rx_traffic_watch_, 8*pdu_size);
          		 wimaxHdr_sdu.direction = UL; 
           
		    	UL_Protocol_Header uph;
		    	mac_sdu->RemoveHeader(uph);
		    	uint16_t proto_num = uph.protocol;
           
           	//	m_rxCallback (this, mac_sdu, proto_num, classifier_list_.lh_first->declassify(wimaxHdr_sdu.GMH.GMH.cid));
			//note there may be a bug here if a classifier is implemented that uses more than the first entry in classifer list
			Mac48Address origin = declassify(wimaxHdr_sdu.GMH.GMH.cid);
			m_traceRx(mac_sdu, origin);//, origin);
           		Receive(mac_sdu, proto_num, addr(), origin);
			mac_sdu_gen = false;
			// Go ahead and delete all the packets pertaining to this MAC SDU from the transmission queue
			p_temp =con->getArqStatus ()->arq_trans_queue_->Dequeue();
			p_temp = 0;
            debug2("ARQ BS cid=%d: To keep a count on how many times Dequeue function is called \n",con->get_cid());
		}  	 
    else {
      debug2("ARQ BS  cid=%d: Something really wrong with the implementation\n",con->get_cid());
      abort();
    }			 	
  }

	end:
#ifdef WIMAX_DEBUG
	printf ("ARQ recv cid=%d: TxQueue=%d\n",con->get_cid(), con->getArqStatus ()->arq_trans_queue_->GetNPackets());
	con->getArqStatus ()->arq_trans_queue_->resetIterator();
	p = con->getArqStatus ()->arq_trans_queue_->getNext ();
	while (p) 
	{
		wimaxHdr= HDR_MAC802_16(p);
		printf ("\tARQ cid=%d Block sn=%d\n",con->get_cid(), wimaxHdr.PSH.data.sn);
		p = con->getArqStatus ()->arq_trans_queue_->getNext ();
	}
	printf ("ARQ cid=%d recv: ReTxQueue=%d\n",con->get_cid(), con->getArqStatus ()->arq_retrans_queue_->GetNPackets());
	con->getArqStatus ()->arq_retrans_queue_->resetIterator();
	p = con->getArqStatus ()->arq_retrans_queue_->getNext ();
	
	while (p) 
	{
		wimaxHdr= HDR_MAC802_16(p);
		printf ("\tARQ cid=%d Block sn=%d\n",con->get_cid(), wimaxHdr.PSH.data.sn);
		p = con->getArqStatus ()->arq_retrans_queue_->getNext ();
	}
#endif
	
	pkt = 0;
	//update_watch(&loss_watch_,0);
	//pktRx_=NULL; 
}
//End RPI

/**
 * Process a packet received by the Mac. Only scheduling related packets should be sent here (BW request, UL_MAP...)
 * @param con The connection by which it arrived
 * @param p The packet to process
 */
void Wimax2BSNetDevice::process_mac_packet (Connection *con, Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  //debug2 ("Wimax2BSNetDevice received packet to process\n");

  //assert (HDR_CMN(p)->ptype()==PT_MAC);
  debug10 ("Wimax2BSNetDevice received MAC packet to process\n");
  
  All_Headers wimaxHdr;
  p->RemoveHeader(wimaxHdr);
  gen_mac_header_t *header = &wimaxHdr.GMH.GMH;
  bw_req_header_t *req = &wimaxHdr.SMH.BRH;

  //check if this is a bandwidth request
  
  if (header->ht == 1) {
	p->AddHeader(wimaxHdr);
    if ( (req->type == 0x3) || (req->type == 0x2) ) {
      process_cdma_req (p);
      return;
    } else {
      process_bw_req (p);
      return;
    }
  }




  //Begin RPI
  if (p->GetSize() == 0)
    {
      if(header->type_arqfb == 1)
	{
	  for(uint16_t i=0; i < wimaxHdr.num_of_acks; i++)
	    {
	      Connection *connection = this->getCManager ()->get_connection(wimaxHdr.arq_ie[i].cid, true);
	      if(connection)
		{ 
		  if(connection->getArqStatus () && connection->getArqStatus ()->isArqEnabled () == 1){
		    debug2("ARQ:BS Feedback Payload Received: Has a feedback: Value of i:%d , Value of number of acks:%d \n" , i, wimaxHdr.num_of_acks);
		    connection->getArqStatus ()->arqRecvFeedback(p, i, connection);
		  }
		}  
	    }
	  
	  p = 0;
	  return ;
	}
    }
  //End RPI
  
  Type_Check tp;
  p->PeekHeader(tp);
  p->AddHeader(wimaxHdr);


  //mac802_16_dl_map_frame *frame = &dmf.mf;
  
  switch (tp.type) {
  case MAC_RNG_REQ: 
    process_ranging_req (p);
    break;
  case MAC_REG_REQ:     
    process_reg_req (p);
    break;
  case MAC_MOB_SCN_REQ:
    if (ctrlagent_) 
      ctrlagent_->process_scan_request (p);
    else
      fprintf (stderr, "Warning: no controler to handle scan request in BS %s\n", HelperFunctions::getMacAddrString(addr()).c_str());
    break;
  case MAC_MOB_MSHO_REQ:
    //process_msho_req (p); //reenable when handover is implemented
    break;
  case MAC_MOB_HO_IND:
    //process_ho_ind (p); //reenable when handover is implemented
    break;
  case MAC_DSA_REQ: 
  case MAC_DSA_RSP: 
  case MAC_DSA_ACK: 
    serviceFlowHandler_->process (p); // rpi changed pktRx_ to p , cause pktRx_ is not in scope. 
    break;
  default:
    debug2 (" unknown packet in BS\n");
  }

  //Packet::free (p);
}

/**
 * Process a RNG-REQ message
 * @param p The packet containing the ranging request information
 */
void Wimax2BSNetDevice::process_ranging_req (Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  UlSubFrame *ulsubframe = getMap()->getUlSubframe();
  All_Headers ah;
  p->RemoveHeader(ah);
  RANGING_REQUEST_FRAME rrfh;
  p->RemoveHeader(rrfh);
  mac802_16_rng_req_frame *req = &rrfh.data;

  debug2 ("BS received ranging request from CID :%d\n", ah.GMH.GMH.cid);

  if (ah.GMH.GMH.cid != INITIAL_RANGING_CID) {
    //process request for DIUC
  } else {
    //here we can make decision to accept the SS or not.
    //for now, accept everybody
    //check if CID already assigned for the SS
    PeerNode *peer = getPeerNode (req->ss_mac_address);
    if (peer==NULL) {
      debug2 ("New peer node requesting ranging ssid :%s\n",HelperFunctions::getMacAddrString(req->ss_mac_address).c_str());
      //Assign Management CIDs
      Connection *basic = new Connection (CONN_BASIC);
      Connection *upbasic = new Connection (CONN_BASIC, basic->get_cid());
      Connection *primary = new Connection (CONN_PRIMARY);
      Connection *upprimary = new Connection (CONN_PRIMARY, primary->get_cid());
      
      //Create Peer information
      peer = new PeerNode (req->ss_mac_address);
      peer->setBasic (upbasic, basic);
      peer->setPrimary (upprimary, primary);
      addPeerNode (peer);
      getCManager()->add_connection (upbasic, IN_CONNECTION);
      getCManager()->add_connection (basic, OUT_CONNECTION);
      getCManager()->add_connection (upprimary, IN_CONNECTION);
      getCManager()->add_connection (primary, OUT_CONNECTION);
      peer->setDIUC (req->req_dl_burst_profile);

  	/*Xingting changes it. Old code does not work in case of CDMA ranging.*/	  
      peer->setchannel(ah.PIH.channel_index);
      //peer->setchannel(req->channel_num);
      //schedule timer in case the node never register
      addtimer17 (req->ss_mac_address);

      //create packet for answers
      Ptr<Packet> rep = getPacket ();
      All_Headers ah_rep;
      rep->RemoveHeader(ah_rep);
      
      //rep->allocdata (sizeof (struct mac802_16_rng_rsp_frame));
      RANGING_RESPONSE_FRAME frame_header;
      mac802_16_rng_rsp_frame *frame = &frame_header.data;
      
      frame->type = MAC_RNG_RSP;
      frame->uc_id = ulsubframe->getChannelID();
      frame->rng_status = RNG_SUCCESS;
      frame->ss_mac_address = req->ss_mac_address;
      frame->basic_cid = basic->get_cid();
      frame->primary_cid = primary->get_cid();
      ah_rep.GMH.GMH.len += frame_header.GetSerializedSize();

      //Enqueue packet
      rep->AddHeader(frame_header);
      rep->AddHeader(ah_rep);
      getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION)->enqueue (rep);
      
      debug2 ("BS created ranging response MAC_RNG_RSP to SSID :%s, CID :%d, primary :%d, size :%d\n", HelperFunctions::getMacAddrString(req->ss_mac_address).c_str(), ah.GMH.GMH.cid, primary->get_cid(), ah.GMH.GMH.len);

      if (cl_head_==NULL) {
		cl_head_ = (new_client_t*)malloc (sizeof (new_client_t));
		cl_tail_ = cl_head_;
      } else {
		cl_tail_->next = (new_client_t*)malloc (sizeof (new_client_t));
		cl_tail_=cl_tail_->next;
      }
      cl_tail_->cid = primary->get_cid();
      cl_tail_->next = NULL;

#ifdef USE_802_21
      send_link_detected (addr(), peer->getAddr(), 1);
#endif

    } else {
      debug2 ("BS received ranging for known station (%s) and create MAC_RNG_RSP\n", HelperFunctions::getMacAddrString(req->ss_mac_address).c_str());
      //reset invited ranging retries for SS
      //create packet for answers
      Connection *basic = peer->getBasic(IN_CONNECTION);
      Connection *primary = peer->getPrimary(IN_CONNECTION);
      Ptr<Packet> rep = getPacket ();
      All_Headers ah_rep;
      RANGING_RESPONSE_FRAME rrf;
      mac802_16_rng_rsp_frame *frame = &rrf.data;
      rep->RemoveHeader(ah_rep);
      
      frame->type = MAC_RNG_RSP;
      frame->uc_id = ulsubframe->getChannelID();
      frame->rng_status = RNG_SUCCESS;
      frame->ss_mac_address = req->ss_mac_address;
      frame->basic_cid = basic->get_cid();
      frame->primary_cid = primary->get_cid();
      ah_rep.GMH.GMH.len += rrf.GetSerializedSize();
      
      rep->AddHeader(rrf);
      rep->AddHeader(ah_rep);
      //Enqueue packet
      getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION)->enqueue (rep);
    }
  }
}

/**
 * Process bandwidth request
 * @param p The request
 */
void Wimax2BSNetDevice::process_bw_req (Ptr<Packet> p)
{ 
  NS_LOG_FUNCTION_NOARGS ();
  All_Headers wimaxHdr;
  p->PeekHeader(wimaxHdr);
  
  bw_req_header_t *req = &wimaxHdr.SMH.BRH;

  debug2 ("received bandwidth request of %d bytes from %d\n", req->br, req->cid); 
  
  //retrieve the CID and update bandwidth request information
  Connection *c =  getCManager()->get_connection (req->cid, IN_CONNECTION);
  assert (c);
  if ( (req->type & 0x7) == 0x001) {
    //aggregate
    c->setBw (req->br & 0x7FFFF); //masks 19 bits
    debug10 ("BS (setBW) Aggregate request on connection %d of %d\n", c->get_cid(), c->getBw());
  } else if ( (req->type & 0x7) == 0x000) {
    //incremental
    c->setBw (c->getBw() + (req->br & 0x7FFFF));
    debug10 ("BS (setBW) Incremental request on connection %d of %d\n", c->get_cid(), c->getBw());

  } 
}

/**
 * Process cdma bandwidth request - this doesn't seem to be called anymore
 * @param p The request
 */
void Wimax2BSNetDevice::process_cdma_req (Ptr<Packet> p)
{
  NS_LOG_FUNCTION_NOARGS ();
  All_Headers wimaxHdr;
  p->PeekHeader(wimaxHdr);

  cdma_req_header_t *req = &wimaxHdr.CRH.data;

  debug2 ("BS (process_cdma_req) received cdma request code :%d from :%d\n", req->code, req->cid);

  //retrieve the CID and update bandwidth request information
  Connection *c =  getCManager()->get_connection (req->cid, IN_CONNECTION);
  assert (c);
  debug10 ("BS (setCDMA), code :%d, top :%d, ssid :%d\n", req->code, req->top, req->br);

  if (req->type == 0x3) {
    c->setCDMA (1);
    c->setCDMA_code (req->code);
    c->setCDMA_top (req->top);
  } else {
    c->setCDMA(2);
    c->setCDMA_SSID_FLAG(req->br, 2);
    c->setCDMA_SSID_CODE(req->br, req->code);
    c->setCDMA_SSID_TOP(req->br, req->top);
    c->setCDMA_SSID_CID(req->br, c->get_cid());
    c->setCDMA_code(0);
    c->setCDMA_top(0);
  }

}

/**
 * Process registration request
 * @param p The request
 */
void Wimax2BSNetDevice::process_reg_req (Ptr<Packet> req)
{ 
  NS_LOG_FUNCTION_NOARGS ();
  All_Headers *wimaxHdr_req = new All_Headers;
  req->PeekHeader(*wimaxHdr_req);
  gen_mac_header_t header_req = wimaxHdr_req->GMH.GMH;
  
  debug2 ("BS received registration request from :%d\n", header_req.cid);

  Ptr<Packet> p;
  
  All_Headers wimaxHdr;
  REGISTRATION_RESPONSE_FRAME rrfh;
  mac802_16_reg_rsp_frame *reg_frame = &rrfh.data;
  PeerNode *peer;

  //create packet for request
  p = getPacket ();
  p->RemoveHeader(wimaxHdr);
  
  reg_frame->type = MAC_REG_RSP;
  reg_frame->response = 0; //OK
  peer = getCManager()->get_connection (header_req.cid, IN_CONNECTION)->getPeerNode();
  Connection *secondary = peer->getSecondary (OUT_CONNECTION);
  if (secondary==NULL) {
    //first time 
    secondary = new Connection (CONN_SECONDARY);
    Connection *upsecondary = new Connection (CONN_SECONDARY, secondary->get_cid());
    getCManager()->add_connection (upsecondary, IN_CONNECTION);
    getCManager()->add_connection (secondary, OUT_CONNECTION);
    peer->setSecondary (upsecondary, secondary);
  }
  reg_frame->sec_mngmt_cid = secondary->get_cid();
  wimaxHdr.GMH.GMH.cid = header_req.cid;
  wimaxHdr.GMH.GMH.len += rrfh.GetSerializedSize();
  debug2 ("BS created registration response MAC_REG_RSP to :%d, secondary :%d, size :%d\n", header_req.cid, secondary->get_cid(), wimaxHdr.GMH.GMH.len);
  
  //Enqueue packet..must be replaced with second line later
  //getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION)->Enqueue (p);
  p->AddHeader(rrfh);
  p->AddHeader(wimaxHdr);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);

  //clear t17 timer for this node
  removetimer17 (peer->getAddr());

#ifdef USE_802_21
  debug2 ("At %f in Mac %s, send link up\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str());
  send_link_up (peer->getAddr(),addr(), -1);
#endif
}

/**
 * Send a neighbor advertisement message
 */
void Wimax2BSNetDevice::send_nbr_adv ()
{
  NS_LOG_FUNCTION_NOARGS ();
  nbr_db_->getNbNeighbor();
  debug2 ("At %f in BS %s send_nbr_adv (nb_neighbor=%d)\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str(), nbr_db_->getNbNeighbor());
  Ptr<Packet> p;
  All_Headers wimaxHdr;
  NEIGHBOR_ADVERTISEMENT_FRAME_HEADER nafh;
  mac802_16_mob_nbr_adv_frame *frame = &nafh.data;
  //PeerNode *peer;

  //create packet for request
  p = getPacket ();
  p->RemoveHeader(wimaxHdr);
  
  frame->type = MAC_MOB_NBR_ADV;
  frame->n_neighbors = nbr_db_->getNbNeighbor();
  frame->skip_opt_field = 0;
  for (int i = 0 ; i < frame->n_neighbors ; i++) {
    frame->nbr_info[i].phy_profile_id.FAindex = 0;
    frame->nbr_info[i].phy_profile_id.bs_eirp = 0;
    frame->nbr_info[i].nbr_bsid= nbr_db_->getNeighbors()[i]->getID();
    frame->nbr_info[i].dcd_included = true;
    memcpy (&(frame->nbr_info[i].dcd_settings), nbr_db_->getNeighbors ()[i]->getDCD(), sizeof(mac802_16_dcd_frame));
    frame->nbr_info[i].ucd_included = true;
    memcpy (&(frame->nbr_info[i].ucd_settings), nbr_db_->getNeighbors ()[i]->getUCD(), sizeof(mac802_16_ucd_frame));
    frame->nbr_info[i].phy_included = false;
  }
  //wimaxHdr.GMH.GMH.len = Mac802_16pkt::getMOB_NBR_ADV_size(frame);
  wimaxHdr.GMH.GMH.len += nafh.GetSerializedSize();
  
 
  p->AddHeader(nafh);
    debug10("finished generating nbr_adv frame, total size = %d and GMH recorded size of %d\n", p->GetSize() + 6, wimaxHdr.GMH.GMH.len);

  p->AddHeader(wimaxHdr);
      debug10("finished generating nbr_adv frame, total size with wimaxHdr = %d\n", p->GetSize());

  
  getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION)->enqueue (p);
  
}
// 
// /**
//  * Process handover request
//  * @param p The request
//  */
// void Wimax2BSNetDevice::process_msho_req (Ptr<Packet> req)
// {
//   All_Headers wimaxHdr_req;
//   req->PeekHeader(wimaxHdr_req);
//   MSHO_REQUEST_FRAME_HEADER mrfh;
//   mac802_16_mob_msho_req_frame *req_frame = &mrfh.data;
//   
//   debug2 ("At %f in Mac %s received handover request from %d\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str(), wimaxHdr_req.GMH.GMH.cid);
// 
//   //check the BS that has stronger power
//   int maxIndex = 0;
//   int maxRssi = 0; //max value
//   debug2("In the MSHO Request message, there are %d new BS full entities.\n", req_frame->n_new_bs_full);
//   for (int i = 0; i < req_frame->n_new_bs_full ; i++) {
//     if (req_frame->bs_full[i].bs_rssi_mean >= maxRssi) {
//       maxIndex = i;
//       maxRssi = req_frame->bs_full[i].bs_rssi_mean;
//     }
//   }
//   //reply with one recommended BS
//   Ptr<Packet> p;
//   struct hdr_cmn *ch;
//   All_Headers wimaxHdr;
//   p->RemoveHeader(wimaxHdr);
//   mac802_16_mob_bsho_rsp_frame *rsp_frame;
// 
//   send_nbr_adv (); //to force update with latest information
// 
//   //create packet for request
//   p = getPacket ();
//   ch = HDR_CMN(p);
//   wimaxHdr = HDR_MAC802_16(p);
//   p->allocdata (sizeof (struct mac802_16_mob_bsho_rsp_frame)+sizeof (mac802_16_mob_bsho_rsp_rec));
//   rsp_frame = (mac802_16_mob_bsho_rsp_frame*) p->accessdata();
//   rsp_frame->type = MAC_MOB_BSHO_RSP;
//   
//   rsp_frame->mode = 0; //HO request
//   rsp_frame->ho_operation_mode = 1; //mandatory handover response
//   rsp_frame->n_recommended = 1;
//   rsp_frame->resource_retain_flag = 0; //release connection information
//   rsp_frame->n_rec[0].neighbor_bsid = req_frame->bs_full[maxIndex].neighbor_bs_index;
//   rsp_frame->n_rec[0].ho_process_optimization=0; //no optimization
// 
//   ch->GMH.GMH.len += Mac802_16pkt::getMOB_BSHO_RSP_size(rsp_frame);
//   wimaxHdr.header.cid = header_req.cid;
//   getCManager()->get_connection (header_req.cid, OUT_CONNECTION)->Enqueue (p);
// }
//  
// /**
//  * Process handover indication
//  * @param p The indication
//  */
// void Wimax2BSNetDevice::process_ho_ind (Ptr<Packet> p)
// {
//   hdr_mac802_16 *wimaxHdr_req = HDR_MAC802_16(p);
//   gen_mac_header_t header_req = wimaxHdr_req->header;
//   //mac802_16_mob_ho_ind_frame *req_frame = 
//   //  (mac802_16_mob_ho_ind_frame*) p->accessdata();
//   
//   debug2 ("At %f in Mac %s received handover indication from %d\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str(), header_req.cid);
//   debug2("xingting is trying to remove all the connections to SS.\n");
// 
//   Connection *head  = getCManager()->get_out_connection();
//   while(head != NULL)
//   {
//   		debug2("HO in out cid %d\n",head->get_cid());
// 		if( (head->getArqStatus() != NULL))
// 		{
// 			debug2("BS %s delete ARQ timer CID is [%d]\n", HelperFunctions::getMacAddrString(addr()).c_str(),head->get_cid());
// 				head->getArqStatus()->cancelTimer();
// 			//delete head->getArqStatus();
// 			//getCManager()->remove_connection (head);
// 			//delete head;
// 		}
// 		head = head->next_entry();
//   }
// }
 
/**
 * Send a scan response to the MN
 * @param rsp The response from the control
 */
void Wimax2BSNetDevice::send_scan_response (mac802_16_mob_scn_rsp_frame *rsp, int cid)
{
  NS_LOG_FUNCTION_NOARGS ();
  //create packet for request
  Ptr<Packet> p = getPacket ();
  All_Headers wimaxHdr;
  p->RemoveHeader(wimaxHdr);
  //p->allocdata (sizeof (struct mac802_16_mob_scn_rsp_frame));
  mac802_16_mob_scn_rsp_frame* rsp_frame = new mac802_16_mob_scn_rsp_frame;
  memcpy (rsp_frame, rsp, sizeof (mac802_16_mob_scn_rsp_frame));
  rsp_frame->type = MAC_MOB_SCN_RSP;
  
  wimaxHdr.GMH.GMH.cid = cid;
  wimaxHdr.GMH.GMH.len += rsp_frame->GetSerializedSize();
  
  //add scanning station to the list
  PeerNode *peer = getCManager()->get_connection (cid, false)->getPeerNode();

  /* The request is received in frame i, the reply is sent in frame i+1
   * so the frame at which the scanning start is start_frame+2
   */
  ScanningStation *sta = new ScanningStation (peer->getAddr(), rsp_frame->scan_duration & 0xFF, 
					      rsp_frame->start_frame+frame_number_+2, 
					      rsp_frame->interleaving_interval & 0xFF,
					      rsp_frame->scan_iteration & 0xFF);
  sta->insert_entry_head (&scan_stations_);
  
  p->AddHeader(*rsp_frame);
  p->AddHeader(wimaxHdr);
  delete rsp_frame;
  //Enqueue packet
  getCManager()->get_connection (cid, OUT_CONNECTION)->enqueue (p);
}

/**** Internal methods ****/

/**
 * Update the given timer and check if thresholds are crossed
 * @param watch the stat watch to update
 * @param value the stat value
 */// 
// void Wimax2BSNetDevice::update_watch (StatWatch *watch, double value)
// {
//   char *name;
// 
// #ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
//   threshold_action_t action = watch->update (value);
// 
//   if (action != NO_ACTION_TH) {
//     link_parameter_type_s param;
//     union param_value old_value, new_value;
// 
//     if (watch == &loss_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_FRAME_LOSS;
//     } else if (watch == &delay_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_PACKET_DELAY;
//     } else if (watch == &jitter_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_PACKET_JITTER;
//     }
//     old_value.data_d = watch->old_average();
//     new_value.data_d = watch->average();
// 
//     send_link_parameters_report (addr(), addr(), param, old_value, new_value);      
//   }
// #else
//   watch->update (value);
// #endif
// 
//   if (watch == &loss_watch_) {
//     name = "loss";
//   } else if (watch == &delay_watch_) {
//     name = "delay";
//   } else if (watch == &jitter_watch_) {
//     name = "jitter";
//   } else {
//     name = "other";
//   }
//   if (print_stats_)
//     debug2 ("At %f in Mac %d, updating watch stats %s: %f\n", NOW, addr(), name, watch->average());
// }
// 
// /**
//  * Update the given timer and check if thresholds are crossed
//  * @param watch the stat watch to update
//  * @param value the stat value
//  */
// void Wimax2BSNetDevice::update_throughput (ThroughputWatch *watch, double size)
// {
//   char *name;
// 
// #ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
//   threshold_action_t action = watch->update (size, NOW);
//   if (action != NO_ACTION_TH) {
//     link_parameter_type_s param;
//     union param_value old_value, new_value;
//     if (watch == &rx_data_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_RX_DATA_THROUGHPUT;
//     } else if (watch == &rx_traffic_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_RX_TRAFFIC_THROUGHPUT;
//     } else if (watch == &tx_data_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_TX_DATA_THROUGHPUT;
//     } else if (watch == &tx_traffic_watch_) {
//       param.link_type = LINK_GENERIC;
//       param.parameter_type = LINK_GEN_TX_TRAFFIC_THROUGHPUT;
//     }
//     old_value.data_d = watch->old_average();
//     new_value.data_d = watch->average();
//     
//     send_link_parameters_report (addr(), addr(), param, old_value, new_value);
//   }
// #else
//   watch->update (size, NOW);
// #endif 
// 
//   if (watch == &rx_data_watch_) {
//     name = "rx_data";
//     rx_data_timer_->resched (watch->get_timer_interval());
//   } else if (watch == &rx_traffic_watch_) {
//     rx_traffic_timer_->resched (watch->get_timer_interval());
//     name = "rx_traffic";
//   } else if (watch == &tx_data_watch_) {
//     tx_data_timer_->resched (watch->get_timer_interval());
//     name = "tx_data";
//   } else if (watch == &tx_traffic_watch_) {
//     tx_traffic_timer_->resched (watch->get_timer_interval());
//     name = "tx_traffic";
//   }
// 
//   if (print_stats_)
//     debug2 ("At %f in Mac %d, updating throughput stats %s: %f\n", NOW, addr(), name, watch->average());
// }
// 

/**
 * Called when a timer expires
 * @param The timer ID
 */
void Wimax2BSNetDevice::expire (timer_id id)
{
  NS_LOG_FUNCTION_NOARGS ();
  switch (id) {
  case WimaxDCDTimerID:
    sendDCD = true;
    debug2 ("At %f in Mac %s DCDtimer expired\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str());
    dcdtimer_->start (macmib_->dcd_interval);
    break;
  case WimaxUCDTimerID:
    sendUCD = true;
    debug2 ("At %f in Mac %s UCDtimer expired\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str());
    ucdtimer_->start (macmib_->ucd_interval);
    break;
  case WimaxMobNbrAdvTimerID:
    send_nbr_adv();
    nbradvtimer_->start (macmib_->nbr_adv_interval);
    break;
  default:
    debug2 ("Warning: unknown timer expired in Wimax2BSNetDevice\n");
  }
}

/**
 * Start a new frame
 */
void Wimax2BSNetDevice::start_ulsubframe ()
{
  NS_LOG_FUNCTION_NOARGS ();
  debug2 ("At %f in Mac %s BS scheduler ulsubframe expires\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str());

  //change PHY state
  getPhy()->setMode (OFDM_RECV);  

  //start handler of ulsubframe
  getMap()->getUlSubframe()->getTimer()->start (0);

  //reschedule for next frame
  ul_timer_->resched (getFrameDuration());
}

/**
 * Start a new frame
 */
void Wimax2BSNetDevice::start_dlsubframe ()
{
  NS_LOG_FUNCTION_NOARGS ();
  debug2 ("At %f in Mac %s BS scheduler dlsubframe expires (frame=%d), r_ss :%lf, r_bs :%lf\n", NOW, HelperFunctions::getMacAddrString(addr()).c_str(), frame_number_++, r_delta_ss, r_delta_bs);
  r_delta_bs = 0;

  getPhy()->SynchronizeAll();

  assert (map_);
  map_->setStarttime(NOW);
  ((Wimax2BSScheduler*)scheduler_)->schedule();

  //update some information  
  if (sendDCD || map_->getDlSubframe()->getCCC()!= dlccc_) {
    sendDCD = false;
    dlccc_ = map_->getDlSubframe()->getCCC();
    //reschedule timer
    //dcdtimer_->stop();
    //dcdtimer_->start (macmib_->dcd_interval);
  }

  if (sendUCD || map_->getUlSubframe()->getCCC()!= ulccc_) {
    sendUCD = false;
    ulccc_ = map_->getUlSubframe()->getCCC();
    //reschedule timer
    //ucdtimer_->stop();
    //ucdtimer_->start (macmib_->ucd_interval);
  }

  //change PHY state
  getPhy()->setMode (OFDM_SEND);

  //start handler of dlsubframe
  debug10("starting dlsubframetimer(0) from BSNetDevice::start_dlsubframe\n");
  if (getMap()->getDlSubframe()->getTimer()->busy()) getMap()->getDlSubframe()->getTimer()->stop();
//  getMap()->getDlSubframe()->replaceTimer();
  	getMap()->getDlSubframe()->getTimer()->start (0);

	//since we need to schedule another timer for the end of the frame now, if the new instantiation of the class is not placed in the same memory location (which, since it's created right after the previous is destroyed, happens frequently, but not always), the function currently being executed from the start(0) will Seg Fault.
	//we therefore simply don't destroy the old one until very slightly in the future, even though we replace it in the DlSubFrame object now.
	//another solution would be to wait to replace the timer and schedule the end-of-frame task until slightly into the future, but the implementation here was simpler
	//also, note to future programmer: deleteSavedTimer should probably be made static

  DlSubFrameTimer *dlsft = getMap()->getDlSubframe()->replaceTimerWithoutDelete();
  Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(phy_->getPS()), &DlSubFrame::deleteSavedTimer, getMap()->getDlSubframe(), dlsft);
  
  //reschedule for next time (frame duration)
  dl_timer_->resched (getFrameDuration());  

  getMap()->getUlSubframe()->replaceTimer(map_->getUlSubframe()->getStarttime()*getPhy()->getPS());
  //debug10("about to resched ul_timer to %f\n", map_->getUlSubframe()->getStarttime()*getPhy()->getPS());
  //ul_timer_->resched (map_->getUlSubframe()->getStarttime()*getPhy()->getPS());

}

/** Add a new Fast Ranging allocation
 * @param time The time when to allocate data
 * @param macAddr The MN address
 */
void Wimax2BSNetDevice::addNewFastRanging (double time, Mac48Address macAddr)
{
  NS_LOG_FUNCTION_NOARGS ();
  //compute the frame where the allocation will be located
  int frame = int ((time-NOW)/getFrameDuration()) +2 ;
  frame += getFrameNumber();
  //debug2 ("Added fast RA for frame %d (current=%d) for time (%f)\n", 
  //	  frame, getFrameNumber(), time);
  FastRangingInfo *info= new FastRangingInfo (frame, macAddr);
  info->insert_entry(&fast_ranging_head_);
}

/**
 * Add a new timer17 in the list. It also performs cleaning of the list
 * @param index The client address
 */
void Wimax2BSNetDevice::addtimer17 (Mac48Address index)
{
  NS_LOG_FUNCTION_NOARGS ();
  //clean expired timers
  T17Element *entry;
  for (entry = t17_head_.lh_first; entry ; ) {
    if (!entry->busy ()) {
      T17Element *tmp = entry;
      entry = entry->next_entry();
      tmp->remove_entry();
      delete (tmp);
    } else {
      entry = entry->next_entry();
    }
  }

  entry = new T17Element (this, index);
  entry->insert_entry (&t17_head_);
}

/**
 * Cancel and remove the timer17 associated with the node
 * @param index The client address
 */
void Wimax2BSNetDevice::removetimer17 (Mac48Address index)
{
  NS_LOG_FUNCTION_NOARGS ();
  //clean expired timers
  T17Element *entry;
  for (entry = t17_head_.lh_first; entry ; entry = entry->next_entry()) {
    if (entry->index ()==index) {
      entry->cancel();
      entry->remove_entry();
      delete (entry);
      break;
    }
  }
}

/** 
 * Finds out if the given station is currently scanning
 * @param nodeid The MS id
 */
bool Wimax2BSNetDevice::isPeerScanning (Mac48Address nodeid)
{
  NS_LOG_FUNCTION_NOARGS ();
  ScanningStation *sta;
  for (sta = scan_stations_.lh_first; sta ; sta = sta->next_entry()) {
    if (sta->getNodeId()==nodeid && sta->isScanning(frame_number_)) {
      //debug2 ("station %d scanning\n", nodeid);
      return true;
    }
  }
  return false;
}


// void Wimax2BSNetDevice::addPowerinfo(hdr_mac802_16 *wimaxHdr,double power_per_subchannel, bool collision )
// {
//  
//   int num_data_subcarrier = getPhy()->getNumSubcarrier(UL_); 
//   int num_subchannel = getPhy()->getNumsubchannels (UL_);
//   int num_symbol_per_slot = getPhy()->getSlotLength (UL_); 
// 
//   if (collision == true) {
// 
//     if(wimaxHdr.phy_info.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition , chk whether broacastcid reqd ir not. 
//       // if((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) > 30)
//       //if(wimaxHdr.phy_info.num_OFDMSymbol >= num_symbol_per_slot)
//       {
// 	if(wimaxHdr.phy_info.num_OFDMSymbol > num_symbol_per_slot) 
// 	  {
//             //numsymbols = (int)  ceil((((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) - 30) % 30));
//             //numsymbols *= num_symbol_per_slot;  
// 	    // for the first num_symbol_per_slot symbols 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = (wimaxHdr.phy_info.subchannel_offset) ; j<=num_subchannel ; j++ )
//                 (basepower_[i][j] = power_per_subchannel);
// 	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//           
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset) + (wimaxHdr.phy_info.num_OFDMSymbol-num_symbol_per_slot) ; i++)
// 	      for(int j = 1 ; j<=num_subchannel; j++ )
//                 (basepower_[i][j] = power_per_subchannel);  
// 
// 	    // last num_symbol_per_slot 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset +wimaxHdr.phy_info.num_OFDMSymbol-num_symbol_per_slot ; i< wimaxHdr.phy_info.OFDMSymbol_offset + wimaxHdr.phy_info.num_OFDMSymbol/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = 1 ; j<=((wimaxHdr.phy_info.num_subchannels - (num_subchannel - (wimaxHdr.phy_info.subchannel_offset-1)) )%num_subchannel) ; j++ )
// 		(basepower_[i][j] = power_per_subchannel); 
//           }
// 	else 
//           {
// 	    int i = wimaxHdr.phy_info.OFDMSymbol_offset;
// 	    int j = wimaxHdr.phy_info.subchannel_offset;
// 	    for (int k=0 ; k < wimaxHdr.phy_info.num_subchannels ; k++) {
// 	      //printf ("Assigning basepower_ [%d][%d]\n", i, j);
// 	      assert (i<(int) floor((getPhy()->getPS()*(int) floor((getFrameDuration()/getPhy()->getPS())))/getPhy()->getSymbolTime()));
// 	      assert (j < getPhy()->getNumsubchannels (UL_));
// 	      basepower_[i][j] = power_per_subchannel; 	    	      
// 	      j++;
// 	      if (j == getPhy()->getNumsubchannels (UL_)) {
// 		j = 0 ; 
// 		i++;
// 	      }
// 	    }
// 	    /*
// 	      for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
// 	      for(int j = (wimaxHdr.phy_info.subchannel_offset) ; j< (wimaxHdr.phy_info.subchannel_offset) + wimaxHdr.phy_info.num_subchannels ; j++ ) {
// 	      printf ("Assigning basepower_ [%d][%d]\n", i, j);
// 	      assert (i<(int) floor((getPhy()->getPS()*(int) floor((getFrameDuration()/getPhy()->getPS())))/getPhy()->getSymbolTime()));
// 	      assert (j < getPhy()->getNumsubchannels (UL_));
// 	      (basepower_[i][j] = power_per_subchannel); 	    
// 	      }
// 	    */
//           }
// 
//       }
// 
//     else
// 
//       {
// 	//numsymbols = (int)  ceil((((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) - 30) % 30));
// 	//numsymbols *= num_symbol_per_slot;  
//         // for the first num_symbol_per_slot symbols 
// 	if(wimaxHdr.phy_info.num_OFDMSymbol > 1) 
// 	  {
//             for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + 1)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
//               for(int j = (wimaxHdr.phy_info.subchannel_offset) ; j<=num_subchannel ; j++ )
// 		(basepower_[i][j] = power_per_subchannel); 
// 	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//           
//             for(int i = wimaxHdr.phy_info.OFDMSymbol_offset+1 ; i<  ((wimaxHdr.phy_info.OFDMSymbol_offset) + wimaxHdr.phy_info.num_OFDMSymbol-1) ; i++)
// 	      for(int j = 1 ; j<=num_subchannel ; j++ )
// 		basepower_[i][j] = power_per_subchannel;
// 
// 	    // last num_symbol_per_slot 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset +wimaxHdr.phy_info.num_OFDMSymbol-1 ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset) + wimaxHdr.phy_info.num_OFDMSymbol/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = 1 ; j<=((wimaxHdr.phy_info.num_subchannels - (num_subchannel - (wimaxHdr.phy_info.subchannel_offset-1)) )%num_subchannel) ; j++ )
// 		(basepower_[i][j] = power_per_subchannel); 
// 
//           }
// 	else
//           {
//           
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + 1)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = (wimaxHdr.phy_info.subchannel_offset) ; j< wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels ; j++ )
// 		(basepower_[i][j] = power_per_subchannel); 	
// 
//           }
//              
//       }
//   } else {
// 
//     if(wimaxHdr.phy_info.num_OFDMSymbol % num_symbol_per_slot == 0)   // chk this condition , chk whether broacastcid reqd ir not. 
//       // if((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) > 30)
//       //if(wimaxHdr.phy_info.num_OFDMSymbol >= num_symbol_per_slot)
//       {
// 	if(wimaxHdr.phy_info.num_OFDMSymbol > num_symbol_per_slot) 
// 	  {
//             //numsymbols = (int)  ceil((((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) - 30) % 30));
//             //numsymbols *= num_symbol_per_slot;  
// 	    // for the first num_symbol_per_slot symbols 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = (wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
//                 (intpower_[i][j] = power_per_subchannel);
// 	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//           
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset) + (wimaxHdr.phy_info.num_OFDMSymbol-num_symbol_per_slot) ; i++)
// 	      for(int j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
//                 (intpower_[i][j] = power_per_subchannel);  
// 
// 	    // last num_symbol_per_slot 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset +wimaxHdr.phy_info.num_OFDMSymbol-num_symbol_per_slot ; i< wimaxHdr.phy_info.OFDMSymbol_offset + wimaxHdr.phy_info.num_OFDMSymbol/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = 0 ; j<((wimaxHdr.phy_info.num_subchannels - (num_subchannel - (wimaxHdr.phy_info.subchannel_offset)) )%num_subchannel)*num_data_subcarrier ; j++ )
// 		(intpower_[i][j] = power_per_subchannel); 
//           }
// 	else 
//           {
// 	    int i = wimaxHdr.phy_info.OFDMSymbol_offset ;
// 	    int j = wimaxHdr.phy_info.subchannel_offset*num_data_subcarrier ;
// 	    for (int k = 0 ; k < wimaxHdr.phy_info.num_subchannels*num_data_subcarrier ; k++) {
// 	      //printf ("Assigning intpower_ [%d][%d]\n", i, j);
// 	      assert (i<(int) floor((getPhy()->getPS()*(int) floor((getFrameDuration()/getPhy()->getPS())))/getPhy()->getSymbolTime()));
// 	      assert (j < getPhy()->getFFT());
// 	      intpower_[i][j] = power_per_subchannel;		
// 	      if (j == getPhy()->getFFT()) {
// 		j = 0 ; 
// 		i++;
// 	      }	      
// 	    }
// 	    /*
// 	      for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + num_symbol_per_slot) ; i++)
//               for(int j = (wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier ; j<(((wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier) + wimaxHdr.phy_info.num_subchannels*num_data_subcarrier) ; j++ ) {
// 	      printf ("Assigning intpower_ [%d][%d]\n", i, j);
// 	      assert (i<(int) floor((getPhy()->getPS()*(int) floor((getFrameDuration()/getPhy()->getPS())))/getPhy()->getSymbolTime()));
// 	      assert (j < getPhy()->getFFT());
// 	      (intpower_[i][j] = power_per_subchannel); 	    
// 	      }
// 	    */
//           }
// 
//       }
// 
//     else
// 
//       {
// 	//numsymbols = (int)  ceil((((wimaxHdr.phy_info.subchannel_offset + wimaxHdr.phy_info.num_subchannels -1) - 30) % 30));
// 	//numsymbols *= num_symbol_per_slot;  
//         // for the first num_symbol_per_slot symbols 
// 	if(wimaxHdr.phy_info.num_OFDMSymbol > 1) 
// 	  {
//             for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + 1)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
//               for(int j = (wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier ; j<num_subchannel*num_data_subcarrier ; j++ )
// 		(intpower_[i][j] = power_per_subchannel); 
// 	    // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
//           
//             for(int i = wimaxHdr.phy_info.OFDMSymbol_offset+1 ; i<  ((wimaxHdr.phy_info.OFDMSymbol_offset) + wimaxHdr.phy_info.num_OFDMSymbol-1) ; i++)
// 	      for(int j = 0 ; j<num_subchannel*num_data_subcarrier ; j++ )
// 		intpower_[i][j] = power_per_subchannel;
// 
// 	    // last num_symbol_per_slot 
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset +wimaxHdr.phy_info.num_OFDMSymbol-1 ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset) + wimaxHdr.phy_info.num_OFDMSymbol/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = 0 ; j<((wimaxHdr.phy_info.num_subchannels - (num_subchannel - (wimaxHdr.phy_info.subchannel_offset)) )%num_subchannel)*num_data_subcarrier ; j++ )
// 		(intpower_[i][j] = power_per_subchannel); 
// 
//           }
// 	else
//           {
//           
// 	    for(int i = wimaxHdr.phy_info.OFDMSymbol_offset ; i<  (wimaxHdr.phy_info.OFDMSymbol_offset + 1)/*wimaxHdr.phy_info.num_OFDMSymbol)*/ ; i++)
// 	      for(int j = (wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier ; j<(((wimaxHdr.phy_info.subchannel_offset)*num_data_subcarrier) + wimaxHdr.phy_info.num_subchannels*num_data_subcarrier) ; j++ )
// 		(intpower_[i][j] = power_per_subchannel); 	
// 
//           }
//              
//       }
//   }
//    
// }





bool Wimax2BSNetDevice::IsCollision (const All_Headers *wimaxHdr_in, double power_subchannel)
{
  NS_LOG_FUNCTION_NOARGS ();
  All_Headers wimaxHdr; 
	Ptr<Packet> mover = getPacket();
	mover->AddHeader(*wimaxHdr_in);
	mover->RemoveHeader(wimaxHdr);
  bool collision = false; 
  uint32_t num_subchannel = getPhy()->getNumsubchannels (UL_);
  uint32_t num_symbol_per_slot = getPhy()->getSlotLength (UL_); 
   
  if(wimaxHdr.PIH.num_OFDMSymbol % num_symbol_per_slot == 0 )   // chk this condition , chk whether broacastcid reqd ir not. 
    // if((wimaxHdr.PIH.subchannel_offset + wimaxHdr.PIH.num_subchannels -1) > 30)
    //if(wimaxHdr.PIH.num_OFDMSymbol >= num_symbol_per_slot)
    {
      if(wimaxHdr.PIH.num_OFDMSymbol > num_symbol_per_slot) {
	  //numsymbols = (int)  ceil((((wimaxHdr.PIH.subchannel_offset + wimaxHdr.PIH.num_subchannels -1) - 30) % 30));
	  //numsymbols *= num_symbol_per_slot;  
	  // for the first num_symbol_per_slot symbols 
          for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset ; i<  (wimaxHdr.PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)
            {
              for(uint32_t j = (wimaxHdr.PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }
	      //              if(collision == true) return collision;  
	    }
	  // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
          for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset+num_symbol_per_slot ; i<  (wimaxHdr.PIH.OFDMSymbol_offset) + (wimaxHdr.PIH.num_OFDMSymbol-num_symbol_per_slot) ; i++)
            {
              for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 
	      //              if(collision == true) return collision; 
            } 
	  // last num_symbol_per_slot 
          for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset +wimaxHdr.PIH.num_OFDMSymbol-num_symbol_per_slot ; i<  wimaxHdr.PIH.OFDMSymbol_offset + wimaxHdr.PIH.num_OFDMSymbol/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)        
            {
	      for(uint32_t j = 1 ; j<=((wimaxHdr.PIH.num_subchannels - (num_subchannel - (wimaxHdr.PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision ;
		  }
	      //              if(collision == true) return collision;    
            }
	}
      else 
	{

	  for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset ; i<  (wimaxHdr.PIH.OFDMSymbol_offset + num_symbol_per_slot)/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)
            {
	      for(uint32_t j = (wimaxHdr.PIH.subchannel_offset) ; j< (wimaxHdr.PIH.subchannel_offset) + wimaxHdr.PIH.num_subchannels ; j++ )
                if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }    
	      //              if(collision == true) return collision;  
	    }  

	}

    }

  else

    {
      //numsymbols = (int)  ceil((((wimaxHdr.PIH.subchannel_offset + wimaxHdr.PIH.num_subchannels -1) - 30) % 30));
      //numsymbols *= num_symbol_per_slot;  
      // for the first num_symbol_per_slot symbols 
      if(wimaxHdr.PIH.num_OFDMSymbol > 1) 
	{
	  for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset ; i<  (wimaxHdr.PIH.OFDMSymbol_offset + 1)/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = (wimaxHdr.PIH.subchannel_offset) ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 
	      //	      if(collision == true) return collision; 
	    }	
	  // except the last num_symbol_per_slot and first num_symbol_per_slot whatever is thr
          
	  for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset+1 ; i<  ((wimaxHdr.PIH.OFDMSymbol_offset) + wimaxHdr.PIH.num_OFDMSymbol-1) ; i++)
	    {
	      for(uint32_t j = 1 ; j<=num_subchannel ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }  
	      //		 if(collision == true) return collision; 
	    }
	  // last num_symbol_per_slot 
          for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset +wimaxHdr.PIH.num_OFDMSymbol-1 ; i<  (wimaxHdr.PIH.OFDMSymbol_offset) + wimaxHdr.PIH.num_OFDMSymbol/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = 1 ; j<=((wimaxHdr.PIH.num_subchannels - (num_subchannel - (wimaxHdr.PIH.subchannel_offset)) )%num_subchannel) ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  }
	      //               if(collision == true) return collision; 
            }
	}
      else
	{
          
	  for(uint32_t i = wimaxHdr.PIH.OFDMSymbol_offset ; i<  (wimaxHdr.PIH.OFDMSymbol_offset + 1)/*wimaxHdr.PIH.num_OFDMSymbol)*/ ; i++)
	    {
	      for(uint32_t j = (wimaxHdr.PIH.subchannel_offset) ; j< wimaxHdr.PIH.subchannel_offset + wimaxHdr.PIH.num_subchannels ; j++ )
		if(basepower_[i][j] != power_subchannel)
		  {
                    collision = true;
		    debug2(" collision is true : i = %d j =%d basepower[i][j] = %.25f power_subchannel =%.25f \n" , i, j , basepower_[i][j], power_subchannel);		
		    return collision;
		  } 	
	      //                if(collision == true) return collision;  
	    }

	}
             
    }

  return collision; 

}

void Wimax2BSNetDevice::DoDispose() {
	Wimax2NetDevice::DoDispose();
}


}//namespace ns3
