/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef MAC802_16BS_H
#define MAC802_16BS_H

#include "ns3/wimax2-net-device.h"
#include "ns3/wimax2-control-agent.h"
#include "ns3/scanningstation.h"
#include "ns3/globalparams_wimax.h"
#include "ns3/queue.h"


//Some configuration elements to enable/disable features under development
#define BWREQ_PATCH

namespace ns3 {

/** Information about a new client */
struct new_client_t {
  int cid; //primary cid of new client
  new_client_t *next;
};




class T17Element;
LIST_HEAD (t17Element_list, T17Element);
struct t17_list_entry {
	T17Element *le_next;
	T17Element **le_prev;
};
/** Object to handle timer t17 */
class T17Element {
 public:
  T17Element (Ptr<Wimax2NetDevice> mac, Mac48Address index) {
    index_ = index;
    timer = new WimaxT17Timer (mac, index);
    timer->start (mac->macmib_->t17_timeout);
  }

  ~T17Element () { delete (timer); }

  Mac48Address index () { return index_; }
  int paused () { return timer->paused(); }
  int busy () { return timer->busy(); }
  void cancel () { timer->stop(); }

  // Chain element to the list
  inline void insert_entry(struct t17Element_list *head) {
    LIST_INSERT_HEAD(head, this, link);
  }
  
  // Return next element in the chained list
  T17Element* next_entry(void) const { return link.le_next; }

  // Remove the entry from the list
  inline void remove_entry() { 
    LIST_REMOVE(this, link); 
  }
 //protected: commented so pybindgen will work
  /*
   * Pointer to next in the list
   */
  t17_list_entry link;


 private:
  Mac48Address index_;
  WimaxT17Timer *timer;
};

class FastRangingInfo;
LIST_HEAD (fastRangingInfo_list, FastRangingInfo);
struct fast_ranging_list_entry {
	FastRangingInfo *le_next;
	FastRangingInfo **le_prev;
};


/** Store information about a fast ranging request to send */
class FastRangingInfo {
 public:
  FastRangingInfo (int frame, Mac48Address macAddr) {
    frame_ = frame;
    macAddr_ = macAddr;
  }

  int frame () { return frame_; }
  Mac48Address macAddr () { return macAddr_; }

  // Chain element to the list
  inline void insert_entry(struct fastRangingInfo_list *head) {
    LIST_INSERT_HEAD(head, this, link);
  }
  
  // Return next element in the chained list
  FastRangingInfo* next_entry(void) const { return link.le_next; }

  // Remove the entry from the list
  inline void remove_entry() { 
    LIST_REMOVE(this, link); 
  }
 //protected: commented so pybindgen will work (no protected member variables allowed)
  
  /*
   * Pointer to next in the list
   */
  fast_ranging_list_entry link;

 private:
  int frame_;
  Mac48Address macAddr_;
};


/**
 * Class implementing IEEE 802_16 State machine at the BS
 */ 
class Wimax2BSNetDevice : public Wimax2NetDevice {
  friend class Wimax2ControlAgent;
  friend class Wimax2BSScheduler;
 public:

static TypeId GetTypeId (void);

  Wimax2BSNetDevice();

  /**
   * Process packets going out
   * @param p The packet to transmit
   */
  void sendDown(Ptr<Packet> p);

  /**
   * Process packets going out
   * @param p The packet to transmit
   */
  void transmit(Ptr<Packet> p);
        
  /**
   * Process incoming packets 
   * @param p The received packet
   */
//  virtual void sendUp(Ptr<Packet> p);
//  virtual void sendUp(Ptr<Packet> p, uint8_t subchannel, Time time);

  /**
   *sam intpower_
   */ 
  //double intpower_[1024]; 
 
  /**
   * Process the packet after receiving last bit
   */
  //void receive();

  // fille the frame with power values 
  void addPowerinfo(All_Headers *wimaxHdr,double power, bool collision );

  //chk for collision 
  bool IsCollision (const All_Headers *wimaxHdr,double power_subchannel);

   
  /**
   * Process the packet after receiving last bit
   *@param p - the packet to be received  RPI
   */
  void receive(Ptr<Packet> p);

#ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
  /* 
   * Configure/Request configuration
   * The upper layer sends a config object with the required 
   * new values for the parameters (or PARAMETER_UNKNOWN_VALUE).
   * The MAC tries to set the values and return the new setting.
   * For examples if a MAC does not support a parameter it will
   * return  PARAMETER_UNKNOWN_VALUE
   * @param config The configuration object
   */ 
  void link_configure (link_parameter_config_t* config);

#endif

//   /**
//    * Determines if the packet payload should be processed by the MAC layer or sent up
//    * @param p - the packet in question
//    */
//   bool isMACPacket(Ptr<Packet> p);

//protected: would start here but the python bindings scanner can't deal with protected variables

  /**
   * Indicates if it is time to send a DCD message
   */
  bool sendDCD;

  /**
   * DL configuration change count
   */
  int dlccc_;
  
  /**
   * Indicates if it is time to send a UCD message
   */
  bool sendUCD;

  /**
   * UL configuration change count
   */
  int ulccc_;


/**
 * The traces used by PyViz that are fired when the MAC layer either sends packets down or gets packets from below
 */
 TracedCallback<Ptr<const Packet>, const Mac48Address& > m_traceRx;
 TracedCallback<Ptr<const Packet>, const Mac48Address& > m_traceTx;


   
 protected:

  /**
   * init the timers and state
   */
  void init ();

  /**
   * Initialize default connection
   */
  void init_default_connections ();

  /**
   * Update the given timer and check if thresholds are crossed
   * @param watch the stat watch to update
   * @param value the stat value
   */
 // void update_watch (StatWatch *watch, double value);

  /**
   * Update the given timer and check if thresholds are crossed
   * @param watch the stat watch to update
   * @param size the size of packet received
   */
 // void update_throughput (ThroughputWatch *watch, double size);

#ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
  /**
   * Poll the given stat variable to check status
   * @param type The link parameter type
   */
  void poll_stat (link_parameter_type_s type);
#endif
   
  /**
   * Called when a timer expires
   * @param The timer ID
   */
  virtual void expire (timer_id id);

  /**
   * Start a new DL subframe
   */
  virtual void start_dlsubframe ();

  /**
   * Process cdma bandwidth request
   * @param p The request
   */
  void process_cdma_req (Ptr<Packet> p);

  /**
   * Start a new UL subframe
   */
  virtual void start_ulsubframe ();

  /** 
   * Finds out if the given station is currently scanning
   * @param nodeid The MS id
   */
  bool isPeerScanning (Mac48Address nodeid);

  /**
   * Set the control agent
   * @param agent The control agent
   */
  //-#- @agent(transfer_ownership=true) -#-
  void setCtrlAgent (Wimax2ControlAgent *agent);

  /** Add a new Fast Ranging allocation
   * @param time The time when to allocate data
   * @param macAddr The MN address
   */
  void addNewFastRanging (double time, Mac48Address macAddr);

  /**
   * Send a scan response to the MN
   * @param rsp The response from the control
   * @cid The CID for the MN
   */
  //-#- @rsp(transfer_ownership=true) -#-
  void send_scan_response (mac802_16_mob_scn_rsp_frame *rsp, int cid);
  

 protected:
	virtual void DoDispose(); 

 private:

  //Begin RPI
  /**
   * The function is used to process the MAC PDU when ARQ,Fragmentation and Packing are enabled
   * @param con The connection by which it arrived
   * @param p The packet to process
   */
  void process_mac_pdu_witharqfragpack (Connection *con, Ptr<Packet> p);
  //End RPI

  /**
   * Process a MAC type packet
   * @param con The connection by which it arrived
   * @param p The packet to process
   */
  void process_mac_packet (Connection *con, Ptr<Packet> p);

  /**
   * Process a RNG-REQ message
   * @param frame The ranging request information
   */
  void process_ranging_req (Ptr<Packet> p);

  /**
   * Process bandwidth request
   * @param p The request
   */
  void process_bw_req (Ptr<Packet> p);

  /**
   * Process bandwidth request
   * @param p The request
   */
  void process_reg_req (Ptr<Packet> p);

  /**
   * Process handover request
   * @param req The request
   */
  void process_msho_req (Ptr<Packet> req);
 
  /**
   * Process handover indication
   * @param p The indication
   */
  void process_ho_ind (Ptr<Packet> p);
 
  /**
   * Send a neighbor advertisement message
   */
  void send_nbr_adv ();

  /**
   * Add a new timer 17 for client
   * @param index The client address
   */
  void addtimer17 (Mac48Address index);
  
  /**
   * Cancel and remove the timer17 associated with the node
   * @param index The client address
   */
  void removetimer17 (Mac48Address index);

  
  /**
   * Pointer to the head of the list of nodes that should
   * perform registration
   */
  struct t17Element_list t17_head_;

  /**
   * Pointer to the head of the list of clients
   */
  struct new_client_t *cl_head_;

  /**
   * Pointer to the tail of the list of clients
   */
  struct new_client_t *cl_tail_;

  /**
   * The index of the last SS that had bandwidth allocation
   */
  int bw_node_index_;
  
  /**
   * The node that had the last bandwidth allocation
   */
  PeerNode *bw_peer_; 

  /**
   * Timer for DCD
   */
  WimaxDCDTimer *dcdtimer_; 

  /**
   * Timer for UCD
   */
  WimaxUCDTimer *ucdtimer_;

  /**
   * Timer for MOB-NBR_ADV messages
   */
  WimaxMobNbrAdvTimer *nbradvtimer_;

  /** 
   * List of station in scanning 
   */
  struct scanningStation_list scan_stations_;
  
  /**
   * The Wimax control for BS synchronization
   */
  Wimax2ControlAgent *ctrlagent_;
  
  /**
   * List of the upcoming Fast Ranging allocation 
   */
  struct fastRangingInfo_list fast_ranging_head_;

  /**
   * Used to hit the PointToPoint style tracing hooks for the visualizer
   */
  Ptr<Queue> m_queue;

};

}//namespace ns3

#endif //MAC802_16BS_H

