
#include "ns3/wimax2-channel.h"
#include "ns3/wimax2-net-device.h"
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/node.h"
#include "ns3/log.h"
#include "ns3/mobility-model.h"
#include "ns3/cost231-propagation-loss-model.h"
using std::cout;

NS_LOG_COMPONENT_DEFINE ("Wimax2Channel");

#define SPEED_OF_LIGHT	 300000000 //meters / second ... should really be 299792458, but this is commonly used


namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (Wimax2Channel);


TypeId 
Wimax2Channel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Wimax2Channel")
    .SetParent<Channel> ()
    .AddConstructor<Wimax2Channel> ()
    ;
  return tid;
}

Wimax2Channel::Wimax2Channel () : num_dl_symbol(0), num_ul_symbol(0), num_dl_subch(0), num_ul_subch(0), isInitialized(false)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_loss = CreateObject<Cost231PropagationLossModel> ();
  collisions_ = NULL;
  frame_duration = Seconds(-1);
}

void Wimax2Channel::DoDispose() {
	m_devices.clear();
	m_loss = 0;
	Channel::DoDispose();
}


void Wimax2Channel::Send (Ptr<Packet> p, Ptr<Wimax2Phy> Sender, uint8_t subchannel) {
	NS_LOG_FUNCTION (p << Sender << subchannel);
	Ptr<MobilityModel> transmitter_mobility = 0, receiver_mobility = 0;
	transmitter_mobility = Sender->GetMac()->GetNode()->GetObject<MobilityModel>();
	debug15("At %f, on subchannel %d, %s is sending a block of 1 Symbol-Time by 1 Subchannel of size %u Bytes over the channel\n", NOW, subchannel, Sender->GetMac()->addr_s(), p->GetSize());
	for (std::vector<Ptr<Wimax2Phy> >::const_iterator i = m_devices.begin(); i != m_devices.end(); ++i) {
		Time travel = Seconds(0);
		Ptr<Wimax2Phy> tmp = *i;
		//we don't want to "transmit" back to the node it came from
		if (tmp == Sender) continue;

		receiver_mobility = tmp->GetMac()->GetNode()->GetObject<MobilityModel>();
		if (transmitter_mobility != 0 && receiver_mobility != 0) {
			travel = Seconds (transmitter_mobility->GetDistanceFrom(receiver_mobility) / SPEED_OF_LIGHT);
			//RxdBm = m_loss->CalcRxPower (TxdBm, transmitter_mobility, receiver_mobility);
		}
		Simulator::ScheduleWithContext (tmp->GetMac()->GetNode()->GetId(), travel, &Wimax2Phy::ReceiveSlot, tmp, p->Copy (), subchannel);
			
	}
}


//void
//Wimax2Channel::Send (Ptr<Packet> p, uint16_t protocol, Mac48Address to, Mac48Address from, Ptr<Wimax2Phy> sender)
//{
//	//Update collision information
//	PHY_INFO_Header pih;
//	p->PeekHeader(pih);
//	bool collision = false;
//	if (pih.direction == 0) { //DL Subframe
//		for (uint32_t i = pih.OFDMSymbol_offset; !collision && i < num_dl_symbol && i < pih.OFDMSymbol_offset + pih.num_OFDMSymbol; i++)
//			for (uint32_t j = pih.subchannel_offset; !collision && j < num_dl_subch && j < pih.subchannel_offset + pih.num_subchannels; j++)
//				if (collisions_[i][j]) 
//					collision = true;
//				else
//					collisions_[i][j] = true;
//		
//
//	} else { //UL Subframe
//
//
//	}
//
//	//get positioning information
//	Ptr<MobilityModel> transmitter_mobility = 0, receiver_mobility = 0;
//	transmitter_mobility = sender->GetMac()->GetNode()->GetObject<MobilityModel>();
////	double RxdBm = 0;
//	
//	//transmit the packet, so that it is recieved at appropriate power by all other devices (many of which may ignore it if they are not scheduled to receive at that time)
//	NS_LOG_FUNCTION (p << protocol << to << from << sender);
//	for (std::vector<Ptr<Wimax2Phy> >::const_iterator i = m_devices.begin (); i != m_devices.end (); ++i)
//	    {
//	      Time travel = Seconds(0);
//	      Ptr<Wimax2Phy> tmp = *i;
//	      if (tmp == sender) continue;
//
//		receiver_mobility = tmp->GetMac()->GetNode()->GetObject<MobilityModel>();
//		if (transmitter_mobility != 0 && receiver_mobility != 0) {
//			travel = Seconds (transmitter_mobility->GetDistanceFrom(receiver_mobility) / SPEED_OF_LIGHT);
//			//RxdBm = m_loss->CalcRxPower (TxdBm, transmitter_mobility, receiver_mobility);
//		}
//
//	      	p->RemoveHeader(pih);
//		cout<<"SENDING PACKET OF SIZE " << p->GetSize() << " OVER CHANNEL TO MAC: " <<tmp->GetMac()->addr_s() << " FROM MAC: " << from << std::endl;
//		p->AddHeader(pih);
//		//tmp->GetMac()->getPhy()->Receive(p->Copy(), protocol, from, to, sender->GetMac(), RxdBm);
//		//Simulator::Schedule(Seconds(0), &Wimax2Phy::Receive, tmp->GetMac()->getPhy(), p->Copy(), protocol, from, to, sender->GetMac(), 0.0);
//	       Simulator::ScheduleWithContext (tmp->GetMac()->GetNode ()->GetId (), travel, &Wimax2Phy::Receive, tmp->GetMac()->getPhy(), p->Copy (), protocol, from, to, sender->GetMac());
//	    }
//}
//
void 
Wimax2Channel::Add (Ptr<Wimax2Phy> device)
{
  m_devices.push_back (device);
  NS_LOG_FUNCTION_NOARGS ();
}

uint32_t 
Wimax2Channel::GetNDevices (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_devices.size ();
}
Ptr<NetDevice> 
Wimax2Channel::GetDevice (uint32_t i) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_devices[i]->GetMac();
}

Ptr<Wimax2Phy>
Wimax2Channel::GetPhy(uint32_t i) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_devices[i];
}

void Wimax2Channel::initializeCollisionDetection() {
	if (isInitialized) return;
	Ptr<Wimax2Phy> p = m_devices[0];
	Ptr<Wimax2NetDevice> m = p->GetMac();
	if (p == 0) return;
	
	num_dl_symbol = m->getMaxDlduration();
	num_ul_symbol = m->getMaxUlduration();
	num_dl_subch = p->getNumsubchannels(DL_);
	num_ul_subch = p->getNumsubchannels(UL_);
	frame_duration = Seconds(m->getFrameDuration());

	uint32_t total_subch = num_dl_subch > num_ul_subch ? num_dl_subch : num_ul_subch;
	uint32_t total_symb = num_dl_symbol + num_ul_symbol;

	collisions_ = new bool*[total_symb];
	for (uint32_t i = 0; i < total_symb; i++) {
		collisions_[i] = new bool[total_subch];
		//we just set the upper few (5 in 10 MHz channel) subchannels / symbol in DL_frame that don't exist to already in use so anything trying to transmit there will cause errors
		for (uint32_t j = 0; j < total_subch; j++) 
			(i <= num_dl_symbol && j > num_dl_subch) ? collisions_[i][j] = false : collisions_[i][j] = true;
	}
	
	Simulator::ScheduleWithContext(Simulator::GetContext(), frame_duration, &Wimax2Channel::resetCollisionMonitor, this);
		
	
}

void Wimax2Channel::resetCollisionMonitor() {
	Ptr<Wimax2Phy> p = m_devices[0];
	uint32_t total_subch = num_dl_subch > num_ul_subch ? num_dl_subch : num_ul_subch;
	uint32_t total_symb = num_dl_symbol + num_ul_symbol;
	if (p == 0) return;
	for (uint32_t i = 0; i < total_symb; i++)
		for (uint32_t j = 0; j < total_subch; j++)
			(i <= num_dl_symbol && j > num_dl_subch) ? collisions_[i][j] = false : collisions_[i][j] = true;
			
	Simulator::ScheduleWithContext(Simulator::GetContext(), frame_duration, &Wimax2Channel::resetCollisionMonitor, this);
}

void Wimax2Channel::sync_preamble() {
	for (std::vector<Ptr<Wimax2Phy> >::const_iterator i = m_devices.begin (); i != m_devices.end (); ++i) {
		Ptr<Wimax2Phy> p = *i;
		p->GetMac()->SynchronizeFrameStart(Simulator::Now());
	}
}


} // namespace ns3
