/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef Wimax2SSNetDevice_H
#define Wimax2SSNetDevice_H

#include "wimax2-net-device.h"
#include "ns3/connectionmanager.h"
#include "ns3/wimax2-serviceflowhandler.h"
#include "ns3/wimaxneighborentry.h"
#include "ns3/ptr.h"

namespace ns3 {

#define BS_NOT_CONNECTED -1 //bs_id when MN is not connected

/** Defines the state of the MAC */
enum Mac802_16State {
  MAC802_16_DISCONNECTED,
  MAC802_16_WAIT_DL_SYNCH,
  MAC802_16_WAIT_DL_SYNCH_DCD,
  MAC802_16_UL_PARAM,
  MAC802_16_RANGING,
  MAC802_16_WAIT_ANON_RNG_RSP,
  MAC802_16_WAIT_RNG_ALLOC,
  MAC802_16_WAIT_RNG_RSP,
//  MAC802_16_SEND_ACTUAL_RNG_REQ,
  MAC802_16_REGISTER,
  MAC802_16_SCANNING,
  MAC802_16_CONNECTED
};

/** Data structure to store MAC state */
struct state_info {
  Mac802_16State state; 
  int bs_id;
  double frameduration;
  int frame_number;
  int channel;
  ConnectionManager * connectionManager;
  ServiceFlowHandler * serviceFlowHandler;
  struct peerNode_list *peer_list;
  int nb_peer;
};

/** The sub state (while connected) */
enum ss_sub_state {
  NORMAL,           //Normal state
  SCAN_PENDING,     //Normal period but pending scanning to start/resume
  SCANNING,         //Currently scanning
  HANDOVER_PENDING, //Normal state but handover to start
  HANDOVER          //Executing handover
};

/** Data structure to store scanning information */
struct scanning_structure {
  struct mac802_16_mob_scn_rsp_frame *rsp; //response from BS
  struct sched_state_info scan_state;     //current scanning state
  struct sched_state_info normal_state;   //backup of normal state
  int iteration;                          //current iteration
  WimaxScanIntervalTimer *scn_timer_;     //timer to notify end of scanning period
  int count;                              //number of frame before switching to scanning  
  ss_sub_state substate;
  WimaxNeighborEntry *nbr; //current neighbor during scanning or handover
  //arrays of rdv timers
  WimaxRdvTimer *rdv_timers[2*MAX_NBR];
  int nb_rdv_timers;
  //handoff information
  Mac48Address serving_bsid;
  int handoff_timeout; //number frame to wait before executing handoff
};

/**
 * Class implementing IEEE 802_16 State Machine at the SS
 */ 
class Wimax2SSNetDevice : public Wimax2NetDevice {
  friend class BwRequest;

 public:

  Wimax2SSNetDevice(Time boot);
  Wimax2SSNetDevice();
  
  static TypeId GetTypeId (void);
  
  /**
   * Interface with the TCL script
   * @param argc The number of parameter
   * @param argv The list of parameters
   */
  //int command(int argc, const char*const* argv);

  /**
   * Set the mac state
   * @param state The new mac state
   */  
  void setMacState (Mac802_16State state);

  /**
   * Return the mac state
   * @return The new mac state
   */  
  Mac802_16State getMacState ();

  /**
   * Creates a snapshot of the MAC's state and reset it
   * @return The snapshot of the MAC's state
   */
  state_info *backup_state ();

  /**
   * Restore the state of the Mac
   * @param state The state to restore
   */
 //-#- @state(transfer_ownership=true) -#-
  void restore_state (state_info *state);  

  /**
   * Process packets going out
   * @param p The packet to transmit
   */
  void transmit(Ptr<Packet> p);

  inline uint8_t get_diuc() { return default_diuc_; }

  /**
   * Process packets going out
   * @param p The packet to transmit
   */
  void sendDown(Ptr<Packet> p);
        
  /**
   * Process incoming packets 
   * @param p The received packet
   * Note - this is no longer implemented by subclasses
   */
  //virtual void sendUp(Ptr<Packet> p, uint8_t subchannel, Time r_time);

  /*
   * Process the packet after receiving last bit
   */
  //void receive();

   /**
   * Process the packet after receiving last bit
   *@param p - the packet to be received  RPI
   */
  void receive(Ptr<Packet> p);

  /* initial random propagation channel
   *rpi
   */
  int  GetInitialChannel();
 
 // fille the frame with power values 
 //-#- @wimaxHdr(transfer_ownership=true) -#-
  void addPowerinfo(All_Headers *wimaxHdr,double power, bool collision );

  //chk for collision

 //-#- @wimaxHdr(transfer_ownership=true) -#-
  bool IsCollision (const All_Headers *wimaxHdr,double power_subchannel);
  
#ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
  /* 
   * Configure/Request configuration
   * The upper layer sends a config object with the required 
   * new values for the parameters (or PARAMETER_UNKNOWN_VALUE).
   * The MAC tries to set the values and return the new setting.
   * For examples if a MAC does not support a parameter it will
   * return  PARAMETER_UNKNOWN_VALUE
   * @param config The configuration object
   */ 
 //-#- @config(transfer_ownership=true) -#-
  void link_configure (link_parameter_config_t* config);
  
  /*
   * Disconnect from the PoA
   */
  void link_disconnect ();
  
  /*
   * Connect to the PoA
   * @param poa The address of PoA
   */
  void link_connect (int poa);
    
  /*
   * Set the operation mode
   * @param mode The new operation mode
   * @return true if transaction succeded
   */
  bool set_mode (mih_operation_mode_t mode); 
#endif

  /*
   * Scan channel
   * @param req the scan request information
   */
 //-#- @req(transfer_ownership=true) -#-
  void link_scan (void *req);
  
  /** 
   * ns3 function - true if connected, false otherwise
   */
  bool IsLinkUp() {return getMacState() == MAC802_16_CONNECTED;}

  /**
   * Returns the MAC address of the BS this SS is currently attached to (or 00:00:00:00:00:00 if it is unattached)
   */
 inline Mac48Address getBSAddr() {return bs_mac_addr;}
 
  /**
   * Stores the MAC Address of the BS to which this SS is currently attached
   */
 inline void setBSAddr(Mac48Address adr) {bs_mac_addr = adr;}
  
  /**
   * used to notify the netdevice to start or stop ignoring incoming packets based on which subframe we're in (in UL subframe we ignore incoming data)
   */

  inline void StartSubframe(bool new_subframe_is_dl) {
  	in_dl_subframe = new_subframe_is_dl; 
  }
  
  /**
   *Processes a Frame Prefix Header after grabbing it from the received grid
   * public so we can call this from the scheduler easily
   */
  void process_FCH();

  protected:

  virtual void DoDispose();
  
  /**
   * init the timers and state
   */
  void init ();
  
  /**
   * Initialize default connection
   */
  void init_default_connections ();
  
  /**
   * Called when lost synchronization
   */
  void lost_synch ();
  
  /**
   * Start a new DL subframe
   */
  void start_dlsubframe ();
  
  /**
   * Start a new UL subframe
   */
  void start_ulsubframe ();
  
  /**
   * Start/Continue scanning
   */
  void resume_scanning ();
  
  /**
   * Pause scanning
   */
  void pause_scanning ();
  
  
  
  /**
   * Update the given timer and check if thresholds are crossed
   * @param watch the stat watch to update
   * @param value the stat value
   */
//  void update_watch (StatWatch *watch, double value); //CAT- uncomment
  
  /**
   * Update the given timer and check if thresholds are crossed
   * @param watch the stat watch to update
   * @param size the size of packet received
   */
//  void update_throughput (ThroughputWatch *watch, double size);//CAT- uncomment

#ifdef USE_802_21 //Switch to activate when using 802.21 modules (external package)
  /**
   * Poll the given stat variable to check status
   * @param type The link parameter type
   */
  void poll_stat (link_parameter_type_s type);
  
#endif
  
  /**
   * Called when a timer expires
   * @param The timer ID
   */
  virtual void expire (timer_id id);


 private:
   //incoming packet block storage
   Ptr<Packet> **incoming;

   /**
    * ns3 tracing hooks
    */

    TracedCallback<Ptr<const Packet>, const Mac48Address& > m_traceRx;
    TracedCallback<Ptr<const Packet>, const Mac48Address& > m_traceTx;

    TracedCallback<Ptr<const Packet>, Mac48Address, int> m_traceSSRx; //int is cid
    TracedCallback<Ptr<const Packet>, Mac48Address, int> m_traceSSTx; //int is cid

    TracedCallback<Ptr<const Packet > > m_ssTxTrace; //when packets come in from higher layers
    
    TracedCallback<Ptr<const Packet> > m_ssRxTrace; //when packets are forwarded up to higher layers
    TracedCallback<Ptr<const Packet> > m_ssTxDropTrace; //packets from higher layer dropped before transmission
    TracedCallback<Ptr<const Packet> > m_ssRxDropTrace; //packets from lower layer dropped before being passed up 

   /**
	* Used to track current subframe (UL or DL).  We ignore any packet recieved during UL subframe
    */
	bool in_dl_subframe;
 
   /**
    * Used to determine whether it is safe to process the UL_MAP, or if that should be delayed very slightly
    * Note - if the simulation is multi-threaded, this system could create a race condition
    */
   bool new_dlmap_processed;
   
   //used to prevent processing retransmissions of UL_MAP and DL_MAP
   bool this_frame_ulmap_processed;
   bool this_frame_dlmap_processed;
   
   //resets this_frame_[d/u]lmap_processed at the start of a new frame
   void reset_repetition_detectors();

//Begin RPI
  /**
    * The function is used to process the MAC PDU when ARQ,Fragmentation and Packing are enabled
    * @param con The connection by which it arrived
    * @param p The packet to process
    */
  void process_mac_pdu_witharqfragpack (Connection *con, Ptr<Packet> p);
//End RPI

  /**
   * Process a MAC type packet
   * @param p The packet to process
   */
  void process_mac_packet (Ptr<Packet> p);

  
  /**
   * Process a DL_MAP message
   * @param frame The dl_map information
   */
  void process_dl_map (DL_MAP_FRAME_HEADER *frame_h, double timeSent);

  /**
   * Process a DCD message
   * @param frame The dcd information
   */
  void process_dcd (DCD_FRAME_HEADER *frame_h);

  /**
   * Process a UL_MAP message
   * @param frame The ul_map information
   */
  void process_ul_map (UL_MAP_FRAME_HEADER *frame_h, double txtime);

  /**
   * Process a UCD message
   * @param frame The ucd information
   */
  void process_ucd (UCD_FRAME_DATA *frame_h);

  /**
   * Process a ranging response message 
   * @param frame The ranging response frame
   */
  void process_ranging_rsp (RANGING_RESPONSE_FRAME *frame_h);
  
  /**
   * Process a registration response message 
   * @param frame The registration response frame
   */
  void process_reg_rsp (REGISTRATION_RESPONSE_FRAME *frame_h);

  /**
   * Schedule a ranging
   */
  void init_ranging ();

  /**
   * Prepare to send a registration message
   */
  void send_registration ();

  /**
   * Send a scanning message to the serving BS
   */
  void send_scan_request ();

  /**
   * Process a scanning response message 
   * @param frame The scanning response frame
   */
  void process_scan_rsp (mac802_16_mob_scn_rsp_frame *frame);  

  /**
   * Process a BSHO-RSP message 
   * @param frame The handover response frame
   */
  void process_bsho_rsp (BSHO_RESPONCE_FRAME_HEADER *frame_h); 
  
  /**
   * Process a BSHO-RSP message 
   * @param frame The handover response frame
   */
  void process_nbr_adv (NEIGHBOR_ADVERTISEMENT_FRAME_HEADER *frame_h);  

  /**
   * Send a MSHO-REQ message to the BS
   */
  void send_msho_req ();

  /**
   * Check rdv point when scanning
   */
  void check_rdv ();

  /**
   * Set the scan flag to true/false
   * param flag The value for the scan flag
   */
  void setScanFlag(bool flag);

  /**
   * return scan flag
   * @return the scan flag
   */
  bool isScanRunning();

  /**
   * A packet buffer used to temporary store a packet 
   * received by upper layer. Used during scanning
   */
  Ptr<Packet> pktBuf_;
  
  /*
   * The state of the MAC
   */
  Mac802_16State state_;
  
  /**
   * Current number of registration retry
   */
  uint32_t nb_reg_retry_;
  
  /**
   * Current number of scan request retry
   */
  uint32_t nb_scan_req_;
  
  /**
   * Timers
   */
  WimaxT1Timer  *t1timer_;
  WimaxT2Timer  *t2timer_;
  WimaxT6Timer  *t6timer_;
  WimaxT12Timer *t12timer_;
  WimaxT21Timer *t21timer_;
  WimaxLostDLMAPTimer *lostDLMAPtimer_;
  WimaxLostULMAPTimer *lostULMAPtimer_;
  WimaxT44Timer *t44timer_;
  
  /** 
   * The scanning information
   */
  struct scanning_structure *scan_info_; 
  
  /**
   * Indicates if scan has been requested
   */
  bool scan_flag_;
  
  /**
   * Default DIUC to use.
   */
  uint8_t default_diuc_;

  /**
   *The MAC address of the BS this SS is currently attached to (or 00:00:00:00:00:00 if it's not attached)
   */
  Mac48Address bs_mac_addr;

  /**
   * This is used to hit the hooks for PyViz
   */
  Ptr<Queue> m_queue;

  /**
   * Used to determine the first block recieved in a new frame
   * so we can find the DL_MAP
   */
  Time start_current_frame;

};

}//namespace

#endif //Wimax2SSNetDevice_H

