/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef NEIGHBORENTRY_H
#define NEIGHBORENTRY_H

#include "ns3/framemap.h"
#include "ns3/wimax2-timer.h"


namespace ns3 {
/* Data structure used to store Mac state information during scanning */
struct sched_state_info {
  struct state_info *state_info; 
  FrameMap *map;
  WimaxT1Timer  *t1timer;
  WimaxT2Timer  *t2timer;
  WimaxT6Timer  *t6timer;
  WimaxT12Timer *t12timer;
  WimaxT21Timer *t21timer;
  WimaxLostDLMAPTimer *lostDLMAPtimer;
  WimaxLostULMAPTimer *lostULMAPtimer;
  WimaxT44Timer *t44timer;
};

/**
 * Store information about a neighboring BS.
 */
class WimaxNeighborEntry {
 public:
  /** 
   * Constructor
   */
  WimaxNeighborEntry (Mac48Address id);

  /**
   * Destructor
   */
  ~WimaxNeighborEntry ();

  /**
   * Return the address of the neighbor of this entry
   * @return the address of the neighbor of this entry
   */
  Mac48Address getID ();

  /**
   * Set the neighbor advertisement message
   * @param frame The advertisment message
   */
  void setNbrAdvMessage (mac802_16_nbr_adv_info *frame);
  
  /**
   * Return the neighbor advertisement message
   * @param frame The advertisment message
   */
  mac802_16_nbr_adv_info * getNbrAdvMessage ();

  /**
   * Set the DCD message received during scanning
   * @param dcd the DCD message received
   */
  void setDCD (DCD_FRAME_HEADER *frame);

  /**
   * Get the DCD message received during scanning
   * @return the DCD message received
   */
  DCD_FRAME_HEADER *getDCD ();  

  /**
   * Set the UCD message received during scanning
   * @param dcd the DCD message received
   */
  void setUCD (UCD_FRAME_DATA *frame);

  /**
   * Get the DCD message received during scanning
   * @return the DCD message received
   */
  UCD_FRAME_DATA *getUCD ();  

  /**
   * Set the UCD message received during scanning
   * @param dcd the DCD message received
   */
  void setRangingRsp (RANGING_RESPONSE_FRAME *frame);

  /**
   * Get the DCD message received during scanning
   * @return the DCD message received
   */
  RANGING_RESPONSE_FRAME *getRangingRsp ();  

  /**
   * Mark the neighbor as being detected
   * @param detected indicate if the neighbor has been detected
   */
  void setDetected (bool detected);

  /**
   * Indicates the neighbor as being detected
   * @return indication if the neighbor has been detected
   */
  bool isDetected ();

  /**
   * Set the MAC state associated with this neighbor
   * @param state
   */
  //void setState (sched_state_info *state);

  /**
   * Get the MAC state associated with this neighbor
   * @return the MAC state associated with this neighbor
   */
  sched_state_info * getState ();

 private:
  /**
   * The MAC address of neighbor
   */
  Mac48Address id_;

  /**
   * The neighbor info adv message
   */
  struct mac802_16_nbr_adv_info *nbr_adv_;

  /**
   * The DCD message received during scanning
   */
  DCD_FRAME_HEADER *dcd_;

  /**
   * The DCD message received during scanning
   */
  UCD_FRAME_DATA *ucd_;

  /** 
   * The ranging response received during scanning
   */
  RANGING_RESPONSE_FRAME *rng_rsp_;

  /**
   * Save the state
   */
  struct sched_state_info state_;

  /**
   * Indicate if the neighbor has been detected
   */
  bool detected_;
};

}//namespace ns3  
#endif
