/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "subframe.h"
#include "contentionslot.h"
#include "ulsubframetimer.h"
#include "dlsubframetimer.h"

namespace ns3 {

/*** Functions of super class ***/

/**
 * Constructor
 * @param map The frame 
 */
SubFrame::SubFrame (FrameMap *map)
{
  assert (map);
  map_ = map;
  ccc_ = 0;
  nbProfile_ = 0;
  LIST_INIT (&profile_list_);
  debug10("profile_list.lh_first: %p\n", &profile_list_);//.lh_first);
}

/**
 * Destructor
 */
SubFrame::~SubFrame ()
{
  removeProfiles();
}

/**
 * Add a profile
 * @param f The frequency of the profile
 * @param enc The encoding used in the profile
 * @return a new profile with the given caracteristics
 */
Profile * SubFrame::addProfile (int f, Ofdm_mod_rate enc)
{
  Profile *p = new Profile (this, f, enc);
  p->insert_entry (&profile_list_);
  nbProfile_++;
  incrCCC();
  return p;
}

/**
 * Remove a profile
 * @param p The profile to remove
 * @return a new profile with the given caracteristics
 */
void SubFrame::removeProfile (Profile *p)
{
  p->remove_entry ();
  nbProfile_--;
  incrCCC();
}

/**
 * Remove all profiles
 */
void SubFrame::removeProfiles ()
{
  for (Profile *p = profile_list_.lh_first; p ; p=profile_list_.lh_first) {
    removeProfile (p);
    delete (p);
  }
}

/**
 * Return the profile with the given IUC
 * @return the profile with the given IUC
 */
Profile * SubFrame::getProfile (int iuc)
{
  Profile *p;
  for (p = profile_list_.lh_first; p ; p=p->next_entry()) {
    if (p->getIUC()==iuc)
      return p;
  }
    return NULL;
}

/**
 * Return the number of profiles for this subframe
 */
int SubFrame::getNbProfile () 
{ 
  return nbProfile_; 
}

/**
 * Return the head of the profile list
 */
Profile * SubFrame::getFirstProfile () 
{ 
  return profile_list_.lh_first; 
}

/**
 * Return the Configuration Change count
 */
int SubFrame::getCCC () 
{ 
  return ccc_; 
}

/**
 * Increment the configuration change count
 * The CCC is modulo 256.
 */
void SubFrame::incrCCC ()
{
  ccc_ = (ccc_+1)%256;
}

/*** end of super class ***/

/**
 * Constructor
 * @param map The frame 
 */
DlSubFrame::DlSubFrame (FrameMap *map): SubFrame(map)
{
  phypdu_ = new DlPhyPdu (map , 0); //no preamble by default
  timer_ = new DlSubFrameTimer (this);
}

/**
 * Destructor
 */
DlSubFrame::~DlSubFrame ()
{
  delete (phypdu_);
  delete (timer_);
}

void DlSubFrame::replaceTimer() {
	delete(timer_);
	timer_ = new DlSubFrameTimer(this);
}

DlSubFrameTimer* DlSubFrame::replaceTimerWithoutDelete() {
	DlSubFrameTimer* old = timer_;
	timer_ = new DlSubFrameTimer(this);
	return old;
}

void DlSubFrame::deleteSavedTimer(DlSubFrameTimer *in) {
	delete in;
}

void DlSubFrame::replaceTimer(double time) {
	delete(timer_);
	timer_ = new DlSubFrameTimer(this);
	timer_ ->start(time);
}

/*** class UlSubFrame ***/

/**
 * Constructor
 * @param map The frame 
 */
UlSubFrame::UlSubFrame (FrameMap *map): SubFrame(map)
{
//  LIST_INIT (&phypdu_list_);
  //phypdu_list_.size() =0;
  timer_ = new UlSubFrameTimer (this);
  bw_req_ = new BwContentionSlot (map);
  ranging_ = new RngContentionSlot (map);
}

BwContentionSlot* UlSubFrame::getBw_req() {
	return bw_req_;
}

/**
 * Destructor
 */
UlSubFrame::~UlSubFrame ()
{
  for (std::list<PhyPdu*>::iterator i = phypdu_list_.begin(); i != phypdu_list_.end(); i++) delete *i;
  phypdu_list_.clear();
  delete (timer_);
  delete (bw_req_);
  delete (ranging_);
}


PhyPdu * UlSubFrame::getFirstPdu() {
	return phypdu_list_.front();
}

std::list<PhyPdu*> * UlSubFrame::getPduList() {
	return &phypdu_list_;
}


/**
 * Add an uplink phy pdu
 * @param pos The position of the PhyPdu in the subframe
 * @param preamble The size of the preamble in units of XX
 * @return newly created PhyPdu
 */
PhyPdu * UlSubFrame::addPhyPdu (uint32_t pos, int preamble)
{
  assert (pos >= 0 && pos <= phypdu_list_.size() );
  UlPhyPdu *p = new UlPhyPdu (map_, preamble);
  if (pos==0) {
    phypdu_list_.push_front(p);
   } else {
    uint32_t index = 0;
    std::list<PhyPdu*>::iterator ps = phypdu_list_.begin();
    while (index < pos) {
      index++;
      ps++;
    }
    phypdu_list_.insert(ps,p);
  }
  std::list<PhyPdu*>::iterator ps = phypdu_list_.begin();
  return p;
}

/**
 * Remove a Phy PDU
 * @param pdu The Phy PDU to remove
 */
void UlSubFrame::removePhyPdu (PhyPdu *p)
{
  for (std::list<PhyPdu*>::iterator i = phypdu_list_.begin(); i != phypdu_list_.end(); i++) {
  	if (*i == p) {
		phypdu_list_.erase(i);
		return;
	}
   }
  //phypdu_list_.size()--;
}

uint32_t UlSubFrame::getNbPdu() {
	return getPduList()->size();
}

/**
 * Return the burst located at the given index
 * @param pos The position of the burst
 */
PhyPdu* UlSubFrame::getPhyPdu(uint32_t pos)
{
  assert (pos >= 0 && pos < phypdu_list_.size() );
  std::list<PhyPdu*>::iterator p = phypdu_list_.begin();
  for (uint32_t i = 0 ; i < pos ; i++) {
  	p++;
  }
  return *p;
}

void UlSubFrame::replaceTimer() {
	delete(timer_);
	timer_ = new UlSubFrameTimer(this);
}

void UlSubFrame::replaceTimer(double time) {
	delete(timer_);
	timer_ = new UlSubFrameTimer(this);
	timer_ ->start(time);
}

} //namespace ns3
