/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#ifndef ULBURST_H
#define ULBURST_H

#include "burst.h"
#include "ns3/mac48-address.h"
#include <cassert>

namespace ns3 {

/* Defines the OFDM UIUC values */
//note that this uses values from the 802.16d spec, and should really be changed to use 802.16e
//in 802.16e, UIUC=12 is used for all CDMA packets (init ranging and bw-req) - see IEEE 802.16e
//spec page 430
enum uiuc_t {
  //0 is reserved
  UIUC_INITIAL_RANGING=1,		//1
  UIUC_REQ_REGION_FULL,			//2
  UIUC_REQ_REGION_FOCUSED,		//3
  UIUC_FOCUSED_CONTENTION_IE,	//4
  //5-12 reserved for burst profiles	
  UIUC_PROFILE_1,				//5
  UIUC_PROFILE_2,				//6
  UIUC_PROFILE_3,				//7
  UIUC_PROFILE_4,				//8
  UIUC_PROFILE_5,				//9
  UIUC_PROFILE_6,				//10
  UIUC_PROFILE_7,				//11
  UIUC_PROFILE_8,				//12
  UIUC_SUBCH_NET_ENTRY=13,		//13
  UIUC_CDMA_ALLOCATION_IE,		//14 - changed from ns2 model - to be used for allocations based on cdma bw requests.  implies that top/code needs to be transmitted
  UIUC_EXT_UIUC					//15
};

/* Defines Midamble repetition interval */
enum midamble_t {
  PREAMBLE_ONLY,
  MIDAMBLE_8_SYMBOLS,  //every 8 data symbols
  MIDAMBLE_16_SYMBOLS, //every 16 data symbols
  MIDAMBLE_32_SYMBOLS  //every 32 data symbols
};

/* List of Extended UIUC */
enum extuiuc_t {
  UIUC_FAST_POWER_CTRL=0,
  UIUC_PHY_MODIFIER,
  UIUC_AAS,
  UIUC_FAST_RANGING,
  UIUC_FAST_TRACKING
};


/**
 * Uplink Burst description
 */
class UlBurst : public Burst
{
public:
  /**
   * Default contructor
   * @param phypdu The PhyPdu where it is located
   */
  UlBurst (PhyPdu *phypdu);

  /**
   * Return the midamble used in the burst
   */
  inline int getMidamble( ) { return midamble_; }
  /**
   * Set the midamble used in the burst
   */
  inline void setMidamble( int midamble ) { midamble_ = midamble; }
  
  /**
   * Return extended uiuc
   */
  inline int getExtendedUIUC () { return extuiuc_; }
  inline void setExtendedUIUC (int new_ex_iuc) { extuiuc_ = new_ex_iuc;}

  /**
   * Set the burst to Fast Ranging
   * @param macAddr address of MN
   * @param uiuc Encoding to use
   */
  void setFastRangingParam (Mac48Address macAddr, int uiuc);
  
  /**
   * Return the mac address for fast ranging
   * @return The mac address for fast ranging
   */
  Mac48Address getFastRangingMacAddr ();

  /**
   * Return the UIUC encoding to use
   * @return The UIUC encoding to use
   */
  int getFastRangingUIUC ();


 private:
  /**
   * The type of midamble included in the burst
   */
  int midamble_;

  /**
   * The extended UIUC
   */
  int extuiuc_;

  /**
   * For Fast Ranging
   */
  int uiuc2_;
  Mac48Address macAddr_; 
};

}
#endif
