/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 * @modified by Chakchai So-In
 */

#include "ulsubframetimer.h"
#include "ns3/framemap.h"
#include "ns3/subframe.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/contentionslot.h"
#include "ns3/wimax2-ss-net-device.h"

namespace ns3 {

/**
 * Creates a timer to handle the subframe transmission
 * @param subframe The UlSubframe
 */
UlSubFrameTimer::UlSubFrameTimer (UlSubFrame *subframe): Wimax2Timer(0), pdu_(0), newphy_(true)
{
  assert (subframe);
  subframe_ = subframe;
  //OFDMSymbol_ =5;
  OFDMSymbol_ = 0;
  /*  if (!mac) {
      mac= subframe_->map_->getMac();
      }*/
  //  debug2 (" start ulduration = %d \n" , mac->getStartUlduration());
  //  debug2 (" start ulduration = %d \n" , mac->getMaxUlduration());
  
  //OFDMSymbol_ =  mac->getStartUlduration();
}

/**
 * Reset the timer
 */
void UlSubFrameTimer::reset ()
{
  debug10("Resetting UlSubFrameTimer to starting conditions for the next frame\n");
  pdu_= NULL;
  newphy_ = true;
  //OFDMSymbol_=5;
  OFDMSymbol_= 3;
  //OFDMSymbol_ =  mac->getStartUlduration();
  if (status()==TIMER_PENDING)
    Simulator::Cancel(intr);
  
  if (mac->getNodeType()==STA_BS) {
	mac->getPhy()->setMode (OFDM_SEND);
  } else {
	mac->getPhy()->setMode (OFDM_RECV);
  }    
}

/**
 * we handle an event the same way we handle expiration
 */
void UlSubFrameTimer::handle() {
	expire ();
}


/**
 * When it expires on a SS, it will schedule all transmissions for 
 * this subframe for that SS with the simulator
 * When it expires on a BS, it does nothing
 */
void UlSubFrameTimer::expire()
{
  if (!mac) {
    mac= subframe_->map_->getMac();
  }
  debug10("At %f, in %s UlSubFrameTimer expiring...\n", NOW, mac->addr_s());

  //debug2("At %f in Mac %s UlsubFrameTimer expires, startulduration :%d\n", NOW, mac->addr_s(), mac->getStartUlduration());

  //A BS will have nothing scheduled to do in the UL Subframe...
  if(mac->getNodeType() == STA_BS ) {
  	mac->getPhy()->setMode(OFDM_RECV);
  	return;
  }

  mac->GetObject<Wimax2SSNetDevice>()->StartSubframe(false);

  //set the phy state to allow sending
  mac->getPhy()->setMode(OFDM_SEND);
  
  //we now look through all PhyPdus, and for each PhyPdu examine each burst for it
  for (uint32_t pdu_index = 0; pdu_index < subframe_->getNbPdu(); pdu_index++) {
  	 pdu_ = subframe_->getPhyPdu(pdu_index);
  	 if (!pdu_) {
  	 	debug10("no pdu, returning\n");
  	 	return; // no uplink burst was allocated
  	 }
  	 //OFDMSymbol_ =  mac->getStartUlduration();
  	 
	 for (int index = 0 ; index < pdu_->getNbBurst(); index++) {
	 
		  UlBurst *burst =  (UlBurst*)pdu_->getBurst(index);
		  int iuc= burst->getIUC() == UIUC_CDMA_ALLOCATION_IE ? burst->getExtendedUIUC() : burst->getIUC();
		  if (iuc == UIUC_EXT_UIUC && burst->getExtendedUIUC()== UIUC_FAST_RANGING)
			 iuc = burst->getFastRangingUIUC ();
		  

		  //set the modulation
		  Ofdm_mod_rate rate = subframe_->getProfile (iuc)->getEncoding();
		  mac->getPhy()->setModulation (rate);

		  PHY_INFO_Header pih;
		  //note these offset parameters are for the first logical symbol
		  //the actual first symbol transmittied may well be earlier in time if transmission spans multiple rows in the OFDM allocation grid
		  pih.subchannel_offset = burst->getSubchannelOffset();
		  pih.OFDMSymbol_offset = burst->getStarttime();
		  pih.num_OFDMSymbol = burst->getDuration();
		  pih.direction = UL_;
		  pih.modulation_ = rate;

		  
		  
		  //contention region packet transmission
		  if((iuc == UIUC_INITIAL_RANGING) || (iuc == UIUC_REQ_REGION_FULL)) {
			  //check if packet to send
			  int q_len = pdu_->getBurst(0)->getQueueLength_packets();
			  debug10 ("\t ul_subframe iuc :%d (ranging or cdma bw_req region) => only one Phypdu, Qburst :%d\n", iuc, q_len);
		
			  //This loop will check how many cdma req enqueued; however, it's only 2 cols or 1 cols for either cdma initial ranging or cdma bw-req so for more than 2 or 1 => need OFDM looping; do it later
			  while(q_len>0) {
				Ptr<Packet> p_tmp;
				p_tmp = burst->dequeue();
				if (p_tmp) {
					 q_len--;
					 
					 PHY_INFO_Header pih;
					 if (iuc == UIUC_INITIAL_RANGING)
						 pih.subchannel_offset = mac->getMap()->getUlSubframe()->getRanging()->getTOP(mac->addr());
					 else {//if it's a bw-req packet, we need to go forward past the init-ranging region
					 	CDMA_REQUEST_HEADER crh;
						p_tmp->PeekHeader(crh);
					 	pih.subchannel_offset = mac->getMap()->getUlSubframe()->getBw_req()->getTOP(crh.data.cid) + subframe_->getPhyPdu(0)->getBurst(0)->getDuration();
					 }
					 pih.OFDMSymbol_offset = 0;
					 pih.num_subchannels = 1;
					 pih.num_OFDMSymbol = mac->getPhy()->getSlotLength(UL_);
					 pih.direction = UL;
		  			 pih.modulation_ = rate;
					 p_tmp->AddHeader(pih);
					 //pih.
					 
					 //debug10 ("\t ul_subframe enqueue CID :%d, #symbol :%d, #symbol_offset :%d, #subchannel :%d, #subchannel_offset :%d, code :%d, top: %d, channel_index :%d, direction :%d\n", wimaxHdr_send.CRH.data.cid, wimaxHdr_send.PIH.num_OFDMSymbol, wimaxHdr_send.PIH.OFDMSymbol_offset, wimaxHdr_send.PIH.num_subchannels, wimaxHdr_send.PIH.subchannel_offset, wimaxHdr_send.CRH.data.code, wimaxHdr_send.CRH.data.top, wimaxHdr_send.PIH.channel_index , wimaxHdr_send.PIH.direction);
					 //debug10 ("\t ul_subframe packet CID :%d, size :%d, time :%f, direction :%d\n", wimaxHdr_send.CRH.data.cid, wimaxHdr_send.GMH.GMH.len, wimaxHdr_send.txtime, wimaxHdr_send.PIH.direction);
			
					 debug10 ("\t ul_subframe -CDMA packet to send, contention/ranging packets\n");
					 //p_tmp->AddHeader(wimaxHdr_send.CRH);
					 //p_tmp->AddHeader(wimaxHdr_send.PIH);
					 
					 //Simulator::Schedule(Seconds(wimaxHdr_send.PIH.OFDMSymbol_offset * mac->getPhy()->getSymbolTime()), &Wimax2NetDevice::transmit, mac, p_tmp);
					 Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(0), &Wimax2NetDevice::transmit, mac, p_tmp);
				}
			 }
		  } else {  //if we're not transmitting a contention-based packet
			  if((newphy_==true) && pdu_== NULL) return; 
			  
			  debug10 ("\t ul_subframe (data regions)\n");
		
			  if (pdu_!=NULL)
			  {
				int len = burst->getQueueLength_packets();
				while(len!=0) {
				  //because we dequeue packets in this loop, position 0 changes on each itteration
				  Ptr<Packet> p = burst->lookup(0);
				  if (PeekPointer(p)) {
					len--;
					debug10("\t loop ul_subframe packet being scheduled for transmission.  logical start = (OFDMSymbol = %d, subchannel = %d), linear length = %d\n", pih.OFDMSymbol_offset, pih.subchannel_offset, pih.num_OFDMSymbol);
								 
					//debug rpi
					p = burst->dequeue();
					p->AddHeader(pih);
				
					//we schedule the transmit function for the start of the subframe every time
					//the phy::Send function, baesd on the PHY_INFO_Header, will schedule each block correctly
					//because of the line wrapping, it's much easier to start phy::Send at the same time for every burst
					Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(0), &Wimax2NetDevice::transmit, mac, p);
				  }
				}
			}	
		  newphy_=false;
		}
	 }
  }
 reset();
}

}//namespace ns3
