/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2008 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "ns3/wimax2-channel.h"
#include "ns3/wimax2-net-device.h"
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-ss-net-device.h"
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/node.h"
#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/uinteger.h"
#include "ns3/trace-source-accessor.h"
using std::cout;

NS_LOG_COMPONENT_DEFINE ("Wimax2Phy");

#define SPEED_OF_LIGHT   300000000


namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED(Wimax2Phy);

TypeId 
Wimax2Phy::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Wimax2Phy")
    .SetParent<Object> ()
    .AddConstructor<Wimax2Phy> ()
    .AddAttribute ("g", "The ratio of cyclic prefix time to useful time.", DoubleValue(0.125), MakeDoubleAccessor(&Wimax2Phy::g_), MakeDoubleChecker<double>())
    .AddAttribute("fbandwidth", "The total radio bandwidth used by this wimax network.", UintegerValue(10e+6), MakeUintegerAccessor(&Wimax2Phy::fbandwidth_), MakeUintegerChecker<uint32_t>())
    .AddTraceSource ("Rx", "Recieve Trace", MakeTraceSourceAccessor (&Wimax2Phy::m_traceRx))
    .AddTraceSource ("Tx", "Transmit Trace", MakeTraceSourceAccessor (&Wimax2Phy::m_traceTx))
    ;
  return tid;
}

TypeId
Wimax2Phy::GetInstanceTypeId() const
{
    return GetTypeId();
}

Wimax2Phy::Wimax2Phy () : Object()
{
 ObjectBase::ConstructSelf(AttributeList());
 //default modulation is BPSK
  //modulation_ = OFDM_BPSK_1_2;
  modulation_ = OFDM_64QAM_3_4;
  ul_index_ = 0;
  dl_index_ = 0;
  
  //default values (and only ones supported in ns2 model)
  setPermutationscheme(PUSC, DL_);
  setPermutationscheme(PUSC, UL_);
  
  state_ = OFDM_IDLE;
  activated_ = true;
  updateFs();
}

Wimax2Phy::Wimax2Phy (Ptr<Wimax2NetDevice> mac)
{
 ObjectBase::ConstructSelf(AttributeList());
 //default modulation is BPSK
  //modulation_ = OFDM_BPSK_1_2;
  modulation_ = OFDM_64QAM_3_4;
  ul_index_ = 0;
  dl_index_ = 0;
  
  //default values (and only ones supported in ns2 model)
  setPermutationscheme(PUSC, DL_);
  setPermutationscheme(PUSC, UL_);
  
  
  state_ = OFDM_IDLE;
  activated_ = true;
  updateFs();
  mac_ = mac;
}

Wimax2Phy::Wimax2Phy(double bw, Permutation_scheme d, Permutation_scheme u) {
 ObjectBase::ConstructSelf(AttributeList());
//default modulation is BPSK
  //modulation_ = OFDM_BPSK_1_2;
  modulation_ = OFDM_64QAM_3_4;
  ul_index_ = 0;
  dl_index_ = 0;
  
  fbandwidth_ = bw;
  setPermutationscheme(d, DL_);
  setPermutationscheme(u, UL_);
    
  state_ = OFDM_IDLE;
  activated_ = true;
  updateFs();
}

Wimax2Phy::~Wimax2Phy() {

}

void Wimax2Phy::ReceiveSlot(Ptr<Packet> p, uint8_t subchannel) {
	if (state_ != OFDM_IDLE && state_ != OFDM_RECV) return;
	if (GetMac()->GetObject<Wimax2SSNetDevice>() != NULL && GetMac()->GetObject<Wimax2SSNetDevice>()->getMacState() == MAC802_16_DISCONNECTED) return;

	debug15("\t In PHY Layer of Station %s, received a packet chunk of size %u on subchannel %u at time %f\n", GetMac()->addr_s(), p->GetSize(), subchannel, NOW);
	GetMac()->sendUp(p, subchannel, Simulator::Now());
	
}



void Wimax2Phy::Send (Ptr<Packet> p) 
{
	state_ = OFDM_SEND;
	if (state_ != OFDM_SEND) {
#ifdef DEBUG_WIMAX
    	std::cout << "Warning: PHY not in sending state. Drop packet.\n";
#endif
    	return;
	}
	
	PHY_INFO_Header h;
	p->RemoveHeader(h);
	m_traceTx(p);

	if (GetMac()->GetObject<Wimax2BSNetDevice>() != 0) {
	  //this is a transmission from a BS, so it's happening in the DL subframe.
	  //Therefore, allocations described in the PIH are in rectangles
	  //we need to divide the packet into blocks that fill a given OFDMA allocation grid position
	  uint32_t total_blocks = h.num_subchannels * h.num_OFDMSymbol;
	  uint32_t max_bytes = (total_blocks / getSlotLength(DL_) ) * getSlotCapacity(h.modulation_,DL_);
	  uint32_t bytes_per_block = getBlockCapacity(h.modulation_, DL_);
	  //note that in some cases, a slot's total data capacity doesn't split evenly into blocks 
	  //(ex: BPSK, slot = 3 B, since it's all integers, 3/2=1, so 2 blocks = 1 slot = 2 B)
	  //we're going to say that if a slot = 2*block + 1, every first block in the slot will have 1 extra byte
	  //this way the totals work out and we don't have to use floating point math for everything

	  //note the cast here is just going from int32_t to uint32_t to stop the compiler from complaining about comparing signed and unsigned ints
	  bool even_blocks = getBlockCapacity(h.modulation_, DL_) * getSlotLength(DL_) == static_cast<uint32_t>(getSlotCapacity(h.modulation_, DL_));

  	  debug2 ("At %f, OFDMA phy sending packet in downlink subframe of size %d. Modulation is %d, cyclic prefix is %f\n", NOW, p->GetSize(), modulation_, g_);
	
	  //make sure the packet is long enough that we can divide it into appropriate chunks to fill the entire allocation, even if it's with no data
	  if (p->GetSize() < max_bytes) {
		  Ptr<Packet> padding = Create<Packet>(max_bytes- p->GetSize());
		  p->AddAtEnd(padding);
	  }

	  //actually divide the packet into what gets sent on each subchannel
	  std::vector<Ptr<Packet> > slot_chunks;
	  uint32_t blocked_off = 0;
	  for(uint32_t i = 0; i < total_blocks; i++) {
	  	if (even_blocks || i % getSlotLength(DL_) == 1) {
			slot_chunks.push_back(p->CreateFragment(blocked_off, bytes_per_block));
			blocked_off += bytes_per_block;
		}
		else  {
			slot_chunks.push_back(p->CreateFragment(blocked_off, bytes_per_block + 1));
			blocked_off += bytes_per_block + 1;
		}
	  }

	  //transmit each subchannel-symbol block at the appropriate time
	  for (uint32_t i = 0; i < h.num_subchannels; i++) 
		for (uint32_t j = 0; j < h.num_OFDMSymbol; j++) {
			//we select the transmission time by finding the beginning of the block in which it will transmit and adding a tiny bit to that
			//to avoid any timing discrepences or rounding errors from causing it to arrive in the wrong slot
			//signal propagation time would usually take care of this, but if one of the nodes doesn't have a mobility model for some reason
			//it can cause issues
	    		double block_start_offset = j * getSymbolTime() + getPS(); 
			//schedule transmission of each block over channel here
			Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(block_start_offset), &Wimax2Channel::Send, channel_, slot_chunks[i*h.num_OFDMSymbol + j], this, i + h.subchannel_offset);
		}    
	} else {
		  uint32_t total_blocks = h.num_OFDMSymbol;
		  uint32_t max_bytes = total_blocks * getBlockCapacity(h.modulation_, UL_);
		  uint32_t bytes_per_block = getBlockCapacity(h.modulation_, UL_);
		  debug2 ("At %f, OFDMA phy sending packet in uplink subframe of size %d. Modulation is %d, cyclic prefix is %f\n", NOW, p->GetSize(), modulation_, g_);

		  //make sure the packet is long enough that we can divide it into appropriate chunks to fill the entire allocation, even if it's with no data
		  if (p->GetSize() < max_bytes) {
			  Ptr<Packet> padding = Create<Packet>(max_bytes- p->GetSize());
			  p->AddAtEnd(padding);
		  }

		  //actually divide the packet into what gets sent on each subchannel
		  std::vector<Ptr<Packet> > slot_chunks;
		  for(uint32_t i = 0; i < total_blocks; i++) {
			slot_chunks.push_back(p->CreateFragment(i*bytes_per_block, bytes_per_block));
		  }
		
		  //this is used to make the current symbol wrap correctly
		  bool on_first_row = true;

		  //transmit each subchannel-symbol block at the appropriate time
		  if (h.OFDMSymbol_offset == 0) {  //we're a contention packet of some kind
			for (int i = 0; i < getSlotLength(UL_); i++) 
				Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(i * getSymbolTime()), &Wimax2Channel::Send, channel_, slot_chunks[i], this, h.subchannel_offset);
		  } else {
		  uint32_t current_block = 0;
			  for (int32_t i = h.subchannel_offset; i < getNumsubchannels(UL_); i++ )
				for (int32_t j = on_first_row ? h.OFDMSymbol_offset : getSlotLength(UL_); j < GetMac()->getMaxUlduration(); j++ ) {
					on_first_row = false;
					double block_start_offset = j * getSymbolTime();
					Simulator::ScheduleWithContext(Simulator::GetContext(), Seconds(block_start_offset), &Wimax2Channel::Send, channel_, slot_chunks[current_block], this, i);
					if (++current_block == total_blocks) return;
				}
		 }
	 }
}

 

/**
 * Change the frequency at which the phy is operating
 * @param freq The new frequency
 */
void Wimax2Phy::setFrequency (double freq)
{
  freq_ = freq;
  lambda_ = SPEED_OF_LIGHT / freq_;
}


/*
 * Activate node
 */
void Wimax2Phy::node_on ()
{
  activated_ = true;
}

/*
 * Deactivate node
 */
void Wimax2Phy::node_off ()
{
  activated_ = false;
}


/**
 * Set the new modulation for the physical layer
 * @param mod The new modulation
 */
void Wimax2Phy::setModulation (Ofdm_mod_rate mod) {
  modulation_ = mod;
}
/**
 * Return the current modulation
 */
Ofdm_mod_rate Wimax2Phy::getModulation () {
  return modulation_;
}


/**
 * Set the new permutation scheme for the physical layer
 * @param perm The new permutation scheme
 * @param dir The direction in which the permutation applies
 */
void Wimax2Phy::setPermutationscheme (Permutation_scheme perm, direction dir) 
{  
  /* Currently only supports PUSC */
  if (perm != PUSC) {
    fprintf (stderr, "Wimax2Phy: error, model currently supports PUSC only\n");
  }

  u_int32_t i = 0;
  if (dir == UL_) {
    ul_perm_ = perm;
    u_int nbRow = sizeof (UL_OFDMA_DATA) / (NB_OFDMA_PHY_PARAM*sizeof (unsigned long));
    for (i = 0; i < nbRow ; i++) {
      if (((u_int32_t)ul_perm_) == UL_OFDMA_DATA[i][OFDMA_INDEX_PERM]
	  && fbandwidth_ == (double) UL_OFDMA_DATA[i][OFDMA_INDEX_BW]) {
	ul_index_ = i;
	break;
      }
    }
    if (i == nbRow) {
      fprintf (stderr, "Wimax2Phy: error did not find match for permutation and bw combination\n");
      exit (1);
    }
  }
  else {
    dl_perm_ = perm;
    u_int nbRow = sizeof (DL_OFDMA_DATA) / (NB_OFDMA_PHY_PARAM*sizeof (unsigned long));
    for (i = 0; i < nbRow ; i++) {
      if ((u_int32_t)dl_perm_ == DL_OFDMA_DATA[i][OFDMA_INDEX_PERM]
	  && fbandwidth_ == (double) DL_OFDMA_DATA[i][OFDMA_INDEX_BW]) {
	dl_index_ = i;
	break;
      }
    }
    if (i == nbRow) {
      fprintf (stderr, "Wimax2Phy: error did not find match for permutation and bw combination\n");
      exit (1);
    }
  }
}


/**
 * get the current permutation scheme
 * @param dir The direction 
 */
Permutation_scheme Wimax2Phy::getPermutationscheme (direction dir) {
  if (dir == UL_)
    return ul_perm_;
  return dl_perm_;
}

/**
 * calculate the FFT size (same of UL and DL)
 * @return the FFT size
 */
int Wimax2Phy::getFFT () {
  return (int) (UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_FFT]);
}

/**
 * calculate the number of subchannels
 * @param dir The direction
 * @return The number of subchannel
 */
int Wimax2Phy::getNumsubchannels (direction dir) {
  if (dir == UL_)
    return UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_NBSUBCH];
  else 
    return DL_OFDMA_DATA[dl_index_][OFDMA_INDEX_NBSUBCH];
}



/**
 * Get the number of subcarrier per subchannel per symbol
 * @param dir The direction
 * @return The number of subcarrier
 */
int Wimax2Phy::getNumSubcarrier (direction dir) {
  if (dir == UL_)
    return UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_NBSUBCAR];
  else 
    return DL_OFDMA_DATA[dl_index_][OFDMA_INDEX_NBSUBCAR];
}

/**
 * Get the number of data subcarrier per subchannel per symbol
 * @param dir The direction
 * @return The numbe of data subcarrier
 */
int Wimax2Phy::getNumDataSubcarrier (direction dir) {
  if (dir == UL_)
    return UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_NBDATASUBCAR];
  else 
    return DL_OFDMA_DATA[dl_index_][OFDMA_INDEX_NBDATASUBCAR];
}

/**
 * Get the number of symbol-times covered by a slot
 * @param dir The direction
 */
int Wimax2Phy::getSlotLength (direction dir) {
  if (dir == UL_)
    return UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_SLOT_LENGTH];
  else 
    return DL_OFDMA_DATA[dl_index_][OFDMA_INDEX_SLOT_LENGTH];
}

/**
 * Get the total number of subcarriers per symbol
 * @param dir The direction
 */
int Wimax2Phy::getNUsed (direction dir) {
  if (dir == UL_)
    return UL_OFDMA_DATA[ul_index_][OFDMA_INDEX_NUSED];
  else 
    return DL_OFDMA_DATA[dl_index_][OFDMA_INDEX_NUSED];
}

Ofdm_mod_rate Wimax2Phy::getModulationScheme(int bytes_per_subchannel_symbol, direction dir) {
	int carriers = getNumDataSubcarrier(dir);

	for (int i = 0; i < 7; i++)
		if (static_cast<int>(getModulationRate( static_cast<Ofdm_mod_rate>(i)) * carriers) == bytes_per_subchannel_symbol)
			return static_cast<Ofdm_mod_rate>(i);

	fprintf(stderr, "Error in Wimax2Phy::getModulationScheme - invalid input, returning QPSK 1/2\n");
	return OFDM_QPSK_1_2;
}

/** 
 * Get the bit rate per symbol and subcarrier for the given modulation
 * which is to say the number of bits in conveyed by one symbol on one subcarrier
 * @param mod The modulation
 */
double Wimax2Phy::getModulationRate (Ofdm_mod_rate mod)
{
  switch (mod) {
  case OFDM_BPSK_1_2:
    return 1.0 * 1 / 2;
  case OFDM_QPSK_1_2:
    return 2.0 * 1 / 2;    
  case OFDM_QPSK_3_4:
    return 2.0 * 3 / 4;
  case OFDM_16QAM_1_2:
    return 4.0 * 1 / 2;
  case OFDM_16QAM_3_4:
    return 4.0 * 3 / 4;
  case OFDM_64QAM_2_3:
    return 6.0 * 2 / 3;
  case OFDM_64QAM_3_4:
    return 6.0 * 3 / 4; 
  default:
    fprintf (stderr, "Error in  Wimax2Phy::getModulationRate modulation unknown\n");
    exit (1);
  }
  return 0.0;
}

double Wimax2Phy::getBitsPerSubchannel(Ofdm_mod_rate mod, direction dir) {
	return getModulationRate(mod) * getNumDataSubcarrier(dir);
}


/**
 * Compute the transmission time for a packet of size sdusize, num subchannels allocated per OFDM symbol, direction,
 * using the given modulation
 */
double Wimax2Phy::getTrxTime (int sdusize, Ofdm_mod_rate mod, int num_subchannels, direction dir)
{
  //we compute the number of symbols required
  int bpsymb = (int) floor((num_subchannels*getNumDataSubcarrier(dir)) * getModulationRate (mod));
  double txtime = (int) (sdusize*8*getSymbolTime ()/bpsymb);

#ifdef DEBUG_WIMAX
  printf ("\tWimax2Phy::getTrxTime bpsymbol %d duration:%f\n", bpsymb, txtime);
#endif

  return txtime;
}


/**
 * Compute the transmission time using OFDM symbol as
 * minimum allocation for a packet of size sdusize and
 * using the given modulation, num subchannels, direction)
 * @param sdusize The packet size
 * @param mod The modulation used
 * @param dir The direction
 */
int Wimax2Phy::getNumSubchannels (int sdusize, Ofdm_mod_rate mod, direction dir) 
{
  //we compute the number of symbols required
  double bpsubchn = floor (getNumDataSubcarrier(dir)*getSlotLength(dir)*getModulationRate (mod));
  int nb_subchannels = (int) ceil(((double)sdusize*8)/bpsubchn);

#ifdef DEBUG_WIMAX
  printf("#_data_subcarrier[%d]\t slot_len[%d]\t modulation[%e]\t bpsubchn[%e]\n", getNumDataSubcarrier(dir),getSlotLength(dir),getModulationRate (mod),bpsubchn);
  printf ("\tWimax2Phy::getNumSubchannels bpsubchannel=%d Nb subchannel  (SDUsize_bits/bpsubchn):%d\n", (int) bpsubchn, nb_subchannels);
#endif

  return (nb_subchannels);
}


/* 
 * Compute the maximum packet size (in Bytes) that can be put in the allocation using
 * the given number of subchannel, modulation, and direction
 * @param numSubchannels The number of subchannel allocated
 * @param mod The modulation rate
 * param dir The direction
 */ 
int Wimax2Phy::getMaxPktSize (int numSubchannels, Ofdm_mod_rate mod, direction dir) {
  //we compute the number of symbols required, note we are taking raw total subchannels here, and relying
  //on the calling code to translate from slots (or rectangual mappings)
  double bpsubchn = floor (getNumDataSubcarrier(dir)*getModulationRate (mod));
  int max_size = (int) (floor((numSubchannels*bpsubchn)/8));
#ifdef DEBUG_WIMAX
  printf ("\tWimax2Phy::getMaxPktSize nbSuchannel=%d mod=%d dir=%d MaxPacketSize %d\n", numSubchannels, mod,dir, max_size);
#endif

  return (max_size);
}

/**
 * Compute the transmission time using OFDM symbol as
 * minimum allocation for a packet of size sdusize and
 * using the given modulation, num subchannels, direction, scheme)
 */
double Wimax2Phy::getTrxSymbolTime (int sdusize, Ofdm_mod_rate mod, int num_subchannels, direction dir) {
  //we compute the number of symbols required
  double bpsymb = (int) floor((num_subchannels*getNumDataSubcarrier(dir)) * getModulationRate (mod));
  double nb_symbols = ceil(((double)sdusize*8)/bpsymb);

  //#ifdef DEBUG_WIMAX
  printf ("Nb symbols=%f\n", nb_symbols);
  //#endif

  return (nb_symbols*getSymbolTime ());
}


/**
 * Return the maximum size in bytes that can be sent for the given 
 * nb of symbols and modulation, and direction. 
 */
int Wimax2Phy::getMaxPktSizeForSymbol (int nbsymbols, Ofdm_mod_rate mod, direction dir) {

  //we compute the bits per symbol (using all used subcarriers)
  double bpsymb = (int) floor((getNUsed(dir)) * getModulationRate (mod));
  int max_size = (int) floor((nbsymbols*bpsymb)/8);

#ifdef DEBUG_WIMAX
  printf ("\tWimax2Phy::getMaxPktSizemax nb symbols=%d mod=%d nb subcarriers= %d packet size = %d\n", nbsymbols, mod, getNUsed(dir), max_size);
#endif
  return max_size;
}


/**
 * Set the new transmitting power
 */
void Wimax2Phy::setTxPower (double power) {
  Pt_ = power;
}
/**
 * Return the current transmitting power
 */
double Wimax2Phy::getTxPower () {
  return getPt();
}

double Wimax2Phy::getPt() {
	return Pt_;
}

/** 
 * Update the PS information
 */
void Wimax2Phy::updateFs () {
  /* The PS=4*Fs with Fs=floor (n.BW/8000)*8000
   * and n=8/7 is channel bandwidth multiple of 1.75Mhz
   * n=86/75 is channel bandwidth multiple of 1.5Mhz
   * n=144/125 is channel bandwidth multiple of 1.25Mhz
   * n=316/275 is channel bandwidth multiple of 2.75Mhz
   * n=57/50 is channel bandwidth multiple of 2.0Mhz
   * n=8/7 for all other cases
   */
  double n; 
    
  if (fbandwidth_ != 10e+6) {
    fprintf (stderr, "Error: Currently only supports 10Mhz bandwidth\n");
    exit (1);
  }

  if (((int) (fbandwidth_ / 1.25)) * 1.25 == fbandwidth_
      || ((int) (fbandwidth_ / 1.5)) * 1.5 == fbandwidth_
      || ((int) (fbandwidth_ / 2.0)) * 2.0 == fbandwidth_
      || ((int) (fbandwidth_ / 2.75)) * 2.75 == fbandwidth_) {
    n = 28.0/25;
  } else {
    n = 8.0/7;
  }

  fs_ = floor (n*fbandwidth_/8000) * 8000;
#ifdef DEBUG_WIMAX
  printf ("Fs updated. Bw=%f, n=%f, new value is %e\n", fbandwidth_, n, fs_);
#endif
}


/*
 * Compute the transmission time for a packet of size sdusize and
 * using the given modulation when ALL subcarriers are used (OFDM type)
 * @param sdusize Size in bytes of the data to send
 * @param mod The modulation to use
 */
double Wimax2Phy::getTrxTime (int sdusize, Ofdm_mod_rate mod) {
  //we compute the number of symbols required
  
  //bits / symbol
  int bpsymb = (int) floor((getNUsed(DL_)) * getModulationRate (mod));

#ifdef DEBUG_WIMAX
  printf ("Nb symbols=%f\n", ((double)sdusize*8)/bpsymb);
#endif

  // # bits * ((sec / symbol) / (bits / symbol)) = # seconds
  return sdusize*8*getSymbolTime ()/bpsymb;
}

/*
 * Compute the transmission time for a packet of size sdusize and
 * using the given modulation
 * @param sdusize Size in bytes of the data to send
 * @param mod The modulation to use
 */
double Wimax2Phy::getTrxSymbolTime (int sdusize, Ofdm_mod_rate mod) {
  //bits / symbol
  int bpsymb = (int) floor((getNUsed(DL_)) * getModulationRate (mod));
  int nb_symbols;

#ifdef DEBUG_WIMAX
  printf ("Nb symbols=%d\n", (int) (ceil(((double)sdusize*8)/bpsymb)));
#endif
  //# bits / (bits / symbol) = # symbols
  nb_symbols = (int) (ceil(((double)sdusize*8)/bpsymb));
  return (nb_symbols*getSymbolTime ());
}


/*
 * Return the maximum size in bytes that can be sent for the given 
 * nb symbols and modulation
 */
//int Wimax2Phy::getMaxPktSize (double nbsymbols, Ofdm_mod_rate mod)
//{
//  int bpsymb = (int) floor((getNUsed(DL_)) * getModulationRate (mod));
//
//  return (int)(nbsymbols*bpsymb)/8;
//}

/**
 * Return the OFDM symbol duration time
 */
double Wimax2Phy::getSymbolTime () 
{ 
  //printf ("fs=%e, Subcarrier spacing=%e\n", fs_, fs_/((double)NFFT));
  //return ceil( OFDM_PRECISION*(1+g_)*((double)NFFT)/fs_)/ OFDM_PRECISION;
  return ((1+g_)*((double)getFFT()))/fs_;
}


/*
 * Set the mode for physical layer
 * The Mac layer is in charge of know when to change the state by 
 * request the delay for the Rx2Tx and Tx2Rx
 */
void Wimax2Phy::setMode (Ofdm_phy_state mode)
{
  state_ = mode;
}

/*
 * Get the mode for physical layer
 */
Ofdm_phy_state Wimax2Phy::getMode ()
{
  return state_;
}


//returns the number of BYTES in a single slot for this direction and modulation rate
int32_t Wimax2Phy::getSlotCapacity(Ofdm_mod_rate rate, direction dir){
  return (int) (getNumDataSubcarrier (dir) * getSlotLength (dir) * getModulationRate(rate))/8 ;
}

uint32_t Wimax2Phy::getBlockCapacity(Ofdm_mod_rate rate, direction dir) {
	return static_cast<uint32_t>(static_cast<double>(getNumDataSubcarrier(dir) * getModulationRate(rate))/ 8.0);
}


//return the maximum block size for a given modulation
int Wimax2Phy::getMaxBlockSize(Ofdm_mod_rate rate){
  int size;
  switch(rate){
  case OFDM_BPSK_1_2:
    size = 10;
    break;
  case OFDM_QPSK_1_2:
    size = 10;
    break;
  case OFDM_QPSK_3_4:
    size = 6;
    break;
  case OFDM_16QAM_1_2:
    size = 5;
    break;
  case OFDM_16QAM_3_4:
    size = 3;
    break;
  case OFDM_64QAM_2_3:
    size = 2;
    break;
  case OFDM_64QAM_3_4:
    size = 2;
    break;
  default:
    size = 2;
    break;
  }

  return size;
}


int Wimax2Phy::getMCSIndex (Ofdm_mod_rate rate , int block_size ) {

  int index; 
  switch(rate){
  case OFDM_QPSK_1_2:
    if( block_size == 1) 
      index = 1;
    else if (block_size == 2)
      index = 2;
    else if (block_size == 3)
      index = 3;
    else if (block_size == 4)
      index = 4;
    else if (block_size == 5)
      index = 5; 
    else if (block_size == 6)
      index = 6;
    else if (block_size == 7)
      index = 6;	
    else if (block_size == 8)
      index = 7;
    else if (block_size == 9)
      index = 8;
    else if (block_size == 10)
      index = 9;       
    break;
  case OFDM_QPSK_3_4:
    if( block_size == 1) 
      index = 10;
    else if (block_size == 2)
      index = 11;
    else if (block_size == 3)
      index = 12;
    else if (block_size == 4)
      index = 13;
    else if (block_size == 5)
      index = 14; 
    else if (block_size == 6)
      index = 15;
    break;
  case OFDM_16QAM_1_2:
    if( block_size == 1) 
      index = 16;
    else if (block_size == 2)
      index = 17;
    else if (block_size == 3)
      index = 18;
    else if (block_size == 4)
      index = 19;
    else if (block_size == 5)
      index = 20;
    break;
  case OFDM_16QAM_3_4:
    if( block_size == 1) 
      index = 21;
    else if (block_size == 2)
      index = 22;
    else if (block_size == 3)
      index = 23;
    break;
    /*    case OFDM_64QAM_1_2:              64QAM1-2, and 5-6 not implemented currently
	  if( block_size == 1) 
	  index = 24;
	  else if (block_size == 2)
	  index = 25;
	  else if (block_size == 2)
	  index = 26;
	  break;
    */    
  case OFDM_64QAM_2_3:
    if( block_size == 1) 
      index = 27;
    else if (block_size == 2)
      index = 28;
    break;
  case OFDM_64QAM_3_4:
    if( block_size == 1) 
      index = 29;
    else if (block_size == 2)
      index = 30;
    break;
    /*     case OFDM_64QAM_5_6:
	   if( block_size == 1) 
	   index = 31;
	   else if (block_size == 2)
	   index = 32;
	   break; 
    */    default:
    index = 1;
    break;
  }

  return index;
}

void Wimax2Phy::DoDispose() {
	Object::DoDispose();	
}

void Wimax2Phy::SynchronizeAll() {
	channel_->sync_preamble();
}

Ptr<Wimax2Channel> Wimax2Phy::getChannel() {return channel_;}
void Wimax2Phy::setChannel(Ptr<Wimax2Channel> chan) {if (chan) channel_ = chan;}


} // namespace ns3
