/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  rouil
 */

#include "ns3/wimax2-serviceflowhandler.h"
#include "ns3/wimax2-net-device.h"
#include "ns3/wimax2-scheduler.h"
#include "ns3/arqstatus.h"

namespace ns3 {

static int TransactionID = 0; 
/* 
 * Create a service flow
 * @param mac The Mac where it is located
 */
ServiceFlowHandler::ServiceFlowHandler ()
{
  LIST_INIT (&flow_head_);
  LIST_INIT (&pendingflow_head_);
  LIST_INIT (&static_flow_head_);
}

/*
 * Set the mac it is located in
 * @param mac The mac it is located in
 */
void ServiceFlowHandler::setMac (Ptr<Wimax2NetDevice> mac)
{
  assert (mac);

  mac_ = mac;
}

/**
 * Process the given packet. Only service related packets must be sent here.
 * @param p The packet received
 */
void ServiceFlowHandler::process (Ptr<Packet>  p) 
{ 
  All_Headers wimaxHdr;
  p->RemoveHeader(wimaxHdr);
  //gen_mac_header_t *header = &wimaxHdr->GMH.GMH;

  //we cast to this frame because all management frame start with
  //a type 
  Type_Check tp;
  p->PeekHeader(tp);
 // mac802_16_dl_map_frame *frame = &dlmf.mf;
  
  p->AddHeader(wimaxHdr);

  switch (tp.type) {
  case MAC_DSA_REQ: 
    processDSA_req (p);
    break;
  case MAC_DSA_RSP: 
    processDSA_rsp (p);
    break;
  case MAC_DSA_ACK: 
    processDSA_ack (p);
    break;
  default: 
    debug2 ("Unknow frame type (%d) in flow handler\n", tp.type);
  }
  //Packet::free (p);
}

/**
 * Add a flow with the given qos
 * @param qos The QoS for the flow
 * @return the created ServiceFlow
 */
Wimax2ServiceFlow* ServiceFlowHandler::addFlow (Wimax2ServiceFlowQoS * qos) {
  return NULL;
}

/**
 * Remove the flow given its id
 * @param id The flow id
 */
void ServiceFlowHandler::removeFlow (int id) {
  
}

/**
 * Send a flow request to the given node
 * @param index The node address
 * @param out The flow direction
 */
void ServiceFlowHandler::sendFlowRequest (Mac48Address index, bool out)
{
  Ptr<Packet> p;
  //struct hdr_cmn *ch;
  All_Headers wimaxHdr;
  mac802_16_dsa_req_frame dsa_frame;
  DSA_REQ_FRAME_HEADER drfh;
  PeerNode *peer;

  //create packet for request
  peer = mac_->getPeerNode(index);  
  p = mac_->getPacket ();
  p->RemoveHeader(wimaxHdr);
  
  //p->allocdata (sizeof (struct mac802_16_dsa_req_frame));
  //dsa_frame = (mac802_16_dsa_req_frame*) p->accessdata();
  dsa_frame.type = MAC_DSA_REQ;
  dsa_frame.uplink = (out && mac_->getNodeType()==STA_MN) || (!out && mac_->getNodeType()==STA_BS) ;
  dsa_frame.transaction_id = TransactionID++;
  debug2(" sampad in send flow request"); 
  if (mac_->getNodeType()==STA_MN) {
    drfh.data = dsa_frame;
    wimaxHdr.GMH.GMH.len += drfh.GetSerializedSize();
  } else {
    //assign a CID and include it in the message
    Connection *data = new Connection (CONN_DATA);

    mac_->getCManager()->add_connection (data, out);
    if (out)
    {
      peer->setOutData (data);
      debug2("set outcoming data connection for mac %s\n", mac_->addr_s());
    }
    else
    {
   	   peer->setInData (data);
   	   debug2("set incoming data connection for mac %s\n", mac_->addr_s());
    }
    dsa_frame.cid = data->get_cid();
    drfh.data = dsa_frame;
    wimaxHdr.GMH.GMH.len += drfh.GetSerializedSize();
  }
  


  wimaxHdr.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  p->AddHeader(drfh);
  p->AddHeader(wimaxHdr);
  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
}

/**
 * process a flow request
 * @param p The received request
 */
void ServiceFlowHandler::processDSA_req (Ptr<Packet> p)
{
  All_Headers wimaxHdr_req;
  p->RemoveHeader(wimaxHdr_req);
  
  debug2 ("At %f in Mac %s received DSA request from %d\n", NOW, HelperFunctions::getMacAddrString(mac_->addr()).c_str(), wimaxHdr_req.GMH.GMH.cid);
  debug2(" sampad in send process DSA_request\n"); 
  Ptr<Packet> rsp;
//  struct hdr_cmn *ch;
  All_Headers wimaxHdr_rsp;
  DSA_REQ_FRAME_HEADER dreq_fh;
  DSA_RSP_FRAME_HEADER drsp_fh;  
  mac802_16_dsa_req_frame *dsa_req_frame = &dreq_fh.data;
  mac802_16_dsa_rsp_frame *dsa_rsp_frame = &drsp_fh.data;
  PeerNode *peer;
  Connection *data;
  Arqstatus *arqstatus;

  //read the request
  p->RemoveHeader(dreq_fh);
  peer = mac_->getCManager ()->get_connection (wimaxHdr_req.GMH.GMH.cid, true)->getPeerNode();
  
  //allocate response
  //create packet for request
  rsp = mac_->getPacket ();
  rsp->RemoveHeader(wimaxHdr_rsp);
  dsa_rsp_frame->type = MAC_DSA_RSP;
  dsa_rsp_frame->transaction_id = dsa_req_frame->transaction_id;
  dsa_rsp_frame->uplink = dsa_req_frame->uplink;
  dsa_rsp_frame->confirmation_code = 0; //OK
  dsa_rsp_frame->staticflow = dsa_req_frame->staticflow;

  if (mac_->getNodeType()==STA_MN) {
    //the message contains the CID for the connection
    data = new Connection (CONN_DATA, dsa_req_frame->cid);
/*
    data->setCDMA(0);
    data->initCDMA();
    data->setPOLL_interval(0);
*/

    data->set_serviceflow(dsa_req_frame->staticflow);
    // We will move all the ARQ information in the service flow to the isArqStatus that maintains the Arq Information.
    if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ) {
    		debug2("ARQ status is enable.\n");
	    arqstatus = new Arqstatus ();
	    data->setArqStatus (arqstatus);
	    data->getArqStatus ()->setArqEnabled (data->get_serviceflow ()->getQoS ()->getIsArqEnabled ());
	    data->getArqStatus ()->setRetransTime (data->get_serviceflow ()->getQoS ()->getArqRetransTime ()); 
	    data->getArqStatus ()->setMaxWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ());
            data->getArqStatus ()->setCurrWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ()); 
	    data->getArqStatus ()->setAckPeriod (data->get_serviceflow ()->getQoS ()->getArqAckPeriod ());
	    // setting timer array for the flow
	    data->getArqStatus ()->arqRetransTimer = new ARQTimer(data);        /*RPI*/
	    debug2("In DSA req STA_MN, generate a ARQ Timer and cid is %d.\n", data->get_cid());
    }  
    if (dsa_req_frame->uplink){
      mac_->getCManager()->add_connection (data, OUT_CONNECTION);
      debug2(" dsa-req-frame being processed and connection being added for uplink node = MN\n"); 
      peer->setOutData (data);
      
    } else {
      mac_->getCManager()->add_connection (data, IN_CONNECTION);
      debug2(" dsa-req-frame being processed and connection being added for downlink node =MN\n"); 
      peer->setInData (data);
    }
    wimaxHdr_rsp.GMH.GMH.len += drsp_fh.GetSerializedSize();
  } else {
    //allocate new connection
    data = new Connection (CONN_DATA);
/*
    data->setCDMA(0);
    data->initCDMA();
    data->setPOLL_interval(0);
*/

    data->set_serviceflow(dsa_req_frame->staticflow);
    // We will move all the ARQ information in the service flow to the isArqStatus that maintains the Arq Information.
    if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ){
    		debug2("Going to set ARQ parameters.\n");
	    arqstatus = new Arqstatus ();
	    data->setArqStatus (arqstatus);
	    data->getArqStatus ()->setArqEnabled (data->get_serviceflow ()->getQoS ()->getIsArqEnabled ());
	    data->getArqStatus ()->setRetransTime (data->get_serviceflow ()->getQoS ()->getArqRetransTime ()); 
	    data->getArqStatus ()->setMaxWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ());
            data->getArqStatus ()->setCurrWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ()); 
	    data->getArqStatus ()->setAckPeriod (data->get_serviceflow ()->getQoS ()->getArqAckPeriod ());
 	    // setting timer array for the flow
	    data->getArqStatus ()->arqRetransTimer = new ARQTimer(data);        /*RPI*/
	    debug2("In DSA req STA_BS, generate a ARQ Timer and cid is %d.\n", data->get_cid());
	    
    }  
    if (dsa_req_frame->uplink) {
      mac_->getCManager()->add_connection (data, IN_CONNECTION);
	debug2(" dsa-req-frame being processed and connection being added for uplink node = not MN\n"); 
      peer->setInData (data);
    } else {
      mac_->getCManager()->add_connection (data, OUT_CONNECTION);
      debug2(" dsa-req-frame being processed and connection being added for downlink node = not MN\n"); 
      peer->setOutData (data);
      debug2("set outcoming data connection for mac %s\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
//Begin RPI
      if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ){	
        //schedule timer based on retransmission time setting
        //mac_->debug(" ARQ Timer Setting Downlink\n");	
        data->getArqStatus ()->arqRetransTimer->Schedule(Seconds(data->getArqStatus ()->getRetransTime ()));
        //mac_->debug(" ARQ Timer Set Downlink\n");
      }    		
//End RPI
    }
    dsa_rsp_frame->cid = data->get_cid();
    wimaxHdr_rsp.GMH.GMH.len += drsp_fh.GetSerializedSize();
  }

  wimaxHdr_rsp.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  rsp->AddHeader(drsp_fh);
  rsp->AddHeader(wimaxHdr_rsp);
  
  
  peer->getPrimary(OUT_CONNECTION)->enqueue (rsp);

}

/**
 * process a flow response
 * @param p The received response
 */
void ServiceFlowHandler::processDSA_rsp (Ptr<Packet> p)
{
  debug2 ("At %f in Mac %s received DSA response\n", NOW, HelperFunctions::getMacAddrString(mac_->addr()).c_str());

  Ptr<Packet> ack;
  All_Headers wimaxHdr_ack;
  All_Headers wimaxHdr_rsp;
  DSA_ACK_FRAME_HEADER dack_fh;
  DSA_RSP_FRAME_HEADER drsp_fh;
  mac802_16_dsa_ack_frame *dsa_ack_frame = &dack_fh.data;
  mac802_16_dsa_rsp_frame *dsa_rsp_frame = &drsp_fh.data;
  Connection *data;
  PeerNode *peer;
  Arqstatus * arqstatus;

  //read the request
  //wimaxHdr_rsp = HDR_Wimax2NetDevice(p);
  p->RemoveHeader(wimaxHdr_rsp);
  p->RemoveHeader(drsp_fh);
  peer = mac_->getCManager ()->get_connection (wimaxHdr_rsp.GMH.GMH.cid, true)->getPeerNode();
  
  //TBD: check if status not OK

  if (mac_->getNodeType()==STA_MN) {
    //the message contains the CID for the connection
    data = new Connection (CONN_DATA, dsa_rsp_frame->cid);
/*
    data->setCDMA(0);
    data->initCDMA();
    data->setPOLL_interval(0);
*/

    data->set_serviceflow(dsa_rsp_frame->staticflow);
    // We will move all the ARQ information in the service flow to the isArqStatus that maintains the Arq Information.
    if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ){
	    arqstatus = new Arqstatus ();
	    data->setArqStatus (arqstatus);
	    data->getArqStatus ()->setArqEnabled (data->get_serviceflow ()->getQoS ()->getIsArqEnabled ());
	    data->getArqStatus ()->setRetransTime (data->get_serviceflow ()->getQoS ()->getArqRetransTime ()); 
	    data->getArqStatus ()->setMaxWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ()); 
	    data->getArqStatus ()->setCurrWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ());	
	    data->getArqStatus ()->setAckPeriod (data->get_serviceflow ()->getQoS ()->getArqAckPeriod ());
	    // setting timer array for the flow
	     data->getArqStatus ()->arqRetransTimer = new ARQTimer(data);             /*RPI*/ 
	     debug2("In DSA rsp STA_MN, generate a ARQ Timer and cid is %d.\n", data->get_cid());
    }
    if (dsa_rsp_frame->uplink) {
      mac_->getCManager()->add_connection (data, OUT_CONNECTION);
      peer->setOutData (data);
      debug2("set outgoing data connection for mac %s\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
//Begin RPI
      if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ){
      	//schedule timer based on retransmission time setting
        //mac_->debug(" ARQ Timer Setting -Uplink\n");
        data->getArqStatus ()->arqRetransTimer->sched(data->getArqStatus ()->getRetransTime ());
        //mac_->debug(" ARQ Timer Set- Uplink\n");
      }   		
//End RPI
    } else {
      mac_->getCManager()->add_connection (data, IN_CONNECTION);
      peer->setInData (data);
      debug2("set incoming data connection for mac %s\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
    }
  }

  //allocate ack
  //create packet for request
  ack = mac_->getPacket ();
  ack->RemoveHeader(wimaxHdr_ack);
  dsa_ack_frame->type = MAC_DSA_ACK;
  dsa_ack_frame->transaction_id = dsa_rsp_frame->transaction_id;
  dsa_ack_frame->uplink = dsa_rsp_frame->uplink;
  dsa_ack_frame->confirmation_code = 0; //OK
  wimaxHdr_ack.GMH.GMH.len += dack_fh.GetSerializedSize();

  wimaxHdr_ack.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
  ack->AddHeader(dack_fh);
  ack->AddHeader(wimaxHdr_ack);
  peer->getPrimary(OUT_CONNECTION)->enqueue (ack);

}

/**
 * process a flow request
 * @param p The received response
 */
void ServiceFlowHandler::processDSA_ack (Ptr<Packet> p)
{
  debug2 ("At %f in Mac %s received DSA ack\n", NOW, HelperFunctions::getMacAddrString(mac_->addr()).c_str());
}


/*
 * Add a static flow
 */
int ServiceFlowHandler::addStaticFlow (bool is_UL, int drate, SchedulingType_t ft, int dsize, int prid, uint8_t arq_en, double arq_ret_time, uint32_t arq_maxwin, uint8_t ack_per, int traffic_pri, int peak_traffic_rate, int min_res_traffic_rate, int req_xmit_pol, int jitter, int sdu_indicator, int min_tol_traffic_rate, int sdusize, int max_burst_size, int said) {

  dir_t dir;
  int32_t datarate = drate;
  SchedulingType_t  flow_type;
  double data_size = dsize;
  u_int16_t period = prid;
  u_int8_t isArqEnabled = arq_en;
  double arq_retrans_time = arq_ret_time;
  u_int32_t arq_max_window = arq_maxwin;
  uint8_t ack_period = ack_per;
  int delay = 0 ;
  int burstsize = 0;


  if (is_UL) dir = UL;
  else dir = DL;

  flow_type = ft;
  
  /* Create the Service Flow Qos object */
  Wimax2ServiceFlowQoS * staticflowqos = new Wimax2ServiceFlowQoS (delay, datarate, burstsize) ;
  staticflowqos->setDataSize(data_size);
  staticflowqos->setPeriod(period);
  staticflowqos->setIsArqEnabled(isArqEnabled);
  debug2("ARQ is enabled for this static flow.\n");
  staticflowqos->setArqRetransTime(arq_retrans_time);
  staticflowqos->setArqMaxWindow(arq_max_window);
  staticflowqos->setArqAckPeriod(ack_period);


  staticflowqos->setTrafficPriority(traffic_pri);
  staticflowqos->setPeakTrafficRate(peak_traffic_rate);
  staticflowqos->setMinReservedTrafficRate(min_res_traffic_rate);
  staticflowqos->setReqTransmitPolicy(req_xmit_pol);
  staticflowqos->setJitter(jitter);
  staticflowqos->setSDUIndicator(sdu_indicator);
  staticflowqos->setMinTolerableTrafficRate(min_tol_traffic_rate);
  staticflowqos->setSDUSize(sdusize);
  staticflowqos->setMaxBurstSize(max_burst_size);
  staticflowqos->setSAID(said);



   /* Create the Static Service Flow */
  Wimax2ServiceFlow * staticflow = new Wimax2ServiceFlow (flow_type, staticflowqos);
  staticflow->setDirection(dir);
 
  /* Add the Service Flow to the Static Flow List*/
  staticflow->insert_entry_head (&static_flow_head_);
  debug2(" service flow static flow created ");
  return 0; //old TCL_OK

}


/*
 * Add a static flow
 * @param argc The number of parameter
 * @param argv The list of parameters
 * this is a remnant of the tcl bindings (hence argc, argv arguments instead of a standard list)
 * but I'm leaving it in becuase it's a cleaner interface that may well be useful when this is 
 * exposed to the scripts
 */
int ServiceFlowHandler::addStaticFlow (int argc, const char*const* argv) 
{
  dir_t dir;
  int32_t datarate = atoi(argv[3]);
  SchedulingType_t  flow_type;
  double data_size = 0.0;
  u_int16_t period = 0;
  u_int8_t isArqEnabled = atoi(argv[7]);
  double arq_retrans_time = atof(argv[8]);
  u_int32_t arq_max_window = atoi(argv[9]);
  u_int8_t ack_period = atoi(argv[10]);
  int delay = 0 ;
  int burstsize = 0;

  if (strcmp(argv[2], "DL") == 0) {
	  dir = DL;
  } else if (strcmp(argv[2], "UL") == 0) {
	  dir = UL;
  } else {
	  return 1; //old TCL_ERROR
  }

  if (strcmp(argv[4], "BE") == 0) {
	  flow_type = SERVICE_BE;
  } else if (strcmp(argv[4], "UGS") == 0) {
	  flow_type = SERVICE_UGS;
	  if (argc != 21) {
		  return 1; //old TCL_ERROR
	  }
	  else {
		  // convert user defined bytes to time
		  //data_size = (atoi(argv[5])<<3)/(double)(mac_->phymib_.getDataRate());
		  period = atoi(argv[6]);
		  data_size = atoi(argv[5]);	
		  
	  }
  } else if (strcmp(argv[4], "ertPS") == 0) {
	  flow_type = SERVICE_ertPS;
	  if (argc != 21) {
		  return 1; //old TCL_ERROR
	  }
	  else {
		  period = atoi(argv[6]);
		  data_size = atoi(argv[5]);	
	  }    
  } else if (strcmp(argv[4], "rtPS") == 0) {
	  flow_type = SERVICE_rtPS;
	  if (argc != 21) {
		  return 1; //old TCL_ERROR
	  }
	  else {
		  period = atoi(argv[6]);
	  }    
  } else if (strcmp(argv[4], "nrtPS") == 0) {
	  flow_type = SERVICE_nrtPS;
	  if (argc != 21) { 
		  return 1; //old TCL_ERROR
	  }
	  else {
		  period = atoi(argv[6]);
	  }   
  } else {
	  return 1; //old TCL_ERROR
  } 
 
  
  /* Create the Service Flow Qos object */
  Wimax2ServiceFlowQoS * staticflowqos = new Wimax2ServiceFlowQoS (delay, datarate, burstsize) ;
  staticflowqos->setDataSize(data_size);
  staticflowqos->setPeriod(period);
  staticflowqos->setIsArqEnabled(isArqEnabled);
  debug2("ARQ is enabled for this static flow.\n");
  staticflowqos->setArqRetransTime(arq_retrans_time);
  staticflowqos->setArqMaxWindow(arq_max_window);
  staticflowqos->setArqAckPeriod(ack_period);


  staticflowqos->setTrafficPriority(atoi(argv[11]));
  staticflowqos->setPeakTrafficRate(atoi(argv[12]));
  staticflowqos->setMinReservedTrafficRate(atoi(argv[13]));
  staticflowqos->setReqTransmitPolicy(atoi(argv[14]));
  staticflowqos->setJitter(atoi(argv[15]));
  staticflowqos->setSDUIndicator(atoi(argv[16]));
  staticflowqos->setMinTolerableTrafficRate(atoi(argv[17]));
  staticflowqos->setSDUSize(atoi(argv[18]));
  staticflowqos->setMaxBurstSize(atoi(argv[19]));
  staticflowqos->setSAID(atoi(argv[20]));



   /* Create the Static Service Flow */
  Wimax2ServiceFlow * staticflow = new Wimax2ServiceFlow (flow_type, staticflowqos);
  staticflow->setDirection(dir);
 
  /* Add the Service Flow to the Static Flow List*/
  staticflow->insert_entry_head (&static_flow_head_);
  debug2(" service flow static flow created ");
  return 0; //old TCL_OK
}

/**
 * Send a flow request to the given node
 * @param index The node address
 * @param uplink The flow direction
 */
void ServiceFlowHandler::init_static_flows (Mac48Address index)
{
  Ptr<Packet> p;
  All_Headers wimaxHdr;
  DSA_REQ_FRAME_HEADER dreq_fh;
  mac802_16_dsa_req_frame *dsa_frame = &dreq_fh.data;
  PeerNode *peer;
  Arqstatus * arqstatus;
  
  debug2(" sampad in init staic flows\n"); 

  for (Wimax2ServiceFlow *n=static_flow_head_.lh_first; n; n=n->next_entry()) {
	  //create packet for request
	  peer = mac_->getPeerNode(index);  
	  p = mac_->getPacket ();
	  p->RemoveHeader(wimaxHdr);
	  dsa_frame->type = MAC_DSA_REQ;
	  if (n->getDirection() == UL ) 
		  dsa_frame->uplink = true;
	  else
		  dsa_frame->uplink = false;
	  dsa_frame->transaction_id = TransactionID++;
	  dsa_frame->staticflow = n;
	  if (mac_->getNodeType()==STA_MN)
		  wimaxHdr.GMH.GMH.len += dreq_fh.GetSerializedSize();
	  else {
		  //assign a CID and include it in the message
		  Connection *data = new Connection (CONN_DATA);
/*
		  data->setCDMA(0);
		  data->initCDMA();
    		  data->setPOLL_interval(0);
*/

		  data->set_serviceflow(n);
		  // We will move all the ARQ information in the service flow to the isArqStatus that maintains the Arq Information.
		  if(data->get_serviceflow ()->getQoS ()->getIsArqEnabled () == 1 ){
		  		debug2("add an ARQ enabled connection.\n");
			  arqstatus = new Arqstatus ();
			  data->setArqStatus (arqstatus);
			  data->getArqStatus ()->setArqEnabled (data->get_serviceflow ()->getQoS ()->getIsArqEnabled ());
			  data->getArqStatus ()->setRetransTime (data->get_serviceflow ()->getQoS ()->getArqRetransTime ()); 
			  data->getArqStatus ()->setMaxWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ());
                          data->getArqStatus ()->setCurrWindow (data->get_serviceflow ()->getQoS ()->getArqMaxWindow ()); 
			  data->getArqStatus ()->setAckPeriod (data->get_serviceflow ()->getQoS ()->getArqAckPeriod ());
			  // setting timer array for the flow
			  data->getArqStatus ()->arqRetransTimer = new ARQTimer(data);      /*RPI*/
			  debug2("In init_static_flow, ARQ connection timer is generated.\n");
		  } 
		  if (n->getDirection() == UL )
		  {
			  mac_->getCManager()->add_connection (data, OUT_CONNECTION);
			  debug2("ARQ UL connection.\n");
		  }
		  else
		  {
			  mac_->getCManager()->add_connection (data, IN_CONNECTION); 
			  debug2("ARQ DL connection.\n");
		  }
		  if (n->getDirection() == UL)
		  {
			  peer->setInData (data);
			  debug2("set incoming data connection for mac %s\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
		  }
		  else
		  {
			  peer->setOutData (data);
			  debug2("set outcoming data connection for mac %s\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
		  }
		  dsa_frame->cid = data->get_cid();
		  wimaxHdr.GMH.GMH.len += dreq_fh.GetSerializedSize();
	  }
 
	  debug2(" sampad in init staic flows before end\n");
	  wimaxHdr.GMH.GMH.cid = peer->getPrimary(OUT_CONNECTION)->get_cid();
	  p->AddHeader(dreq_fh);
	  p->AddHeader(wimaxHdr);
	  peer->getPrimary(OUT_CONNECTION)->enqueue (p);
  }
}

}//namespace ns3
// End RPI

