import math
import ns3
import goocanvas
from visualizer.base import Link, transform_distance_canvas_to_simulation

class Wimax2Link(Link):
    def __init__(self, parent_canvas_item, ss, mac):
        self.node1 = ss
        self.mac = mac 
        self.node2 = None # bs 
        self.canvas_item = goocanvas.Group(parent=parent_canvas_item)
        self.invisible_line = goocanvas.Polyline(parent=self.canvas_item,
                                                 line_width=100.0,
                                                 visibility=goocanvas.ITEM_HIDDEN)
        self.visible_line = goocanvas.Polyline(parent=self.canvas_item,
                                              line_width=1.0,
                                              stroke_color_rgba=0x00000FFFF,
                                              line_dash=goocanvas.LineDash([2.0, 2.0 ]))
        self.invisible_line.props.pointer_events = (goocanvas.EVENTS_STROKE_MASK
                                                    |goocanvas.EVENTS_FILL_MASK
                                                    |goocanvas.EVENTS_PAINTED_MASK)
        self.canvas_item.set_data("pyviz-object", self)
        self.canvas_item.lower(None)
        self.set_bs(None)

    def set_bs(self, bs):
        if bs is self.node2:
            return
        if self.node2 is not None:
            self.node2.remove_link(self)
        self.node2 = bs 
        if self.node2 is None:
            self.canvas_item.set_property("visibility", goocanvas.ITEM_HIDDEN)
        else:
            self.node2.add_link(self)
            self.canvas_item.set_property("visibility", goocanvas.ITEM_VISIBLE)
        self.update_points()

    def update_points(self):
        if self.node2 is None:
            return
        pos1_x, pos1_y = self.node1.get_position()
        pos2_x, pos2_y = self.node2.get_position()
        points = goocanvas.Points([(pos1_x, pos1_y), (pos2_x, pos2_y)])
        self.visible_line.set_property("points", points)
        self.invisible_line.set_property("points", points)
        
    def destroy(self):
        self.canvas_item.destroy()
        self.node1 = None
        self.node2 = None

    def tooltip_query(self, tooltip):
        pos1_x, pos1_y = self.node1.get_position()
        pos2_x, pos2_y = self.node2.get_position()
        dx = pos2_x - pos1_x
        dy = pos2_y - pos1_y
        d = transform_distance_canvas_to_simulation(math.sqrt(dx*dx + dy*dy))
        mac = self.mac
        tooltip.set_text(("WiMAX link between SS Node %s and BS Node %s; distance=%.2f m.\n")
                         % (mac.addr(), mac.getBSAddr(), d))


class Wimax2LinkMonitor(object):
    def __init__(self, dummy_viz):
        self.base_stations = {} # bsid -> node
        self.stations = [] # list of (ss_netdevice, viz_node, wimax_link)

    def scan_nodes(self, viz):
        for (ss_netdevice, viz_node, wimax_link) in self.stations:
            wimax_link.destroy()

        self.access_points = {}
        self.stations = []

        for node in viz.nodes.itervalues():
            ns3_node = ns3.NodeList.GetNode(node.node_index)
            for devI in range(ns3_node.GetNDevices()):
                wimax_mac = ns3_node.GetDevice(devI)
                if not isinstance(wimax_mac, ns3.Wimax2NetDevice):
                    continue
                if isinstance(wimax_mac, ns3.Wimax2SSNetDevice):
                    wimax_link = Wimax2Link(viz.links_group, node, wimax_mac)
                    self.stations.append((wimax_mac, node, wimax_link))
                elif isinstance(wimax_mac, ns3.Wimax2BSNetDevice):
                    bssid = ns3.Mac48Address.ConvertFrom(wimax_mac.GetAddress())
                    self.access_points[str(bssid)] = node

    def simulation_periodic_update(self, viz):
        for (ss_netdevice, viz_node, wimax_link) in self.stations:
            if not ss_netdevice.IsLinkUp():
                wimax_link.set_bs(None)
                continue
            bssid = str(ss_netdevice.getBSAddr())
            if bssid == '00:00:00:00:00:00':
                wimax_link.set_bs(None)
                continue
            bs = self.access_points[bssid]
            wimax_link.set_bs(bs)

    def update_view(self, viz):
        for (dummy_ss_netdevice, dummy_viz_node, wimax_link) in self.stations:
            if wimax_link is not None:
                wimax_link.update_points()


def register(viz):
    link_monitor = Wimax2LinkMonitor(viz)
    viz.connect("simulation-periodic-update", link_monitor.simulation_periodic_update)
    viz.connect("update-view", link_monitor.update_view)
    viz.connect("topology-scanned", link_monitor.scan_nodes)
