/* This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * <BR>
 * We would appreciate acknowledgement if the software is used.
 * <BR>
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * </PRE></P>
 * @author  Richard Rouil
 * @modified by  Chakchai So-In
 */

#include "ns3/wimax2-bs-scheduler.h"
#include "ns3/burst.h"
#include "ns3/dlburst.h"
#include "ns3/ulburst.h"
#include "ns3/random-variable.h"
#include "ns3/wimax2-control-agent.h"
#include "ns3/wimax2-bs-net-device.h"
#include "ns3/wimax2-net-device.h"
#include "ns3/dlsubframetimer.h"
#include <stdlib.h>
#include <fstream>

#define FRAME_SIZE 0.005
#define CODE_SIZE 256
#define CDMA_6SUB 6
#define MAX_CONN 2048

//Scheduler allocates CBR every frame 
#define UGS_AVG	

namespace ns3 {

//This structure is used for virtual allocation and "index_burst" is used for #counting bursts
int index_burst=0;
struct virtual_burst {
  int alloc_type; //0 = DL_MAP, 1 = UL_MAP, 2 = DCD, 3 = UCD, 4 = other broadcast, 5 = DL_burst, 6 = UL_burst
  int cid;
  int n_cid;
  int iuc;
  Ofdm_mod_rate mod;
  bool preamble;
  int symboloffset;
  int suboffset;
  int numsymbol;
  int numsub;   
  int numslots; 
  float byte;     
  int rep;      
  int dl_ul;    //0 = DL, 1 = UL
  int ie_type;
  bool mapped;
} virtual_alloc[MAX_MAP_IE], ul_virtual_alloc[MAX_MAP_IE];

struct allocation_listing {
	int32_t height;
	int32_t width;
	int32_t wasted;
};

struct top_region_listing {
	int32_t height;
	int32_t width;
	int32_t slot_offset;
	int32_t subchannel_offset;
	bool valid; //we store regions in a vector, which doesn't support fast deletion, so this is flagged to ignore
};




/*
 * Create a scheduler
 */
Wimax2BSScheduler::Wimax2BSScheduler () : Wimax2Scheduler ()
{
  debug2 ("Wimax2BSScheduler created\n");
  default_mod_ = OFDM_BPSK_1_2;
  Repetition_code_ = 4; //this and next two should really be set from macmib
  init_contention_size_ = 4;
  bw_req_contention_size_ = 4;

//  contention_size_ = MIN_CONTENTION_SIZE;

  nextDL_ = -1;
  nextUL_ = -1;  

  mt_ = eOCSA;

  allocation_file.open("allocation_information.tx");
  maximum_number_dl_bursts = 107; // maximum number of IE's that will fit if the DL_MAP is repeated only once, has mod OFDM_64QAM_3_4 (current max), and fills the entire 28 subchannels * 2 symbols
  total_overallocations = new uint64_t[maximum_number_dl_bursts];
  total_unmapped_blocks = new uint64_t[maximum_number_dl_bursts];
  total_unmapped_bursts = new uint64_t[maximum_number_dl_bursts];
  total_empty_needed_blocks = new uint64_t[maximum_number_dl_bursts];
  total_number_of_frames = new uint64_t[maximum_number_dl_bursts];
  max_dl_bursts = 0;
}

Wimax2BSScheduler::~Wimax2BSScheduler() {

	delete [] total_overallocations;
	delete [] total_unmapped_blocks;
	delete [] total_unmapped_bursts;
	delete [] total_empty_needed_blocks;
	delete [] total_number_of_frames;
	//this is used to dump final statistics, which it probably shouldn't be
}
 
/**
 * Return the MAC casted to Wimax2BSScheduler
 * @return The MAC casted to Wimax2BSScheduler
 */
Ptr<Wimax2BSNetDevice> Wimax2BSScheduler::getMac()
{
  return mac_->GetObject<Wimax2BSNetDevice>();
}


/**
 * Initializes the scheduler
 */
void Wimax2BSScheduler::init ()
{
   debug10("in bs-scheduler::init()\n");
   Wimax2Scheduler::init();

  // If the user did not set the profiles by hand, let's do it
  // automatically
  if (getMac()->getMap()->getDlSubframe()->getProfile (DIUC_PROFILE_1)==NULL) {
    //#ifdef SAM_DEBUG
    debug2 ("Adding profiles\n");
    //#endif
    Profile *p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_BPSK_1_2);
    p->setIUC (DIUC_PROFILE_1);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_QPSK_1_2);
    p->setIUC (DIUC_PROFILE_2);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_QPSK_3_4);
    p->setIUC (DIUC_PROFILE_3);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_16QAM_1_2);
    p->setIUC (DIUC_PROFILE_4);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_16QAM_3_4);
    p->setIUC (DIUC_PROFILE_5);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_64QAM_2_3);
    p->setIUC (DIUC_PROFILE_6);
    p = getMac()->getMap()->getDlSubframe()->addProfile ((int)round((getMac()->getPhy()->getFreq()/1000)), OFDM_64QAM_3_4);
    p->setIUC (DIUC_PROFILE_7);

    //in order for contention regions to fit in the allocation slots, they must be QPSK 1/2 or higher
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_QPSK_1_2);
    p->setIUC (UIUC_INITIAL_RANGING);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_QPSK_1_2);
    p->setIUC (UIUC_REQ_REGION_FULL);  

    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_BPSK_1_2);
    p->setIUC (UIUC_PROFILE_1);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_QPSK_1_2);
    p->setIUC (UIUC_PROFILE_2);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_QPSK_3_4);
    p->setIUC (UIUC_PROFILE_3);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_16QAM_1_2);
    p->setIUC (UIUC_PROFILE_4);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_16QAM_3_4);
    p->setIUC (UIUC_PROFILE_5);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_64QAM_2_3);
    p->setIUC (UIUC_PROFILE_6);
    p = getMac()->getMap()->getUlSubframe()->addProfile (0, OFDM_64QAM_3_4);
    p->setIUC (UIUC_PROFILE_7);
  }

  //init contention slots
  ContentionSlot *slot = getMac()->getMap()->getUlSubframe()->getRanging ();
  //debug10("%p->%p->%p->%p = %p\n",getMac(), getMac()->getMap(), getMac()->getMap()->getUlSubframe(), getMac()->getMap()->getUlSubframe()->getRanging(), slot);
//  slot->setSize (getInitRangingopportunity ());
  slot->setSize (getMac()->macmib_->init_contention_size);
  slot->setBackoff_start (getMac()->macmib_->rng_backoff_start);
  slot->setBackoff_stop (getMac()->macmib_->rng_backoff_stop);
  
  slot = getMac()->getMap()->getUlSubframe()->getBw_req ();
//  slot->setSize (getBWopportunity ());
  slot->setSize (getMac()->macmib_->bw_req_contention_size);
  slot->setBackoff_start (getMac()->macmib_->bw_backoff_start);
  slot->setBackoff_stop (getMac()->macmib_->bw_backoff_stop);

}

/**
 * Compute and return the bandwidth request opportunity size
 * @return The bandwidth request opportunity size
 */
int Wimax2BSScheduler::getBWopportunity ()
{
  int nbPS = BW_REQ_PREAMBLE * getMac()->getPhy()->getSymbolPS();
  //add PS for carrying header
  nbPS += (int) round((getMac()->getPhy()->getTrxTime (HDR_MAC802_16_SIZE, getMac()->getMap()->getUlSubframe()->getProfile(UIUC_REQ_REGION_FULL)->getEncoding())/getMac()->getPhy()->getPS ()));
  debug2 ("BWopportunity size=%d\n", nbPS);
  debug10 ("BWopportunity size 1 oppo :%d nbPS, BW_PREAMBLE :%d + HDR_MAC802.16 :%d, with UIUC_REQ_REGION_FULL :%d\n", nbPS, BW_REQ_PREAMBLE, HDR_MAC802_16_SIZE, UIUC_REQ_REGION_FULL);
  return nbPS;
}

/**
 * Compute and return the initial ranging opportunity size
 * @return The initial ranging opportunity size
 */
int Wimax2BSScheduler::getInitRangingopportunity ()
{
  int nbPS = INIT_RNG_PREAMBLE * getMac()->getPhy()->getSymbolPS();
  debug2("nbPS = %d " , nbPS);  
  //add PS for carrying header
  nbPS += (int) round((getMac()->getPhy()->getTrxTime(RNG_REQ_SIZE+HDR_MAC802_16_SIZE, getMac()->getMap()->getUlSubframe()->getProfile(UIUC_INITIAL_RANGING)->getEncoding())/getMac()->getPhy()->getPS()));
  debug2 ("Init ranging opportunity size=%d\n", nbPS);
  debug10 ("InitRanging opportunity size 1 oppo :%d nbPS, INIT_RNG_PREAMBLE :%d + RNG_REQ_SIZE :%d + HDR_MAC802.16 :%d, with UIUC_INITIAL_RANGING :%d\n", nbPS, INIT_RNG_PREAMBLE, RNG_REQ_SIZE, HDR_MAC802_16_SIZE, UIUC_INITIAL_RANGING);
  return nbPS;  
}



// This function is used to increase #DL_MAP_IEs (DL_MAP size)
int Wimax2BSScheduler::increase_dl_map_ie(int num_of_entries, int totalslots, int num_ie) {

  debug10("Increasing #DL_MAP_IEs.  num_of_entires: %d, totalslots: %d, num_ie: %d\n",num_of_entries, totalslots, num_ie);

  Ptr<Wimax2Phy> phy = mac_->getPhy();
  DL_MAP_IE sz;
  DL_MAP_FRAME_HEADER dmhf;
  dmhf.mf.nb_ies = num_of_entries;
  
  int total_slots_per_rep = ceil((double)dmhf.GetSerializedSize() / (double) phy->getSlotCapacity(virtual_alloc[0].mod, DL_));
  while ((total_slots_per_rep * virtual_alloc[0].rep) > (phy->getNumsubchannels(DL_) - 2)) { //if we can't fit in the reserved first column...
  	if (virtual_alloc[0].mod == OFDM_64QAM_3_4) {
		virtual_alloc[0].rep -= 2;
	} else {
		virtual_alloc[0].mod = static_cast<Ofdm_mod_rate>  (static_cast<int>(virtual_alloc[0].mod) + 1);
		total_slots_per_rep = ceil((double)dmhf.GetSerializedSize() / (double) phy->getSlotCapacity(virtual_alloc[0].mod, DL_));
	}
  }

  virtual_alloc[0].iuc = getDIUCProfile(virtual_alloc[0].mod);


  int slots_sofar = total_slots_per_rep * virtual_alloc[0].rep;
  for(int i = 1; i<num_of_entries; i++){
    slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }

  int32_t new_total_bytes =  total_slots_per_rep * virtual_alloc[0].rep * phy->getSlotCapacity(virtual_alloc[0].mod, DL_);
  int32_t new_total_slots =  total_slots_per_rep * virtual_alloc[0].rep;
  

  if ( slots_sofar <= totalslots) {
        debug10 ("Check_DLMAP. Old DL map byte :%f, slots :%d, New DL map byte :%d, slots :%d, #ie :%d\n", virtual_alloc[0].byte, virtual_alloc[0].numslots, new_total_bytes, new_total_slots, num_ie);

	virtual_alloc[0].byte = new_total_bytes;
	virtual_alloc[0].numslots = new_total_slots;
	return slots_sofar;
  } else {
  	debug10("slots_sofar: %d, totalslots: %d\n", slots_sofar, totalslots);
	return -1;
  }

}


// This function is used to check if the allocation is feasible (enough slots for both DL_MAP_IE and its own data allocation
int Wimax2BSScheduler::overallocation_withdlmap(int num_of_entries, int totalslots, int ownslots) {

  Ptr<Wimax2Phy> phy = mac_->getPhy();
  int slots_sofar = 0;
  float dl_ie_size = DL_MAP_IE_SIZE;
//padding to byte based
  float dl_map_increase_byte = virtual_alloc[0].byte + dl_ie_size;
  int dl_map_increase_slots = virtual_alloc[0].rep*(int)ceil((ceil)(dl_map_increase_byte)/(double)phy->getSlotCapacity(virtual_alloc[0].mod,DL_));

  debug10 ("Check_DLMAP.From array DL MAP byte :%f, slots :%d, IE_SIZE :%f, slotcapacity :%d\n", virtual_alloc[0].byte, virtual_alloc[0].numslots, dl_ie_size, phy->getSlotCapacity(virtual_alloc[0].mod,DL_));
  debug10 ("\tNew DL MAP byte :%f, slots :%d\n", dl_map_increase_byte, dl_map_increase_slots);

  for(int i = 1; i<num_of_entries; i++){
	debug2 ("\t index :%d, numslots :%d, accum :%d\n", i, virtual_alloc[i].numslots, slots_sofar);
        slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }

  int slots_temp = slots_sofar;
  slots_sofar = slots_sofar + ownslots + dl_map_increase_slots;
  if ( slots_sofar <= totalslots) {
        debug10 ("Check_DLMAP. Old DL map byte :%f, slots :%d, New DL map byte :%f, slots :%d, available burst slot size :%d\n", virtual_alloc[0].byte, virtual_alloc[0].numslots, dl_map_increase_byte, dl_map_increase_slots, ownslots);

	virtual_alloc[0].byte = dl_map_increase_byte;
	virtual_alloc[0].numslots = dl_map_increase_slots;
	return ownslots;
  } else {
	if ( (slots_temp + dl_map_increase_slots) <= totalslots ) {
        	debug10 ("Check_DLMAP. Old DL map byte :%f, slots :%d, New DL map byte :%f, slots :%d, available slot burst size :%d\n", virtual_alloc[0].byte, virtual_alloc[0].numslots, dl_map_increase_byte, dl_map_increase_slots, totalslots - slots_temp - dl_map_increase_slots);
		virtual_alloc[0].byte = dl_map_increase_byte;
		virtual_alloc[0].numslots = dl_map_increase_slots;
		return (totalslots - slots_temp - dl_map_increase_slots);
	} else {
		return -1;
	}
  }
}


// This function is used to calculate maximum number of connection with the increase of DL_MAP_IE
int Wimax2BSScheduler::max_conn_withdlmap(int num_of_entries, int totalslots) {

  Ptr<Wimax2Phy> phy = mac_->getPhy();
  int slots_sofar = 0;
  float dl_ie_size = DL_MAP_IE_SIZE;
  int max_conn = 0;
  float dl_map_increase_byte = 0; if (dl_map_increase_byte);
  int dl_map_increase_slots = 0; if (dl_map_increase_slots);
  float virtual_increase_byte = virtual_alloc[0].byte;
  int virtual_increase_slots = virtual_alloc[0].numslots;

  for(int i = 0; i<num_of_entries; i++) {
//	debug2 ("\t index :%d, numslots :%d, accum :%d\n", i, virtual_alloc[i].numslots, slots_sofar);
	slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }
  int slots_data = slots_sofar - virtual_alloc[0].numslots;

  debug10 ("Check_DLMAP. Max supported conn total_dl_slots :%d, sofar slots :%d, dlmap slots :%d, dlmap byte :%f, available slots :%d\n", totalslots, slots_sofar, virtual_alloc[0].numslots, virtual_alloc[0].byte, totalslots-slots_sofar);

  while (1) {
	virtual_increase_byte += dl_ie_size;
	int virtual_increase_byte_ceil = (int)ceil(virtual_increase_byte);
	virtual_increase_slots = virtual_alloc[0].rep*ceil((double)(virtual_increase_byte_ceil)/(double)phy->getSlotCapacity(virtual_alloc[0].mod,DL_));
	if ( (virtual_increase_slots + slots_data) >= totalslots ) {
	  break;
	} else {
	  max_conn++;
	}
  }

  return max_conn;
}


// This function is used to calculate number of available slots after #DL_MAP_IEs are taken into account.
int Wimax2BSScheduler::freeslots_withdlmap_given_conn(int num_of_entries, int totalslots, int newconn) {

  Ptr<Wimax2Phy> phy = mac_->getPhy();
  int slots_sofar = 0;
  float dl_ie_size = DL_MAP_IE_SIZE;
  int max_conn = 0; if (max_conn);
  float dl_map_increase_byte = 0; if (dl_map_increase_byte);
  int dl_map_increase_slots = 0; if (dl_map_increase_slots);
  float virtual_increase_byte = virtual_alloc[0].byte;
  int virtual_increase_slots = virtual_alloc[0].numslots;
  int total_slots_withdlmap_nodata = 0; if (total_slots_withdlmap_nodata);

  for(int i = 0; i<num_of_entries; i++) {
//	debug2 ("\t index :%d, numslots :%d, accum :%d\n", i, virtual_alloc[i].numslots, slots_sofar);
	slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }
  int slots_data = slots_sofar - virtual_alloc[0].numslots;

  for (int i=0; i<newconn; i++) {
	virtual_increase_byte += dl_ie_size;
  }
  int virtual_increase_byte_ceil = (int)ceil(virtual_increase_byte);
  virtual_increase_slots = virtual_alloc[0].rep*ceil((double)(virtual_increase_byte_ceil)/(double)phy->getSlotCapacity(virtual_alloc[0].mod,DL_));

  debug10 ("Check_DLMAP. #Free slots newconn :%d, sofar slots :%d, dlmap slots :%d, dlmap byte :%f, newdlmap slots :%d, newdlmap byte :%d\n", newconn, slots_sofar, virtual_alloc[0].numslots, virtual_alloc[0].byte, virtual_increase_slots, virtual_increase_byte_ceil);

  return totalslots - slots_data - virtual_increase_slots;
}


// This function is used check #available slots excluing DL_MAP_IE
// return #request slots if feasible else return #available slots
int Wimax2BSScheduler::overallocation_withoutdlmap(int num_of_entries, int totalslots, int ownslots) {

  int slots_sofar = 0;

  for(int i = 0; i<num_of_entries; i++){
        slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }

  int slots_temp = slots_sofar;
  slots_sofar = slots_sofar + ownslots;
  if ( slots_sofar <= totalslots) {
	return ownslots;
  } else {
	return (totalslots-slots_temp);
  }
}


// This function is a traditional bubble sort.
// Note that "sort_field" can be ascending/descending order
void Wimax2BSScheduler::bubble_sort (int arrayLength, con_data_alloc array[], int sort_field) {
   int i, j, flag = 1;
   con_data_alloc temp;

   for (int k = 0; k < arrayLength; k++) {
//        debug10 ("-Before i :%d, cid :%d, counter :%d, req_slots :%d, mod :%d\n", k, array[k].cid, array[k].counter, array[k].req_slots, (int)array[k].mod_rate);
   }

   if (sort_field == 0) {
   for (i = 1; (i <= arrayLength) && flag; i++) {
       flag = 0;
       for (j = 0; j < (arrayLength -1); j++) {
           if (array[j+1].req_slots < array[j].req_slots) {
              temp.cid = array[j].cid;
              temp.direction = array[j].direction;
              temp.mod_rate = array[j].mod_rate;
              temp.req_slots = array[j].req_slots;
              temp.grant_slots = array[j].grant_slots;
              temp.weight = array[j].weight;
              temp.counter = array[j].counter;

              array[j].cid = array[j+1].cid;
              array[j].direction = array[j+1].direction;
              array[j].mod_rate = array[j+1].mod_rate;
              array[j].req_slots = array[j+1].req_slots;
              array[j].grant_slots = array[j+1].grant_slots;
              array[j].weight = array[j+1].weight;
              array[j].counter = array[j+1].counter;

              array[j+1].cid = temp.cid;
              array[j+1].direction = temp.direction;
              array[j+1].mod_rate = temp.mod_rate;
              array[j+1].req_slots = temp.req_slots;
              array[j+1].grant_slots = temp.grant_slots;
              array[j+1].weight = temp.weight;
              array[j+1].counter = temp.counter;
              flag = 1;
           }//end if
       }//end 2nd for
     }//end 1st for
   }//end sort_field == 0

   if (sort_field == 1) {
   for (i = 1; (i <= arrayLength) && flag; i++) {
       flag = 0;
       for (j = 0; j < (arrayLength -1); j++) {
           if (array[j+1].counter > array[j].counter) {
              temp.cid = array[j].cid;
              temp.direction = array[j].direction;
              temp.mod_rate = array[j].mod_rate;
              temp.req_slots = array[j].req_slots;
              temp.grant_slots = array[j].grant_slots;
              temp.weight = array[j].weight;
              temp.counter = array[j].counter;

              array[j].cid = array[j+1].cid;
              array[j].direction = array[j+1].direction;
              array[j].mod_rate = array[j+1].mod_rate;
              array[j].req_slots = array[j+1].req_slots;
              array[j].grant_slots = array[j+1].grant_slots;
              array[j].weight = array[j+1].weight;
              array[j].counter = array[j+1].counter;

              array[j+1].cid = temp.cid;
              array[j+1].direction = temp.direction;
              array[j+1].mod_rate = temp.mod_rate;
              array[j+1].req_slots = temp.req_slots;
              array[j+1].grant_slots = temp.grant_slots;
              array[j+1].weight = temp.weight;
              array[j+1].counter = temp.counter;
              flag = 1;
           }//end if
       }//end 2nd for
     }//end 1st for
   }//end sort_field == 1

   for (int k = 0; k < arrayLength; k++) {
//        debug10 ("-After i :%d, cid :%d, counter :%d, req_slots :%d, mod :%d\n", k, array[k].cid, array[k].counter, array[k].req_slots, (int)array[k].mod_rate);
   }

   return;   //arrays are passed to functions by address; nothing is returned
}


// This function is used to check if the virutall allocation exists
int Wimax2BSScheduler::doesvirtual_allocexist(int num_of_entries, int cid) {
  for(int i = 0; i<num_of_entries-1; i++){
	if (virtual_alloc[i].cid == cid) {
		return cid;
	}
  }
  return -1;
}


// This function is used to add the virtual burst regardless of DL_MAP_IE
int Wimax2BSScheduler::addslots_withoutdlmap(int num_of_entries, int byte, int slots, int cid) {
  for(int i = 0; i<num_of_entries-1; i++){
	if (virtual_alloc[i].cid == cid) {
		virtual_alloc[i].byte += byte;
		virtual_alloc[i].numslots += slots;
		return i;
	}
  }
  return -1;
}


// This function is used caculate #slots used so far
int Wimax2BSScheduler::check_overallocation(int num_of_entries){

  int slots_sofar = 0;
  for(int i = 0; i<num_of_entries; i++){
        slots_sofar = slots_sofar + virtual_alloc[i].numslots;
  }
  return slots_sofar;
}


/**
 * Schedule bursts/packets
 */
void Wimax2BSScheduler::schedule ()
{
	debug10("entering (BS-S) %s schedule\n", mac_->addr_s());

  //The scheduler will perform the following steps:
  //1-Clear UL map
  //2-Allocate CDMA-Ranging region for initial ranging (2 OFDMs) and bw-req (1 OFDMs)
  //  In this version we reserved a whold column resulting 5-subchannel waste
  //3-Allocate UL data allocation (all-cid) per-MS allocation
  //  In this version we do not consider horizontal stripping yet
  //4-Clear DL map
  //5-Allocate Burst for Broadcast message
  //  In this version we allocate one burst for each broadcast message type say DL_MAP, UL_MAP, DCD, UCD, other other broadcast message
  //6-Allocate DL data allocation (all-cid) per-cid allocation
  //7-Assign burst -> physical allocation
  //Note that, we do not simulate FCH in this version. We will do that in next version.

  Ptr<Packet> p;
  //struct hdr_cmn *ch;
  All_Headers ah_p;
  DlBurst *db;
  PeerNode *peer;
  int Channel_num=0;

  // We will try to Fill in the ARQ Feedback Information now...
  Ptr<Packet>  ph = 0;
  PeerNode * peernode;
  Connection * basic ;
  Connection * OutData;
  Ptr<Packet>  pfb = 0;
  All_Headers wimaxHdrMap ;
  
  uint16_t   temp_num_of_acks = 0;
  bool out_datacnx_exists = false; 
  for (Connection *n= mac_->getCManager ()->get_in_connection (); 
       n; n=n->next_entry()) {
	if(n->getArqStatus () != NULL && n->getArqStatus ()->isArqEnabled() == 1) {
		if(!(n->getArqStatus ()->arq_feedback_queue_) || (n->getArqStatus ()->arq_feedback_queue_->GetNPackets() == 0)) 
			continue; 
		else {
		  peernode = n->getPeerNode ();
		  if(peernode->getOutData () != NULL && peernode->getOutData ()->queueLength () != 0 && getMac()->arqfb_in_dl_data_) {
			out_datacnx_exists = true;
		  }
		  if(out_datacnx_exists == false) {
			//debug2("ARQ BS: Feedback in Basic Cid \n");
			basic = peernode->getBasic (OUT_CONNECTION);
			pfb = mac_->getPacket ();
			pfb->RemoveHeader(wimaxHdrMap);
			wimaxHdrMap.GMH.GMH.cid = basic->get_cid ();	
			wimaxHdrMap.num_of_acks = 0;
		  }
		  else {
			debug2("ARQ BS : Feedback in data Cid \n");
			OutData = peernode->getOutData ();
			pfb = OutData->dequeue ();
			pfb->RemoveHeader(wimaxHdrMap);
			if(wimaxHdrMap.GMH.GMH.type_arqfb == 1) {
				debug2("ARQ BS: Feedback already present, do nothing \n");
				pfb->AddHeader(wimaxHdrMap);
				OutData->enqueue_head (pfb);
				continue;
			  }			
			wimaxHdrMap.num_of_acks = 0;
		  }
			  //Dequeue is only once as one feedback in 5ms frame is permitted 
		  ph = n->getArqStatus ()->arq_feedback_queue_->Dequeue();
		  wimaxHdrMap.GMH.GMH.type_arqfb = 1;
		  All_Headers wimaxHdr;
		  
		  ph->RemoveHeader(wimaxHdr);
		  
		  for (temp_num_of_acks = 0; temp_num_of_acks < wimaxHdr.num_of_acks; temp_num_of_acks++)
			{	
			  wimaxHdrMap.arq_ie[temp_num_of_acks].cid = wimaxHdr.arq_ie[temp_num_of_acks].cid;
			  wimaxHdrMap.arq_ie[temp_num_of_acks].last = wimaxHdr.arq_ie[temp_num_of_acks].last;
			  wimaxHdrMap.arq_ie[temp_num_of_acks].ack_type = wimaxHdr.arq_ie[temp_num_of_acks].ack_type;
			  wimaxHdrMap.arq_ie[temp_num_of_acks].fsn = wimaxHdr.arq_ie[temp_num_of_acks].fsn;
			}    
		  wimaxHdrMap.num_of_acks = wimaxHdr.num_of_acks;
		  wimaxHdrMap.GMH.GMH.len += wimaxHdr.num_of_acks * HDR_MAC802_16_ARQFEEDBK_SIZE;
		  debug2("ARQ : In Wimax2BSScheduler: Enqueueing an feedback cid: %d arq_ie->fsn:%d \n", wimaxHdr.arq_ie[0].cid,  wimaxHdr.arq_ie[0].fsn);
			  if(out_datacnx_exists == false) {	
				// If I am here then the Ack packet has been created, so we will enqueue it in the Basic Cid
				pfb->AddHeader(wimaxHdrMap);
				basic->enqueue (pfb);
			  }
			  else
				pfb->AddHeader(wimaxHdrMap);
				OutData->enqueue_head (pfb);  	
			  }
	  }
  }

  peer = mac_->getPeerNode_head();
  if(peer)
    for (int i=0; i<mac_->getNbPeerNodes() ;i++) {
      //peer->setchannel(++Channel_num);
      peer->setchannel((peer->getchannel())+1);
 
      if(peer->getchannel()>999) peer->setchannel(2); 

      peer = peer->next_entry();
    }

  if(Channel_num>999) Channel_num = 2;//GetInitialChannel();

  /// rpi random channel allocation 

  Ptr<Wimax2Phy> phy = mac_->getPhy();
  FrameMap *map = mac_->getMap();
  int nbPS = (int) floor((mac_->getFrameDuration()/phy->getPS()));
#ifdef DEBUG_WIMAX
  assert (nbPS*phy->getPS()<=mac_->getFrameDuration()); //check for rounding errors
#endif

  int nbPS_left = nbPS - mac_->phymib_.rtg - mac_->phymib_.ttg;
  int nbSymbols = (int) floor((phy->getPS()*nbPS_left)/phy->getSymbolTime());  // max num of OFDM symbols available per frame. 
    
  assert (nbSymbols*phy->getSymbolTime()+(mac_->phymib_.rtg + mac_->phymib_.ttg)*phy->getPS() < mac_->getFrameDuration());
  int dlduration = DL_PREAMBLE;                             //number of symbols currently used for downlink
  int maxdlduration = (int) (nbSymbols * (1.0 - dlratio_)); //number of symbols for downlink
  

/////////////////////
// In next version, proper use of DL:UL symbols will be calculated.
//  int nbSymbols_wo_preamble = nbSymbols - DL_PREAMBLE;
//  int nbSymbols_wo_preamble = nbSymbols;
//
//  int virtual_maxdld = (int) (nbSymbols_preamble / (1.0/dlratio_)); //number of symbols for downlink
//  int virtual_maxuld = nbSymbols_wo_preamble - virtual_maxdld;
//  int counter_maxdld = virtual_maxdld;
//
//  for (int i=counter_maxdld; i<nbSymbols_wo_preamble; i++) {
//     if ( ((virtual_maxdld % 2) == 0) && ((virtual_maxuld % 3) == 0) ) {
//	break;
//     } else {
//
//     }
//  }
/////////////////////

  int subchannel_offset = 0;
  int nbdlbursts = 0;
  int maxulduration = nbSymbols - maxdlduration;                //number of symbols for uplink
  int ulduration = 0;           		                //number of symbols currently used for uplink
  int nbulpdus = 0;
  int total_dl_subchannel_pusc = 30;
  int total_ul_subchannel_pusc = 35;
  int total_dl_slots_pusc = floor((double)(maxdlduration-DL_PREAMBLE)/2)*total_dl_subchannel_pusc;
  int total_ul_slots_pusc = floor((double)maxulduration/3)*total_ul_subchannel_pusc;
  	if (total_ul_slots_pusc); //quiet, compiler
  int numie;
  int number_ul_ie = 0;
  //int number_dl_ie = 0;
  
  debug10 ("\n---------------------------------------------------\n");
//| | | |       => symbols
//0 1 2 3       => dlduration
  debug10 ("Start BS Scheduling: TotalSymbols :%d, MAXDL :%d, MAXUL :%d, Preamble :%d\n", nbSymbols, maxdlduration, maxulduration, DL_PREAMBLE);

#ifdef DEBUG_WIMAX
  assert ((nbSymbols*phy->getSymbolPS()+mac_->phymib_.rtg + mac_->phymib_.ttg)*phy->getPS()< mac_->getFrameDuration());
  assert (maxdlduration*phy->getSymbolTime()+mac_->phymib_.rtg*phy->getPS()+maxulduration*phy->getSymbolTime()+mac_->phymib_.ttg*phy->getPS() < mac_->getFrameDuration());
#endif

  mac_->setMaxDlduration (maxdlduration);
  mac_->setMaxUlduration (maxulduration);

  //============================UL allocation=================================
  //1 and 2 - Clear Ul_MAP and Allocate inital ranging and bw-req regions
  int ul_subframe_subchannel_offset = 0;
  int total_UL_num_subchannel = phy->getNumsubchannels(UL_);
  	if (total_UL_num_subchannel); //quiet, compiler
	
  mac_->getMap()->getUlSubframe()->setStarttime (maxdlduration*phy->getSymbolPS()+mac_->phymib_.rtg);

  while (mac_->getMap()->getUlSubframe()->getNbPdu()>0) {
      PhyPdu *pdu = mac_->getMap()->getUlSubframe()->getPhyPdu(0);
      pdu->removeAllBursts();
      mac_->getMap()->getUlSubframe()->removePhyPdu(pdu);
      delete (pdu);
  }

  //Set contention slots for cdma initial ranging
  UlBurst *ub = (UlBurst*)mac_->getMap()->getUlSubframe()->addPhyPdu (nbulpdus++,0)->addBurst (0);
  ub->setIUC (UIUC_INITIAL_RANGING);
  //we're going to allocate a box of one slot duration in width by one subchannel high for each contention slot here
  //since we only need to send 6 Bytes for a CDMA header, and even the lowest possible modulation allows for that in
  //2 Symbols (and we have 3), this is fine
  int int_rng_num_sub = init_contention_size_; //set to 4 in this verison.
  int int_rng_num_sym = mac_->getPhy()->getSlotLength(UL_);
  ub->setDuration (int_rng_num_sym);
  ub->setStarttime (ulduration); 	
  ub->setSubchannelOffset (0);
  ub->setnumSubchannels(int_rng_num_sub);
  ul_subframe_subchannel_offset = (0);
  debug10 ("UL.Initial Ranging, contention_size :%d oppurtunities, 1 opportunity = 6sub*3symbols, ulduration :%d, initial_rng_duration :%d spans :%d slot(s)\n", init_contention_size_, ulduration, int_rng_num_sub, ulduration + int_rng_num_sym);

  //the end of the contention slots = the start of the UL Subframe + 3 symbol times (1 slot time)
  Time contention_endtime = Seconds( mac_->getMap()->getUlSubframe()->getStarttime() * mac_->getPhy()->getPS() + mac_->getPhy()->getSymbolTime() * mac_->getPhy()->getSlotLength(UL_));
  for (int i = 0; i < int_rng_num_sub; i++) {
  	debug10("Scheduling the end of INIT_RANGING contention slot %d for %f\n", i, contention_endtime.GetSeconds() + NOW);
  	PHY_INFO_Header *pih_cont = new PHY_INFO_Header();
	pih_cont->num_OFDMSymbol = mac_->getPhy()->getSlotLength(UL_);
	pih_cont->num_subchannels = 1;
	pih_cont->subchannel_offset = i;
	pih_cont->OFDMSymbol_offset = 0;
	pih_cont->direction = UL;
	Simulator::ScheduleWithContext(Simulator::GetContext(), contention_endtime, &Wimax2NetDevice::assemble_incoming_packet, mac_, pih_cont);
  }

  ulduration += int_rng_num_sym; 

  //Set contention slots for cdma bandwidth request
  ub = (UlBurst*)mac_->getMap()->getUlSubframe()->addPhyPdu (nbulpdus++,0)->addBurst (0);
  ub->setIUC (UIUC_REQ_REGION_FULL);
  int bw_rng_num_sub = bw_req_contention_size_;	//bw_req_contention_size is set to 4 in this version
  int bw_rng_num_sym = mac_->getPhy()->getSlotLength(UL_);
  
  ub->setDuration (bw_rng_num_sym);
  ub->setStarttime (ulduration);
  ub->setSubchannelOffset (0);
  ub->setnumSubchannels(bw_rng_num_sub);
  ul_subframe_subchannel_offset = (int_rng_num_sub);
  debug10 ("UL.Bw-req, contention_size :%d, 1opportunity = 6sub*1symbols, ulduration :%d, bw-req_duration :%d spans %d slot(s)\n", bw_req_contention_size_, ulduration, bw_rng_num_sym, ulduration + bw_rng_num_sym);

  for (int i = 0; i < bw_rng_num_sub; i++) {
  	debug10("Scheduling the end of CDMA_BW_REQ contention slot %d for %f\n", i, contention_endtime.GetSeconds());
  	PHY_INFO_Header *pih_cont = new PHY_INFO_Header();
	pih_cont->num_OFDMSymbol = mac_->getPhy()->getSlotLength(UL_);
	pih_cont->num_subchannels = 1;
	pih_cont->subchannel_offset = i+int_rng_num_sub;
	pih_cont->OFDMSymbol_offset = 0;
	pih_cont->direction = UL;
	//since they're all in the same  column, the end time is the same as the init ranging slots
	Simulator::ScheduleWithContext(Simulator::GetContext(), contention_endtime, &Wimax2NetDevice::assemble_incoming_packet, mac_, pih_cont);
  }
  ulduration += bw_rng_num_sym; 

  //mac_->setStartUlduration (ulduration);

  peer = mac_->getPeerNode_head();

  if(ulduration > maxulduration ) { debug2 (" not enough UL symbols to allocate \n " ); }

  // This cdma_flag is used in case there is no peer yet however the allocation for cdma_transmission oppportunity needed to be allocated by the BS-UL scheduler
  int cdma_flag = 0;
  Connection *con1;
  con1 = mac_->getCManager()->get_in_connection();
  while (con1!=NULL){
    if(con1->get_category() == CONN_INIT_RANGING) {
        if (con1->getCDMA()>0) {
           cdma_flag = 2;
        }
    }
    con1 = con1->next_entry();
  }

  //Call ul_stage2 to allocate the uplink resource
  if ( (peer)  || (cdma_flag>0) ) {
    if(maxulduration > ulduration) {
	
	debug2 ("UL.Before going to ul_stage2 (data allocation) Frame duration :%f, PSduration :%e, Symboltime :%e, nbPS :%d, rtg :%d, ttg :%d, nbPSleft :%d, nbSymbols :%d\n", mac_->getFrameDuration(), phy->getPS(), phy->getSymbolTime(), nbPS, mac_->phymib_.rtg, mac_->phymib_.ttg, nbPS_left, nbSymbols);
	
	debug2 ("\tmaxdlduration :%d, maxulduration :%d, numsubchannels :%d, ulduration :%d\n", maxdlduration, (maxulduration), phy->getNumsubchannels(/*phy->getPermutationscheme (),*/ UL_), ulduration);


	//VERTICAL_STRIPPING is no longer supported
	//mac802_16_ul_map_frame * ulmap = ul_stage2 (mac_->getCManager()->get_in_connection (), phy->getNumsubchannels( UL_), (maxulduration-ulduration), ulduration, VERTICAL_STRIPPING );
	mac802_16_ul_map_frame *ulmap = ul_stage2 (mac_->getCManager()->get_in_connection(), phy->getNumsubchannels( UL_ ), maxulduration, ulduration, HORIZONTAL_STRIPPING);

// we don't support horizontal stripping in this version
// mac802_16_ul_map_frame * ulmap = ul_stage2 (mac_->getCManager()->get_in_connection (), phy->getNumsubchannels( UL_), maxulduration-ulduration,ulduration, HORIZONTAL_STRIPPING );

// From UL_MAP_IE, we map the allocation into UL_BURST
	number_ul_ie = (int)ulmap->nb_ies;
	for (numie = 0 ; numie < (int) ulmap->nb_ies ; numie++) {
	    mac802_16_ofdma_ulmap_ie ulmap_ie = ulmap->ies[numie].ie;
	    ub = (UlBurst*) mac_->getMap()->getUlSubframe()->addPhyPdu (nbulpdus++,0)->addBurst (0);
	    
#ifdef SAM_DEBUG
	    debug2 ("In side ulmap adding to burst,  UL_MAP_IE.UIUC = %d, UL_MAP_IE = %d", ulmap_ie.uiuc, ulmap->nb_ies);
#endif
	    ub->setCid (ulmap_ie.cid);
	    ub->setIUC (ulmap_ie.uiuc);
	    //ub->setStarttime (ulmap_ie.symbol_offset);
	    ub->setDuration (ulmap_ie.duration * 3); //*3 becuase bursts store in raw symbols not slots
	    //ub->setSubchannelOffset (ulmap_ie.subchannel_offset);
	    //ub->setnumSubchannels (ulmap_ie.num_of_subchannels);
	    if (ulmap_ie.uiuc == UIUC_CDMA_ALLOCATION_IE) {
	    	ub->setExtendedUIUC(ulmap_ie.cdma_ie.ie.uiuc);
	//    	ub->setIUC(ulmap_ie.cdma_ie.ie.uiuc);
		ub->setB_CDMA_TOP (ulmap_ie.cdma_ie.ie.subchannel);
		ub->setB_CDMA_CODE (ulmap_ie.cdma_ie.ie.code);	    
	    }


	    debug2("UL.Data region (Addburst): duration (in slots) [%d]\n", ulmap_ie.duration);
        debug2("   Addburst cdma code :%d, cdma top :%d\n", ub->getB_CDMA_CODE(), ub->getB_CDMA_TOP());
	}
	free (ulmap);
    }
  }


  //now that we've allocated everything, we need to schedule the packet reassembly
  Ptr<Packet> pum = map->getUL_MAP();
  UL_MAP_FRAME_HEADER um;
  All_Headers ah_um;
  pum->RemoveHeader(ah_um);
  pum->RemoveHeader(um);
  
  uint16_t total_symbols= 0;
  //we want to schedule this for just after the end of the last symbol
  Time subframe_endtime = mac_->start_of_frame + Seconds( ( mac_->getMap()->getUlSubframe()->getStarttime() * mac_->getPhy()->getPS() ) +  mac_->getMaxUlduration() * mac_->getPhy()->getSymbolTime() + phy->getPS()) - Simulator::Now();
  
  for (uint16_t i = 0; i < um.mf.nb_ies; i++) {
  	if (um.mf.ies[i].ie.uiuc == 1 || um.mf.ies[i].ie.uiuc == 2) continue; //this is contention...
	PHY_INFO_Header *pih = new PHY_INFO_Header();
	pih->num_OFDMSymbol = um.mf.ies[i].ie.duration * phy->getSlotLength(UL_);
	pih->OFDMSymbol_offset = total_symbols % (mac_->getMaxUlduration() - 3) + 3; //we don't map into the first slot-column, it's reserved for contention
	pih->subchannel_offset = floor((double)total_symbols / (double)(mac_->getMaxUlduration() - phy->getSlotLength(UL_))); //the slot length is subtracted to avoid taking the column of symbols reserved for contention into account
	pih->direction = UL;
  	total_symbols+= um.mf.ies[i].ie.duration * phy->getSlotLength(UL_);;
	debug10("At %f, scheduling assembly of UL packet starting at block (symbol = %d, subchannel = %d) and going %d symbols for end of frame at time %f\n", NOW, pih->OFDMSymbol_offset, pih->subchannel_offset, pih->num_OFDMSymbol, NOW + subframe_endtime.GetSeconds());
	//we just process everything at the end of the frame (except contention packets), as most of the packets will wrap around anyway
	Simulator::ScheduleWithContext(Simulator::GetContext(), subframe_endtime, &Wimax2NetDevice::assemble_incoming_packet, mac_,  pih);
  }



  //============================DL allocation=================================
  debug10 ("DL.Scheduler: FrameDuration :%5.4f, PSduration :%e, symboltime :%e, nbPS :%d, rtg :%d, ttg :%d, nbPSleft :%d, nbSymbols :%d, dlratio_ :%5.2f\n", mac_->getFrameDuration(), phy->getPS(), phy->getSymbolTime(), nbPS, mac_->phymib_.rtg, mac_->phymib_.ttg, nbPS_left, nbSymbols, dlratio_);

  map->getDlSubframe()->getPdu()->removeAllBursts();
  bzero(virtual_alloc, MAX_MAP_IE*sizeof(virtual_burst));
  index_burst = 0;
  
  All_Headers ah_sizes; //used to get sizes of the various headers that need to be taken into account for size calculations

//1. Virtual DL_MAP
//Note that, we virtually allocate all burst into "virtual_alloc[]", then we will physically allocate/map those allocation into bursts/physical later. The reason is because we have variable part of DL_MAPs
//"index_burst" is the #total_burst including each DL_MAP, UL_MAP, DCD, UCD, other broadcast message and each CID allocation
//In this version, the allocation is per CID/burst for downlink
  DL_MAP_FRAME_HEADER dmfh;
  int fixed_byte_dl_map = dmfh.GetSerializedSize();
  int rep_fixed_dl = Repetition_code_;
  Ofdm_mod_rate fixed_dl_map_mod = map->getDlSubframe()->getProfile (map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC())->getEncoding();
  int fixed_dl_map_slots = (int) rep_fixed_dl*ceil((double)fixed_byte_dl_map/(double)phy->getSlotCapacity(fixed_dl_map_mod, DL_)); 
  virtual_alloc[index_burst].alloc_type = 0;
  virtual_alloc[index_burst].cid = BROADCAST_CID;
  virtual_alloc[index_burst].n_cid = 0;
  virtual_alloc[index_burst].iuc = map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC();
  virtual_alloc[index_burst].preamble = true;
  virtual_alloc[index_burst].numslots = fixed_dl_map_slots;
  virtual_alloc[index_burst].mod = fixed_dl_map_mod;
  virtual_alloc[index_burst].byte = fixed_byte_dl_map;
  virtual_alloc[index_burst].rep =  Repetition_code_;
  virtual_alloc[index_burst].dl_ul = 0;
  virtual_alloc[index_burst].ie_type = 0;
  index_burst++;

//2. Virtual UL_MAP:
  Ptr<Packet> ulm_temp = map->getUL_MAP();
  UL_MAP_FRAME_HEADER umfh;
  ulm_temp->RemoveHeader(ah_p);
  ulm_temp->RemoveHeader(umfh);
  int fixed_byte_ul_map = umfh.GetSerializedSize() + ah_p.GMH.GetSerializedSize();
  int rep_fixed_ul = Repetition_code_;
  Ofdm_mod_rate fixed_ul_map_mod = map->getDlSubframe()->getProfile (map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC())->getEncoding();
  int fixed_ul_map_slots = (int) rep_fixed_ul*ceil((double)fixed_byte_ul_map/(double)phy->getSlotCapacity(fixed_ul_map_mod, DL_)); 
  virtual_alloc[1].mod = fixed_ul_map_mod;
  while (fixed_ul_map_slots > phy->getNumsubchannels(DL_)){
	virtual_alloc[1].mod = static_cast<Ofdm_mod_rate>  (static_cast<int>(virtual_alloc[1].mod) + 1);
  	fixed_ul_map_slots = (int) rep_fixed_ul*ceil((double)fixed_byte_ul_map/(double)phy->getSlotCapacity(virtual_alloc[1].mod, DL_)); 
	
  }
  virtual_alloc[index_burst].alloc_type = 1;
  virtual_alloc[index_burst].cid = BROADCAST_CID;
  virtual_alloc[index_burst].n_cid = 0;
  virtual_alloc[index_burst].iuc = getDIUCProfile(virtual_alloc[0].mod);
  virtual_alloc[index_burst].preamble = true;
  virtual_alloc[index_burst].numslots = fixed_ul_map_slots;
  virtual_alloc[index_burst].mod = fixed_ul_map_mod;
//  virtual_alloc[index_burst].byte = fixed_byte_ul_map;
  virtual_alloc[index_burst].rep =  Repetition_code_;
  virtual_alloc[index_burst].dl_ul = 0;
  virtual_alloc[index_burst].ie_type = 0;
  int add_ie_to_dlmap = increase_dl_map_ie(index_burst, total_dl_slots_pusc, 1);
  if (add_ie_to_dlmap <0 ) {
	debug10 ("Panic: not enough space for UL_MAP\n");
	exit(1);
  }
  index_burst++;

//3. Virtual DCD
  if (getMac()->sendDCD || map->getDlSubframe()->getCCC()!= getMac()->dlccc_) {
      p = map->getDCD();
      p->RemoveHeader(ah_p);
      int byte_dcd = ah_p.GMH.GMH.len;
      debug10("DCD length: %d\n",byte_dcd);
      int rep_dcd = 1;
      Ofdm_mod_rate dcd_mod = map->getDlSubframe()->getProfile (map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC())->getEncoding();
      int dcd_slots = (int) rep_dcd*ceil((double)byte_dcd/(double)phy->getSlotCapacity(dcd_mod, DL_)); 
      virtual_alloc[index_burst].alloc_type = 2;
      virtual_alloc[index_burst].cid = BROADCAST_CID;
      virtual_alloc[index_burst].n_cid = 1;
      virtual_alloc[index_burst].iuc = map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC();
      virtual_alloc[index_burst].preamble = false;
      virtual_alloc[index_burst].numslots = dcd_slots;
      virtual_alloc[index_burst].mod = dcd_mod;
      virtual_alloc[index_burst].byte = byte_dcd;
      virtual_alloc[index_burst].rep =  rep_dcd;
      virtual_alloc[index_burst].dl_ul = 0;
      virtual_alloc[index_burst].ie_type = 0;
      int add_ie_to_dlmap = increase_dl_map_ie(index_burst, total_dl_slots_pusc, 1);
      if (add_ie_to_dlmap <0 ) {
		debug10 ("Panic: not enough space for DCD\n");
		exit(1);
	  } else {
		debug10 ("DCD size :%d bytes, %d slots\n", ah_p.GMH.GMH.len, dcd_slots);
	  }
      index_burst++;
  }

//4. Virtual UCD
  if (getMac()->sendUCD || map->getUlSubframe()->getCCC()!= getMac()->ulccc_) {
      p = map->getUCD();
      p->PeekHeader(ah_p);
      int byte_ucd = ah_p.GMH.GMH.len;
      int rep_ucd = 1;
      Ofdm_mod_rate ucd_mod = map->getDlSubframe()->getProfile (map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC())->getEncoding();
      int ucd_slots = (int) rep_ucd*ceil((double)byte_ucd/(double)phy->getSlotCapacity(ucd_mod, DL_)); 
      virtual_alloc[index_burst].alloc_type = 3;
      virtual_alloc[index_burst].cid = BROADCAST_CID;
      virtual_alloc[index_burst].n_cid = 1;
      virtual_alloc[index_burst].iuc = map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC();
      virtual_alloc[index_burst].preamble = false;
      virtual_alloc[index_burst].numslots = ucd_slots;
      virtual_alloc[index_burst].mod = ucd_mod;
      virtual_alloc[index_burst].byte = byte_ucd;
      virtual_alloc[index_burst].rep =  rep_ucd;
      virtual_alloc[index_burst].dl_ul = 0;
      virtual_alloc[index_burst].ie_type = 0;
      int add_ie_to_dlmap = increase_dl_map_ie(index_burst, total_dl_slots_pusc, 1);
      if (add_ie_to_dlmap <0 ) {
		debug10 ("Panic: not enough space for UCD\n");
		exit(1);
      } else {
		debug10 ("UCD size :%d bytes, %d slots\n", ah_p.GMH.GMH.len, ucd_slots);
      }
      index_burst++;
  }
 
  int check_dl_slots = check_overallocation(index_burst);

  debug10 ("DL.broadcast messages (slots): (v)Fixed_DL_MAP :%d, (v)UL_MAP :%d, DCD :%d, UCD :%d, MaxDL :%d\n", virtual_alloc[0].numslots, virtual_alloc[1].numslots, virtual_alloc[2].numslots, virtual_alloc[3].numslots, total_dl_slots_pusc);
  debug10 ("\t(bytes): (v)Fixed_DL_MAP :%f, (v)UL_MAP :%f, DCD :%f, UCD :%f\n", virtual_alloc[0].byte, virtual_alloc[1].byte, virtual_alloc[2].byte, virtual_alloc[3].byte);

  if (check_dl_slots > total_dl_slots_pusc) {
     debug10("Panic : not enough dl_slots\n");
     exit(1);
  } 


//std::cerr << "Start Index: " << index_burst << "\t";
//5. Virtual Other broadcast
//note that packing and fragmentation are disabled for broadcast packets
 
uint32_t requested_broadcast_allocation_count = 0;
  if (mac_->getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION)->queueByteLength()>0) {
      
      Ofdm_mod_rate bc_mod = map->getDlSubframe()->getProfile (map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC())->getEncoding();

      int slot_br = 0; 
      if (slot_br); //quiet, compiler
      Connection *c_tmp;
      Ptr<Packet> np;
      All_Headers ah_np;
      c_tmp = mac_->getCManager()->get_connection (BROADCAST_CID, OUT_CONNECTION);
      debug10 ("Retrive connection :%d, qlen :%d\n", c_tmp->get_cid(), c_tmp->queueLength());
      uint32_t slot_tally = 0;
	      
      for (int j_p = 0; j_p<c_tmp->queueLength(); j_p++) {
      	      np = c_tmp->queueLookup(j_p);
	      if (!np) break;
	      np->PeekHeader(ah_np);

		
	      virtual_alloc[index_burst].alloc_type = 4;
	      virtual_alloc[index_burst].cid = BROADCAST_CID;
	      virtual_alloc[index_burst].n_cid = 1;
	      virtual_alloc[index_burst].iuc = map->getDlSubframe()->getProfile (DIUC_PROFILE_2)->getIUC();
	      virtual_alloc[index_burst].preamble = false;
	      virtual_alloc[index_burst].numslots = ceil( (double)ah_np.GMH.GMH.len / (double)phy->getSlotCapacity(bc_mod,DL_));
	      virtual_alloc[index_burst].byte = ah_np.GMH.GMH.len;
	      virtual_alloc[index_burst].rep =  0; //only transmits once
	      virtual_alloc[index_burst].dl_ul = 0;
	      virtual_alloc[index_burst].ie_type = 0;
	      
	      int add_ie_to_dlmap = increase_dl_map_ie(index_burst, total_dl_slots_pusc, 1);
	      if (add_ie_to_dlmap <0 ) {
		      virtual_alloc[index_burst].alloc_type = 0;
		      virtual_alloc[index_burst].cid = 0;
		      virtual_alloc[index_burst].n_cid = 0;
		      virtual_alloc[index_burst].iuc = 0;
		      virtual_alloc[index_burst].preamble = false;
		      virtual_alloc[index_burst].numslots = 0;
		      virtual_alloc[index_burst].byte = 0;;
		      virtual_alloc[index_burst].rep =  0;
		      virtual_alloc[index_burst].dl_ul = 0;
		      virtual_alloc[index_burst].ie_type = 0;
		      debug10 ("Problem: not enough space for other broadcast message, clearing this virtual_alloc\n");
	      } else {
		      slot_tally += virtual_alloc[index_burst].numslots;
		      index_burst++;
		      requested_broadcast_allocation_count++;
	      }
      }


      debug10 ("DL.other_broadcast messages (slots): %d, (bytes): %f\n", virtual_alloc[index_burst].numslots, virtual_alloc[index_burst].byte);
      index_burst++;
    }
//End of Virtual allocation for all broadcast message
//Grabbing a virtual map for other messages

//first we check if everything can fit, and if not we randomly select a set of bursts to send
// -- NOTE -- if/when other QoS models than BE are implemented, this is where it should be done

uint32_t oldest_frame = mac_->getFrameNumber() - 1;
uint32_t requested_allocation_count = 0;
for (Connection *c_con = mac_->getCManager()->get_out_connection (); c_con; c_con = c_con->next_entry()) {
	if (c_con->queueByteLength() <= 0) {
		continue;
	}
	if (! (c_con && c_con->getPeerNode() && c_con->getPeerNode()->getDIUC())) continue;
	if (c_con->get_cid() == BROADCAST_CID) continue;

	if (c_con->queueLength() > 0 && c_con->getLastFrameTransmitted() < oldest_frame) oldest_frame = c_con->getLastFrameTransmitted();
	requested_allocation_count++;
}

//we search through all connections, allocating space based on which haven't transmitted in the longest time
//the LastFrameTransmitted prameter is updated when the bursts are filled intransfer_packets_for_dl_subframe
for (uint32_t fr = oldest_frame; fr < static_cast<uint32_t>(mac_->getFrameNumber()); fr++) {
	for (Connection *c_con = mac_->getCManager()->get_out_connection (); c_con; c_con = c_con->next_entry()) {
		if (c_con->queueByteLength() <= 0) {
			continue;
		}
		if (! (c_con && c_con->getPeerNode() && c_con->getPeerNode()->getDIUC())) continue;
		  Ofdm_mod_rate con_mod = map->getDlSubframe()->getProfile (c_con->getPeerNode()->getDIUC())->getEncoding();

		if (c_con->getLastFrameTransmitted() != fr) continue; //we've either already gotten it, or we'll get it on a later pass

	      Ptr<Packet> np;
	      All_Headers ah_np;
		//since this is just the virtual mapping, we need to look at the queues, but we don't
		//actually remove packets from them

	      uint32_t actual_byte_tally = 0;
	      if (c_con->queueLength() == 0) actual_byte_tally = 0;
	      else if (c_con->queueLength() == 1) {
		np = c_con->queueLookup(0);
		np->PeekHeader(ah_np);
		actual_byte_tally = ah_np.GMH.GMH.len;
		if (ah_np.GMH.GMH.type_frag == 1) actual_byte_tally += ah_np.FSH.GetSerializedSize();
	      }
	      else if (c_con->queueLength() > 1) {
		      actual_byte_tally = ah_np.GMH.GetSerializedSize();
		      for (int j_p = 0; j_p<c_con->queueLength(); j_p++) {
				//since this is just the virtual mapping, we need to look at the queues, but we don't
				//actually remove packets from them
				np = c_con->queueLookup(j_p);
				if ( PeekPointer(np) != NULL ) {
				    np->PeekHeader(ah_np);
				    actual_byte_tally += ah_np.GMH.GMH.len + ah_np.PSH.GetSerializedSize();// - ah_np.GMH.GetSerializedSize();
				}
		      }
	      }

		
	      int32_t bc_slots = ceil((double)actual_byte_tally/(double)phy->getSlotCapacity(con_mod,DL_)); 
	      dlduration = 0;
	      //figure out how many slots have been used by bursts not in reserved areas at the start
	      for (int i = getMac()->getFrameNumber() % 10 != 0 ? 2 : 3; i < index_burst; i++) dlduration += virtual_alloc[i].numslots;
	      //figure out the most blocks we could possibly use outside of the reserved start-areas 
	      int32_t max_alloc = getMac()->getFrameNumber() % 10 == 0 ? (maxdlduration - 7) * phy->getNumsubchannels(DL_) : (maxdlduration - 5) * phy->getNumsubchannels(DL_);
	      //convert blocks to slots
	      max_alloc /= phy->getSlotLength(DL_);

	      if (dlduration + bc_slots > max_alloc) {
		//we don't actually have room for this, throw as large an allocation as we could potentially fit in
		//the downlink mapping probably won't include all of everything, but it might
//		std::cerr << "START: DLDuration: " << dlduration << " BC_SLOTS: " << bc_slots << " max_alloc: " << max_alloc << " byte tally: " <<actual_byte_tally;
		bc_slots = max_alloc - dlduration;
		if (bc_slots < 0) continue;
		actual_byte_tally = phy->getMaxPktSize(bc_slots * phy->getSlotLength(DL_), con_mod, DL_);
//		std::cerr << " MODIFIED: BC_SLOTS: " << bc_slots << "  byte_tally: " << actual_byte_tally << "\n";
	      }
	      virtual_alloc[index_burst].alloc_type = 5;
	      virtual_alloc[index_burst].cid = c_con->get_cid();
	      virtual_alloc[index_burst].n_cid = 1;
	      virtual_alloc[index_burst].iuc = getDIUCProfile(con_mod);
	      virtual_alloc[index_burst].preamble = false;
	      virtual_alloc[index_burst].numslots = bc_slots;
	      virtual_alloc[index_burst].byte = actual_byte_tally;
	      virtual_alloc[index_burst].rep =  1;
	      virtual_alloc[index_burst].dl_ul = 0;
	      virtual_alloc[index_burst].ie_type = 0;
	      int add_ie_to_dlmap = increase_dl_map_ie(index_burst, total_dl_slots_pusc, 1);
	      if (add_ie_to_dlmap < 0 ) {
		      virtual_alloc[index_burst].alloc_type = 0;
		      virtual_alloc[index_burst].cid = 0;
		      virtual_alloc[index_burst].n_cid = 0;
		      virtual_alloc[index_burst].iuc = 0;
		      virtual_alloc[index_burst].preamble = false;
		      virtual_alloc[index_burst].numslots = 0;
		      virtual_alloc[index_burst].byte = 0;;
		      virtual_alloc[index_burst].rep =  0;
		      virtual_alloc[index_burst].dl_ul = 0;
		      virtual_alloc[index_burst].ie_type = 0;
		      debug10 ("Problem: not enough space for other message, clearing this virtual_alloc\n");
	      } else {
		      debug10 ("DL.other messages (slots): %d, (bytes): %f\n", virtual_alloc[index_burst].numslots, virtual_alloc[index_burst].byte);
		      index_burst++;
	      }

	}
}
  ////std::cerr << "Requested Allocations: " << requested_allocation_count << "\t granted allocations: " << index_burst - start_index << std::endl;

  int index_burst_before_data = index_burst;
  	if (index_burst_before_data); //quiet, compiler

  dlduration = ceil((double)check_overallocation(index_burst)/(double)total_dl_subchannel_pusc);

  // from here start ofdma DL -- call the scheduler API- here DLMAP    ------------------------------------
  peer = mac_->getPeerNode_head();

  subchannel_offset = 0;
//  mac802_16_dl_map_frame *dl_map;

  //We Are now going to map the bursts onto physical slots in the OFDMA allocation grid
  //The FCH, DL_MAP, UL_MAP, and if applicable for this frame DCD and UCD are scheduled in the first two columns
  //This is to ensure that they are always included in transmission.
  int virtual_symbol_offset = DL_PREAMBLE + getMac()->getPhy()->getSlotLength(DL_); // = 3
  int virtual_subchannel_offset = 0;
  	if (virtual_subchannel_offset); //quiet, compiler
 // int virtual_accum_slots = 0;


  //first we stick the FCH in.  This is always in the same position, with the same encoding.
  Frame_Prefix_Header fph;
  fph.DL_MAP_length = virtual_alloc[0].numslots;
  //the repetition coding for the FCH uses the following map:
  //	00 = 1 total repetitions 
  //	01 = 2 total repetitions
  //	10 = 4 total repetitions
  //	11 = 6 total repetitions
  fph.repetition_coding = virtual_alloc[0].rep / 2;
  fph.used_subchannel_bmp = 63; //all used
	
  PHY_INFO_Header prefix_pih;
  prefix_pih.num_subchannels = 2;
  prefix_pih.num_OFDMSymbol = 2;
  prefix_pih.subchannel_offset = 0;
  prefix_pih.OFDMSymbol_offset = DL_PREAMBLE;
  prefix_pih.modulation_ = OFDM_QPSK_1_2;

  Ptr<Packet> fph_p = Create<Packet>();
  //this will always have a repetition of 2, so we just add it twice here
  fph_p->AddHeader(fph);
  fph_p->AddHeader(fph);
  fph_p->AddHeader(prefix_pih);

  Simulator::ScheduleWithContext(Simulator::GetContext(), Simulator::Now() - getMac()->start_of_frame + Seconds (getMac()->getPhy()->getSymbolTime()), &Wimax2BSNetDevice::transmit, getMac(), fph_p);

  //Now we're going to do the DL_MAP, since that is also a custom assignment (first column, under the FCH
      db = reinterpret_cast<DlBurst*>(map->getDlSubframe()->getPdu()->addBurst (nbdlbursts++));
      db->setCid (virtual_alloc[0].cid);
      db->setIUC (virtual_alloc[0].iuc);
      db->setPreamble(virtual_alloc[0].preamble); 

      db->setStarttime(DL_PREAMBLE);
      db->setSubchannelOffset(prefix_pih.num_subchannels); //will always be 2
      db->setnumSubchannels(virtual_alloc[0].numslots);
      db->setDuration(getMac()->getPhy()->getSlotLength(DL_)); //will always be 2

 //and Now the UL Map, and UCD/DCD if applicable, to ensure they're transmitted on time regardless of the mapping algorithm

 	uint16_t subch_left;
	uint16_t slots_in = 1;
 	
      db = reinterpret_cast<DlBurst*>(map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++));
      db->setCid(virtual_alloc[1].cid);
      db->setIUC (virtual_alloc[1].iuc);
      db->setPreamble(virtual_alloc[1].preamble); 
      db->setStarttime(DL_PREAMBLE + getMac()->getPhy()->getSlotLength(DL_));
      db->setSubchannelOffset(0); //will always be 0
      db->setnumSubchannels(virtual_alloc[1].numslots);
      db->setDuration(getMac()->getPhy()->getSlotLength(DL_)); //will always be 2

      subch_left = getMac()->getPhy()->getNumsubchannels(DL_) - db->getnumSubchannels();
      slots_in = 1;



int start = 2;
  //now the DCD and UCD if they are transmistted this frame
      if (getMac()->sendDCD || map->getDlSubframe()->getCCC()!= getMac()->dlccc_) {
	      debug2("DL_MAP.FILL.Adding (DCD) burst %d to existing DL_MAP column %d, starting %d subchannels down\n", 2,virtual_symbol_offset, virtual_subchannel_offset);
      	      start++;
	      db = reinterpret_cast<DlBurst*>(map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++));
	      db->setCid(virtual_alloc[2].cid);
	      db->setIUC (virtual_alloc[2].iuc);
	      db->setPreamble(virtual_alloc[2].preamble); 

	      if (subch_left > virtual_alloc[2].numslots) {
		      db->setStarttime(DL_PREAMBLE + getMac()->getPhy()->getSlotLength(DL_));
		      db->setSubchannelOffset(virtual_alloc[1].numslots); //will always be 2
		      db->setnumSubchannels(virtual_alloc[2].numslots);
		      subch_left -= virtual_alloc[2].numslots;
	      } else {
		      slots_in++;
		      db->setStarttime(DL_PREAMBLE + 2*getMac()->getPhy()->getSlotLength(DL_));
		      db->setSubchannelOffset(0);
		      db->setnumSubchannels(virtual_alloc[2].numslots);
		      subch_left = getMac()->getPhy()->getNumsubchannels(DL_) - virtual_alloc[2].numslots;
	      }
	      db->setDuration(getMac()->getPhy()->getSlotLength(DL_)); //will always be 2
		

      }

      if (getMac()->sendUCD || map->getUlSubframe()->getCCC()!= getMac()->ulccc_) {
	      debug2("DL_MAP.FILL.Adding (UCD) burst %d to existing DL_MAP column %d, starting %d subchannels down\n",3,virtual_symbol_offset, virtual_subchannel_offset);
      	      start++;
	      db = reinterpret_cast<DlBurst*>(map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++));
	      db->setCid(virtual_alloc[3].cid);
	      db->setIUC (virtual_alloc[3].iuc);
	      db->setPreamble(virtual_alloc[3].preamble); 

		if (subch_left > virtual_alloc[3].numslots) {
		      db->setStarttime(DL_PREAMBLE + slots_in * getMac()->getPhy()->getSlotLength(DL_));
		      db->setSubchannelOffset(getMac()->getPhy()->getNumsubchannels(DL_) - subch_left); //will always be 2
		      db->setnumSubchannels(virtual_alloc[3].numslots);
		} else {
		      slots_in++;
		      db->setStarttime(DL_PREAMBLE + slots_in * getMac()->getPhy()->getSlotLength(DL_));
		      db->setSubchannelOffset(0);
		      db->setnumSubchannels(virtual_alloc[3].numslots);
		}

	      db->setDuration(getMac()->getPhy()->getSlotLength(DL_)); //will always be 2

      }



      int start_region_offset = db->getDuration() + db->getStarttime();
      virtual_symbol_offset = start_region_offset;

      //first we sort the remaining allocations by size
      //just going to use a simple insertion sort, but something like quicksort would obviously be nicer 
      std::list<virtual_burst> sorted_allocs;

      //std::cerr << "Index Burst: " << index_burst << " start: " << start;

   if (start < index_burst ) {
	      sorted_allocs.push_front(virtual_alloc[start++]);

	      for (int i = start; i < index_burst ; i++) {
	         if (virtual_alloc[i].numslots == 0) continue;
	      	 bool assigned = false;
		 for (std::list<virtual_burst>::iterator j = sorted_allocs.begin(); j != sorted_allocs.end(); ++j) {
			if (virtual_alloc[i].byte > (*j).byte) {
				sorted_allocs.insert(j, virtual_alloc[i]);
				assigned = true;
				break;
			}
		 }
		 if (!assigned) sorted_allocs.push_back(virtual_alloc[i]);
	      }
  }


#ifdef DEBUG_WIMAX
	cout << "Allocation Sizes:\n";
	int alloc_num = 0;
  	for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
		cout << "Allocation " << alloc_num++ << " has size " << i->numslots << std::endl;
	}
	cout << std::endl;
#endif


  switch (mt_) {

  case OCSA:
  { 
  	  uint16_t first_unusable_slot = (mac_->getMaxDlduration()) / phy->getSlotLength(DL_);
  	  uint16_t start_slot = start_region_offset / phy->getSlotLength(DL_);
  	  uint16_t mappable_width = first_unusable_slot - start_slot; if (mappable_width); //quiet, compiler
  	  uint16_t mappable_height = phy->getNumsubchannels(DL_);

	  std::vector <std::vector <allocation_listing> > listings_list;
  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
 	  	std::vector<allocation_listing> possibilities;
		//uint16_t min_width = i->numslots < mappable_height ? 1 : ceil ((double)i->numslots / (double) mappable_height);
		uint16_t min_width = ceil ((double)i->numslots / (double) mappable_height);
		uint16_t max_width = i->numslots > mappable_width ? mappable_width : i->numslots;

		//we add possible allocations in increasing width order
		for (int w = min_width; w < max_width; w++) {
			allocation_listing nl;
			nl.width = w;
			nl.height = ceil ( (double)i->numslots / (double)w);
			nl.wasted = nl.width * nl.height - i->numslots;
			possibilities.push_back(nl);
		}
		if (min_width == max_width) {
			allocation_listing nl;
			nl.width = min_width;
			nl.height = ceil ( (double)i->numslots / (double)min_width);
			nl.wasted = nl.width * nl.height - i->numslots;
			possibilities.push_back(nl);
		}
		listings_list.push_back(possibilities);
 	  } 
  	  
	  uint16_t remaining_width = mappable_width;
	  uint16_t sorted_alloc_count = 0;
  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
  		if ((*i).mapped) continue; //if this burst has already been mapped we don't want to map it again...
		
  		db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  		db->setCid(i->cid);
  		db->setIUC(i->iuc);
  		db->setPreamble(i->preamble);
  		int16_t w_i, h_i;

		int16_t best_allocation = listings_list[sorted_alloc_count][0].wasted;
		int16_t ideal_alloc = 0;
	
		//given the total width left, find the lowest waste allocation that will fit
		//if multiple allocations tie for lowest waste, take the narrowest
		for (uint16_t j = 0; j < listings_list[sorted_alloc_count].size(); j++) {
			if (listings_list[sorted_alloc_count][j].wasted < best_allocation) {
				ideal_alloc = j;
				best_allocation = listings_list[sorted_alloc_count][j].wasted;
			}
		}

  		w_i = listings_list[sorted_alloc_count][ideal_alloc].width;
  		h_i = listings_list[sorted_alloc_count][ideal_alloc].height;
		remaining_width -= w_i;

		
		
  		db->setStarttime(remaining_width * phy->getSlotLength(DL_) + DL_PREAMBLE);
  		db->setDuration(w_i * phy->getSlotLength(DL_));
  
  		db->setnumSubchannels(h_i);
  		db->setSubchannelOffset(phy->getNumsubchannels(DL_) - h_i);
  		(*i).mapped = true;

		//now we check to see if we can fit any additional allocations above this one
  		int16_t filled_subchannels = h_i;
		int16_t sorted_alloc_count_j = sorted_alloc_count + 1;
  		std::list<virtual_burst> ::iterator j = i;
  		for (j++; j != sorted_allocs.end(); j++) {
			if (j->mapped) {
				sorted_alloc_count_j++;
				continue;
			}
			//check if any of the possible allocations will fit
			//if several will, get the one with minimum height
			int16_t best_alloc_index;
			int16_t best_alloc_height;
			for (uint16_t k = 0; k < listings_list[sorted_alloc_count_j].size(); k++) {
				if (listings_list[sorted_alloc_count_j][k].width <= w_i && listings_list[sorted_alloc_count_j][k].height + filled_subchannels <= mappable_height && listings_list[sorted_alloc_count_j][k].height < best_alloc_height) {
					best_alloc_index = k;
					best_alloc_height = listings_list[sorted_alloc_count_j][k].height;
				}

			}

			int16_t w_j = listings_list[sorted_alloc_count_j][best_alloc_index].width;
			int16_t h_j = listings_list[sorted_alloc_count_j][best_alloc_index].height;

			//we have the best allocation shape now, so map that in
  			db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  			db->setCid(j->cid);
  			db->setIUC(j->iuc);
  			db->setPreamble(j->preamble);
  		
  			db->setStarttime(remaining_width * phy->getSlotLength(DL_) + DL_PREAMBLE);
  			db->setDuration(w_j * phy->getSlotLength(DL_));
  
  			db->setnumSubchannels(h_j);
  			db->setSubchannelOffset(phy->getNumsubchannels(DL_) - (filled_subchannels + h_j));
			j->mapped = true;
			sorted_alloc_count_j++;
			filled_subchannels += h_j;
		}
		

		sorted_alloc_count++;
	  }

	  
  	break;
  }
  case eOCSA:
    { //eOCSA case braces to allow variable declaration without issues
  
  	  //now that we have everything sorted, do eOCSA

  	  int16_t first_unusable_slot = (mac_->getMaxDlduration()) / phy->getSlotLength(DL_);
  	  int16_t start_slot = start_region_offset / phy->getSlotLength(DL_);
  	  int16_t mappable_width = first_unusable_slot - start_slot;  if (mappable_width); //quiet, compiler
  	  int16_t mappable_height = phy->getNumsubchannels(DL_);
  	  int16_t current_region_start = first_unusable_slot; //because we'll pull it back from here on the first allocation
  	  int16_t current_region_end = first_unusable_slot;
  
	  uint16_t overallocations = 0;
	  uint16_t allocations = 0;

  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
  		if ((*i).mapped) continue; //if this burst has already been mapped we don't want to map it again...
  
  		db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  		db->setCid(i->cid);
  		db->setIUC(i->iuc);
  		db->setPreamble(i->preamble);
  		int16_t w_i, h_i;
  
  		w_i = ceil( (double)((*i).numslots) / (double)mappable_height);
  		h_i = ceil ((double)((*i).numslots) / double(w_i));

		if (current_region_start - w_i < start_slot) {
  			map->getDlSubframe()->getPdu ()->removeBurst(db);
  			nbdlbursts--;
			continue;
		}
  		current_region_end = current_region_start;
  		current_region_start -= w_i;

  
  		//the start time and duration fields of DlBursts are stored in units of symbol-times, not slots
  		db->setStarttime(DL_PREAMBLE + current_region_start * phy->getSlotLength(DL_));
  		db->setDuration(w_i * phy->getSlotLength(DL_));
  
  		db->setnumSubchannels(h_i);
  		db->setSubchannelOffset(phy->getNumsubchannels(DL_) - h_i);
  		(*i).mapped = true;

		overallocations += w_i * h_i - i->numslots;
		allocations += w_i * h_i;
  
  		//now check if anything will fit on top of this column
  		int16_t filled_subchannels = h_i;
  		std::list<virtual_burst> ::iterator j = i;
  		for (j++; j != sorted_allocs.end(); j++) {
			if (j->mapped) continue;
  			int16_t w_j, h_j;
  			h_j = ceil ((double)(j->numslots) / (double)w_i);
  			w_j = ceil ((double)(j->numslots) / (double)h_j);
  
  			if (w_j > w_i) continue;
  			if (h_j + filled_subchannels > mappable_height) continue;
  
  			//if we're still in the loop, this will fit.
  			db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  			db->setCid(j->cid);
  			db->setIUC(j->iuc);
  			db->setPreamble(j->preamble);
  		
  			db->setStarttime(current_region_start * phy->getSlotLength(DL_) + DL_PREAMBLE);
  			db->setDuration(w_j * phy->getSlotLength(DL_));
  
  			db->setnumSubchannels(h_j);
  			db->setSubchannelOffset(phy->getNumsubchannels(DL_) - (filled_subchannels + h_j));
  			(*j).mapped = true;
  
  			filled_subchannels += h_j;
			overallocations += w_j * h_j - j->numslots;
			allocations += w_j * h_j;
  		}
  
  
  
  		//there's a bug where occasionally DL bursts aren't properly removed for an extra frame
		//and are reallocatd with size 0, this isn't the ideal way to fix it, but it's the fasstest
  		if (db->getDuration() == 0 || db->getnumSubchannels() == 0) {
  			  map->getDlSubframe()->getPdu ()->removeBurst(db);
  			  nbdlbursts--;
  			  continue;
  		}
  
  	  }

	
	  //collect information for output about allocation sizes

	  uint16_t total_size = 0;
	  uint16_t unmapped_blocks = 0;
	  uint16_t num_bursts = 0;
	  uint16_t unmapped_bursts = 0;
  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
	  	num_bursts++;
		total_size += i->numslots;	
		if (!i->mapped){
			unmapped_blocks += i->numslots;
			unmapped_bursts++;
		}
	  }
	  uint16_t empty_needed_blocks = mappable_height * mappable_width - allocations;

	  total_overallocations[num_bursts] += overallocations;
	  total_unmapped_blocks[num_bursts] += unmapped_blocks;
	  total_unmapped_bursts[num_bursts]  += unmapped_bursts;
	  if (unmapped_bursts > 0) total_empty_needed_blocks[num_bursts] += empty_needed_blocks;
	  if (num_bursts > max_dl_bursts) max_dl_bursts = num_bursts;
	  total_number_of_frames[num_bursts]++;

	  //std::cerr << " actually mapped: " << num_bursts - unmapped_bursts << " and unmapped: " << unmapped_bursts << std::endl;


	  //each line will contain [Frame Number] [maximum subframe size] [number of bursts] [sum of all burst sizes] [overallocations] [blocks in allocation grid not assigned] [unmapped bursts] [blocks in bursts that are not mapped]
	  if (total_size > 0)
	  	allocation_file << getMac()->frame_number_ << "\t" << mappable_height * mappable_width << "\t" << num_bursts << "\t" << total_size << "\t" << overallocations << "\t" << empty_needed_blocks << "\t" << unmapped_bursts << "\t" << unmapped_blocks << std::endl;

	
    } //eOCSA case braces to allow variable declaration without issues
    break;
    case mOCSA:
    { //mOCSA case braces to allow variable declaration without issues
  
  	  int16_t first_unusable_slot = (mac_->getMaxDlduration()) / phy->getSlotLength(DL_);
  	  int16_t start_slot = start_region_offset / phy->getSlotLength(DL_);
  	  int16_t mappable_width = first_unusable_slot - start_slot;  if (mappable_width); //quiet, compiler
  	  int16_t mappable_height = phy->getNumsubchannels(DL_);
  	  int16_t current_region_start = first_unusable_slot; //because we'll pull it back from here on the first allocation
  	  int16_t current_region_end = first_unusable_slot;
 	  std::vector<top_region_listing> top_regions;  
 	
	  uint16_t overallocations = 0;
	  uint16_t allocations = 0;

	  //first we allocate the bottom regions, using the same algorithm as eOCSA
  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
  		if ((*i).mapped) continue; //if this burst has already been mapped we don't want to map it again...
  
  		db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  		db->setCid(i->cid);
  		db->setIUC(i->iuc);
  		db->setPreamble(i->preamble);
  		int16_t w_i, h_i;
  
  		w_i = ceil( (double)((*i).numslots) / (double)mappable_height);
  		h_i = ceil ((double)((*i).numslots) / double(w_i));

		if (current_region_start - w_i < start_slot) { //we can't fit this burst...
  			map->getDlSubframe()->getPdu ()->removeBurst(db);
  			nbdlbursts--;
			continue;
		}
	 
		current_region_end = current_region_start;
  		current_region_start -= w_i;

  
  		//the start time and duration fields of DlBursts are stored in units of symbol-times, not slots
  		db->setStarttime(current_region_start * phy->getSlotLength(DL_) + DL_PREAMBLE);
  		db->setDuration(w_i * phy->getSlotLength(DL_));
  
  		db->setnumSubchannels(h_i);
  		db->setSubchannelOffset(phy->getNumsubchannels(DL_) - h_i);
  		(*i).mapped = true;
		
		overallocations += w_i * h_i - i->numslots;
		allocations += w_i * h_i;


  		//there's a bug where occasionally DL bursts aren't properly removed for an extra frame
		//and are reallocatd with size 0, this isn't the ideal way to fix it, but it's the fasstest
  		if (db->getDuration() == 0 || db->getnumSubchannels() == 0) {
  			  map->getDlSubframe()->getPdu ()->removeBurst(db);
  			  nbdlbursts--;
  			  continue;
  		}

		//now we record how much space is left above this column
		top_region_listing new_top;
		new_top.height = mappable_height - h_i;
		new_top.width = w_i;
		new_top.slot_offset = current_region_start;
		new_top.subchannel_offset = 0; //since subchannels are mapped top-down
		new_top.valid = true;
		top_regions.push_back(new_top);
//		std::cerr << "New Top Region (frame " << mac_->getFrameNumber() << " top # " << top_regions.size() << "): height: " << new_top.height << " width " << new_top.width << " slot_offset " << new_top.slot_offset << " subchannel offset " << new_top.subchannel_offset << std::endl;
 	}

	//add a new "top region" for any completely unallocated slot-columns, so the space can be merged in later as needed

	if (current_region_start > start_slot) {
		top_region_listing start_top;
		start_top.height = mappable_height;
		start_top.width = current_region_start - start_slot;
		start_top.slot_offset = start_slot;
		start_top.subchannel_offset = 0;
		start_top.valid = true;
		top_regions.push_back(start_top);
//		std::cerr << "Remainder Top Region (frame " << mac_->getFrameNumber() << " top # " << top_regions.size() << "): height: " << start_top.height << " width " << start_top.width << " slot_offset " << start_top.slot_offset << " subchannel offset " << start_top.subchannel_offset << std::endl;
	}

	//now we examine the regions above these initial columns
	//we want to combine adjacent regions, but because adjacent bottom-allocations
	//may be of different heights, this could result in smaller regions, which we want to avoid

	//we therefore check if combining adjacent columns and creating a single allocation space of
	//the sum of the widths and the smaller height will result in a larger region than either of the originals
	//(the remaining larger-height region will be mapped into it's own new region
	     
	for (uint16_t i = 0; i < top_regions.size(); i++) {
		if (!top_regions[i].valid) continue;
		int16_t i_area = top_regions[i].width * top_regions[i].height;
		for (uint16_t j = 0; j < top_regions.size(); j++) {
			if (i == j) continue;
			if (!top_regions[j].valid) continue;
			int16_t j_area = top_regions[j].width * top_regions[j].height;

			//check if they're adjacent
			if (top_regions[i].slot_offset  == top_regions[j].slot_offset + top_regions[j].width) { //if they're adjacent width-wise
				//check what the height-overlap is
				
				int16_t start_height = top_regions[i].subchannel_offset > top_regions[j].subchannel_offset ? top_regions[i].subchannel_offset : top_regions[j].subchannel_offset;
				int16_t end_height = top_regions[i].subchannel_offset + top_regions[i].height < top_regions[j].subchannel_offset + top_regions[j].height ? top_regions[i].subchannel_offset + top_regions[i].height : top_regions[j].subchannel_offset + top_regions[j].height;
				int16_t combined_area = (end_height - start_height) * (top_regions[i].width + top_regions[j].width);

				//if we have a larger combined area, merge the regions, and create a new region of the leftover area
				if (combined_area > i_area && combined_area > j_area) {
					top_region_listing merged;
					top_region_listing above;
					top_region_listing below;

					//if either of these regions is used, it starts with a copy from a current region, where valid will be true
					above.valid = false;
					below.valid = false;

					merged.height = end_height - start_height;
					merged.width = top_regions[i].width + top_regions[j].width;
					merged.slot_offset = top_regions[i].slot_offset < top_regions[j].slot_offset ? top_regions[i].slot_offset : top_regions[j].slot_offset;
					merged.subchannel_offset = start_height;
					merged.valid = true;
					
					if (top_regions[i].subchannel_offset < start_height) {
						above = top_regions[i];
						above.height = start_height - top_regions[i].subchannel_offset;
					} else if (top_regions[j].subchannel_offset < start_height) {
						above = top_regions[j];
						above.height = start_height - top_regions[j].subchannel_offset;
					}

					if (top_regions[i].subchannel_offset + top_regions[i].height > end_height) {
						below = top_regions[i];
						below.subchannel_offset = end_height;
						below.height = top_regions[i].subchannel_offset + top_regions[i].height - end_height;
					} else if (top_regions[j].subchannel_offset + top_regions[j].height > end_height) {
						below = top_regions[j];
						below.subchannel_offset = end_height;
						below.height = top_regions[j].subchannel_offset + top_regions[j].height - end_height;
					}
#ifdef DEBUG_WIMAX
					cout << "\n~+~+ MERGING REGIONS +~+~\n";
					cout << "Regions merged:\n";
				
					cout << "First: height: " << top_regions[i].height << " width " << top_regions[i].width << " slot_offset " << top_regions[i].slot_offset << " subchannel offset " << top_regions[i].subchannel_offset << std::endl;
					cout << "Second: height: " << top_regions[j].height << " width " << top_regions[j].width << " slot_offset " << top_regions[j].slot_offset << " subchannel offset " << top_regions[j].subchannel_offset << std::endl;
				
					cout << "New combined regions: \n";
					cout << "Merged : height: " << merged.height << " width " << merged.width << " slot_offset " << merged.slot_offset << " subchannel offset " << merged.subchannel_offset << std::endl;
					if (above.valid) cout << "Above: height: " << above.height << " width " << above.width << " slot_offset " << above.slot_offset << " subchannel offset " << above.subchannel_offset << std::endl;
					if (below.valid) cout << "Below: height: " << below.height << " width " << below.width << " slot_offset " << below.slot_offset << " subchannel offset " << below.subchannel_offset << std::endl;

					cout<< "\n";
#endif
					top_regions[i] = merged;
					//since we changed the region's size, area must be recomputed
					i_area = top_regions[i].width * top_regions[i].height;

					if (above.valid) top_regions[j] = above;
					//if just below is valid
					else if (below.valid) top_regions[j] = below;
					
					if (above.valid && below.valid) top_regions.push_back(below);
					
					if (!above.valid && !below.valid) top_regions[j].valid = false;
				} //end merge
			} //end checking adjacent regions
		} //end comparison loop (j)
	} //end outisde region loop (i)

#ifdef DEBUG_WIMAX
	debug10("Region List:\n");
	uint32_t valid_num = 0;
	for (uint32_t i = 0; i < top_regions.size(); i++) {
		if (top_regions[i].valid)
			debug2(" Region %d, height %d, width %d, slot offset %d, subchannel offset %d\n", valid_num++, top_regions[i].height, top_regions[i].width, top_regions[i].slot_offset, top_regions[i].subchannel_offset);
	}
#endif

	//now we map the remaining bursts into the set of available regions using the minimum-height priority mapping system employed for the top-of-column allocations in eOCSA
	for (uint16_t i = 0; i < top_regions.size(); i++) {
		if (!top_regions[i].valid ) continue;
  	  	for (std::list<virtual_burst>::iterator j = sorted_allocs.begin(); j != sorted_allocs.end(); ++j) {
			if (j->mapped) continue;
			int16_t h_j = ceil ( (double)j->numslots / (double)top_regions[i].width);

			if (h_j > top_regions[i].height) {
				continue;
			}

			int16_t w_j = ceil ( (double)j->numslots / (double)h_j);

			//if we're still here, this will fit, so we map it in
  			db = (DlBurst*) map->getDlSubframe()->getPdu()->addBurst(nbdlbursts++);
  			db->setCid(j->cid);
  			db->setIUC(j->iuc);
  			db->setPreamble(j->preamble);
  		
  			db->setStarttime(top_regions[i].slot_offset * phy->getSlotLength(DL_) + DL_PREAMBLE);
  			db->setDuration(w_j * phy->getSlotLength(DL_));
  
  			db->setnumSubchannels(h_j);
  			db->setSubchannelOffset(top_regions[i].subchannel_offset);
  			(*j).mapped = true;
		
			//we take note of any space left over.  Possibly combine with the leftover from the last allocation.
			top_region_listing side;
			side.height = h_j;
			side.width = top_regions[i].width - w_j;
			side.slot_offset = top_regions[i].slot_offset + w_j;
			side.subchannel_offset = top_regions[i].subchannel_offset;;
			side.valid = true;

			//check if this side-region is the same as the one above
			uint16_t l = top_regions.size() - 1;
			if (top_regions[l].slot_offset == side.slot_offset && top_regions[l].width == side.width && top_regions[l].subchannel_offset + top_regions[l].height == side.subchannel_offset) {
				top_regions[l].height += side.height;
			} else if (side.height * side.width > 0) {
				top_regions.push_back(side);
			}

			overallocations += w_j * h_j - j->numslots;
			allocations += w_j * h_j;
  
			top_regions[i].subchannel_offset += h_j;
			top_regions[i].height -= h_j;
			if (top_regions[i].height == 0) break; //this region is full

			


			//there's a bug where occasionally DL bursts aren't properly removed for an extra frame
			//and are reallocatd with size 0, this isn't the ideal way to fix it, but it's the fasstest
			if (db->getDuration() == 0 || db->getnumSubchannels() == 0) {
				  map->getDlSubframe()->getPdu ()->removeBurst(db);
				  nbdlbursts--;
				  continue;
			}
		}
	}


  
	  //collect information for output about allocation sizes

	  uint16_t total_size = 0;
	  uint16_t unmapped_blocks = 0;
	  uint16_t num_bursts = 0;
	  uint16_t unmapped_bursts = 0;
  	  for (std::list<virtual_burst>::iterator i = sorted_allocs.begin(); i != sorted_allocs.end(); ++i) {
	  	num_bursts++;
		total_size += i->numslots;	
		if (!i->mapped) {
			unmapped_blocks += i->numslots;
			unmapped_bursts++;
			debug2("\t Unmapped burst %d has size %d\n", unmapped_bursts, i->numslots);
		}
	  }

	  if (num_bursts > max_dl_bursts) max_dl_bursts = num_bursts;

	  uint16_t empty_needed_blocks = mappable_height * mappable_width - allocations;

	  total_overallocations[num_bursts] += overallocations;
	  total_unmapped_blocks[num_bursts] += unmapped_blocks;
	  total_unmapped_bursts[num_bursts]  += unmapped_bursts;
	  if (unmapped_bursts > 0) total_empty_needed_blocks[num_bursts] += empty_needed_blocks;
	  total_number_of_frames[num_bursts]++;

	  //each line will contain [Frame Number] [maximum subframe size] [number of bursts] [sum of all burst sizes] [overallocations] [unused space] [unmapped blocks]
	  if (total_size > 0)
	  	allocation_file << getMac()->frame_number_ << "\t" << mappable_height * mappable_width << "\t" << num_bursts << "\t" << total_size << "\t" << overallocations << "\t" << empty_needed_blocks << "\t" << unmapped_bursts << "\t" << unmapped_blocks << std::endl;
  
    } //mOCSA case braces to allow variable declaration without issues
  default:
  	//only here to avoid compiler complaining about "OFDM" allocation type not being supported
	break;
  }




//  for (int i=start; i<index_burst; i++) {
//      db = (DlBurst*) map->getDlSubframe()->getPdu ()->addBurst (nbdlbursts++);
//      db->setCid (virtual_alloc[i].cid);
//      db->setIUC (virtual_alloc[i].iuc);
//      db->setPreamble(virtual_alloc[i].preamble); 
//      
//      int num_allocation_slots = virtual_alloc[i].numslots;
//      if (virtual_subchannel_offset  + num_allocation_slots <= total_dl_subchannel_pusc) {
//        //if we can fit it all in this column, stick it there
//      	debug2("DL_MAP.FILL.Adding burst %d to existing DL_MAP column %d, starting %d subchannels down\n",i,virtual_symbol_offset, virtual_subchannel_offset);
//        db->setStarttime(virtual_symbol_offset); 
//        db->setSubchannelOffset(virtual_subchannel_offset);
//      	db->setnumSubchannels(num_allocation_slots);
//      	db->setDuration(2);
//      	virtual_subchannel_offset += num_allocation_slots;
//      	if (virtual_subchannel_offset == total_dl_subchannel_pusc) {
//      		virtual_symbol_offset+=2;
//      		virtual_subchannel_offset = 0;
//      	}
//      }
//      else if (num_allocation_slots <= total_dl_subchannel_pusc) {
//      	//if it all fits in one column, fill that in and keep track of how much room is left at the end
//      	if(! (virtual_symbol_offset == DL_PREAMBLE && virtual_subchannel_offset == 0)) virtual_symbol_offset+=2; //we're leaving the top of the last column empty and starting a new one, so move this over by 2
//
//	debug2("DL_MAP.FILL.Adding burst %d to new (single) DL_MAP column %d\n",i,virtual_symbol_offset+1);
//      	virtual_subchannel_offset = num_allocation_slots;
//      	db->setStarttime(virtual_symbol_offset); 
//        db->setSubchannelOffset(0);
//      	db->setnumSubchannels(num_allocation_slots);
//      	db->setDuration(2);
//      	if (num_allocation_slots == total_dl_subchannel_pusc) {
//      		virtual_symbol_offset+=2;
//      		virtual_subchannel_offset = 0;
//      	}
//      }
//      else {
//      	//fill in as many columns as we need, and leave whatever's left on the last column empty
//	debug2("DL_MAP.FILL.Adding burst %d to new (multiple) DL_MAP columns %d to %d\n",i,virtual_symbol_offset+1, (int)(virtual_symbol_offset+1+ceil(num_allocation_slots/ total_dl_subchannel_pusc) ));
//      	virtual_symbol_offset+=2;
//      	db->setStarttime(virtual_symbol_offset); 
//        db->setDuration(ceil( (double)num_allocation_slots / (double)total_dl_subchannel_pusc));
//        if (db->getDuration() % 2 == 0) db->setDuration(db->getDuration() + 1); //we allocate in sets of 2 cols, but have initial offset 3
//        db->setSubchannelOffset(0);
//        db->setnumSubchannels(total_dl_subchannel_pusc);
//        virtual_symbol_offset += ceil((double)num_allocation_slots / (double)total_dl_subchannel_pusc) + 1;        
//      }
//      
//      
//      //there's a bug where some messages are not being correctly flushed from the queues after transmission, but their trnamission length is getting set to 0, so not actual retransmission occurs.  This prevents that
//      if (db->getDuration() == 0 || db->getnumSubchannels() == 0) {
//      	  debug10(" numsymbol: %d, numsub: %d\n", db->getDuration(), db->getnumSubchannels());
//		  map->getDlSubframe()->getPdu ()->removeBurst(db);
//		  nbdlbursts--;
//		  continue;
//      }
//
//      debug10("\t\t\t\t new duration (in symbols) for burst of index %d is %d\n", i, db->getDuration());
//
//      virtual_accum_slots += virtual_alloc[i].numslots;
//
//      debug10("DL.add bursts: index :%d, cid :%d, iuc :%d, subchannel_offset :%d, symbol_offset :%d, num_slots (num_subchannel) :%d, num_symbol :%d, accum_slots :%d\n", i, virtual_alloc[i].cid, virtual_alloc[i].iuc, virtual_subchannel_offset, virtual_symbol_offset, virtual_alloc[i].numslots, db->getDuration(), virtual_accum_slots);
//   }

/*
//This comment is the previous version of adding DL burst
  if (number_dl_ie > 0) {
      for (numie = 0; numie < (int) dl_map->nb_ies ; numie++) {
           mac802_16_dlmap_ie  dlmap_ie = dl_map->ies[numie];

	   debug10 ("DL.1.addburst, #bursts %d, CID :%d, DIUC :%d, SymbolOffset :%d, #symbols :%d, SubchannelOffset :%d, #subchannels :%d\n", nbdlbursts, dlmap_ie.cid, dlmap_ie.diuc,dlmap_ie.symbol_offset, dlmap_ie.num_of_symbols, dlmap_ie.subchannel_offset, dlmap_ie.num_of_subchannels); 
	   addDlBurst (nbdlbursts++,dlmap_ie.cid, dlmap_ie.diuc,dlmap_ie.symbol_offset, dlmap_ie.num_of_symbols, dlmap_ie.subchannel_offset, dlmap_ie.num_of_subchannels); 
      }
      free (dl_map);
  }
*/

#ifdef DEBUG_WIMAX
  assert (dlduration <= maxdlduration);
#endif
  Ptr<Packet> copier;


  // Now transfer the packets to the physical bursts starting with broadcast messages
  // In this version, we can directly map burst information from virtual burst into physical burst.
  int b_data = 0;
  int max_data = 0;
  //All_Headers wimaxHdr;
  double txtime2 = 0;
  int offset = 0;
  int subchannel_offset_wimaxhdr = 0;
  int num_burst_before_data = 0;
  Burst *b;
  int i_burst = 0;

  b = map->getDlSubframe()->getPdu ()->getBurst (i_burst);
  i_burst++;
  b_data  = 0;
  txtime2 = 0;
  offset  = 0;
  subchannel_offset_wimaxhdr = 0;
  max_data = phy->getMaxPktSize (b->getDuration(), mac_->getMap()->getDlSubframe()->getProfile (b->getIUC())->getEncoding(), DL_)-b_data;

  p = map->getDL_MAP();
  
  p->RemoveHeader(ah_p);

  //if we don't perfectly mesh with the slots / repetition, pad it out so everything lines up right when splitting the repetition on the SS
  int pad_ammount = (virtual_alloc[0].byte / virtual_alloc[0].rep) - p->GetSize();
  if (pad_ammount > 0) {
	Ptr<Packet> pad = Create<Packet>(pad_ammount);
	p->AddAtEnd(pad);
  }
	//virtual_alloc[0].byte = total_slots_per_rep * virtual_alloc[0].rep * getPhy()->getSlotCapacity(virtual_alloc[0].mod, DL_);
	//virtual_alloc[0].numslots = total_slots_per_rep * virtual_alloc[0].rep;
 
  copier = p->Copy();
  for (int i = 1; i < virtual_alloc[0].rep; i++) p->AddAtEnd(copier->Copy());
  offset = b->getStarttime( );


  debug10 ("DL/UL.start transfer packets into burst => symbol offset of DLMAP_start at :%d, subchannel offset at :%d\n", offset, subchannel_offset_wimaxhdr);

  ulduration = 0;

  Ofdm_mod_rate dlul_map_mod = map->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
  
   //   ah_p.PIH.channel_index = 1; //broadcast packet
    
  debug2("The length of the queue of burst is [%d]\n",b->getQueueLength_packets());

  b->enqueue(p);      //enqueue the DL_MAP to the burst
  num_burst_before_data ++;
  
  debug2("The length of the queue of burst is [%d]\n",b->getQueueLength_packets());
  //  b->setDuration(b->getDuration()-txtime_s);
//  b_data += ch->size();
  b_data += p->GetSize();

  debug2("DL/UL.wimaxHdr:DL-MAP --- Mod[%d]\t size[%d]\t symbol offset[%d]\t symbol num[%d]\t subchannel offset[%d]\t subchannel num[%d]\n", dlul_map_mod, p->GetSize(), b->getStarttime(), b->getDuration(), b->getSubchannelOffset(), b->getnumSubchannels());

  b = map->getDlSubframe()->getPdu ()->getBurst (i_burst);
  i_burst++;
  offset = b->getStarttime( );  

  p = map->getUL_MAP();
  All_Headers ahp;
  p->RemoveHeader(ahp);
  p->AddHeader(ahp.GMH);
  
  num_burst_before_data++;
  
  //take the repetition coding into account 
  copier = p->Copy();
  for (int i = 1; i < virtual_alloc[1].rep; i++) p->AddAtEnd(copier->Copy());

  debug10 ("DL/UL.wimaxHdr:UL-MAP --- Mod[%d]\t size[%d]\t symbol offset[%d]\t symbol num[%d]\t subchannel offset[%d]\t subchannel num[%d]\n", dlul_map_mod, p->GetSize(), b->getStarttime(), b->getDuration(), b->getSubchannelOffset(), b->getnumSubchannels());

  b->enqueue(p);      //enqueue into burst
  debug2("The length of the queue of burst is [%d]\n",b->getQueueLength_packets());
  b_data += p->GetSize();

  if (getMac()->sendDCD || map->getDlSubframe()->getCCC()!= getMac()->dlccc_) {
    b = map->getDlSubframe()->getPdu ()->getBurst (i_burst);
    i_burst++;
	
    p = map->getDCD();
    
    num_burst_before_data++;
    All_Headers ahp;
    p->RemoveHeader(ahp);
    p->AddHeader(ahp.GMH);
    
    //take the repetition factor into account
    copier = p->Copy();
    for (int i = 1; i < virtual_alloc[2].rep; i++) p->AddAtEnd(copier->Copy());
    
    dlul_map_mod = map->getDlSubframe()->getProfile (b->getIUC())->getEncoding();

    debug10 ("DL/UL.wimaxHdr:DCD --- Mod[%d]\t size[%d]\t symbol offset[%d]\t symbol num[%d]\t subchannel offset[%d]\t subchannel num[%d]\n", dlul_map_mod, p->GetSize(), b->getStarttime(), b->getDuration(), b->getSubchannelOffset(), b->getnumSubchannels());

    b->enqueue(p);      //enqueue into burst
    b_data += p->GetSize();
  }

  
  if (getMac()->sendUCD || map->getUlSubframe()->getCCC()!= getMac()->ulccc_) {
    b = map->getDlSubframe()->getPdu ()->getBurst (i_burst);
    i_burst++;

    p = map->getUCD();

    num_burst_before_data++;
    All_Headers ahp;
    p->RemoveHeader(ahp);

    UCD_FRAME_DATA ufd;
    p->RemoveHeader(ufd);
    p->AddHeader(ufd);
    p->PeekHeader(ufd);



    p->AddHeader(ahp.GMH);
    
    //take the repetition factor into account
    copier = p->Copy();
    for (int i = 1; i < virtual_alloc[3].rep; i++) p->AddAtEnd(copier->Copy());

    
    dlul_map_mod = map->getDlSubframe()->getProfile (b->getIUC())->getEncoding();
    
    debug10 ("DL/UL.wimaxHdr:UCD --- Mod[%d]\t size[%d]\t symbol offset[%d]\t symbol num[%d]\t subchannel offset[%d]\t subchannel num[%d]\n", dlul_map_mod, p->GetSize(), b->getStarttime(), b->getDuration(), b->getSubchannelOffset(), b->getnumSubchannels());

    b->enqueue(p);      //enqueue into burst
    b_data += p->GetSize();
  }

  //Now get the other bursts
  debug2("BS scheduler is going to handle other bursts (not DL/UL_MAP, DCD/UCD), #bursts = %d\n", map->getDlSubframe()->getPdu ()->getNbBurst() - num_burst_before_data);
  for (int index = num_burst_before_data ; index < map->getDlSubframe()->getPdu ()->getNbBurst(); index++) {
      Burst *b = map->getDlSubframe()->getPdu ()->getBurst (index);

      Connection *c=mac_->getCManager ()->get_connection (b->getCid(),OUT_CONNECTION);
      debug2("DL/UL.other CID [%d]\n", b->getCid());

      if (b && c && b->getDuration()> 0 && b->getnumSubchannels() > 0) {
	debug2 ("DL/UL.check CID :%d, flag: FRAG :%d, PACK :%d, ARQ: %p\n", b->getCid(), c->isFragEnable(), c->isPackingEnable(), c->getArqStatus ()); 
	transfer_packets_for_dl_subframe(c, b);

      }
      //Begin RPI
//      if (c!=NULL) {
//	  debug10 ("DL/UL.before transfer_packets1 (other data) to burst_i :%d, CID :%d, b_data (bytes_counter) :%d\n", index, b->getCid(), b_data);
//	  if (c->isFragEnable() && c->isPackingEnable() &&  (c->getArqStatus () != NULL) && (c->getArqStatus ()->isArqEnabled() == 1)) {
//	     debug2("DL/UL.BSSscheduler is goting to transfer packet with fragackarq.\n");
//	     b_data = transfer_packets_with_fragpackarq (c, b, b_data); /*RPI*/
//	  } else {	
//	     b_data = transfer_packets1(c, b, b_data);
//	  }
//       }
//       debug10 ("\nDL/UL.after transfer_packets1 (other data) to burst_i :%d, CID :%d, b_data (update_counter) :%d\n", index, b->getCid(), b_data);
//       debug10 ("Dl/UL.the length of the queue of burst is [%d]\n", b->getQueueLength_packets());
  }//end loop ===> transfer bursts

  //Print the map
//#ifdef DEBUG_WIMAX
  debug2 ("\n==================BS %s Subframe============================\n", HelperFunctions::getMacAddrString(mac_->addr()).c_str());
  mac_->getMap()->print_frame();
  debug2 ("===========================================================\n");
//#endif
	//mac_->getMap()->getDlSubframe()->getTimer()->start(mac_->getMap()->getStarttime());
}

/**   
 * Add a downlink burst with the given information
 * @param burstid The burst number
 * @param c The connection to add
 * @param iuc The profile to use
 * @param dlduration current allocation status
 * @param the new allocation status
 */
/*       // rpi removed this dl burst ----------------------------------------------------------------------------
	 int Wimax2BSScheduler::addDlBurst (int burstid, Connection *c, int iuc, int dlduration, int maxdlduration)
	 {
	 double txtime; //tx time for some data (in second)
	 int txtime_s;  //number of symbols used to transmit the data
	 OFDMPhy *phy = mac_->getPhy();

	 //add a burst for this node
	 DlBurst *db = (DlBurst*) mac_->getMap()->getDlSubframe()->getPdu ()->addBurst (burstid);
	 db->setCid (c->get_cid());
	 db->setIUC (iuc);
	 db->setStarttime (dlduration);

	 txtime = phy->getTrxSymbolTime (c->queueByteLength(), mac_->getMap()->getDlSubframe()->getProfile (db->getIUC())->getEncoding());
	 txtime += c->queueLength() * TX_GAP; //add small gaps between packets to send
	 txtime_s = (int) ceil(txtime/phy->getSymbolTime ()); //in units of symbol
	 if (txtime_s < maxdlduration-dlduration) {
	 db->setDuration (txtime_s);
	 dlduration += db->getDuration ()+1; //add 1 OFDM symbol between bursts
	 } else {
	 //fill up the rest
	 db->setDuration (maxdlduration-dlduration);
	 dlduration = maxdlduration;
	 }
	 return dlduration;
	 }

*/ // rpi removed this dl burst ----------------------------------------------------------------------------

// rpi added adddlburst for ofdma ------------------------------------------------------------------

/**
 * Add a downlink burst with the given information
 * @param burstid The burst number
 * @param c The connection to add
 * @param iuc The profile to use
 * @param dlduration current allocation status
 * @param the new allocation status
 */
void Wimax2BSScheduler::addDlBurst (int burstid, int cid, int iuc, int ofdmsymboloffset, int numofdmsymbols, int subchanneloffset, int numsubchannels)
{
  debug2("Adding DlBurst with cid %d and iuc %d \n", cid, iuc);
  DlBurst *db = (DlBurst*) mac_->getMap()->getDlSubframe()->getPdu ()->addBurst (burstid);
  db->setCid (cid);
  db->setIUC (iuc);
  db->setStarttime (ofdmsymboloffset);
  db->setDuration (numofdmsymbols);
  db->setSubchannelOffset (subchanneloffset);
  db->setnumSubchannels (numsubchannels);
  
  debug2("Data Burst:----symbol_offset[%d]\t symbol_num[%d]\t subchannel_offset[%d]\t subchannel_num[%d]\n",
	 ofdmsymboloffset, numofdmsymbols,subchanneloffset,numsubchannels);

}

int Wimax2BSScheduler::doesMapExist(int cid, int *cid_list, int num_of_entries){
  for(int i = 0; i<num_of_entries; ++i){
    if(cid_list[i]==cid)
      return i;
  }
  return -1;
}
		 
//dl_stage2 (Virtual Allocation) for data allocation (all CIDs)
mac802_16_dl_map_frame * Wimax2BSScheduler::dl_stage2 (Connection *head, int input_subchannel_offset, int total_subchannels, int total_symbols, int symbol_start, int stripping, int total_dl_slots_pusc) {
  struct mac802_16_dl_map_frame *dl_map;
  struct DL_MAP_IE *ies;
  Connection *con;
  int i, ie_index, temp_index;
       if (temp_index); //quiet compiler
  int num_of_symbols, num_of_subchannels, num_of_slots;
       if (num_of_symbols || num_of_subchannels); //quiet compiler
  double allocationsize;
  int freeslots;
  int symbol_offset = symbol_start;
  int subchannel_offset = input_subchannel_offset;
  int subchannel_start = 0;
       if (subchannel_start); //quiet compiler
  int num_of_data_connections = 0;
  int total_dl_free_slots = 0;
  ConnectionType_t contype;
  SchedulingType_t schedtype;
  Ptr<Wimax2NetDevice>  mac_ = getMac ();
  Ofdm_mod_rate mod_rate;

  int slots_per_con[MAX_MAP_IE];
       if (slots_per_con[0]); //quiet compiler
  int cid_list[MAX_MAP_IE];
       if (cid_list[0]); //quiet compiler
  int diuc_list[MAX_MAP_IE];
       if (diuc_list[0]); //quiet compiler


  int leftOTHER_slots=0;
  int needmore_con[3]={0,0,0};
       if (needmore_con[0]); //quiet compiler
  int needmore_c=0;
       if (needmore_c); //quiet compiler
  int ori_symbols = total_symbols;
  	if (ori_symbols); //quiet, compiler
  int req_slots_tmp1 = 0;
  	if (req_slots_tmp1); //quiet, compiler
  int return_cid_tmp = 0;
       if (return_cid_tmp); //quiet compiler

  dl_map = (struct mac802_16_dl_map_frame *) malloc(sizeof(struct mac802_16_dl_map_frame));
  dl_map->type = MAC_DL_MAP;
  dl_map->bsid = mac_->addr(); // its called in the mac_ object

  ies = dl_map->ies;
  ie_index = 0; //0 is for the ul-map
  int temp_index_burst = index_burst;

  if ((total_symbols%2)==1) total_symbols=total_symbols-1;
  freeslots = total_subchannels * floor(total_symbols/2);
  total_dl_free_slots = freeslots;

  debug10 ("DL2,****FRAME NUMBER**** :%d, Total_DL_Slots :%d, FreeSlots(TotalSub*TotalSym/2) :%d, TotalSub :%d, Ori_TotalSymbol :%d, TotalSymbol :%d, StartSymbol :%d\n", mac_->getFrameNumber(), total_dl_slots_pusc, total_dl_free_slots, total_subchannels, ori_symbols, total_symbols, symbol_start );

//1. (1st priority) Allocate management message (basic, primary, secondary) 
  con=head;
  while(con!=NULL){
    if(con->get_category()==CONN_DATA) ++num_of_data_connections;
    con = con->next_entry();
  }

  for(i=0;i<3;++i){
    con = head;
    if(i==0) 		contype = CONN_BASIC;
    else if(i==1) 	contype = CONN_PRIMARY;
    else 		contype = CONN_SECONDARY;
    while(con!=NULL) {
	if(con->get_category() == contype) {
	    allocationsize = con->queueByteLength();

  	    int con_byte = con->queueByteLength();
            int rep = 1;
	    if ( con_byte > 0 ) {
               Ofdm_mod_rate con_mod = mac_->getMap()->getDlSubframe()->getProfile (con->getPeerNode()->getDIUC())->getEncoding();
               int con_slots = (int) rep*ceil((double)con_byte/(double)mac_->getPhy()->getSlotCapacity(con_mod, DL_)); 
	       int virtual_alloc_exist = doesvirtual_allocexist(index_burst, con->get_cid());
	       int add_slots = 0;
	       if (virtual_alloc_exist > 0) {
		  add_slots = overallocation_withoutdlmap (index_burst, total_dl_slots_pusc, con_slots);
		  int t_index = addslots_withoutdlmap(index_burst, con_byte, con_slots, con->get_cid());
		  	if (t_index); //quiet, compiler
	    	  debug10 ("DL Add more slots into existing burst contype(%d), index :%d, cid :%d, byte :%f, slots :%d\n", contype, t_index, con->get_cid(), virtual_alloc[t_index].byte, virtual_alloc[t_index].numslots);
	       } else {
		  add_slots = overallocation_withdlmap (index_burst, total_dl_slots_pusc, con_slots);
		  if (add_slots > 0) {
            	 virtual_alloc[index_burst].alloc_type = 5;
  	    	     virtual_alloc[index_burst].cid = con->get_cid();
  	    	     virtual_alloc[index_burst].n_cid = 1;
  	    	     //virtual_alloc[index_burst].iuc = getDIUCProfile(con_mod);
  	    	     virtual_alloc[index_burst].iuc = getDIUCProfile(OFDM_QPSK_1_2);
  	    	     virtual_alloc[index_burst].numslots = add_slots;
  	    	     virtual_alloc[index_burst].byte = (add_slots * mac_->getPhy()->getSlotCapacity(con_mod, DL_));
  	    	     virtual_alloc[index_burst].rep =  rep;
  	    	     virtual_alloc[index_burst].dl_ul = 0;
  	    	     virtual_alloc[index_burst].ie_type = 0;
	    	     debug10 ("DL Add new burst contype(%d), index :%d, cid :%d, byte :%f, slots :%d\n", contype, index_burst, con->get_cid(), virtual_alloc[index_burst].byte, virtual_alloc[index_burst].numslots);
  	    	     index_burst++;
		  }
	      }
	    }//end if q > 0

	    debug10 ("DL2.management msg, contype(%d), index :%d, CID :%d, Q-Bytes :%d, allocationsize :%d\n", contype, index_burst, con->get_cid(), con->queueByteLength(), con_byte);
	}
	con = con->next_entry();
    }//end con!=null
  }//end for loop

//2. Calculate #of active connections (ugs, ertps, rtps, nrtps, be) 
  int conn_per_schetype[6]={0,0,0,0,0,0};
  int con_ugs_all  =0; int con_ertps_all=0; int con_rtps_all =0;  int con_nrtps_all=0; int con_be_all   =0;

  for (i=0;i<5;++i) {
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;    
    else if(i==3)       schedtype = SERVICE_nrtPS; 
    else                schedtype = SERVICE_BE;

    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){
	if(schedtype==SERVICE_UGS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[0]++;
	  con_ugs_all++;
	} else if (schedtype==SERVICE_ertPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[1]++;
	  con_ertps_all++;
	} else if (schedtype==SERVICE_rtPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[2]++;
	  con_rtps_all++;
	} else if (schedtype==SERVICE_nrtPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[3]++;
	  con_nrtps_all++;
	} else if (schedtype==SERVICE_BE) {
	  if (con->queueByteLength() > 0) conn_per_schetype[4]++;
	  con_be_all++;
	} else {
	  conn_per_schetype[5]++;
	}
      }
      con = con->next_entry();
    }
  }

  int active_conn_except_ugs = 0;
  for (int i=1; i<5; i++) active_conn_except_ugs =+ conn_per_schetype[i];
  
  debug10 ("DL2, Active UGS <%d>, ertPS <%d>, rtPS <%d>, nrtPS <%d>, BE <%d>\n", conn_per_schetype[0],conn_per_schetype[1], conn_per_schetype[2], conn_per_schetype[3], conn_per_schetype[4]);
  debug10 ("\tAll UGS <%d>, ertPS <%d>, rtPS <%d>, nrtPS <%d>, BE <%d>\n", con_ugs_all, con_ertps_all, con_rtps_all, con_nrtps_all, con_be_all);

//3. (2nd priority) Allocate UGS 
  int next_ie_index = ie_index;
       if (next_ie_index); //quiet compiler
  //***************
  leftOTHER_slots = freeslots;
  //UGS Scheduling
  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    
    while (con!=NULL) {
      if (con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){
        req_slots_tmp1 = 0;
        int grant_slots = 0;
        num_of_slots = 0; 
        mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();

        if (schedtype==SERVICE_UGS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  if (con->queueByteLength() > 0) {
#ifdef UGS_AVG
          allocationsize = (int) ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
#endif
#ifndef UGS_AVG
          int tmp_getpoll = con->getPOLL_interval();
          if ( (tmp_getpoll%sfqos->getPeriod())== 0 ) {
                allocationsize = ceil(sfqos->getDataSize());
                con->setPOLL_interval(0);
          } else {
                allocationsize = 0;
          }
          tmp_getpoll++;
          con->setPOLL_interval(tmp_getpoll);
#endif
	  } else {
	     allocationsize = 0;
	  }

//ARQ aware scheduling
	  int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
	  int t_bytes = 0;
	  double ori_grant = allocationsize;
	  	if (ori_grant); //quiet, compiler
	  if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
		t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
		if (t_bytes > 0) {
			allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
		}
		debug10 ("\tARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
		allocationsize = allocationsize + 14;	//optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub
		
	  } 
//End ARQ aware scheduling

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));;
	  if (con->queueByteLength() <=0) { 
		allocationsize = 0;
		num_of_slots = 0;
	  } else {
		num_of_slots = grant_slots;
	  }

          Ofdm_mod_rate con_mod = mac_->getMap()->getDlSubframe()->getProfile (con->getPeerNode()->getDIUC())->getEncoding();
          int con_slots = num_of_slots; 
          int con_byte = allocationsize; 
	  int virtual_alloc_exist = doesvirtual_allocexist(index_burst, con->get_cid());
	  int add_slots = 0;
	  if (virtual_alloc_exist > 0) {
		add_slots = overallocation_withoutdlmap (index_burst, total_dl_slots_pusc, con_slots);
		addslots_withoutdlmap(index_burst, con_byte, con_slots, con->get_cid());
	  } else {
		add_slots = overallocation_withdlmap (index_burst, total_dl_slots_pusc, con_slots);
		if (add_slots > 0) {
            	   virtual_alloc[index_burst].alloc_type = 5;
  	    	   virtual_alloc[index_burst].cid = con->get_cid();
  	    	   virtual_alloc[index_burst].n_cid = 1;
  	    	   virtual_alloc[index_burst].iuc = getDIUCProfile(con_mod);
  	    	   virtual_alloc[index_burst].numslots = add_slots;
  	    	   virtual_alloc[index_burst].byte = (add_slots * mac_->getPhy()->getSlotCapacity(con_mod, DL_));
  	    	   virtual_alloc[index_burst].rep =  1;
  	    	   virtual_alloc[index_burst].dl_ul = 0;
  	    	   virtual_alloc[index_burst].ie_type = 0;
  	    	   index_burst++;
	   	}
	  }

	  debug10 ("DL2.UGS, DataSize :%f, period :%d, numslots :%d, CID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), con_slots, con->get_cid(), con->getPeerNode()->getDIUC());
	  debug10 ("\tQ-bytes :%d, grant-bytes :%ld\n", con->queueByteLength(), (long int)( con_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }//end UGS
      }
      con = con->next_entry();
    }//end con != NULL
  }//end UGS


//4. (3rd priority) Allocate to Others 
//In this version, we validate the result only on best effort, we do not test rtPS, ertPS, and nrtPS yet.
//Fair allocation
//Find out how many connections can be supported given variable DL MAP
  int max_conn_dlmap = max_conn_withdlmap(index_burst, total_dl_slots_pusc);
  int active_shared_conn = active_conn_except_ugs;
  if (active_shared_conn > max_conn_dlmap) active_shared_conn = max_conn_dlmap;

  double weighted_isp = 1;
       if (weighted_isp); //quiet compiler
  int max_conn = MAX_CONN;
  int fair_share[max_conn];
  int dl_all_con=0;
       if (dl_all_con); //quiet compiler

  con_drr dl_con_data[MAX_CONN];
  con_data_alloc con_data[max_conn];
  con_data_alloc con_data_final[max_conn];

  bzero(fair_share, max_conn*sizeof(int));
  bzero(con_data, max_conn*sizeof(con_data_alloc));
  bzero(con_data_final, max_conn*sizeof(con_data_alloc));
  bzero(dl_con_data, max_conn*sizeof(con_drr));

  for (int i=0; i<MAX_CONN; i++) {
     dl_con_data[i].cid = -2;
     dl_con_data[i].quantum = 0;
     dl_con_data[i].counter = 0;
  }

//Find out #request_slots for each connection
 int j_con = 0;
 for (i=0; i<5; ++i){
     con = head;
     if(i==0)        schedtype = SERVICE_UGS;
     else if(i==1)   schedtype = SERVICE_ertPS;
     else if(i==2)   schedtype = SERVICE_rtPS;
     else if(i==3)   schedtype = SERVICE_nrtPS;
     else            schedtype = SERVICE_BE;
     if ( i==0 ) continue;      

     while(con != NULL) {

        if (con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype) {
	   Ofdm_mod_rate mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();

           int withfrag = 0;
/*           if  (con->getFragmentBytes()>0) {
              withfrag = con->queueByteLength() - con->getFragmentBytes() + 4;
           } else {
              withfrag = con->queueByteLength();;
           }*/

		int length = 0;
		if (con->getArqStatus () != NULL && con->getArqStatus ()->isArqEnabled() == 1)
		{
			if(con->getArqStatus()->arq_retrans_queue_->GetNBytes()>0 ||con->getArqStatus()->arq_trans_queue_->GetNBytes() >0 )
			{
				debug2("ARQ retrans queue length is %d\n\n",con->getArqStatus()->arq_retrans_queue_->GetNBytes() );
				if(con->getArqStatus()->arq_retrans_queue_->GetNBytes()>0) 
					length =con->getArqStatus()->arq_retrans_queue_->GetNBytes();
				else
					length = con->getArqStatus()->arq_trans_queue_->GetNBytes();
			}
			else
			{
				length = con->queueByteLength();
			}
		}
		else
		{
			length = con->queueByteLength();
		}	

           if  (con->getFragmentBytes()>0) 
	    {
              withfrag = length - con->getFragmentBytes() + 4;
		
           } 
           else 
           {
              withfrag = length+4;
           }
	    debug2("ARQ retrans queue got allocation size %d\n\n", withfrag);
//ARQ aware scheduling
           int arq_enable_f = 0;
	   if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	   }
           int t_bytes = 0;
           double ori_grant = withfrag;
           	if (ori_grant); //quiet, compiler
           if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)withfrag % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        withfrag = withfrag + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 ("\t ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%d, all_header_included :%d\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, withfrag, withfrag + 14);
                withfrag = withfrag + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

           }
//End ARQ

           int req_slots = ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
           if (req_slots > 0) {
              debug10 ("Con-Request byte :%d, slotcapa :%d, reqslot :%d\n", con->queueByteLength(), mac_->getPhy()->getSlotCapacity(mod_rate, DL_), req_slots );
              Ptr<Packet> np;
              if ( (np = con->queueLookup(0)) != NULL ) {
              	 All_Headers ah_np;
              	 np->PeekHeader(ah_np);
                 int p_size = ah_np.GMH.GMH.len;
                 if (p_size); //quiet, compiler
                 debug10 ("CON CID :%d, q->byte :%d, q->len :%d, packet_size :%d, frag_no :%d, frag_byte :%d, frag_stat :%d\n", con->get_cid(), con->queueByteLength(), con->queueLength(), p_size, con->getFragmentNumber(), con->getFragmentBytes(), (int)con->getFragmentationStatus() );
              }

              Ofdm_mod_rate t_mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();
              con_data[j_con].cid = con->get_cid();
              con_data[j_con].direction = DL_;
              con_data[j_con].mod_rate = t_mod_rate;
              con_data[j_con].req_slots = req_slots;
              con_data[j_con].req_bytes = withfrag;
              con_data[j_con].grant_slots = 0;
              con_data[j_con].weight = 0;
              con_data[j_con].counter = 0;
              j_con++;
          } else {
/*
            for (int j=0; j<MAX_CONN; j++) {
              if (dl_con_data[j].cid == con->get_cid()) {
                 dl_con_data[j].counter = 0;
                 break;
              }
            }
*/
          }//end else
        }
        con = con->next_entry();
     }
 }

 if (j_con == active_conn_except_ugs) debug10 ("DL2.other QoS flows OK :num_con :%d\n", j_con);
 else  debug10 ("Error Panic\n");

//Sorting and make sure #request slots is correct
 if (j_con>1)  {     
	for (int i=0; i<j_con; i++) { 
		debug10 ("DL2.other QoS flows, Req_slots Before i :%d, cid :%d, counter :%d, req_slots :%d, mod :%d\n", i, con_data[i].cid, con_data[i].counter, con_data[i].req_slots, (int)con_data[i].mod_rate); 		
	}
        bubble_sort(j_con, con_data, 0);
        for (int i=0; i<j_con; i++) { 
		debug10 ("DL2.other QoS flows, Req_slots After i :%d, cid :%d, counter :%d, req_slots :%d, mod :%d\n", i, con_data[i].cid, con_data[i].counter, con_data[i].req_slots, (int)con_data[i].mod_rate); 
	}
 }

 int t_freeslots = 0;
 int j_need = j_con;
 if (j_need > max_conn_dlmap) j_need = max_conn_dlmap;
 if (j_need > 0 ) t_freeslots = freeslots_withdlmap_given_conn(index_burst, total_dl_slots_pusc, j_need);

//Algorithm is here
//Note that in this version, we do slot-fair allocation or different MCSs will get the same throughput
//Next version, we will include weighted fair allocation
 int j_count = j_need;
 for (int i=0; i<j_need; i++) {
        if ( (t_freeslots <=0) ) break;
        fair_share[i] = floor(t_freeslots/j_count);
        if (fair_share[i] >= con_data[i].req_slots) {
           con_data[i].grant_slots = con_data[i].req_slots;
        } else {
           con_data[i].grant_slots = fair_share[i];
        }
	debug10 ("DL2.other QoS flows, Freeslots :%d, fairshare[%d] ;%d, grant_slots :%d\n", t_freeslots, i, fair_share[i], con_data[i].grant_slots);
        t_freeslots = t_freeslots - con_data[i].grant_slots;
	j_count--;
 }

 for (int i=0; i<j_need; i++) {
     debug10 ("\tAfter fair share ->Con :%d, req_slots :%d, grant_slots :%d\n", con_data[i].cid, con_data[i].req_slots, con_data[i].grant_slots);
 }

//Assign granted slots to each connection (virtual burst)
 for (int i=0; i<j_need; i++) {
    if (con_data[i].grant_slots >=0 ) {
       Ofdm_mod_rate con_mod = con_data[i].mod_rate;
       int con_slots = con_data[i].grant_slots; 
       int con_byte = (con_slots * mac_->getPhy()->getSlotCapacity(con_mod, DL_));
       debug10 ("OTHERS, i ;%d, numslots :%d, bytes :%d, CID :%d\n", i, con_slots, con_byte, con_data[i].cid);

       int virtual_alloc_exist = doesvirtual_allocexist(index_burst, con_data[i].cid);
       int add_slots = 0;
       if (add_slots); //quiet compiler
       if (virtual_alloc_exist > 0) {
	  int add_slots = overallocation_withoutdlmap (index_burst, total_dl_slots_pusc, con_slots);
	  	if (add_slots); //quiet, compiler
	  addslots_withoutdlmap(index_burst, con_byte, con_slots, con_data[i].cid);
       } else {
	  int add_slots = overallocation_withdlmap (index_burst, total_dl_slots_pusc, con_slots);
	  if (add_slots > 0) {
       	     virtual_alloc[index_burst].alloc_type = 5;
      	     virtual_alloc[index_burst].cid = con_data[i].cid; 
	     virtual_alloc[index_burst].n_cid = 1;
  	     virtual_alloc[index_burst].iuc = getDIUCProfile(con_mod);
  	     virtual_alloc[index_burst].numslots = add_slots;
  	     virtual_alloc[index_burst].byte = (add_slots * mac_->getPhy()->getSlotCapacity(con_mod, DL_));
  	     virtual_alloc[index_burst].rep =  1;
  	     virtual_alloc[index_burst].dl_ul = 0;
  	     virtual_alloc[index_burst].ie_type = 0;
  	     index_burst++;
	   }
       }

     debug10 ("DL2.other QoS flows, index :%d, grant_numslots :%d, grant_bytes :%ld, CID :%d\n", i, con_slots, (long int)( con_slots*mac_->getPhy()->getSlotCapacity(con_data[i].mod_rate, DL_)), con_data[i].cid);
    }

 }

/*
// This comment is the previous allocation for all QoS classes but we do not consider the variable part of DL_MAP.
// Users can apply some portions of the code here for other QoSs.
// 4. (3rd priority) Allocate other flows (except UGS) fairly (BE only)
  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;
    else if(i==3)       schedtype = SERVICE_nrtPS;
    else                schedtype = SERVICE_BE;
    
    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){

        req_slots_tmp1 = 0;
        return_cid_tmp = 0;
        int grant_slots = 0;
        num_of_slots = 0; 

        return_cid_tmp = 0;
        temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
        if(temp_index < 0) return_cid_tmp = -1;
        else return_cid_tmp = con->get_cid();
        mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();


        if(schedtype==SERVICE_UGS){
        } else {
//Fair allocation

	}//end else UGS

        if(schedtype == SERVICE_ertPS) {

	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
          if (con->queueByteLength() > 0) {
             allocationsize = (int) ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
          } else {
             allocationsize = 0;
          }

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          } 
	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));;

	  if (con->queueByteLength() <=0) num_of_slots = 0;
	  else num_of_slots = grant_slots;

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for ertPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;


	  debug10 ("DL.Check1.3.ertPS, DataSize :%f, period :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
//	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofertPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[1]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }
        
        
        if(schedtype == SERVICE_rtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
          int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();
          }
	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) allocationsize = (int) ceil((double)(sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  else  allocationsize = 0;

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for rtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[0]++;

	  debug10 ("DL.Check1.3.rtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getMinReservedRate(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
//	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[2]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));
        
        }


        if(schedtype == SERVICE_nrtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }

	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) allocationsize = (int) ceil((sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  else  allocationsize = 0;

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for nrtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[1]++;

	  debug10 ("DL.Check1.3.nrtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getMinReservedRate(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
//	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofnrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[3]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }

        if(schedtype==SERVICE_BE){
	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }

	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) needmore_con[2]++;

	  debug10 ("DL.Check1.3.BE, Peer-CID :%d, returnCID :%d, DIUC :%d\n", con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tQ-bytes :%d\n", con->queueByteLength());
	  con = con->next_entry();
	  continue;
        }

        if(num_of_slots> 0) {
	  temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);

	  if(temp_index < 0){
	    temp_index = ie_index;
	    cid_list[temp_index] = con->get_cid();
	    slots_per_con[temp_index] = num_of_slots;
	    diuc_list[temp_index] = getDIUCProfile(mod_rate);
	    ++ie_index;

	    debug10 ("DL.Check1.3: MapNotExist New_CID :%d, return index :-1, cid_list[%d] :%d,  #entry (ie_index)++ :%d\n", con->get_cid(), temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else if(temp_index<MAX_MAP_IE) {
	    slots_per_con[temp_index] += num_of_slots;

	    debug10 ("DL.Check1.3: MapExist Peer_CID :%d, return index :%d, cid_list[%d] :%d,  #entry (ie_index) :%d\n", con->get_cid(), temp_index, temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else {
	    freeslots += num_of_slots;                      //return back the slots
	    leftOTHER_slots += num_of_slots;                //return back the slots
	  }

	  debug10 ("DL.Check1.3, First Assign (ugs/ertps/rtps/nrtps/no be): CID(%d), Schetype :%d, Numslots :%d, Freeleft :%d, StoreSlots[%d] :%d, mod_rate :%d, DIUC :%d\n", cid_list[temp_index], schedtype, num_of_slots, freeslots, temp_index, slots_per_con[temp_index], mod_rate, diuc_list[temp_index]);
        }

      }
      con = con->next_entry();
    }//end con != NULL

  }//end for loop 5 Qos
*/



/*
// This comment is to allocate all connections fairly
  //Assign left-over to rtPS, ertPS, and BE fairly
  int share_next_slots = 0;
  for (int i=0;i<3;i++) 
    needmore_c = needmore_c+needmore_con[i];
  if (needmore_c>0) 
    share_next_slots = (int) floor(leftOTHER_slots/needmore_c);

  int first_assign = 0;

  while (needmore_c>0 && freeslots>0) {

    share_next_slots = (int) floor(freeslots/needmore_c);

    debug10 ("DL.Check1.4, (Check still need more here): Needmore Conn :%d, Free :%d, Sharenext :%d\n", needmore_c, freeslots, share_next_slots);

    for(i=0;i<5;++i){
      con = head;
      if(i==0)        schedtype = SERVICE_UGS;
      else if(i==1)   schedtype = SERVICE_ertPS;
      else if(i==2)   schedtype = SERVICE_rtPS;
      else if(i==3)   schedtype = SERVICE_nrtPS;
      else            schedtype = SERVICE_BE;
    
      if ( (i==0) || (i==1) ) continue;

      first_assign = 0;
      while(con!=NULL) {
	debug2 ("Rich: Con %p cid=%d\n", con, con->get_cid());
	if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype) {

	  mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();
	  temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
	  if(temp_index < 0) return_cid_tmp = -1;
	  else return_cid_tmp = temp_index;

	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }
	  int req_slots = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_)); 

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    if (return_cid_tmp == -1) {
	      debug10 ("DL.Check1.4.rtps/nrtps, No_CID(%d), n_Conn :%d, Free :%d\n", return_cid_tmp, needmore_c, freeslots);
	      con = con->next_entry();
	      continue;
	    }
	    if ( req_slots<=slots_per_con[return_cid_tmp] ) {
	      debug10 ("DL.Check1.4.rtPS/nrtPS, CID(%d), <=MinSlots: n_Conn :%d, REQ-SLOTS :%d, GRANT-1SLOTS :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, req_slots, slots_per_con[return_cid_tmp], freeslots);
	      con = con->next_entry();
	      continue;
	    }
	  }

	  int t_num_of_slots = 0;

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    first_assign = slots_per_con[return_cid_tmp];

	    if (req_slots <= (first_assign + share_next_slots) ) {           
	      t_num_of_slots = req_slots;
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);
	      debug10 ("DL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		if (freeslots>0) { 
		  t_num_of_slots = (first_assign + 1);
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
		t_num_of_slots =  (first_assign + share_next_slots);            
		freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	    }
	    debug10 ("DL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNeedMore: n_Conn :%d, Previous-Slots :%d, Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, first_assign, t_num_of_slots, freeslots);
	  } else { //BE
	    if(req_slots<1) {
	      con = con->next_entry();
	      continue;
	    }
	    temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
	    if(temp_index < 0){
	      temp_index = ie_index;
	      cid_list[temp_index] = con->get_cid();
	      slots_per_con[temp_index] = 0;
	      diuc_list[temp_index] = getDIUCProfile(mod_rate);
	      ++ie_index;
	      debug10 ("DL.Check1.4.BE, CID(%d), Initial_BE: n_Conn :%d, REQ-SLOTS :%d, Free :%d\n", cid_list[temp_index], needmore_c, req_slots, freeslots);

	    } else {
	    }// Richard.

	    first_assign = slots_per_con[temp_index];
	    if (req_slots <= (first_assign + share_next_slots) ) {           
	      t_num_of_slots = req_slots;
	      slots_per_con[temp_index] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);

	      debug10 ("DL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		if (freeslots>0) { 
		  t_num_of_slots = (first_assign + 1);
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
		t_num_of_slots =  (first_assign + share_next_slots);            
		freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[temp_index] = t_num_of_slots;

	      debug10 ("DL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    }
	    //}//richard
	  }    //else BE

	}//else CONNDATA
	con = con->next_entry();
      }//while
    }//for

  }//while
*/

  int total_slots_sofar = check_overallocation(temp_index_burst);
  subchannel_offset = (total_slots_sofar%total_subchannels);
  symbol_offset = DL_PREAMBLE + floor(total_slots_sofar/total_subchannels);//*2;
  debug10 ("DL2.end of dl_stage2, Sum allocated slots :%d, Suboffset :%d, Symoffset :%d\n", total_slots_sofar, subchannel_offset, symbol_offset);
  
/*
// In this version, we do the stripping outside the dl_stage2
  if(stripping == VERTICAL_STRIPPING){
    for(i=0; i<index_burst-temp_index_burst; ++i){
      ies[i].ie.cid = virtual_alloc[i+temp_index_burst].cid;
      num_of_slots = virtual_alloc[i+temp_index_burst].numslots;
      ies[i].ie.diuc = virtual_alloc[i+temp_index_burst].iuc;
      ies[i].ie.subchannel_offset = subchannel_offset;
      ies[i].ie.symbol_offset = symbol_offset;
      num_of_subchannels = num_of_slots;
      num_of_symbols = (int) ceil((double)(subchannel_offset + num_of_subchannels)/total_subchannels)*2;
      ies[i].ie.num_of_symbols = num_of_symbols;
      ies[i].ie.num_of_subchannels = num_of_subchannels;

      debug10 ("DL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid = %d, #Slots :%d, DIUC :%d\n", ie_index, i, ies[i].ie.cid, num_of_slots, ies[i].ie.diuc);
      debug10 ("\tIE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d, #IE_Subchannel :%d\n", subchannel_offset, subchannel_start, ies[i].ie.subchannel_offset, ies[i].ie.num_of_subchannels);
      debug10 ("\tIE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*2] :%d\n", symbol_offset, symbol_start, ies[i].ie.symbol_offset, subchannel_offset, num_of_subchannels, total_subchannels, ies[i].ie.num_of_symbols);

      subchannel_offset = (subchannel_offset + num_of_subchannels)%(total_subchannels);
      symbol_offset += num_of_symbols - 2;
    }
  }
*/

/*
// In this version, we do the stripping outside the dl_stage2
  if(stripping == VERTICAL_STRIPPING){
    for(i=next_ie_index;i<ie_index;++i){
      ies[i].ie.cid = cid_list[i];
      num_of_slots = slots_per_con[i];
      ies[i].ie.diuc = diuc_list[i];
      ies[i].ie.subchannel_offset = subchannel_offset + subchannel_start;
      ies[i].ie.symbol_offset = symbol_offset + symbol_start;
//      debug10 ("1.#Sub :%d, #slot :%d, symbol :%d\n", num_of_subchannels, num_of_slots, num_of_symbols);
      num_of_subchannels = num_of_slots;
      num_of_symbols = (int) ceil((double)(subchannel_offset + num_of_subchannels)/total_subchannels)*2;
      ies[i].ie.num_of_symbols = num_of_symbols;
//      debug10 ("2.Before #Sub :%d, #slot :%d, ies[%d].#sub :%d, .#sym :%d\n", num_of_subchannels, num_of_slots, i, ies[i].ie.num_of_subchannels, ies[i].ie.num_of_symbols);
      ies[i].ie.num_of_subchannels = num_of_subchannels;
//      debug10 ("3.After #Sub :%d, #slot :%d, ies[%d].#sub :%d, .#sym :%d\n", num_of_subchannels, num_of_slots, i, ies[i].ie.num_of_subchannels, ies[i].ie.num_of_symbols);

      debug10 ("DL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid = %d, #Slots :%d, DIUC :%d\n", ie_index, i, ies[i].ie.cid, num_of_slots, ies[i].ie.diuc);
      debug10 ("\tIE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d, #IE_Subchannel :%d\n", subchannel_offset, subchannel_start, ies[i].ie.subchannel_offset, ies[i].ie.num_of_subchannels);
      debug10 ("\tIE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*2] :%d\n", symbol_offset, symbol_start, ies[i].ie.symbol_offset, subchannel_offset, num_of_subchannels, total_subchannels, ies[i].ie.num_of_symbols);

      subchannel_offset = (subchannel_offset + num_of_subchannels)%(total_subchannels);
      symbol_offset += num_of_symbols - 2;
    }
  }
*/


  ie_index = index_burst - temp_index_burst;

  dl_map->nb_ies = ie_index;
  return dl_map;
}

/* DL Scheduler, Added by Ritun, Modifed by Chakchai */
/*
//This is the previous version of dl_stage2 (do not consider the variable part of DL_MAP
//mac802_16_dl_map_frame * Wimax2BSScheduler::dl_stage2(Connection *head, int total_subchannels, int total_symbols, int symbol_start, int  stripping){
mac802_16_dl_map_frame * Wimax2BSScheduler::dl_stage2(Connection *head, int input_subchannel_offset,  int total_subchannels, int total_symbols, int symbol_start, int  stripping){
  struct mac802_16_dl_map_frame *dl_map;
  struct mac802_16_dlmap_ie *ies;
  Connection *con;
  int i, ie_index, temp_index;
  int num_of_symbols, num_of_subchannels, num_of_slots;
  double allocationsize;
  int freeslots;
  int symbol_offset = 0;
  //int subchannel_offset = 0;
  int subchannel_offset = input_subchannel_offset;
  int subchannel_start = 0;
  int num_of_data_connections = 0;
  ConnectionType_t contype;
  SchedulingType_t schedtype;
  Mac802_16 *  mac_ = getMac ();
  Ofdm_mod_rate mod_rate;

  int slots_per_con[MAX_MAP_IE];
  int cid_list[MAX_MAP_IE];
  int diuc_list[MAX_MAP_IE];

  int leftOTHER_slots=0;
  int needmore_con[3]={0,0,0};
  int needmore_c=0;
  int ori_symbols = total_symbols;
  int req_slots_tmp1 = 0;
  int return_cid_tmp = 0;

  dl_map = (struct mac802_16_dl_map_frame *) malloc(sizeof(struct mac802_16_dl_map_frame));
  dl_map->type = MAC_DL_MAP;
  dl_map->bsid = mac_->addr(); // its called in the mac_ object

  ies = dl_map->ies;
  ie_index = 0; //0 is for the ul-map

  //2symbols * 2 clusters = 4: -> 1 slot, since we use #sub_channel (which is 14*2 subcarriers), total_symbols -> horizontal
  //Check it again => can't be devided by 2
  if ((total_symbols%2)==1) total_symbols=total_symbols-1;

//Chakchai next version
//freeslots = (total_subchannels-subchannel_offset)+ (total_subchannels*(total_symbols-2)/2);
  freeslots = total_subchannels * total_symbols/2;

  debug10 ("DL.Check1.0, ****FRAME NUMBER**** :%d, FreeSlots(TotalSub*TotalSym/2) :%d, TotalSub :%d, Ori_TotalSymbol :%d, TotalSymbol :%d, StartSymbol :%d\n", mac_->getFrameNumber(), freeslots, total_subchannels, ori_symbols, total_symbols, symbol_start );

//  debug10 ("ARQ block :%d\n", getMac()->arq_block_size_);

  con=head;
  while(con!=NULL){
    if(con->get_category()==CONN_DATA)
      ++num_of_data_connections;
    con = con->next_entry();
  }

  for(i=0;i<3;++i){
    con = head;
    if(i==0) 		contype = CONN_BASIC;
    else if(i==1) 	contype = CONN_PRIMARY;
    else 		contype = CONN_SECONDARY;

    
    while(con!=NULL)
      {
	if(con->get_category() == contype)
	  {
	    debug2("xingting==  connection id [%d]\n",con->get_cid());
	    allocationsize = con->queueByteLength();
	    mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();
	    num_of_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	    debug10 ("DL.Check1.1.contype(%d), CID :%d, Q-Bytes :%d, allocationsize :%f, #slots :%d\n", contype, con->get_cid(), con->queueByteLength(), allocationsize, num_of_slots);
	    debug10 ("\tfree slots :%d, left-over :%d\n", freeslots, freeslots-num_of_slots);

	    if(freeslots < num_of_slots) num_of_slots = freeslots;
	    freeslots -= num_of_slots;

	    if(ie_index<MAX_MAP_IE && num_of_slots > 0 && stripping == HORIZONTAL_STRIPPING)
	      {
		ies[ie_index].cid = con->get_cid();
		ies[ie_index].n_cid = 1;
		ies[ie_index].diuc = getDIUCProfile(mod_rate);
      		num_of_subchannels = num_of_slots;
      		ies[ie_index].subchannel_offset = subchannel_offset + subchannel_start;
      		ies[ie_index].symbol_offset = symbol_offset + symbol_start;
      		ies[ie_index].num_of_subchannels = num_of_subchannels;
      		num_of_symbols = (int) ceil( (double)(symbol_offset + num_of_slots) / (total_symbols) );
      		ies[ie_index].num_of_symbols = num_of_symbols;

      		subchannel_offset += num_of_symbols;
      		symbol_offset = (symbol_offset + num_of_slots) % (total_symbols);

		++ie_index;
	      }
	    else if(ie_index<MAX_MAP_IE && num_of_slots > 0 && stripping == VERTICAL_STRIPPING)
	      {
		ies[ie_index].cid = con->get_cid();
		ies[ie_index].n_cid = 1;
		ies[ie_index].diuc = getDIUCProfile(mod_rate);

		ies[ie_index].subchannel_offset = (subchannel_offset + subchannel_start) % total_subchannels;
		ies[ie_index].symbol_offset = symbol_offset + symbol_start;

		num_of_subchannels = num_of_slots;
		num_of_symbols = (int) ceil((double)(subchannel_offset + num_of_subchannels)/total_subchannels)*2;
		debug2 ("In DL scheduler, number of symbols = %d\n", num_of_symbols);
		ies[ie_index].num_of_symbols = num_of_symbols;
		ies[ie_index].num_of_subchannels = num_of_subchannels;

		debug10 ("DL.Check1.2: ie_index(MAX=60) :%d, ies[%d].cid :%d, #Slots :%d, UIUC :%d, IE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d\n", ie_index, i, ies[ie_index].cid, num_of_slots, ies[ie_index].diuc, subchannel_offset, subchannel_start, ies[ie_index].subchannel_offset);
		debug10 ("DL.Check1.2: IE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, .#Subchannel (=#slots) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*2] :%d\n", symbol_offset, symbol_start, ies[ie_index].symbol_offset, num_of_subchannels, subchannel_offset, num_of_subchannels, total_subchannels, num_of_symbols);

		subchannel_offset = (subchannel_offset + num_of_subchannels)%(total_subchannels);
		symbol_offset += num_of_symbols - 2;
		++ie_index;
	      }
	  }
	con = con->next_entry();
      }//end con!=null
  }//end for loop


  int next_ie_index = ie_index;
  //Assign slot for ugs/ertps/rtps/nrtps/be
  //Find out how many connection for each type of traffic

  int conn_per_schetype[6]={0,0,0,0,0,0};
  conn_per_schetype[0]=0; conn_per_schetype[1]=0; conn_per_schetype[2]=0; conn_per_schetype[3]=0; conn_per_schetype[4]=0; conn_per_schetype[5]=0;

  int con_ugs_all=0;
  int con_ertps_all=0;
  int con_rtps_all=0;  
  int con_nrtps_all=0;
  int con_be_all=0;

  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;    
    else if(i==3)       schedtype = SERVICE_nrtPS; 
    else                schedtype = SERVICE_BE;

    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){
	if(schedtype==SERVICE_UGS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[0]++;
	  con_ugs_all++;
	} else if (schedtype==SERVICE_ertPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[1]++;
	  con_ertps_all++;
	} else if (schedtype==SERVICE_rtPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[2]++;
	  con_rtps_all++;
	} else if (schedtype==SERVICE_nrtPS) {
	  if (con->queueByteLength() > 0) conn_per_schetype[3]++;
	  con_nrtps_all++;
	} else if (schedtype==SERVICE_BE) {
	  if (con->queueByteLength() > 0) conn_per_schetype[4]++;
	  con_be_all++;
	} else {
	  conn_per_schetype[5]++;
	}
      }
      con = con->next_entry();
    }
  }
  debug10 ("DL, Active UGS <%d>, ertPS <%d>, rtPS <%d>, nrtPS <%d>, BE <%d>, OTHER <%d>\n", conn_per_schetype[0],conn_per_schetype[1], conn_per_schetype[2], conn_per_schetype[3], conn_per_schetype[4], conn_per_schetype[5] );
  debug10 ("    All    UGS <%d>, ertPS <%d>, rtPS <%d>, nrtPS <%d>, BE <%d>\n", con_ugs_all, con_ertps_all, con_rtps_all, con_nrtps_all, con_be_all);

  leftOTHER_slots = freeslots;
  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;
    else if(i==3)       schedtype = SERVICE_nrtPS;
    else                schedtype = SERVICE_BE;

    
    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){

        req_slots_tmp1 = 0;
        return_cid_tmp = 0;
        int grant_slots = 0;
        num_of_slots = 0; 

        return_cid_tmp = 0;
        temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
        if(temp_index < 0) return_cid_tmp = -1;
        else return_cid_tmp = con->get_cid();
        mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();


        if(schedtype==SERVICE_UGS){
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  if (con->queueByteLength() > 0) {

#ifdef UGS_AVG
          allocationsize = (int) ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
#endif
#ifndef UGS_AVG
          int tmp_getpoll = con->getPOLL_interval();
          if ( (tmp_getpoll%sfqos->getPeriod())== 0 ) {
                allocationsize = ceil(sfqos->getDataSize());
                con->setPOLL_interval(0);
          } else {
                allocationsize = 0;
          }
          tmp_getpoll++;
          con->setPOLL_interval(tmp_getpoll);
#endif

	  } else {
	     allocationsize = 0;
	  }

	  int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
	  int t_bytes = 0;
	  double ori_grant = allocationsize;
	  if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
		t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
		if (t_bytes > 0) {
			allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
		}
		debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
		allocationsize = allocationsize + 14;	//optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub
		
	  } 

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));;

	  if (con->queueByteLength() <=0) num_of_slots = 0;
	  else num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for UGS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  debug10 ("DL.Check1.3.UGS, DataSize :%f, period :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofUGS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[0]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }

        if(schedtype == SERVICE_ertPS) {

	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
          if (con->queueByteLength() > 0) {
             allocationsize = (int) ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
          } else {
             allocationsize = 0;
          }

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          } 
	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));;

	  if (con->queueByteLength() <=0) num_of_slots = 0;
	  else num_of_slots = grant_slots;

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for ertPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;


	  debug10 ("DL.Check1.3.ertPS, DataSize :%f, period :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofertPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[1]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }
        
        
        if(schedtype == SERVICE_rtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
          int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();
          }
	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) allocationsize = (int) ceil((double)(sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  else  allocationsize = 0;

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for rtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[0]++;

	  debug10 ("DL.Check1.3.rtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getMinReservedRate(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[2]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));
        
        }


        if(schedtype == SERVICE_nrtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }

	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) allocationsize = (int) ceil((sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  else  allocationsize = 0;

          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, DL_));
	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for nrtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[1]++;

	  debug10 ("DL.Check1.3.nrtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getMinReservedRate(), grant_slots, con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofnrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[3]);
	  debug10 ("\tQ-bytes :%d, Grant-bytes :%ld\n", con->queueByteLength(), (long int)( grant_slots*mac_->getPhy()->getSlotCapacity(mod_rate, DL_) ));

        }

        if(schedtype==SERVICE_BE){
	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }

	  req_slots_tmp1 = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_));

	  if (req_slots_tmp1>0) needmore_con[2]++;

	  debug10 ("DL.Check1.3.BE, Peer-CID :%d, returnCID :%d, DIUC :%d\n", con->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tQ-bytes :%d\n", con->queueByteLength());
	  con = con->next_entry();
	  continue;
        }

        freeslots = leftOTHER_slots;

        if(num_of_slots> 0) {
	  temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);

	  if(temp_index < 0){
	    temp_index = ie_index;
	    cid_list[temp_index] = con->get_cid();
	    slots_per_con[temp_index] = num_of_slots;
	    diuc_list[temp_index] = getDIUCProfile(mod_rate);
	    ++ie_index;

	    debug10 ("DL.Check1.3: MapNotExist New_CID :%d, return index :-1, cid_list[%d] :%d,  #entry (ie_index)++ :%d\n", con->get_cid(), temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else if(temp_index<MAX_MAP_IE) {
	    slots_per_con[temp_index] += num_of_slots;

	    debug10 ("DL.Check1.3: MapExist Peer_CID :%d, return index :%d, cid_list[%d] :%d,  #entry (ie_index) :%d\n", con->get_cid(), temp_index, temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else {
	    freeslots += num_of_slots;                      //return back the slots
	    leftOTHER_slots += num_of_slots;                //return back the slots
	  }

	  debug10 ("DL.Check1.3, First Assign (ugs/ertps/rtps/nrtps/no be): CID(%d), Schetype :%d, Numslots :%d, Freeleft :%d, StoreSlots[%d] :%d, mod_rate :%d, DIUC :%d\n", cid_list[temp_index], schedtype, num_of_slots, freeslots, temp_index, slots_per_con[temp_index], mod_rate, diuc_list[temp_index]);
        }

      }
      con = con->next_entry();
    }//end con != NULL
  }//end for loop 5 Qos


  //Assign left-over to rtPS, ertPS, and BE fairly
  int share_next_slots = 0;
  for (int i=0;i<3;i++) 
    needmore_c = needmore_c+needmore_con[i];
  if (needmore_c>0) 
    share_next_slots = (int) floor(leftOTHER_slots/needmore_c);

  int first_assign = 0;

  while (needmore_c>0 && freeslots>0) {

    share_next_slots = (int) floor(freeslots/needmore_c);

    debug10 ("DL.Check1.4, (Check still need more here): Needmore Conn :%d, Free :%d, Sharenext :%d\n", needmore_c, freeslots, share_next_slots);

    for(i=0;i<5;++i){
      con = head;
      if(i==0)        schedtype = SERVICE_UGS;
      else if(i==1)   schedtype = SERVICE_ertPS;
      else if(i==2)   schedtype = SERVICE_rtPS;
      else if(i==3)   schedtype = SERVICE_nrtPS;
      else            schedtype = SERVICE_BE;
    
      if ( (i==0) || (i==1) ) continue;

      first_assign = 0;
      while(con!=NULL) {
	debug2 ("Rich: Con %p cid=%d\n", con, con->get_cid());
	if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype) {

	  mod_rate = mac_->getMap()->getDlSubframe()->getProfile(con->getPeerNode()->getDIUC())->getEncoding();
	  temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
	  if(temp_index < 0) return_cid_tmp = -1;
	  else return_cid_tmp = temp_index;

	  int withfrag = 0;
          if  (con->getFragmentBytes()>0) {
               withfrag = con->queueByteLength() - con->getFragmentBytes() + 2;
          } else {
               withfrag = con->queueByteLength();;
          }
	  int req_slots = (int) ceil((double)withfrag/(double)mac_->getPhy()->getSlotCapacity(mod_rate, DL_)); 

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    if (return_cid_tmp == -1) {
	      debug10 ("DL.Check1.4.rtps/nrtps, No_CID(%d), n_Conn :%d, Free :%d\n", return_cid_tmp, needmore_c, freeslots);
	      con = con->next_entry();
	      continue;
	    }
	    if ( req_slots<=slots_per_con[return_cid_tmp] ) {
	      debug10 ("DL.Check1.4.rtPS/nrtPS, CID(%d), <=MinSlots: n_Conn :%d, REQ-SLOTS :%d, GRANT-1SLOTS :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, req_slots, slots_per_con[return_cid_tmp], freeslots);
	      con = con->next_entry();
	      continue;
	    }
	  }

	  int t_num_of_slots = 0;

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    first_assign = slots_per_con[return_cid_tmp];

	    if (req_slots <= (first_assign + share_next_slots) ) {           
	      t_num_of_slots = req_slots;
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);
	      debug10 ("DL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		if (freeslots>0) { 
		  t_num_of_slots = (first_assign + 1);
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
		t_num_of_slots =  (first_assign + share_next_slots);            
		freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	    }
	    debug10 ("DL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNeedMore: n_Conn :%d, Previous-Slots :%d, Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, first_assign, t_num_of_slots, freeslots);
	  } else { //BE
	    if(req_slots<1) {
	      con = con->next_entry();
	      continue;
	    }
	    temp_index = doesMapExist(con->get_cid(), cid_list, ie_index);
	    if(temp_index < 0){
	      temp_index = ie_index;
	      cid_list[temp_index] = con->get_cid();
	      slots_per_con[temp_index] = 0;
	      diuc_list[temp_index] = getDIUCProfile(mod_rate);
	      ++ie_index;
	      debug10 ("DL.Check1.4.BE, CID(%d), Initial_BE: n_Conn :%d, REQ-SLOTS :%d, Free :%d\n", cid_list[temp_index], needmore_c, req_slots, freeslots);

	    } else {
	    }// Richard.

	    first_assign = slots_per_con[temp_index];
	    if (req_slots <= (first_assign + share_next_slots) ) {           
	      t_num_of_slots = req_slots;
	      slots_per_con[temp_index] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);

	      debug10 ("DL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		if (freeslots>0) { 
		  t_num_of_slots = (first_assign + 1);
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
		t_num_of_slots =  (first_assign + share_next_slots);            
		freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[temp_index] = t_num_of_slots;

	      debug10 ("DL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    }
	    //}//richard
	  }    //else BE

	}//else CONNDATA
	con = con->next_entry();
      }//while
    }//for

  }//while

  if(stripping == HORIZONTAL_STRIPPING){
    for(i=next_ie_index;i<ie_index;++i){
      ies[i].ie.cid = cid_list[i];
      num_of_slots = slots_per_con[i];
      ies[i].ie.diuc = diuc_list[i];

      num_of_subchannels = num_of_slots;
      ies[i].ie.subchannel_offset = subchannel_offset + subchannel_start;
      ies[i].ie.symbol_offset = symbol_offset + symbol_start;
      ies[i].ie.num_of_subchannels = num_of_subchannels;
      num_of_symbols = (int) ceil( (double)(symbol_offset + num_of_slots) / (total_symbols) );
      ies[i].ie.num_of_symbols = num_of_symbols;

      debug10 ("DL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid :%d, #Slots :%d, DIUC :%d\n", ie_index, i, ies[i].ie.cid, num_of_slots, ies[i].ie.diuc);
      debug10 ("\tIE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d, #IE_Subchannel :%d\n", subchannel_offset, subchannel_start, ies[i].ie.subchannel_offset, ies[i].ie.num_of_subchannels);
      debug10 ("\tIE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*3] :%d\n", symbol_offset, symbol_start, ies[i].ie.symbol_offset, subchannel_offset, num_of_subchannels, total_subchannels, ies[i].ie.num_of_symbols);
      subchannel_offset += num_of_symbols;
      symbol_offset = (symbol_offset + num_of_slots) % (total_symbols);

    }
  }

  if(stripping == VERTICAL_STRIPPING){
    for(i=next_ie_index;i<ie_index;++i){
      ies[i].ie.cid = cid_list[i];
      num_of_slots = slots_per_con[i];
      ies[i].ie.diuc = diuc_list[i];
      ies[i].ie.subchannel_offset = subchannel_offset + subchannel_start;
      ies[i].ie.symbol_offset = symbol_offset + symbol_start;
//      debug10 ("1.#Sub :%d, #slot :%d, symbol :%d\n", num_of_subchannels, num_of_slots, num_of_symbols);
      num_of_subchannels = num_of_slots;
      num_of_symbols = (int) ceil((double)(subchannel_offset + num_of_subchannels)/total_subchannels)*2;
      ies[i].ie.num_of_symbols = num_of_symbols;
//      debug10 ("2.Before #Sub :%d, #slot :%d, ies[%d].#sub :%d, .#sym :%d\n", num_of_subchannels, num_of_slots, i, ies[i].ie.num_of_subchannels, ies[i].ie.num_of_symbols);
      ies[i].ie.num_of_subchannels = num_of_subchannels;
//      debug10 ("3.After #Sub :%d, #slot :%d, ies[%d].#sub :%d, .#sym :%d\n", num_of_subchannels, num_of_slots, i, ies[i].ie.num_of_subchannels, ies[i].ie.num_of_symbols);

      debug10 ("DL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid = %d, #Slots :%d, DIUC :%d\n", ie_index, i, ies[i].ie.cid, num_of_slots, ies[i].ie.diuc);
      debug10 ("\tIE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d, #IE_Subchannel :%d\n", subchannel_offset, subchannel_start, ies[i].ie.subchannel_offset, ies[i].ie.num_of_subchannels);
      debug10 ("\tIE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*2] :%d\n", symbol_offset, symbol_start, ies[i].ie.symbol_offset, subchannel_offset, num_of_subchannels, total_subchannels, ies[i].ie.num_of_symbols);

      subchannel_offset = (subchannel_offset + num_of_subchannels)%(total_subchannels);
      symbol_offset += num_of_symbols - 2;
    }
  }



  dl_map->nb_ies = ie_index;
  return dl_map;
}
*/

/* returns the DIUC profile associated with a current modulation and coding scheme
 * Added by Ritun
 */
diuc_t Wimax2BSScheduler::getDIUCProfile(Ofdm_mod_rate rate){
  Profile *p;

  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_1);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_1;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_2);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_2;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_3);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_3;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_4);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_4;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_5);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_5;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_6);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_6;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_7);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_7;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_8);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_8;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_9);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_9;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_10);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_10;
  p = getMac()->getMap()->getDlSubframe()->getProfile(DIUC_PROFILE_11);
  if(p->getEncoding() == rate)
    return DIUC_PROFILE_11;
  
  return DIUC_PROFILE_1;
}

/* Added by Ritun
 */
uiuc_t Wimax2BSScheduler::getUIUCProfile(Ofdm_mod_rate rate){
  Profile *p;

  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_1);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_1;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_2);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_2;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_3);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_3;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_4);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_4;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_5);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_5;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_6);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_6;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_7);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_7;
  p = getMac()->getMap()->getUlSubframe()->getProfile(UIUC_PROFILE_8);
  if(p->getEncoding() == rate)
    return UIUC_PROFILE_8;
  
  return UIUC_PROFILE_1;
}

/* 
 *	Allocates space in the uplink subframe
 *	Before this is called above, we verify that everything we're trying to allocate can fit in the
 *	subframe, so we do not need to check that here
 * 	note : total_slots should be total_symbols
 */

struct mac802_16_ul_map_frame * Wimax2BSScheduler::ul_stage2(Connection *head, int total_subchannels, int total_slots, int slot_start, int stripping) {

  struct mac802_16_ul_map_frame *ul_map;
  struct UL_MAP_IE *ies;
  Connection *con;
//  uint16_t max_allocation_size = 2047; //based on GMH's length field of 11 Bytes
  int i, ie_index, temp_index;
  int num_of_slots; //, num_of_symbols, num_of_subchannels;
  double allocationsize;
  int freeslots;
//  int symbol_offset = 0;
//  int subchannel_offset = 0;
//  int subchannel_start = 0;
  ConnectionType_t contype;
  SchedulingType_t schedtype;
  int slots_per_con[MAX_MAP_IE];
  int cid_list[MAX_MAP_IE];
  int uiuc_list[MAX_MAP_IE];
  int cdma_flag_list[MAX_MAP_IE];
  uint8_t cdma_code_list[MAX_MAP_IE];
  uint8_t cdma_top_list[MAX_MAP_IE];

  Ofdm_mod_rate mod_rate;

  int leftOTHER_slots=0;
  int needmore_con[3]={0,0,0};
  int needmore_c=0;
  int ori_symbols = total_slots;
  	if (ori_symbols); //quiet, compiler
  int req_slots_tmp1 = 0;
  int return_cid_tmp = 0;
  num_of_slots=0;
  int tmp_cdma_code_top[CODE_SIZE][CODE_SIZE];

  for(int i=0;i<CODE_SIZE;i++){
    for(int j=0;j<CODE_SIZE;j++){
      tmp_cdma_code_top[i][j] = 0;
    }
  }

  //total slots - 1 because first is reserved for contention
  freeslots = total_subchannels * (total_slots - getMac()->getPhy()->getSlotLength(UL_)) / getMac()->getPhy()->getSlotLength(UL_);
  leftOTHER_slots = freeslots;

  debug10 ("UL.Check1.0, ****FRAME NUMBER**** :%d, FreeSlots(TotalSub*TotalSym) :%d, TotalSub :%d, Original Num Slots(maxul-ulduration) :%d, TotalSlots :%d, StartSlot :%d\n", mac_->getFrameNumber(), freeslots, total_subchannels, ori_symbols, total_slots, slot_start );

  ul_map = (struct mac802_16_ul_map_frame *) malloc(sizeof(struct mac802_16_ul_map_frame));
  ul_map->type = MAC_UL_MAP;

  ies = ul_map->ies;
  ie_index = 0;

  bzero(slots_per_con, MAX_MAP_IE*sizeof(int));
  bzero(cid_list, MAX_MAP_IE*sizeof(int));
  for (int i=0; i<MAX_MAP_IE; i++) {
      cdma_code_list[i] = 0;
      cdma_top_list[i] = 0;
      cdma_flag_list[i] = 0;
  }

//Check if cdma_init_ranging request
  con = head;
  while (con!=NULL) {
    if(con->get_category() == CONN_INIT_RANGING) {
        if (con->getCDMA() == 2) {
           int begin_code = 0;
           int begin_top = 0;
           int begin_flag = 0;
           for (int i = 0; i < MAX_SSID; i++) {
              begin_flag = con->getCDMA_SSID_FLAG(i);
              begin_code = con->getCDMA_SSID_CODE(i);
              begin_top = con->getCDMA_SSID_TOP(i);
              if (begin_flag == 0) continue;
              for (int j = i+1; j < MAX_SSID; j++) {
                if (con->getCDMA_SSID_FLAG(j) == 0) continue;
                if ( (begin_code == con->getCDMA_SSID_CODE(j)) && (begin_top == con->getCDMA_SSID_TOP(j)) ) {
                   debug10 ("=Collission CDMA_INIT_RNG_REQ (ssid i :%d and ssid j :%d), CDMA_flag i :%d and CDMA_flag j:%d, CDMA_code i :%d, CDMA_top i :%d\n", con->getCDMA_SSID_SSID(i), con->getCDMA_SSID_SSID(j), con->getCDMA_SSID_FLAG(i), con->getCDMA_SSID_FLAG(j), con->getCDMA_SSID_CODE(i), con->getCDMA_SSID_TOP(i));
                   con->setCDMA_SSID_FLAG(j, 0);
                   con->setCDMA_SSID_FLAG(i, 0);
                }
              }
           }
        }

        con->setCDMA(0);
        break;
    }
    con = con->next_entry();
  }

  con = head;
  while (con!=NULL) {
    if(con->get_category() == CONN_INIT_RANGING){
      for (int i = 0; i<MAX_SSID; i++) {
         int cdma_flag = con->getCDMA_SSID_FLAG(i);
         if (cdma_flag > 0) {
            

/*
//0 to 6
debug10 (" bpsk1/2 :%d, qpsk1/2 :%d, qpsk3/4 :%d, 16qam1/2 :%d, 16qam3/4 :%d, 64qam2/3 :%d, 64qam3/4 :%d\n", OFDM_BPSK_1_2,   OFDM_QPSK_1_2,   OFDM_QPSK_3_4,  OFDM_16QAM_1_2, OFDM_16QAM_3_4,  OFDM_64QAM_2_3, OFDM_64QAM_3_4);
  
//5 to 11
debug10 (" uiuc_p1 :%d, uiuc_p2 :%d, uiuc_p3 :%d, uiuc_p4 :%d, uiuc_p5 :%d, uiuc_p6 :%d, uiuc_p7 :%d\n", UIUC_PROFILE_1, UIUC_PROFILE_2, UIUC_PROFILE_3, UIUC_PROFILE_4, UIUC_PROFILE_5, UIUC_PROFILE_6, UIUC_PROFILE_7);

//1 to 7
debug10 (" diuc_p1 :%d, diuc_p2 :%d, diuc_p3 :%d, diuc_p4 :%d, diuc_p5 :%d, diuc_p6 :%d, diuc_p7 :%d\n", DIUC_PROFILE_1, DIUC_PROFILE_2, DIUC_PROFILE_3, DIUC_PROFILE_4, DIUC_PROFILE_5, DIUC_PROFILE_6, DIUC_PROFILE_7);

for (int i = 0; i <7; i++ ) {
  debug10 ("mod_rate :%d, UIUC_Profile :%d\n", i, (int)getUIUCProfile((Ofdm_mod_rate)i) );
}
*/

            mod_rate = (Ofdm_mod_rate)1;
	   //mod_rate = mac_->getMap()->getUlSubframe()->getProfile(con->getPeerNode()->getDIUC()-DIUC_PROFILE_1+UIUC_PROFILE_1)->getEncoding();
//            allocationsize =  RNG_REQ_SIZE+GENERIC_HEADER_SIZE;
            allocationsize =  RNG_REQ_SIZE; //should really be changed to a GetSerialziedSize call for futureproofing, but this doesn't change for the moment
            num_of_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
            con->setCDMA_SSID_FLAG(i,0);
            debug10("=> Allocate init_ranging_msg opportunity for ssid :%d, code :%d, top :%d, size :%f\n", con->getCDMA_SSID_SSID(i), con->getCDMA_SSID_CODE(i), con->getCDMA_SSID_TOP(i), allocationsize);


            if(freeslots < num_of_slots) num_of_slots = freeslots;

            freeslots -= num_of_slots;

            if (num_of_slots > 0) {
               int temp_index = ie_index++;
               cid_list[temp_index] = con->get_cid();
               slots_per_con[temp_index] = num_of_slots;
               uiuc_list[temp_index] = getUIUCProfile(mod_rate);
               cdma_code_list[temp_index] = con->getCDMA_SSID_CODE(i);
               cdma_top_list[temp_index] = con->getCDMA_SSID_TOP(i);
               cdma_flag_list[temp_index] = 1;
            }
	    con->setCDMA_SSID_FLAG(i, 0);
            con->setCDMA_SSID_CODE(i, 0);
            con->setCDMA_SSID_TOP(i, 0);
         }
      }
    }
    con = con->next_entry();
  }


  
//Check if cdma_bandwidth_ranging request
  for(i=0;i<4;++i) {
    con = head;
    if(i==0) 	  contype = CONN_BASIC;
    else if(i==1) contype = CONN_PRIMARY;
    else if(i==2) contype = CONN_SECONDARY;
    else contype = CONN_DATA;

    while(con!=NULL){
      if(con->get_category() == contype){
		if (con->getCDMA() == 1) {
		  tmp_cdma_code_top[(int)con->getCDMA_code()][(int)con->getCDMA_top()]++;
		}
		
		if (con->getCDMA()>0) {
		  debug10 ("=Contype :%d, CDMA_flag :%d, CDMA_code :%d, CDMA_top :%d, CDMA_code_top++ :%d\n", contype, con->getCDMA(), con->getCDMA_code(), con->getCDMA_top(), tmp_cdma_code_top[(int)con->getCDMA_code()][(int)con->getCDMA_top()]);
		}
      }
      con = con->next_entry();
    }

  }//end for

  for(i=0;i<4;++i) {
    con = head;
    if(i==0) 	  contype = CONN_BASIC;
    else if(i==1) contype = CONN_PRIMARY;
    else if(i==2) contype = CONN_SECONDARY;
    else contype = CONN_DATA;

    while(con!=NULL){
      if(con->get_category() == contype){
//      	debug10 ("=Contype :%d, CDMA_flag :%d, CDMA_code :%d, CDMA_top :%d\n", contype, con->getCDMA(), con->getCDMA_code(), con->getCDMA_top());
	if (con->getCDMA() == 1) {
	  if ( tmp_cdma_code_top[(int)con->getCDMA_code()][(int)con->getCDMA_top()] >1 ) {
	    con->setCDMA(0);
	    debug10 ("=Collission CDMA_BW_REQ, Contype :%d, CDMA_flag :%d, CDMA_code :%d, CDMA_top :%d, CDMA_code_top :%d\n", contype, con->getCDMA(), con->getCDMA_code(), con->getCDMA_top(), tmp_cdma_code_top[(int)con->getCDMA_code()][(int)con->getCDMA_top()]);
	  }
	}
      }
      con = con->next_entry();
    }

  }//end for 

//if we have an outgoing CDMA packet, this loop will determine it's size
  for(i=0;i<4;++i) {
    con = head;
    if(i==0) 	  contype = CONN_BASIC;
    else if(i==1) contype = CONN_PRIMARY;
    else if(i==2) contype = CONN_SECONDARY;
    else contype = CONN_DATA;
    
    while(con!=NULL){
      if(con->get_category() == contype){

        mod_rate = (Ofdm_mod_rate)1;
        num_of_slots = 0;
        allocationsize = 0;
        
	if (con->getCDMA() == 1) {
	  allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
	  int tmp_num_poll = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  	if (tmp_num_poll); //quiet compiler
	  num_of_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  debug10 ("\tUL.Check1.1.contype(%d), Polling CDMA :%f, numslots :%d, freeslot :%d, CID :%d\n", contype, allocationsize, num_of_slots, freeslots, con->get_cid());
	  con->setCDMA(0);
	}//end getCDMA

	if (freeslots < num_of_slots) num_of_slots = freeslots;

        freeslots -= num_of_slots;

	if(num_of_slots> 0){
//	  temp_index = doesMapExist(con->getPeerNode()->getBasic(IN_CONNECTION)->get_cid(), cid_list, ie_index); //conn, ie list, number of ies
	  temp_index = doesMapExist(con->get_cid(), cid_list, ie_index); //conn, ie list, number of ies
	  if(temp_index < 0){
	    temp_index = ie_index;
//	    cid_list[temp_index] = con->getPeerNode()->getBasic(IN_CONNECTION)->get_cid();
            cid_list[temp_index] = con->get_cid();
	    slots_per_con[temp_index] = num_of_slots;
	    uiuc_list[temp_index] = getUIUCProfile(mod_rate);
            cdma_code_list[temp_index] = con->getCDMA_code();
            cdma_top_list[temp_index] = con->getCDMA_top();
            cdma_flag_list[temp_index] = 1;
	    ++ie_index;
	  }
	  else if(temp_index < MAX_MAP_IE)	 slots_per_con[temp_index] += num_of_slots;
	  else	 				 freeslots += num_of_slots;

	  debug10 ("UL.Check1.2, Polling CDMA (basic/pri/sec): CID(%d), Contype(B:5, P:6, S:7, D:8) :%d, Numslots :%d, Free :%d, allo :%e, StoreSlots[%d] :%d\n", cid_list[temp_index], contype, num_of_slots, freeslots, allocationsize, temp_index, slots_per_con[temp_index]);

	}
      }
      con = con->next_entry();
    }
  }


  //Assign to basic, primary, secondary, and data now
  for(i=0;i<4;++i) {
    con = head;
    if(i==0) 	  contype = CONN_BASIC;
    else if(i==1) contype = CONN_PRIMARY;
    else if(i==2) contype = CONN_SECONDARY;
    else 	  contype = CONN_DATA;
    
    while(con!=NULL){
      if(con->get_category() == contype){

	mod_rate = mac_->getMap()->getUlSubframe()->getProfile(con->getPeerNode()->getDIUC()-DIUC_PROFILE_1+UIUC_PROFILE_1)->getEncoding();
	if (mod_rate > OFDM_64QAM_3_4) {
	  mod_rate = OFDM_64QAM_3_4;
	}

//Chakchai Modified 22 May 2008
        num_of_slots = 0;
        allocationsize = 0;
//
        
	if (con->getBw() > 0 ) {
	  allocationsize = con->getBw();
//	  if (allocationsize > max_allocation_size) allocationsize = max_allocation_size;
	  num_of_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));

	  debug10 ("\tBwreq Still>0 :%f, numslots :%d, freeslot :%d\n", allocationsize, num_of_slots, freeslots);
	} else {
	  if (con->getCDMA() == 1) {
	    allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
	    int tmp_num_poll = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	    int still_setbw = (int) ceil(con->getBw()/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));

	    if (tmp_num_poll>still_setbw) {
	      num_of_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	      debug10 ("\tPolling CDMA :%f, numslots :%d, freeslot :%d\n", allocationsize, num_of_slots, freeslots);
	    } else {
	      debug10 ("\tStill set BW Not Polling\n");
	    }
	  }//end getCDMA
	  con->setCDMA(0);
	}

	if (num_of_slots>0) {
	  debug10 ("UL.Check1.1.contype(%d), allocationsize (con->getBW) :%f, #slots :%d\n", contype, allocationsize, num_of_slots);
	  debug10 ("\tfree slots :%d, left-over :%d\n", freeslots, freeslots-num_of_slots);
	}

	if ( (allocationsize>0) || (num_of_slots>0) ) debug2 ("=> Allocation size %f, num of slots %d\n", allocationsize, num_of_slots);

	if (freeslots < num_of_slots) num_of_slots = freeslots;

        freeslots -= num_of_slots;

#ifdef BWREQ_PATCH
	//decrement bandwidth requested by allocation
	int new_bw = con->getBw() > (num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)) ? con->getBw() - (num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)) : 0;
	debug2 ("\tCtrl CID= %d Req= %d Alloc= %d Left=%d\n", 
		con->get_cid(), con->getBw(), 
		num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_),
		new_bw);
	con->setBw(new_bw);
#endif

	if(num_of_slots> 0){
	  temp_index = doesMapExist(con->getPeerNode()->getBasic(IN_CONNECTION)->get_cid(), cid_list, ie_index); //conn, ie list, number of ies
//          temp_index = doesMapExist(con->get_cid(), cid_list, ie_index); //conn, ie list, number of ies

	  if(temp_index < 0){
	    temp_index = ie_index;
//	    cid_list[temp_index] = con->getPeerNode()->getBasic(IN_CONNECTION)->get_cid();
            cid_list[temp_index] = con->get_cid();

	    slots_per_con[temp_index] = num_of_slots;
	    uiuc_list[temp_index] = getUIUCProfile(mod_rate);
	    ++ie_index;
	  }
	  else if(temp_index < MAX_MAP_IE)	 slots_per_con[temp_index] += num_of_slots;
	  else	 				 freeslots += num_of_slots;

	  debug10 ("UL.Check1.2, Init Assign (basic/pri/sec): CID(%d), Contype(Basic:5, Primary:6, Secondary:7, Data :8) :%d, Numslots :%d, Free :%d, allo :%e, StoreSlots[%d] :%d\n", cid_list[temp_index], contype, num_of_slots, freeslots, allocationsize, temp_index, slots_per_con[temp_index]);

	}
      }
      con = con->next_entry();
    }
  }


  leftOTHER_slots = freeslots;
  int conn_per_schetype[6]={0,0,0,0,0,0};
  conn_per_schetype[0]=0;
  conn_per_schetype[1]=0;
  conn_per_schetype[2]=0;
  conn_per_schetype[3]=0;
  conn_per_schetype[4]=0;
  conn_per_schetype[5]=0;

  int con_ertps_all=0;
  int con_rtps_all=0;
  int con_nrtps_all=0;
  int con_be_all=0;

  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;
    else if(i==3)       schedtype = SERVICE_nrtPS;
    else                schedtype = SERVICE_BE;

    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){
	mod_rate = mac_->getMap()->getUlSubframe()->getProfile(con->getPeerNode()->getDIUC()-DIUC_PROFILE_1+UIUC_PROFILE_1)->getEncoding();

	if(schedtype==SERVICE_UGS) {
	  conn_per_schetype[0]++;
	}
	else if (schedtype==SERVICE_ertPS) {
	  if (con->getBw() > 0) conn_per_schetype[1]++;
	  con_ertps_all++;

	} else if (schedtype==SERVICE_rtPS) {
	  if (con->getBw() > 0) conn_per_schetype[2]++;
	  con_rtps_all++;
	} else if (schedtype==SERVICE_nrtPS) {
	  if (con->getBw() > 0) conn_per_schetype[3]++;
	  con_nrtps_all++;
	} else if (schedtype==SERVICE_BE) {
	  if (con->getBw() > 0) conn_per_schetype[4]++;
	  con_be_all++;
	} else {
	  conn_per_schetype[5]++;
	}
      }
      con = con->next_entry();
    }
  }

  debug10 ("UL, Active -UGS <%d>, -ertPS <%d>, -rtPS <%d>, -nrtPS <%d>, -BE <%d>, -OTHER <%d>\n", conn_per_schetype[0],conn_per_schetype[1], conn_per_schetype[2], conn_per_schetype[3], conn_per_schetype[4], conn_per_schetype[5] );
  debug10 ("    All     UGS <%d>,  ertPS <%d>,  rtPS <%d>,  nrtPS <%d>,  BE <%d>\n", conn_per_schetype[0], con_ertps_all, con_rtps_all, con_nrtps_all, con_be_all);



  for(i=0;i<5;++i){
    con = head;
    if(i==0)            schedtype = SERVICE_UGS;
    else if(i==1)       schedtype = SERVICE_ertPS;
    else if(i==2)       schedtype = SERVICE_rtPS;
    else if(i==3)       schedtype = SERVICE_nrtPS;
    else                schedtype = SERVICE_BE;

    
    while(con!=NULL){
      if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype){
	//Richard: we must look up the rate again
	mod_rate = mac_->getMap()->getUlSubframe()->getProfile(con->getPeerNode()->getDIUC()-DIUC_PROFILE_1+UIUC_PROFILE_1)->getEncoding();
	if (mod_rate > OFDM_16QAM_3_4) {
	  mod_rate = OFDM_16QAM_3_4;
	}

        req_slots_tmp1 = 0;
        return_cid_tmp = 0;
        int grant_slots = 0;
        num_of_slots = 0; 

        return_cid_tmp = 0;
        temp_index = doesMapExist(con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list, ie_index);
        if(temp_index < 0) return_cid_tmp = -1;
        else return_cid_tmp = con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid();

        if(schedtype==SERVICE_UGS){
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	

#ifdef UGS_AVG
          allocationsize = (int) ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
#endif
#ifndef UGS_AVG
          int tmp_getpoll = con->getPOLL_interval();
          if ( (tmp_getpoll%sfqos->getPeriod())== 0 ) {
                allocationsize = ceil(sfqos->getDataSize());
                con->setPOLL_interval(0);
          } else {
                allocationsize = 0;
 	  }
          tmp_getpoll++;
          con->setPOLL_interval(tmp_getpoll);
#endif
	 
          int arq_enable_f = 0;
	  if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) ) {
	     arq_enable_f = 1;
	  }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          	if (ori_grant); //quiet, compiler	
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));

	  num_of_slots = grant_slots;

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for UGS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;


	  debug10 ("UL.Check1.3.UGS, DataSize :%f, period :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), grant_slots, con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofUGS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[0]);

	}

    if(schedtype == SERVICE_ertPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  req_slots_tmp1 = (int) ceil((double)con->getBw()/(double)mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  int issue_pol = 0;

/*
	  if (req_slots_tmp1>0) allocationsize = (int) ceil(sfqos->getDataSize()/sfqos->getPeriod());
	  else  allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
*/

          int tmp_getpoll = con->getPOLL_interval();
          if (req_slots_tmp1>0) {
                        allocationsize = ceil((double)sfqos->getDataSize()/(double)sfqos->getPeriod());
                        debug10 ("\tPoll_ertPS: No polling (bw-req>0), lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());
                        con->setPOLL_interval(0);
          } else  {
                if ( (tmp_getpoll%sfqos->getPeriod())== 0 ) {
                        debug10 ("\tPoll_ertPS(yes): Issues unicast poll, lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());

                        allocationsize = GENERIC_HEADER_SIZE;                                        //for explicit polling
                        con->setPOLL_interval(0);
			issue_pol = 1;
                } else {
                        debug10 ("\tPoll_ertPS(no): Don't issues unicast poll, lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());
                        allocationsize = 0;
                }
                debug10 ("\tPoll_ertPS: Current polling_counter :%d, update_polling :%d, period :%d\n", tmp_getpoll, tmp_getpoll+1, sfqos->getPeriod());
                tmp_getpoll++;
                con->setPOLL_interval(tmp_getpoll);
          }

          int arq_enable_f = 0;
          if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) && (issue_pol==0) ) {
             arq_enable_f = 1;
          }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          	if (ori_grant); //quiet, compiler
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  num_of_slots = grant_slots;

#ifdef BWREQ_PATCH
	  //decrement bandwidth requested by allocation	  
	  debug2 ("\tertPS: CID= %d Req= %d Alloc= %d Left=%d\n", 
		  con->get_cid(), con->getBw(), 
		  num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_),
		  con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
	  con->setBw(con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
#endif

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for ertPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;


	  debug10 ("UL.Check1.3.ertPS, DataSize :%f, period :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d\n", sfqos->getDataSize(), sfqos->getPeriod(), grant_slots, con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofertPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[1]);

	}
	
    if(schedtype == SERVICE_rtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  req_slots_tmp1 = (int) ceil((double)con->getBw()/(double)mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  int issue_pol = 0;

/*
	  if (req_slots_tmp1>0) allocationsize = (int) ceil((sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  else  allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
*/

          int tmp_getpoll = con->getPOLL_interval();
          if (req_slots_tmp1>0) {
                        allocationsize = ceil((sfqos->getMinReservedRate()*FRAME_SIZE)/8);
                        debug10 ("\tPoll_rtPS: No polling (bw-req>0), lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());
                        con->setPOLL_interval(0);
          } else  {
                if ( (tmp_getpoll % sfqos->getPeriod())== 0 ) {
                        debug10 ("\tPoll_rtPS(yes): Issues unicast poll, lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());

                        allocationsize = GENERIC_HEADER_SIZE;                                        //for explicit polling
                        con->setPOLL_interval(0);
			issue_pol = 1;
                } else {
                        debug10 ("\tPoll_rtPS(no): Don't issue unicast poll, lastpoll :%d, period :%d\n", tmp_getpoll, sfqos->getPeriod());
                        allocationsize = 0;
                }
                debug10 ("\tPoll_rtPS: Current polling_counter :%d, update_polling :%d, period :%d\n", tmp_getpoll, tmp_getpoll+1, sfqos->getPeriod());
                tmp_getpoll++;
                con->setPOLL_interval(tmp_getpoll);
          }

          int arq_enable_f = 0;
          if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) && (issue_pol==0) ) {
             arq_enable_f = 1;
          }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          	if (ori_grant); //quiet, compiler
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub
          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  num_of_slots = grant_slots;

#ifdef BWREQ_PATCH
	  //decrement bandwidth requested by allocation
	  debug2 ("\trtPS: CID= %d Req= %d Alloc= %d Left=%d\n", 
		  con->get_cid(), con->getBw(), 
		  num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_),
		  con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
	  con->setBw(con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
#endif
	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for rtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[0]++;

	  debug10 ("UL.Check1.3.rtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d, bw_req_header :%d, getBw :%d\n", sfqos->getMinReservedRate(), grant_slots, con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC(), GENERIC_HEADER_SIZE, con->getBw());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[2]);
	
	}


   if(schedtype == SERVICE_nrtPS) {
	  Wimax2ServiceFlow *sf = con->get_serviceflow();
	  Wimax2ServiceFlowQoS *sfqos = sf->getQoS();
	  int issue_pol = 0;
	  req_slots_tmp1 = (int) ceil((double)con->getBw()/(double)mac_->getPhy()->getSlotCapacity(mod_rate, UL_));

	  if (req_slots_tmp1>0) {
	    allocationsize = (int) ceil((sfqos->getMinReservedRate()*FRAME_SIZE)/8);
	  } else  {
	    if (con->getCDMA() == 1) {
	      if (GENERIC_HEADER_SIZE > con->getBw()) {
		allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
		issue_pol = 1;
	      } 
	      con->setCDMA(0);
	    } else {
	      allocationsize = 0;
	    }
	  }

          int arq_enable_f = 0;
          if ((con->getArqStatus () != NULL) && (con->getArqStatus ()->isArqEnabled() == 1) && (issue_pol == 0)) {
             arq_enable_f = 1;
          }
          int t_bytes = 0;
          double ori_grant = allocationsize;
          	if (ori_grant); //quiet, compiler
          if ( (arq_enable_f == 1) && (allocationsize > 0) ) {
                t_bytes = (int)allocationsize % (int)getMac()->arq_block_size_;
                if (t_bytes > 0) {
                        allocationsize = allocationsize + (getMac()->arq_block_size_ - t_bytes);
                }
                debug10 (" ARQ enable CID :%d, arq_block :%d, requested_size :%f, arq_block_boundary_size :%f, all_header_included :%f\n", con->get_cid(), getMac()->arq_block_size_, ori_grant, allocationsize, allocationsize + (double)14);
                allocationsize = allocationsize + 14;   //optional => MAC+frag_sub+pack_sub+gm_sub+mesh_sub

          }

	  grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	  num_of_slots = grant_slots;

#ifdef BWREQ_PATCH
	  //decrement bandwidth requested by allocation
	  debug2 ("UL.Check1.3.nrtPS: CID= %d Req= %d Alloc= %d Left=%d\n", 
		  con->get_cid(), con->getBw(), 
		  num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_),
		  con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
	  con->setBw(con->getBw()-(num_of_slots*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
#endif

	  if (leftOTHER_slots<=num_of_slots) {
	    debug10 ("Fatal Error: There is not enough resource for nrtPS\n");
	    exit(1);
	  } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	  if (grant_slots<req_slots_tmp1) needmore_con[1]++;

	  debug10 ("UL.Check1.3.nrtPS, MinReservedRate :%d, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d, getBw :%d\n", sfqos->getMinReservedRate(), grant_slots, con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC(), con->getBw());
	  debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofnrtPS :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[3]);

	}

	if(schedtype==SERVICE_BE){
	  req_slots_tmp1 = (int) ceil((double)con->getBw()/(double)mac_->getPhy()->getSlotCapacity(mod_rate, UL_));

	  if (req_slots_tmp1>0) needmore_con[2]++;

	  //may effect QoS for other classes => may need more sophicicate scheduler
  	  if (con!=NULL && con->getCDMA() == 1) {
	    if (GENERIC_HEADER_SIZE > con->getBw()) {
	      allocationsize = GENERIC_HEADER_SIZE; 					//for explicit polling
	      grant_slots = (int) ceil(allocationsize/mac_->getPhy()->getSlotCapacity(mod_rate, UL_));
	      num_of_slots = grant_slots;
	    }
	    con->setCDMA(0);

	    if (leftOTHER_slots<=num_of_slots) {
	      debug10 ("Fatal Error: There is not enough resource for BE polling\n");
	      // exit(1);
	      num_of_slots = leftOTHER_slots;
	    } else leftOTHER_slots=leftOTHER_slots-num_of_slots;

	    debug10 ("UL.Check1.3.BE polling, PRE-GRANT-SLOTS :%d, Peer-CID :%d, returnCID :%d, DIUC :%d, bw_req_header :%d, setCDMA :%d\n", grant_slots, con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), return_cid_tmp, con->getPeerNode()->getDIUC(), GENERIC_HEADER_SIZE, con->getCDMA());
	    debug10 ("\tAllocatedSLots :%d, BeforeFree :%d, LeftforOTHER :%d, NumofBE :%d\n", num_of_slots, freeslots, leftOTHER_slots, conn_per_schetype[4]);

	  } 

	}//end BE

 	freeslots = leftOTHER_slots;

        if(num_of_slots> 0) {
	  temp_index = doesMapExist(con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list, ie_index);
//          temp_index = doesMapExist(con->get_cid(), cid_list, ie_index); //conn, ie list, number of ies

	  if(temp_index < 0){
	    temp_index = ie_index;
//	    cid_list[temp_index] = con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid();
            cid_list[temp_index] = con->get_cid();

	    slots_per_con[temp_index] = num_of_slots;
	    uiuc_list[temp_index] = getUIUCProfile(mod_rate);
	    ++ie_index;

	    debug10 ("UL_Check1.3: MapNotExist New_CID :%d, return index :-1, cid_list[%d] :%d,  #entry (ie_index)++ :%d\n", con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else if(temp_index<MAX_MAP_IE) {
	    slots_per_con[temp_index] += num_of_slots;

	    debug10 ("UL.Check1.3: MapExist Peer_CID :%d, return index :%d, cid_list[%d] :%d,  #entry (ie_index) :%d\n", con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), temp_index, temp_index, cid_list[temp_index], ie_index);
	    debug10 ("\tNumSlots1.2 :%d\n", num_of_slots);

	  } else {
	    freeslots += num_of_slots; 			//return back the slots
	    leftOTHER_slots += num_of_slots; 		//return back the slots
	  }

	  debug10 ("UL.Check1.3, First Assign (ugs/ertps/rtps/nrtps/no be): CID(%d), Schetype :%d, Numslots :%d, Freeleft :%d, StoreSlots[%d] :%d, mod_rate :%d, UIUC :%d\n", cid_list[temp_index], schedtype, num_of_slots, freeslots, temp_index, slots_per_con[temp_index], mod_rate, uiuc_list[temp_index]);
        }



      }
      con = con->next_entry();
    }//end con != NULL
  }//end for loop 5 Qos


  //Allocate left-over to rtPS, nrtPS, and BE fairly
  int share_next_slots = 0;
  for (int i=0;i<3;i++) needmore_c = needmore_c+needmore_con[i];
  if (needmore_c>0) share_next_slots = (int) floor(leftOTHER_slots/needmore_c);

  int first_assign = 0;

  while (needmore_c>0 && freeslots>0) {

    share_next_slots = (int) floor(freeslots/needmore_c);

    debug10 ("UL.Check1.4, (Check still need more here): Needmore Conn :%d, Free :%d, Sharenext :%d\n", needmore_c, freeslots, share_next_slots
	     );

    for(i=0;i<5;++i){
      con = head;
      if(i==0)        schedtype = SERVICE_UGS;
      else if(i==1)   schedtype = SERVICE_ertPS;
      else if(i==2)   schedtype = SERVICE_rtPS;
      else if(i==3)   schedtype = SERVICE_nrtPS;
      else            schedtype = SERVICE_BE;
    
      if ( (i==0) || (i==1) ) continue;

      first_assign = 0;
      while(con!=NULL) {

	if(con->get_category() == CONN_DATA && con->get_serviceflow()->getScheduling() == schedtype) {

	  mod_rate = mac_->getMap()->getUlSubframe()->getProfile(con->getPeerNode()->getDIUC()-DIUC_PROFILE_1+UIUC_PROFILE_1)->getEncoding();
	  if (mod_rate > OFDM_16QAM_3_4) {
	    mod_rate = OFDM_16QAM_3_4;
	  }

	  temp_index = doesMapExist(con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list, ie_index);
//          temp_index = doesMapExist(con->get_cid(), cid_list, ie_index); //conn, ie list, number of ies

	  if(temp_index < 0) return_cid_tmp = -1;
	  else return_cid_tmp = temp_index;

	  int req_slots = 0; 
	  req_slots = (int) ceil((double)con->getBw()/(double)mac_->getPhy()->getSlotCapacity(mod_rate, UL_)); 

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    if (return_cid_tmp == -1) {
	      debug10 ("UL.Check1.4.rtps/nrtps, No_CID(%d), n_Conn :%d, Free :%d\n", return_cid_tmp, needmore_c, freeslots);
	      con = con->next_entry();
	      continue;
	    }
	    if ( req_slots<=slots_per_con[return_cid_tmp] ) {
	      debug10 ("UL.Check1.4.rtPS/nrtPS, CID(%d), <=MinSlots: n_Conn :%d, REQ-SLOTS :%d, GRANT-1SLOTS :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, req_slots, slots_per_con[return_cid_tmp], freeslots);
	      con = con->next_entry();
	      continue;
	    }
	  }

	  int t_num_of_slots = 0;

	  if ( (schedtype==SERVICE_rtPS) || (schedtype==SERVICE_nrtPS) ) {
	    first_assign = slots_per_con[return_cid_tmp];

	    if (req_slots <= (first_assign + share_next_slots) ) {           
	      t_num_of_slots = req_slots;
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);
	      debug10 ("UL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		if (freeslots>0) { 
		  t_num_of_slots = (first_assign + 1);
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
		t_num_of_slots =  (first_assign + share_next_slots);            
		freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[return_cid_tmp] = t_num_of_slots;
	    }
	    debug10 ("UL.Check1.4.rtps/nrtps, CID(%d), rtps/nrtpsNeedMore: n_Conn :%d, Previous-Slots :%d, Grant-Slots :%d, Free :%d\n", cid_list[return_cid_tmp], needmore_c, first_assign, t_num_of_slots, freeslots);
	  } else { //BE
	    if(req_slots<1) {
	      con = con->next_entry();
	      continue;
	    }
	    temp_index = doesMapExist(con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list, ie_index);
//            temp_index = doesMapExist(con->get_cid(), cid_list, ie_index); //conn, ie list, number of ies

	    if(temp_index < 0){
	      temp_index = ie_index;
	      cid_list[temp_index] = con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid();
//              cid_list[temp_index] = con->get_cid();

	      slots_per_con[temp_index] = 0;
	      uiuc_list[temp_index] = getUIUCProfile(mod_rate);
	      ++ie_index;
	      debug10 ("UL.Check1.4.BE, CID(%d), Initial_BE: n_Conn :%d, REQ-SLOTS :%d, Free :%d, getBW :%d\n", cid_list[temp_index], needmore_c, req_slots, freeslots, con->getBw());

	    } else {
	    }

	    first_assign = slots_per_con[temp_index];
	    int bw_assigned = 0; //amount we are assigning in the round
#ifdef BWREQ_PATCH
	    //req_slots is updated each round since we update remain Bw to allocate
	    if (req_slots <= share_next_slots ) {
	      bw_assigned = req_slots;
	      t_num_of_slots = req_slots;
	      slots_per_con[temp_index] += req_slots;
	      needmore_c--;
	      freeslots = freeslots - req_slots;
#else
	    if (req_slots <= (first_assign + share_next_slots) ) {
	      bw_assigned = req_slots;
	      t_num_of_slots = req_slots;
	      slots_per_con[temp_index] = t_num_of_slots;
	      needmore_c--;
	      freeslots = freeslots - (req_slots-first_assign);
#endif

	      debug10 ("UL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENoNeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    } else {
	      if (share_next_slots==0) {
		//there is less slot than connections left..assign slot 1 by 1.
		if (freeslots>0) { 
		  bw_assigned = 1;
		  t_num_of_slots = (first_assign + 1); //assign one more slot
		  freeslots = freeslots - 1;
		} else t_num_of_slots = first_assign;
	      } else {
			bw_assigned = share_next_slots;
			t_num_of_slots =  (first_assign + share_next_slots); //assign its share of slots
			freeslots = freeslots - share_next_slots;
	      }
	      slots_per_con[temp_index] = t_num_of_slots;

	      debug10 ("UL.Check1.4.BE, Peer-CID(%d), returnCID(%d), BENeedMore: n_Conn :%d, Pre-Grant-Slots :%d, Free :%d\n", con->getPeerNode()->getBasic(OUT_CONNECTION)->get_cid(), cid_list[temp_index], needmore_c, t_num_of_slots, freeslots);
	    }

#ifdef BWREQ_PATCH
	    //update Bw requested left after allocation
	    debug2 ("\tBE: CID= %d Req= %d Slot cap=%d slots= %d Alloc= %d Left=%d\n", 
		    con->get_cid(), con->getBw(), mac_->getPhy()->getSlotCapacity(mod_rate, UL_), req_slots,
		    bw_assigned*mac_->getPhy()->getSlotCapacity(mod_rate, UL_),
		    con->getBw()-(bw_assigned*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));
	    con->setBw(con->getBw()-(bw_assigned*mac_->getPhy()->getSlotCapacity(mod_rate, UL_)));	    
#endif

	  }	//else BE

	}//else CONNDATA
	con = con->next_entry();
      }//while
    }//for

  }//while

  if(stripping == HORIZONTAL_STRIPPING) {
  	//keept track of whether the vertically mapped contention regions have already been mapped so it's safe to go on to others.  these should always be mapped first so nothing is wrapped to new rows of symbols overlapping with them
  	bool init_ranging_region_mapped = false;
  	bool req_region_mapped = false;
  	//store the current position (next (subchannel,symbol) to be used)
  	int running_slot_offset = 0;
  	int running_subchannel_offset = 0;
  	//define the rectangle taken up by the initial ranging and bw_req contention regions in the UL subframe
  	int contention_region_width = 0;
  	int contention_region_height = 0;
  	
  	//we're first going to deal with the vertically mapped initial request and cmda_bw_req regions
  	//then the horiztonally mapped everything else
    for(i=0; i<ie_index; i++) {
      ies[i].ie.repetition_coding = 0; //not implemented.  according to spec should only be implemented for QPSK burst profile
      
      if (uiuc_list[i] == UIUC_INITIAL_RANGING) {
      	init_ranging_region_mapped = true;
      	ies[i].ie.cid = cid_list[i];
      	ies[i].ie.duration = slots_per_con[i];
      	ies[i].ie.uiuc = uiuc_list[i];

	//note, contention region is defined to be entirely in first column.  This code was written assuming it could span multiple columns.
	//it is being left more or less as is in case that becomes possible at some point.  However, contention region width and height are being
	//set to the correct values and not changed so this functions correctly
	contention_region_width = 1;
	contention_region_height = total_subchannels;
      	
      	//if we can fit it all in the first symbol time (accross several subchannels), great, do that
//      	if (num_of_slots <= total_subchannels) {
//      		contention_region_width = 1;
//      		contention_region_height = ies[i].ie.duration;
//      	}
//      	//otherwise figure out how much room it will take and store that information
//      	else {
//      		contention_region_width = ceil ( (double)ies[i].ie.duration / (double)total_subchannels );
//      		contention_region_height = ies[i].ie.duration % total_subchannels; //height in the last (at least) partially used column
//      	}
      	
      } else if (uiuc_list[i] == UIUC_REQ_REGION_FULL) {
      
      	req_region_mapped = true;
      	if (!init_ranging_region_mapped) {
      		debug10("Initial ranging region appears to be missing... this may cause problems\n");
      	}
		ies[i].ie.cid = cid_list[i];
      	ies[i].ie.duration = slots_per_con[i];
      	ies[i].ie.uiuc = uiuc_list[i];


	//see comment above  contention region size being set for INIT ranging region
      	
      	//we look at the following possibilities
      	//1 - this fits in the same column as the intial ranging region ends
      	//2 - we have to add one or more columns to fit this in

//      	if (contention_region_height + num_of_slots <= total_subchannels) {
//      		//1
//      		//contention_region_width is unchanged
//			contention_region_height += ies[i].ie.duration;
//	} else {
//		//2 and 3
//		
//		int remaining_height = total_subchannels - contention_region_height;
//		int remaining_slots = ies[i].ie.duration - remaining_height;
//		contention_region_width += ceil ( (double)remaining_slots / (double)total_subchannels );
//		contention_region_height = remaining_slots % total_subchannels;
//	}

      } else { //the horizontally striped regions
		  
		  //need to readd support for cdma_ie to the data structure
		  //also update print statements below to calculate the correct values for display
		  
		  
		  ies[i].ie.cid = cid_list[i];
		  ies[i].ie.duration = slots_per_con[i];
		  
		  
		  if (cdma_flag_list[i]>0) {
			 ies[i].ie.cdma_ie.ie.subchannel = cdma_top_list[i];
			 ies[i].ie.cdma_ie.ie.code = cdma_code_list[i];
			 ies[i].ie.uiuc = UIUC_CDMA_ALLOCATION_IE;
			 ies[i].ie.cdma_ie.ie.uiuc = uiuc_list[i];
		  } else {
		  	 ies[i].ie.uiuc = uiuc_list[i];
			 ies[i].ie.cdma_ie.ie.subchannel = 0;
			 ies[i].ie.cdma_ie.ie.code = 0;
		  }
	
		  debug10 ("UL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid :%d, #Slots :%d, UIUC :%d, Duration :%d\n", ie_index, i, ies[i].ie.cid, ies[i].ie.duration/3, ies[i].ie.uiuc, ies[i].ie.duration);
		  debug10("\tStart Position: slot_offset[%d], subchannel_offset[%d]\n", running_slot_offset, running_subchannel_offset);
		  
		  running_slot_offset += ies[i].ie.duration;
		  //repeatedly subtract the width of the available portion of the subframe until we fit on a
		  //given row.  this will vary in length by 1 depending on whether we are above or below the
		  //contention regions
		  
		  while (running_slot_offset > (ies[i].ie.duration - contention_region_width) ) {
		  	if (total_subchannels - running_subchannel_offset > contention_region_height)
		  		running_slot_offset -= ies[i].ie.duration - (contention_region_width - 1);
		  	else 
		  		running_slot_offset -= ies[i].ie.duration - contention_region_width;
		  	running_subchannel_offset++;
		  }
		  
		  debug10("End Position: subchannel_offset[%d], slot_offset[%d]\n", running_slot_offset, running_subchannel_offset );

	
	/*
		  subchannel_offset += num_of_subchannels - 1;
		  symbol_offset = (symbol_offset + num_of_symbols)%(total_slots);
	*/
	 }
    }
  }

  if(stripping == VERTICAL_STRIPPING) {
  debug2 ("BSScheduler::ul_stage2 called with vertical striping selected, which is no longer supported - MAP WILL NOT BE GENERATED\n");
//     for(i=0;i<ie_index;++i){
//       ies[i].ie.cid = cid_list[i];
//       num_of_slots = slots_per_con[i];
//       ies[i].ie.uiuc = uiuc_list[i];
//       ies[i].ie.subchannel_offset = subchannel_offset + subchannel_start;
//       ies[i].ie.symbol_offset = symbol_offset + slot_start;
//       num_of_subchannels = num_of_slots;
//       
//       num_of_symbols = (int) ceil((double)(subchannel_offset + num_of_subchannels)/total_subchannels)*3;
//       ies[i].ie.num_of_symbols = num_of_symbols;
//       ies[i].ie.num_of_subchannels = num_of_subchannels;
// 
//       if (cdma_flag_list[i]>0) {
//          ies[i].ie.cdma_ie.ie.subchannel = cdma_top_list[i];
//          ies[i].ie.cdma_ie.ie.code = cdma_code_list[i];
//       } else {
//          ies[i].ie.cdma_ie.ie.subchannel = 0;
//          ies[i].ie.cdma_ie.ie.code = 0;
//       }
// 
//       debug10 ("UL.Check1.5: ie_index(MAX=60) :%d, ies[%d].cid :%d, #Slots :%d, UIUC :%d\n", ie_index, i, ies[i].ie.cid, num_of_slots, ies[i].ie.uiuc);
//       debug10 ("\tIE_Subchannel_offset (sub_offset<%d>+sub_start<%d>) :%d, #IE_Subchannel :%d\n", subchannel_offset, subchannel_start, ies[i].ie.subchannel_offset, ies[i].ie.num_of_subchannels);
//       debug10 ("\tIE_Symbol_offset (sym_offset<%d>+sym_start<%d>) :%d, #Symbols [ceil((sub_offset<%d>+#Subchannel<%d>)/total_sub<%d>)*3] :%d\n", symbol_offset, slot_start, ies[i].ie.symbol_offset, subchannel_offset, num_of_subchannels, total_subchannels, ies[i].ie.num_of_symbols);
// 
//       subchannel_offset = (subchannel_offset + num_of_subchannels)%(total_subchannels);
//       symbol_offset += num_of_symbols - 3;
// 
//     }
  }

  ul_map->nb_ies = ie_index;
  return ul_map;
}


/*
	This function transfers packets from the queue in a connection to a burst.
	If that connection supports fragmentation and/or packing, those are employed
	to maximize utilization of granted bandwidth.

	Upon completion of this function, any transfered packets will have been removed
	from the Connection's queue and added in final form (all headers correct) to the
	Burst.  The return value is the total number of bytes transfered.

	The following is an outline of what specifically happens:
	1. If there are no packets in the Connection's queue, return


	2. If Fragmentation and Packing are both disabled, add the first packet in
	   the connection queue if possible, then return
	

	3. If Fragmentation is Enabled and Packing is not:
		3.A If the entire first packet will fit, add it and return
		3.B If the entire first packet will not fit, add the part that does
	
	4. If Packing is Enabled and Fragmentation is not:
		4.A If the entire first packet will not fit, return
		4.B Add the first packet
		4.C remove that packet from the connection, so the next is now in position 1
		4.D Goto 4.A
	
	5.  If Packing and Fragmentation are BOTH enabled:
		5.A If the entire first packet will not fit, add what will and return
		5.B If the entire first packet fits and there is not enough froom for another, add it and return
		5.C Add the first packet entirely and remove it from the connection queue, so the next is now in position 1
		5.D Goto 5.A
*/

uint32_t Wimax2BSScheduler::transfer_packets_for_dl_subframe(Connection *c, Burst *b) {
	debug2 ("At %f, Entering Wimax2BSScheduler::transfer_packets_for_dl_subframe\n", NOW);
	//1. If there are no packets in the Connection's queue, return
	if (c->queueLength() == 0) return 0;

	Ofdm_mod_rate modulation;
	if (c->get_cid() == BROADCAST_CID) modulation = getMac()->getMap()->getDlSubframe()->getProfile(b->getIUC())->getEncoding();
	else modulation = getMac()->getMap()->getDlSubframe()->getProfile (c->getPeerNode()->getDIUC())->getEncoding();
	uint32_t remaining_bytes_in_burst = getMac()->getPhy()->getMaxPktSize( b->getDuration() * b->getnumSubchannels() , modulation, DL_);

	//2. If Fragmentation and Packing are both disabled, add the first packet in
	//   the connection queue if possible, then return
	
	if ((! c->isFragEnable() ) &&  (! c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		
		//if our one packet won't fit...
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			p->AddHeader(ah);
			c->enqueue_head(p);
			return 0;
		}
		//otherwise, add it
	
		//since Fragmentation is disabled, this cannot be a later portion of a packet
		p->AddHeader(ah.GMH);
		b->enqueue(p);
		c->setLastFrameTransmitted(mac_->getFrameNumber());
		return ah.GMH.GMH.len;
	}


	//3. If Fragmentation is Enabled and Packing is not:
	//	3.A If the entire first packet will fit, add it and return
	//	3.B If the entire first packet will not fit, add the part that does

	if (( c->isFragEnable() ) &&  (! c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		

		//if we've already fragmented this packet, we'll need a fragmentation subheader
		//this is taken into account for total packet size here to ensure we accurately
		//check if the packet can fit unfragmented or not
		if (ah.GMH.GMH.type_frag == 1) {
			ah.GMH.GMH.len += ah.FSH.GetSerializedSize();
		}
		
		//if our one packet won't fit, fragment it
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			
			//set up the fragmentation header

			//if we're already fragmented
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_CONT;
			} else {
				ah.FSH.data.fc = FRAG_FIRST;
			}
			ah.FSH.data.fsn = 0; //used by ARQ, not currently supported
			ah.FSH.data.sn = c->getFragNumber();
			c->incrementFragNumber();

			//Fragment the packet
			Ptr<Packet> transmission_fragment = p->CreateFragment(0, remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize());

			//set the fields in the GMH to the correct values
			ah.GMH.GMH.type_frag = 1;
			ah.GMH.GMH.len = remaining_bytes_in_burst;

			//add the headers and enqueue
			transmission_fragment->AddHeader(ah.FSH);
			transmission_fragment->AddHeader(ah.GMH);
			c->setLastFrameTransmitted(mac_->getFrameNumber());
			b->enqueue(transmission_fragment);

			p->RemoveAtStart(remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize());
			ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
			p->AddHeader(ah);
			c->enqueue_head(p);
			return remaining_bytes_in_burst;
		}
		//otherwise, add it whole


		 //if we're already fragmented, but we can fit the rest of the packet here
		if (ah.GMH.GMH.type_frag == 1) {
			ah.FSH.data.fc = FRAG_LAST;
			ah.FSH.data.fsn = 0;
			ah.FSH.data.sn = c->getFragNumber();
			c->incrementFragNumber();
			p->AddHeader(ah.FSH);
		}

		p->AddHeader(ah.GMH);
		b->enqueue(p);
		c->setLastFrameTransmitted(mac_->getFrameNumber());
		return ah.GMH.GMH.len;
	}

	//4. If Packing is Enabled and Fragmentation is not:
	//	4.A If the entire first packet will not fit, return
	//	4.B Add the first packet
	//	4.C remove that packet from the connection, so the next is now in position 1
	//	4.D Goto 4.A

	if ((! c->isFragEnable() ) &&  ( c->isPackingEnable() )) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		
		//if our first packet won't fit, we can't do anything
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			p->AddHeader(ah);
			c->enqueue_head(p);
			return 0;
		}

		//we can fit at least one packet in, determine if we can fit more than one, so we know to take Packing Subheaders into account for size

		Ptr<Packet> p_next = c->dequeue();
		All_Headers ah_next;
		if (p_next != 0) {
		    p_next->PeekHeader(ah_next);
		    c->enqueue_head(p_next);
		}

		//if we can only fit our first packet in the burst (or there is only one packet), add it and return
		if (p_next == 0 || ah.GMH.GMH.len + ah.GMH.GMH.len + 2 * ah.PSH.GetSerializedSize() > remaining_bytes_in_burst) {
			ah.GMH.GMH.type_pck = 0;
			p->AddHeader(ah.GMH);
			b->enqueue(p);
			c->setLastFrameTransmitted(mac_->getFrameNumber());
			return ah.GMH.GMH.len;
		}
	
		//we are going to pack two or more packets into this burst
		Ptr<Packet> p_burst= getMac()->getPacket();
		All_Headers ah_burst;
		p_burst->RemoveHeader(ah_burst);

		ah_burst.GMH.GMH.type_pck = 1;
		remaining_bytes_in_burst -= ah.GMH.GetSerializedSize();
		

		while ( ah.GMH.GMH.len - ah.GMH.GetSerializedSize() + ah.PSH.GetSerializedSize() <= remaining_bytes_in_burst) {
			//set up the packing subheader
			ah.PSH.data.fc = FRAG_NOFRAG;
			ah.PSH.data.sn = c->incrementPackNumber();
			ah.PSH.data.length = ah.GMH.GMH.len - ah.GMH.GetSerializedSize() + ah.PSH.GetSerializedSize();
			p->AddHeader(ah.PSH);
			p_burst->AddAtEnd(p);
			remaining_bytes_in_burst -= ah.PSH.data.length;


			//get the next packet to examine
			p= c->dequeue();
			//if we're out of packets, stop looping
			if (p== 0) break;
			//otherwise, pull off the All_Headers to get size information for next loop
			p->RemoveHeader(ah);
		}

		//if p_next isn't null, then the last packet dequeued from the connection couldn't fit, so we need to add it back on
		//so that it will be sent in a future frame
		if (p != 0) {
			p->AddHeader(ah);
			c->enqueue_head(p);
		}


		//Generate the GMH for the accumulator packet that we've been copying everything into
		ah_burst.GMH.GMH.type_pck = 1;
		ah_burst.GMH.GMH.len = p_burst->GetSize() + ah_burst.GMH.GetSerializedSize();
		ah_burst.GMH.GMH.cid = c->get_cid();

		p_burst->AddHeader(ah_burst.GMH);
		b->enqueue(p);
		c->setLastFrameTransmitted(mac_->getFrameNumber());
		return ah_burst.GMH.GMH.len;
	}


	//5.  If Packing and Fragmentation are BOTH enabled:
	//	5.A If the entire first packet will not fit, add what will and return
	//	5.B If the entire first packet fits, add it and remove from the queue so we have a new first, go to 5.A
	//	5.C If part of the first packet fits, add what will
	//	5.D Set up the accumulator packet's header (GMH for the whole burst), and return

	//note that if we apply both fragmentation and packing to the same burst, we only use packing subheaders.  the fc field on these
	//works the same as the fc field on the fragmentation subheaders, so using both is not necessary.
	if (c->isFragEnable() && c->isPackingEnable()) {
		Ptr<Packet> p = c->dequeue();
		All_Headers ah;
		p->RemoveHeader(ah);
		
		
		//if our first packet won't fit, we'll fragment it
		//since part of the first fragment takes the whole burst, we won't have room to pack more in
		if (ah.GMH.GMH.len > remaining_bytes_in_burst) {
			//set up the fragmentation header

			//if we're already fragmented
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_CONT;
			} else {
				ah.FSH.data.fc = FRAG_FIRST;
			}
			ah.FSH.data.fsn = 0; //used by ARQ, not currently supported
			ah.FSH.data.sn = c->incrementFragNumber();

			//Fragment the packet
			Ptr<Packet> transmission_fragment = p->CreateFragment(0, remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize());

			//set the fields in the GMH to the correct values
			ah.GMH.GMH.type_frag = 1;
			ah.GMH.GMH.type_pck = 0;
			ah.GMH.GMH.len = remaining_bytes_in_burst;

			//add the headers and enqueue
			transmission_fragment->AddHeader(ah.FSH);
			transmission_fragment->AddHeader(ah.GMH);
			b->enqueue(transmission_fragment);
			c->setLastFrameTransmitted(mac_->getFrameNumber());

			p->RemoveAtStart(remaining_bytes_in_burst - ah.FSH.GetSerializedSize() - ah.GMH.GetSerializedSize());
			ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
			p->AddHeader(ah);
			c->enqueue_head(p);
			return p->GetSize();
		}

		//if we make it this far, we know that we can fit at least one full packet
		//we now want to loop on adding new packets to the burst until either we 
		//run out of room in the burst or out of packets in the Connection's queue

		//before we loop we check how much data we do have.  if it's enough for the first packet and no 
		//others (or if there's only one packet, which has already been dequeued), we add it and return
		//without adding packing subheader(s)
		uint32_t bytes_to_next_payload = c->queueLength() > 0 ? p->GetSize() + ah.PSH.GetSerializedSize() * 2: p->GetSize() + ah.PSH.GetSerializedSize();
		if (remaining_bytes_in_burst <= bytes_to_next_payload || c->queueLength() == 0) {
			//just add the first packet
			if (ah.GMH.GMH.type_frag == 1) {
				ah.FSH.data.fc = FRAG_LAST;
				ah.FSH.data.fsn = 0;
				ah.FSH.data.sn = c->incrementFragNumber();
				ah.GMH.GMH.type_frag = 1;
				p->AddHeader(ah.FSH);
			}
			p->AddHeader(ah.GMH);
			b->enqueue(p);
			c->setLastFrameTransmitted(mac_->getFrameNumber());
			return p->GetSize();
		}

		//looks like we're going to pack a few in

		Ptr<Packet> p_burst = getMac()->getPacket();
		All_Headers ah_burst;
		p_burst->RemoveHeader(ah_burst);

		ah_burst.GMH.GMH.type_pck = 1;
		ah_burst.GMH.GMH.type_frag = 0;

		remaining_bytes_in_burst -= ah.GMH.GetSerializedSize();
		bool partial_packet_left = true; //used to know if we should try to add a frag to the end of the allocation
		bool first_loop = true;

		debug10("Packing several packets into burst\n");

		while (remaining_bytes_in_burst > bytes_to_next_payload) {
			if (first_loop && ah.GMH.GMH.type_frag == 1) {
				ah.PSH.data.fc = FRAG_LAST;
				ah_burst.GMH.GMH.type_frag = 1;
			} else {
				ah.PSH.data.fc = FRAG_NOFRAG;
			}
			first_loop = false;

			ah.PSH.data.sn = c->incrementPackNumber();
			ah.PSH.data.length = p->GetSize() + ah.PSH.GetSerializedSize();
			p->AddHeader(ah.PSH);
			p_burst->AddAtEnd(p);
			debug10("\t Added packet of size %d (including packing subheader) to burst\n", ah.PSH.data.length);
			remaining_bytes_in_burst -= p->GetSize();

			if (c->queueLength() > 0 && remaining_bytes_in_burst > ah.PSH.GetSerializedSize()) p = c->dequeue();
			else {
				partial_packet_left = false;
				break;
			}
			p->RemoveHeader(ah);
			//bytes_to_next_payload = c->queueLength() > 0 ? p->GetSize() + ah.PSH.GetSerializedSize() * 2: p->GetSize() + ah.PSH.GetSerializedSize();
			bytes_to_next_payload = p->GetSize() + ah.PSH.GetSerializedSize();
		}

		//we've added as many full packets as we can.  check if we can add part of this one still
		if (partial_packet_left && p != 0 && remaining_bytes_in_burst > ah.PSH.GetSerializedSize()) {
			if (remaining_bytes_in_burst > ah.PSH.GetSerializedSize()) {
				if (ah.GMH.GMH.type_frag == 1) {
					ah.PSH.data.fc = FRAG_CONT;
					ah_burst.GMH.GMH.type_frag = 1;
				} else {
					ah.GMH.GMH.type_frag = 1;
					ah.PSH.data.fc = FRAG_FIRST;
					ah_burst.GMH.GMH.type_frag = 1;
				}
				ah.PSH.data.sn = c->incrementFragNumber();
				ah.PSH.data.length = remaining_bytes_in_burst;
				
				//there is a boundary condition that's causing us to rarely get here despite having room for another full packet (but just barely)
				//this can probably be fixed above, but can be fixed here faster...
				if (remaining_bytes_in_burst -ah.FSH.GetSerializedSize() < p->GetSize()) {
					Ptr<Packet> transmission_fragment = p->CreateFragment(0, remaining_bytes_in_burst - ah.FSH.GetSerializedSize());

					transmission_fragment->AddHeader(ah.PSH);
					p_burst->AddAtEnd(transmission_fragment);
					debug10("\t Added packet fragment of size %d (of %d) (including packing subheader) to burst\n", ah.PSH.data.length, p->GetSize());


					p->RemoveAtStart(remaining_bytes_in_burst - ah.PSH.GetSerializedSize());
					ah.GMH.GMH.len = p->GetSize() + ah.GMH.GetSerializedSize();
					p->AddHeader(ah);
					c->enqueue_head(p);
				} else {
					p->AddHeader(ah.PSH);
					p_burst->AddAtEnd(p);
					debug10("\t Added packet of size %d (including packing subheader) to burst\n", ah.PSH.data.length);
				}
			} else {
				debug10("\t Could not fit even a fragment of the next packet into the burst.  Placing that packet at the head of the queue\n");
				p->AddHeader(ah);
				c->enqueue_head(p);
			}
		}

		//set up the GMH for the accumulator packet

		ah_burst.GMH.GMH.len = p_burst->GetSize() + ah_burst.GMH.GetSerializedSize();
		ah_burst.GMH.GMH.cid = c->get_cid();
		p_burst->AddHeader(ah_burst.GMH);
		b->enqueue(p_burst);
		c->setLastFrameTransmitted(mac_->getFrameNumber());
		return p_burst->GetSize();
	}
	//should really never get here...
	return 0;
	

}

} //namespace ns3
